/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        acitscsi.h

    Purpose:

        Header file for declaring all SCSI data structures. This file is
        included by ACIT library and OSMs to resolve SCSI data structures
        that are shared across ACIT and OSMs.
        

--*/



#define ACIT_SCSI_MAX_CDB_LENGTH                    (16)


/*
 * SCSI Command Opcodes
 */
#define ACIT_SCSI_CMD_TUR                           (0x00)
#define ACIT_SCSI_CMD_INQUIRY                       (0x12)
#define ACIT_SCSI_CMD_REPORTLUNS                    (0xa0)


/*
 * Other SCSI Definition
 */
#define ACIT_SCSI_CMD_INQUIRY_EVPD_BIT              (0x01)

/*
typedef struct _ACIT_LUN
{
    union
    {
        ACIT_UEXACT8            lun_bytes[8];
        ACIT_UEXACT16           lun_leveled[4];
        ACIT_U64                lun_complete;
    } u;
} ACIT_LUN, *PACIT_LUN;
*/

// fixed format sense data

// valid response_code
#define SENSE_DATA_CURRENT_ERRORS       (0x70)
#define SENSE_DATA_DEFERRED_ERRORS      (0x71)

// sense key code
#define SENSE_KEY_NO_SENSE              (0)
#define SENSE_KEY_RECOVERED_ERROR       (1)
#define SENSE_KEY_NOT_READY             (2)
#define SENSE_KEY_MEDIUM_ERROR          (3)
#define SENSE_KEY_HARDWARE_ERROR        (4)
#define SENSE_KEY_ILLEGAL_REQUEST       (5)
#define SENSE_KEY_UNIT_ATTENTION        (6)
#define SENSE_KEY_DATA_PROTECT          (7)
#define SENSE_KEY_BLANK_CHECK           (8)
#define SENSE_KEY_VENDOR_SPECIFIC       (9)
#define SENSE_KEY_COPY_ABORTED          (10)
#define SENSE_KEY_ABORTED_CMD           (11)
#define SENSE_KEY_OBSOLETE              (12)
#define SENSE_KEY_VOLUME_OVERFLOW       (13)
#define SENSE_KEY_MISCOMPARE            (14)
#define SENSE_KEY_RESERVED              (15)

// ASC and ASCQ handled
#define ASC_REPORT_LUN_DATA_HAS_CHANGED (0x3F)
#define ASCQ_REPORT_LUN_DATA_HAS_CHANGED (0x0E)

typedef struct _ACIT_SCSI_SENSE_DATA
{
    ACIT_UEXACT8    response_code:7;
    ACIT_UEXACT8    valid:1;

    ACIT_UEXACT8    obsolete_1;
    
    ACIT_UEXACT8    sense_key:4;
    ACIT_UEXACT8    rsvd_1:1;
    ACIT_UEXACT8    ili:1;
    ACIT_UEXACT8    eom:1;
    ACIT_UEXACT8    filemark:1;

    ACIT_UEXACT8    information[4];
    ACIT_UEXACT8    addi_sense_len;
    ACIT_UEXACT8    cmd_specific_info[4];
    ACIT_UEXACT8    asc;
    ACIT_UEXACT8    ascq;
    ACIT_UEXACT8    field_repl_unit_code;

    ACIT_UEXACT8    sense_key_specific_1:7;
    ACIT_UEXACT8    sense_key_specific_valid:1;
    ACIT_UEXACT8    sense_key_specific_2[2];

} ACIT_SCSI_SENSE_DATA, *PACIT_SCSI_SENSE_DATA;

/*
 * The following is brought over from acitapi.h
 * acitapi.h could'nt have these since the SG
 * file doesnt support these definitions
 */
/*
 * Do definition checks to make sure OSM has exported all required definitions
 */
#ifndef ACIT_CFG_DEBUG_BUILD
#error ACIT_CFG_DEBUG_BUILD definition is missing, (see acitcfg.h)
#endif
#ifndef ACIT_CFG_CPU_LITTLE_ENDIAN
#error ACIT_CFG_CPU_LITTLE_ENDIAN definition is missing, (see acitcfg.h)
#endif
#ifndef ACIT_CFG_VASE_ENABLED
#error ACIT_CFG_VASE_ENABLED definition is missing, (see acitcfg.h)
#endif
#ifndef ACIT_CFG_SIMULATION
#error ACIT_CFG_SIMULATION definition is missing, (see acitcfg.h)
#endif

#ifndef ACIT_CFG_MAX_CONTIGUOUS_MEM
#error ACIT_CFG_MAX_CONTIGUOUS_MEM definition is missing, (see acitcfg.h)
#else
#if ACIT_CFG_MAX_CONTIGUOUS_MEM < (128 * 1024)
#error unsupported ACIT_CFG_MAX_CONTIGUOUS_MEM value. need atleast 128KB , (see acitcfg.h)
#endif
#endif


