/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


/*!
@file
   blade.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/initiator/acit/blade.h,v 1.8 2006/02/28 02:02:09 prabalm Exp $
*/
#ifndef __blade_h__
#define __blade_h__

/* PCI definitions for BE ASIC */
#define BE_PCI_VENDORID                             0x19A2
#define BE_PCI_VENDORID_STR                         "19A2"
#define BE_PCI_DEVICEID0                            0x0212
#define BE_PCI_DEVICEID0_STR                        "0212"

#define BE_PCI_DEVICEID1                            0x0702
#define BE_PCI_DEVICEID1_STR                        "0702"

#define BE_PCI_DEVICEID2                            0x0703
#define BE_PCI_DEVICEID2_STR                        "0703"

#define BE_PCI_DEVICEID3                            0x0222
#define BE_PCI_DEVICEID3_STR                        "0222"

#define BE_PCI_DEVICEID4                            0x0712
#define BE_PCI_DEVICEID4_STR                        "0712"

#define BE_MAX_R2T_CREDITS                          (128)
#define BE_MAX_ITTS_PER_INVALIDATE                  (128)

// Maximum length a Scatter Gather Element can hold
#define BE_MAX_BYTES_PER_FRAG                       (64*1024)

#define BE_MAX_SGE_PER_SG                           (256)

// Max number of SG elements supported by ARM fw
#define BE_MAX_SGE_IN_MCC_WRB                       (19)

// 64Bytes WRB times 256 commands per connection
#define BE_MAX_WRB_BYTES_PER_CXN                    (16*1024)

#define CEV_TIMER_MAX                               (250)

#if defined (BE_EMULATION)// || (FPGA_ARM)
#define CRI_FROM_CID(pc, cid)                           (ACIT_UEXACT16)(cid)
#define CID_FROM_CRI(pc, cri)                           (ACIT_UEXACT16)(cri)
#else
//#define CRI_FROM_CID(pc, cid)                           (cid-pc->fw_config.iscsi_cid_start)
//#define CID_FROM_CRI(pc, cri)                           (cri+pc->fw_config.iscsi_cid_start)
#define CRI_FROM_CID(pc, cid)                           (ACIT_UEXACT16)((cid-pc->fw_config.iscsi_cid_start)/2)
#define CID_FROM_CRI(pc, cri)                           (ACIT_UEXACT16)(cri*2+pc->fw_config.iscsi_cid_start)
#endif

#define SGLICD_FROM_ITT(itt)                            (ACIT_UEXACT16)(itt & 0xfff)

#if defined (BE_EMULATION)// || (FPGA_ARM)
#define BE_MAX_CXNS                                 (512)
#else
#ifdef ENABLE_512_TARGETS
#define BE_MAX_CXNS                                 (512)
#else
#define BE_MAX_CXNS                                 (128)
#endif
#endif

#define BE_MAX_TARGETS                              (BE_MAX_CXNS)
#define BE_MAX_CMD_PER_CXN                          (256)
#define BE_MAX_ICDS                                 (2048/2)
#define BE_MAX_IOS                                  (BE_MAX_ICDS - BE_MAX_CXNS)

/* This is size of MCC ring divided by the size of MCC_WRB */
#define MAX_MCC_WRBS                                (128)

//
// Set to 1 though BE has 2 physical ports. The reason being the ARM
// abstracts the knowledge of physical ports and transparently fails
// over IOs across ports
//
#define MAX_PHYS_PORTS                              (1)

//
// hardware needs the async PDU buffers to be posted in multiples of 8
// So have atleast 8 of them by default
//
#define BE_MIN_ASYNC_ENTRIES                        (8)

//#define BE_MCC_PAYLOAD_THRESHOLD                    (236)
// Max payload size for embedded MCC WRB not including the IOCTL header
#define BE_MCC_PAYLOAD_THRESHOLD                    (220)
// move to .sg files

#define TPOST_SEMAPHORE_OFFSET                      (0x4ac)
#define TPRE_SEMAPHORE_OFFSET                       (0x8ac)
#define RXULP0_SEMAPHORE_OFFSET                     (0x10ac)
#define RXULP1_SEMAPHORE_OFFSET                     (0x14ac)
#define RXULP2_SEMAPHORE_OFFSET                     (0x18ac)
#define TXULP0_SEMAPHORE_OFFSET                     (0x1cac)
#define TXULP1_SEMAPHORE_OFFSET                     (0x20ac)
#define TXULP2_SEMAPHORE_OFFSET                     (0x24ac)
#define MPU_SEMAPHORE_OFFSET                        (0xac)

//
// Post SGL pages takes parameters in terms of pages. BE understands
// only 4K size pages
//
#define BE_PAGE_SIZE                                (4096)

#endif

