/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        hwi.h

    Purpose:

        Header file for declaring all hardware interface functions 
        

--*/

#define GET_HWI_CONTROLLER_WS(pc)           (pc->phwi_ws)
#define HWI_GET_CQ_ID(pc, msg)              ((PHWI_CONTROLLER_WS)(GET_HWI_CONTROLLER_WS(pc)))->partition[msg].cq.id
#define HWI_GET_DEF_BUFQ_ID(pc)             ((PHWI_CONTROLLER_WS)(GET_HWI_CONTROLLER_WS(pc)))->default_pdu_buf.id
#define HWI_GET_DEF_HDRQ_ID(pc)             ((PHWI_CONTROLLER_WS)(GET_HWI_CONTROLLER_WS(pc)))->default_pdu_hdr.id
#define HWI_GET_ASYNC_PDU_CTX(phwi)         (phwi->phwic->pasync_ctx)
/*
 * This defines the number of SA SGLs that can be constructed from
 * ACIT's memory descriptor. This is needed since ACIT supports a
 * variable number of descriptors and multiple descriptors need to
 * be constructed into a SGList for BECLib
 */ 
#define HWI_MAX_SGL_SUPPORTED               8

typedef enum {
    HWH_TYPE_IO = 1,
    HWH_TYPE_LOGOUT = 2,
    HWH_TYPE_TMF = 3,
    HWH_TYPE_NOP = 4,
    HWH_TYPE_INVALID = 0xFFFFFFFF
} HWH_TYPE_ENUM, *PHWH_TYPE_ENUM;

typedef struct _WRB_HANDLE {

    ACIT_DC_LIST   link;
    HWH_TYPE_ENUM   type; 
    ACIT_UEXACT16   wrb_index;
    ACIT_UEXACT16   unused_count;

    union {

        PACIT_IO_HANDLE         pio_handle;
        PACIT_LOGOUT_HANDLE     plogout_handle;
        PACIT_TMF_HANDLE        ptmf_handle;
        PACIT_NOP_HANDLE        pnop_handle;
    };
    PISCSI_WRB    pwrb;

} WRB_HANDLE, *PWRB_HANDLE;

typedef struct _SGL_HANDLE {

    ACIT_DC_LIST   link;
    ACIT_UEXACT32   sgl_index;
    PISCSI_SGE  pfrag;
} SGL_HANDLE, *PSGL_HANDLE;

#define HWI_FRAG_DESC_SZ    (sizeof(ISCSI_SGE)* BE_MAX_SGE_PER_SG * BE_MAX_ICDS)
#define HWI_MEM_NUM_FRAG    ACIT_CEIL(HWI_FRAG_DESC_SZ, ACIT_CFG_MAX_CONTIGUOUS_MEM)

#define HWI_WRB_DESC_SZ     (BE_MAX_WRB_BYTES_PER_CXN * BE_MAX_CXNS)
#define HWI_MEM_NUM_WRB     ACIT_CEIL(HWI_WRB_DESC_SZ, ACIT_CFG_MAX_CONTIGUOUS_MEM)

#define HWI_MEM_NUM_WRBH    (BE_MAX_CXNS/8)

// These have to be allocated based on the connection count
typedef struct _BE_CXN_QUEUE_OBJECT
{
    BE_ISCSI_WRB_QUEUE_OBJECT   be_wrbq;

}BE_CXN_QUEUE_OBJECT, *PBE_CXN_QUEUE_OBJECT;

// These have to be allocated based on msix messages
typedef struct _HWI_INT_QUEUE_OBJECTS
{
    BE_EQ_OBJECT    be_eq;
    BE_CQ_OBJECT    be_cq;

}HWI_INT_QUEUE_OBJECTS, *PHWI_INT_QUEUE_OBJECTS;


typedef struct _HWI_CONTEXT_MEMORY
{
    SA_DEV                      sa_dev;
    BE_CHIP_OBJECT              be_chip;
    BE_FUNCTION_OBJECT          be_function;

    BE_ERROR_OBJECT             be_error_obj;

    // Bug 21307 :
    //struct 
    //{
    //    BE_EQ_OBJECT                be_eq;
    //    BE_CQ_OBJECT                be_cq;

    //} partition[ACIT_CFG_MAX_PARTITIONS];
    PHWI_INT_QUEUE_OBJECTS      phwi_queue;

    BE_CQ_OBJECT                be_mcc_cq;
    BE_MCC_OBJECT               be_mcc;

    BE_DEFAULT_PDU_QUEUE_OBJECT be_def_hdrq;
    BE_DEFAULT_PDU_QUEUE_OBJECT be_def_bufq;

    //BE_ISCSI_WRB_QUEUE_OBJECT   be_wrbq[BE_MAX_CXNS];

    // Bug 21307 :
    // Pointer to ISCSI_WRB_QUEUE_OBJECT. As connections 
    // increase an array definition like above puts
    // pressure on overall uncached memory. So allocate
    // WRB Queue Object based on config (load time) params
    //
    PBE_CXN_QUEUE_OBJECT        pbe_wrbq;

    PBE_MCC_WRB_CONTEXT         pbe_mcc_context;

    PHWI_ASYNC_PDU_CONTEXT       pasync_ctx;

     // Track what needs to be cleaned up.
    struct {
        ACIT_UEXACT32 lib_init       : 1;
        ACIT_UEXACT32 chip_function  : 1;
        ACIT_UEXACT32 zero_buffer    : 1;
        ACIT_UEXACT32 pt             : 1;
        ACIT_UEXACT32 sgl_pages      : 1;
        ACIT_UEXACT32 templates      : 1;
        ACIT_UEXACT32 eq             : 1;
        ACIT_UEXACT32 cq             : 1;

        ACIT_UEXACT32 num_wrbq       : 6;
        ACIT_UEXACT32 mcc_cq         : 1;
        ACIT_UEXACT32 mcc            : 1;

        ACIT_UEXACT32 async_regn     : 1;
        ACIT_UEXACT32 def_bufq       : 1;
        ACIT_UEXACT32 def_hdrq       : 1;
        ACIT_UEXACT32 mailbox        : 1;
        ACIT_UEXACT32 jell           : 1;
        ACIT_UEXACT32 r2tarray       : 1;

        ACIT_UEXACT32 reserved       : 9;
    } initialized;

}HWI_CONTEXT_MEMORY, *PHWI_CONTEXT_MEMORY;

typedef struct _HWI_WRB_HANDLE_CONTEXT
{
    ACIT_DC_LIST                wrb_handle_list;
    ACIT_DC_LIST                wrb_handle_drvr_list;
    PWRB_HANDLE                 pwrb_handle_base;
    PISCSI_WRB                  plast_wrb;
    ACIT_UEXACT32               initialized;
    ACIT_UEXACT16               wrb_handles_available;
    ACIT_UEXACT16               cid;

}HWI_WRB_CONTEXT, *PHWI_WRB_CONTEXT;

typedef struct _HWI_CONTROLLER_WS
{
    PACIT_CONTROLLER_WS         pacit_ws;
    PACIT_VOID                  pscratchpad_base;
    
    ACIT_DC_LIST                io_sgl_list;    // sgl handles allocated for IOs
    ACIT_DC_LIST                eh_sgl_list;    // sgl handles allocated for error handling (TMF/logout/NOP)

    PSGL_HANDLE                 psgl_handle_base;

    PACIT_UEXACT8               sge_base[HWI_MEM_NUM_FRAG];// points to start of scatter gather elements for all connections
    PACIT_UEXACT8               wrb_base[HWI_MEM_NUM_WRB]; // points to start of wrb's for all connections
    ACIT_UEXACT32               sge_bytes_per_io;  
    ACIT_UEXACT32               wrb_bytes_per_cxn;
    ACIT_UEXACT32               max_cxns_per_descriptor;
    ACIT_UEXACT32               max_sgl_per_descriptor;
    ACIT_UEXACT32               max_wrbh_per_descriptor;

    ACIT_UEXACT32               wrb_mem_index;
    ACIT_UEXACT32               frag_mem_index;
    ACIT_UEXACT32               wrbh_mem_index;

    ACIT_UEXACT16               io_sgl_handles_available;
    ACIT_UEXACT16               eh_sgl_handles_available;

    HWI_WRB_CONTEXT             wrb_context[BE_MAX_CXNS];

    PMCC_WRB                    pmcc_wrb_base;

    struct {

        SA_RING                     eq; // event queue ring
        SA_RING                     cq; // completion queue ring

    } partition[ACIT_CFG_MAX_PARTITIONS];

    /* 
     * If this is a per-connection ring, then we need to allocate 
     * ACIT_MAX_NUM_CXNs of these..
     */
    SA_RING                     default_pdu_hdr;
    SA_RING                     default_pdu_buf;

    PHWI_CONTEXT_MEMORY         phwic;

    //ACIT_UEXACT16               cq_errors[CXN_KILLED_CMND_DATA_NOT_ON_SAME_CONN];


    ACIT_UEXACT32               nop_icd;

#ifdef BE_EMULATION
#ifdef BE_RXIF_CREATE
    BE_RXIF_OBJECT              rxif_object[2];
#endif
#endif

}HWI_CONTROLLER_WS, *PHWI_CONTROLLER_WS;

ACIT_UEXACT32
hwi_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size);

ACIT_UEXACT32
hwi_init_controller_ws(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_init_mailbox_descriptor(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_U8
hwi_disable_irq(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_U8
hwi_enable_irq(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_U32
hwi_process_irq(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_U32
hwi_process_msi_irq(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_process_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_purge_eq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_process_mcc_eq_from_init (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_process_cq (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_VOID
hwi_process_mcc_cq(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_U32
hwi_init_hiber_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle);

ACIT_U32
hwi_init_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle);

ACIT_VOID
hwi_stop_controller(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_register_arm_async_events(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_check_and_log_ue(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
hwi_check_health(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           log_error);

ACIT_VOID
hwi_check_interrupts(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           message_id);

ACIT_UEXACT32
hwi_init_beclib(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              osm_handle);

ACIT_VOID
hwi_cleanup_beclib(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_STATUS
hwi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_VOID
hwi_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle);

ACIT_STATUS
hwi_alloc_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_VOID
hwi_free_logout_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle);

ACIT_STATUS
hwi_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_VOID
hwi_free_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle);

ACIT_UEXACT32
hwi_alloc_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    ACIT_UEXACT32           cri,
    PACIT_NOP_HANDLE        poptional_nop_handle);

ACIT_VOID
hwi_free_nop_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle);

PACIT_IO_HANDLE
hwi_io_handle_from_itt(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cri, 
    ACIT_UEXACT32           itt);

PBE_FUNCTION_OBJECT
hwi_get_be_pfo(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_fill_mcc_wrb_sgl(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference);

ACIT_STATUS
hwi_fill_non_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len,
    PACIT_VOID              presponse_hdr,
    ACIT_UEXACT32           response_hdr_len,
    PACIT_VOID              posm_reference);

ACIT_STATUS
hwi_fill_embedded_wrb(
    PACIT_CONTROLLER_WS     pcontroller,
    PMCC_WRB                pwrb,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len);

#if 0
ACIT_VOID
hwi_validate_pending_ios(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);
#endif

ACIT_VOID
hwi_build_sa_sgl (
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    OUT PSA_SGL             sgl);

ACIT_VOID
hwi_build_sa_sgl_by_offset (
    IN  PACIT_CONTROLLER_WS pcontroller,
    IN  ACIT_UEXACT32       mem_descriptor_index,
    IN  ACIT_UEXACT32       offset,
    IN  ACIT_UEXACT32       size,
    OUT PSA_SGL             sgl);

STATIC INLINE
ACIT_UEXACT32
hwi_get_ring_size(
    ACIT_UEXACT32           num_elements)
{
    ACIT_UEXACT32 size = 0;

    OSM_ASSERT(num_elements);

    while(!(num_elements & 0x1))
    {
        size++;
        num_elements = (num_elements >> 1);
    }

    return (size + 1);
}

STATIC INLINE 
ACIT_VOID
hwi_ring_eq_db(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           id, 
    ACIT_UEXACT32           clr_interrupt,
    ACIT_UEXACT32           num_processed, 
    ACIT_UEXACT8            rearm, 
    ACIT_UEXACT8            event) 
{
    EQ_DB doorbell;

    doorbell.dw = 0; // initialize

    doorbell.qid   = id;
    doorbell.ClearInt = clr_interrupt;
    doorbell.rearm = rearm;
    doorbell.event = event;
    doorbell.num_popped = num_processed;

    ACIT_PD_WRITE (pcontroller, cq_db, doorbell);
}

STATIC INLINE 
ACIT_VOID
hwi_ring_cq_db(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           id, 
    ACIT_UEXACT32           num_processed, 
    ACIT_UEXACT8            rearm, 
    ACIT_UEXACT8            event) 
{
    CQ_DB doorbell;

    doorbell.dw = 0; // initialize

    doorbell.qid   = id;
    doorbell.rearm = rearm;
    doorbell.event = event;
    doorbell.num_popped = num_processed;

    ACIT_PD_WRITE (pcontroller, cq_db, doorbell);
}

ACIT_VOID
hwi_get_pci_link_info(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_VOID
hwi_cleanup_after_ue(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_start_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
hwi_poll_for_ue_recovery(
    PACIT_CONTROLLER_WS     pcontroller);
