/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

    iscsiapi.h - iSCSI Layer Interface Routines

    Purpose:

        File for internal iSCSI implementation.
        

--*/

/*
 * Private iSCSI implementation
 */
ACIT_UEXACT32
iscsi_init_connection_information (
    PACIT_TARGET_WS         ptarget,
    PCONNECTION_DATA        pcxn_data,
    ACIT_UEXACT32           is_first_cxn);

ACIT_UEXACT32
iscsi_reset_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn);

ACIT_UEXACT32
iscsi_delete_session_info(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

INLINE
ACIT_UEXACT32
iscsi_update_sn(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    ACIT_UEXACT32           max_cmd_sn,
    ACIT_UEXACT32           exp_cmd_sn);

ACIT_UEXACT32
iscsi_update_sn_by_offset(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    ACIT_UEXACT32           max_cmd_sn,
    ACIT_UEXACT32           exp_cmd_sn);

ACIT_UEXACT16
iscsi_get_conn_index(
    PISCSI_TARGET_WS        ptarget,
    PISCSI_CONNECTION       pcxn);

STATIC INLINE
PISCSI_CONNECTION 
iscsi_get_cxn_from_pool(
    PACIT_CONTROLLER_WS         pcontroller,
    ACIT_UEXACT32               cxn_index)
{
    PISCSI_CONTROLLER_WS        piscsi_controller;
    ACIT_UEXACT32               base_index, offset;
    PISCSI_CONNECTION           pcxn;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    base_index = cxn_index/(piscsi_controller->max_cxnpool_per_desc - 1);
    offset = cxn_index % piscsi_controller->max_cxnpool_per_desc;

    pcxn = (PISCSI_CONNECTION)((PACIT_UEXACT8)piscsi_controller->connections_pool[base_index]+
                 offset*sizeof(ISCSI_CONNECTION));

    OSM_ASSERT(pcxn);

    return pcxn;
}

STATIC INLINE
PISCSI_CONNECTION 
iscsi_get_cxn (
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cid) 
{        
    PISCSI_CONNECTION       pcxn = ACIT_NULL;
    ACIT_UEXACT32           cri;
#if defined (BE_EMULATION) //|| (FPGA_ARM)
    PISCSI_CONTROLLER_WS    piscsi_controller;
    ACIT_UEXACT32           index;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    for(index = 0; index < pcontroller->params.num_cxn; index++)
    {
        pcxn = piscsi_controller->connection_table[index];
        if((pcxn != ACIT_NULL) && (pcxn->cid == cid))
        {
            break;
        }
    }
    OSM_ASSERT(pcxn);
    return pcxn;
#else
    cri = CRI_FROM_CID(pcontroller, cid);
    pcxn = GET_ISCSI_CXN(pcontroller, cri);

    //OSM_ASSERT (pcxn);
    return pcxn;
#endif
}

ACIT_UEXACT16
iscsi_get_reopen_flag(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_UEXACT16
iscsi_get_cfg_flag(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);
