/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        mgmtapi.h

    Purpose:

        Header file for Management module
                

--*/

/* Timeout value in seconds for internal IOCTLs issued by Management module */
#define MGMT_INTERNAL_IOCTL_DEFAULT_TIMEOUT     30

/* Timeout value in seconds for error handling IOCTLs */
#define MGMT_EH_IOCTL_TIMEOUT                   10 

/* Timeout value in seconds for reopen session IOCTLs */
#define MGMT_REOPEN_SESSION_TIMEOUT             6

/* These timeouts are given by the firmware team */
#define MGMT_IOCTL_OFFLOAD_TIMEOUT              10
#define MGMT_IOCTL_LOGIN_TIMEOUT                250
#define MGMT_IOCTL_GET_ALL_SESSION_ID_TIMEOUT   10
#define MGMT_IOCTL_BOOT_READ_WRITE_TIMEOUT      250

#define MGMT_STATUS     ACIT_UEXACT8

#define MGMT_RST_RECV   1
#define MGMT_FIN_RECV   2

#define MGMT_SGL_ICD_STATUS_MASK                0x000000ff
#define MGMT_SGL_ICD_BYTE_COUNT_MASK            0xffffff00
#define MGMT_SGL_ICD_BYTE_COUNT_SHIFT           8

struct _MGMT_IOCTL_HANDLE;

typedef MGMT_STATUS (* PMGMT_POST)(PACIT_CONTROLLER_WS      pcontroller,
                                     struct _MGMT_IOCTL_HANDLE     *pioctl_handle,
                                     PIOCTL_RESPONSE_HEADER pioctl_hdr,
                                     PACIT_VOID             pdata_buffer);


#define MGMT_CONTROLLER_BUSY_IOCTL_HANDLE       (1<<0)
#define MGMT_CONTROLLER_BUSY_IOCTL_BUFFER       (1<<1)

//  The IOCTL buffer and IOCTL handle busy and free lists. Buffers and handles are managed
// differently because IOCTLs originating outside ACIT have the buffers already allocated.

typedef struct _MGMT_CONTROLLER_WS {

    ACIT_DC_LIST     ioctl_buffer_free_list;
    ACIT_DC_LIST     ioctl_buffer_busy_list;
    ACIT_DC_LIST     ioctl_handle_free_list;
    ACIT_DC_LIST     ioctl_handle_busy_list;

    PACIT_CONTROLLER_WS pacit_ws;


    PACIT_VOID       pioctl_handle_base;
    PACIT_VOID       pioctl_buffer_base;

    PACIT_IOCTL_REQUEST porig_ioctl_request;
    
    ACIT_UEXACT32    ioctl_handles_available;
    ACIT_UEXACT32    ioctl_buffers_available;

    ACIT_UEXACT32    flags;
    ACIT_UEXACT32    busy;
    PACIT_VOID       ununsed;

}MGMT_CONTROLLER_WS, *PMGMT_CONTROLLER_WS;


//  This is the Target representation in MGMT layer. 
typedef struct _MGMT_TARGET_WS {
    PACIT_VOID      pacit_ws;
    PACIT_VOID      dummy;
}MGMT_TARGET_WS, *PMGMT_TARGET_WS;

/* MGMT IOCTL flags */

#define MGMT_IOCTLH_FREE                (1<<0)

/* These get set while issuing the IOCTL */
#define MGMT_IOCTLH_POLLED_MODE         (1<<1)
#define MGMT_IOCTLH_INTERNAL            (1<<2)
#define MGMT_IOCTLH_RESPONSE_BUFFER     (1<<3)

/* These are the resource management flags */
#define MGMT_IOCTLH_INTERNAL_RESOURCE   (1<<4)
#define MGMT_IOCTLH_EXPLICIT_BUFFER     (1<<5)
#define MGMT_IOCTLH_SUBMITTED           (1<<6)
#define MGMT_IOCTLH_COMPLETED           (1<<7)

#define MGMT_IOCTL_BUFFER_FLAG_FREE                (1<<0)

typedef struct _MGMT_INTERNAL_IOCTL_BUFFER_HANDLE {
    ACIT_DC_LIST            controller_link;
    ACIT_DC_LIST            target_link;
    ACIT_UEXACT32           flags;
    IOCTL_HEADER            ioctl_header;
    union {

        OFFLOAD_SESSION_PARAMS offload_session;
        ACIT_UEXACT8 get_all_session_ids[GET_ALL_SESSION_ID_PARAMS_MAXSIZE];
        ACIT_UEXACT8 session_ids[GET_ALL_SESSION_ID_PARAMS_MAXSIZE];
        IOCTL_ISCSI_INI_DRIVER_INVALIDATE_CONNECTION conn_invalidate;
        IOCTL_COMMON_TCP_UPLOAD conn_upload;
#if 0
        IOCTL_ISCSI_REOPEN_A_CONN_PARAMS conn_reopen;
#endif
        IOCTL_ISCSI_REOPEN_A_SESSION_PARAMS session_reopen;
        INVALIDATE_COMMANDS_PARAMS invalidate_cmds;
        SCSI_READ_WRITE_PARAMS_IN scsi_read_write_cmd;
        IOCTL_COMMON_GET_CNTL_ATTRIBUTES controller_attribs;
        OFFLOAD_SESSION_PARAMS_V1 offload_session_v1;

    }data_buffer;
}MGMT_INTERNAL_IOCTL_BUFFER_HANDLE, *PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE;

typedef struct _MGMT_IOCTL_HANDLE {
    ACIT_DC_LIST         controller_link;
    PACIT_CONTROLLER_WS  pcontroller;
    PACIT_TARGET_WS      ptarget;
    PACIT_IOCTL_REQUEST  pioctl_request;
    PMGMT_INTERNAL_IOCTL_BUFFER_HANDLE pmgmt_buf_handle;
    PACIT_VOID           pmgmt_buffer;
    PACIT_VOID           pmgmt_buffer_h;
    PMCC_WRB             pmcc_wrb;
    PMGMT_POST           pmgmt_post_routine;
    ACIT_UEXACT32        index;
    ACIT_UEXACT32        ticks;
    ACIT_UEXACT32        timeout;
    ACIT_UEXACT32        flags;
    ACIT_UEXACT32        payload_length;

}MGMT_IOCTL_HANDLE, *PMGMT_IOCTL_HANDLE;

#define GET_MGMT_CONTROLLER_WS(pcontroller)    (pcontroller->pmgmt_ws)
#define GET_MGMT_TARGET_WS(ptarget)            (ptarget->pmgmt_ws)

#define IS_BOOT_SESSION_OPEN(pboot)                                 \
    (pboot->boot_session_handle != 0xffff &&                        \
       pboot->boot_session_count)

ACIT_UEXACT32
mgmt_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size);

ACIT_UEXACT32
mgmt_init_mem_descriptors(
    PACIT_CONTROLLER_WS         pcontroller);

ACIT_UEXACT32
mgmt_init_controller_ws(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
mgmt_init_const_mem_desc(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
mgmt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
mgmt_init_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT16           connection_handle,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           cq_id,
    ACIT_UEXACT16           def_hdr_id,
    ACIT_UEXACT16           def_data_id,
    PPHYS_ADDR              ptemplate_address,
    ACIT_UEXACT32           sync);

ACIT_UEXACT32
mgmt_init_lun_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun);

ACIT_UEXACT32
mgmt_init_io_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_WS             pio);

ACIT_UEXACT32
mgmt_init_tmf_ws(
    PACIT_CONTROLLER_WS     pcontroller,    
    PACIT_TMF_WS            ptmf);

ACIT_UEXACT32
mgmt_stop_controller(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           stop_type);

MGMT_STATUS
mgmt_get_link_status(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled);

ACIT_UEXACT32
mgmt_discover_targets(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled);

ACIT_UEXACT32
mgmt_discover_boot_target(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_process_ioctl_tick(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_issue_function_reset(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_check_supported_fw(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_get_fw_config(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_set_fragnum_bits(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           num_bits);

MGMT_STATUS
mgmt_epfw_cleanup(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT16           chute);

MGMT_STATUS
mgmt_modify_eq_delay(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           pid,
    ACIT_UEXACT32           eq_id,
    ACIT_UEXACT32           new_delay);

MGMT_STATUS
mgmt_modify_eq_delay_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_get_addi_attribs(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_process_ioctl_login(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_ioctl_login_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_process_ioctl_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_ioctl_logout_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_process_ioctl_modify_sessionloginkeys(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_ioctl_modify_sessionloginkeys_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_process_ioctl_modify_connloginkeys(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_ioctl_modify_connloginkeys_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);
/*
// No MCS support
MGMT_STATUS
mgmt_process_ioctl_addconn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_ioctl_addconn_post(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_process_ioctl_closeconn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);
*/
ACIT_UEXACT8
mgmt_process_set_tcpconfig(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_process_set_nw_params_post(
    PACIT_CONTROLLER_WS     pcontroller, 
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_process_get_nic_statistics(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_nop_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    PMGMT_POST              post_routine);

ACIT_UEXACT8
mgmt_process_external_ioctl_default(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_start_external_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    PACIT_VOID              ptarget,
    PMGMT_POST              post_routine);

ACIT_UEXACT32
mgmt_start_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_start_hiber_io_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_start_hiber_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_REQUEST        pio_request);

MGMT_STATUS
mgmt_get_all_session_id(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled);

MGMT_STATUS
mgmt_get_all_session_id_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_offload_session(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    POFFLOAD_SESSION_IN     poffload_session_in,
    ACIT_UEXACT32           is_polled);

MGMT_STATUS
mgmt_offload_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  pioctl_hdr,
    PACIT_VOID              pdata_buffer);

// No MCS support
#if 0
MGMT_STATUS
mgmt_offload_cxn(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           cq_id,
    ACIT_UEXACT16           def_hdr_id,
    ACIT_UEXACT16           def_data_id,
    PPHYS_ADDR              ptemplate_address,
    ACIT_UEXACT32           is_polled);

MGMT_STATUS
mgmt_offload_cxn_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  pioctl_hdr,
    PACIT_VOID              pdata_buffer);

#endif

MGMT_STATUS
mgmt_get_boot_session(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           discover);

MGMT_STATUS
mgmt_discover_boot_target_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_get_boot_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_complete_external_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

ACIT_UEXACT32
mgmt_return_ioctl(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    MGMT_STATUS             status);

MGMT_STATUS
mgmt_delete_nonpersistent_targets(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_login_all_targets(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           is_polled);

MGMT_STATUS
mgmt_invalidate_connection(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT16           issue_reset,
    ACIT_UEXACT16           savecfg_flag);

MGMT_STATUS
mgmt_invalidate_connection_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_upload_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT16           cid,
    ACIT_UEXACT32           upload_flag);

MGMT_STATUS
mgmt_upload_connection_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_logout_fw_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

MGMT_STATUS
mgmt_peek_icd_inv_resources(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_invalidate_icds(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_INVALIDATE_HANDLE pinv_handle);

MGMT_STATUS
mgmt_invalidate_icds_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_target_upload_complete
(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           session_handle,
    PACIT_IOCTL_REQUEST     pioctl_request);

MGMT_STATUS
mgmt_all_targets_close_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request);

#if 0
MGMT_STATUS
mgmt_reopen_cxn(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT32           cmd_sn);

MGMT_STATUS
mgmt_reopen_cxn_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);
#endif

MGMT_STATUS
mgmt_reopen_sessions(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           reopen_type,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT8            domain,
    ACIT_UEXACT32           is_polled);

MGMT_STATUS
mgmt_reopen_a_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_reopen_all_session_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

MGMT_STATUS
mgmt_save_nonpersistent_targets(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
mgmt_return_pending_ioctls
(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget);

ACIT_VOID
mgmt_return_ioctl_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    ACIT_UEXACT32           status);

ACIT_VOID
mgmt_fill_ioctl_hdr_version(
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            version);

ACIT_VOID
mgmt_fill_ioctl_hdr(
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            subsystem,
    ACIT_UEXACT8            page_code,
    ACIT_UEXACT8            domain,
    ACIT_UEXACT32           transfer_length,
    ACIT_UEXACT32           timeout);

ACIT_VOID
mgmt_fill_drvr_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    ACIT_UEXACT8            page_code,
    PACIT_VOID              ppayload,
    ACIT_UEXACT32           payload_len);

PMGMT_IOCTL_HANDLE
mgmt_alloc_ioctl_handle(
   PMGMT_CONTROLLER_WS      pmgmt_controller);

ACIT_VOID
mgmt_free_ioctl_handle(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle);

MGMT_STATUS
mgmt_alloc_ioctl_buffer(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle);

ACIT_VOID
mgmt_free_ioctl_buffer(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   PMGMT_IOCTL_HANDLE       pioctl_handle);

#define MGMT_IOCTL_POLLED               (1<<0)
#define MGMT_IOCTL_NON_POLLED           (1<<1)
#define MGMT_IOCTL_INTERNAL             (1<<2)
#define MGMT_IOCTL_EXTERNAL             (1<<3)
#define MGMT_IOCTL_RESPONSE_BUFFER      (1<<4)
#define MGMT_IOCTL_NO_RESPONSE_BUFFER   (1<<5)

#define MGMT_SET_ALLOCATION_FLAGS(poll, internal, resp_b) (poll | internal | resp_b)

PMGMT_IOCTL_HANDLE
mgmt_alloc_ioctl_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IOCTL_REQUEST     pioctl_request,
    ACIT_UEXACT32           payload_length,
    ACIT_UEXACT32           flags);

ACIT_VOID
mgmt_free_ioctl_resources(
   PACIT_CONTROLLER_WS      pcontroller,
   PMGMT_IOCTL_HANDLE       pioctl_handle);

ACIT_UEXACT32
mgmt_flag_busy_controller(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   ACIT_UEXACT32            reason);

ACIT_UEXACT32
mgmt_flag_ready_controller(
   PMGMT_CONTROLLER_WS      pmgmt_controller,
   ACIT_UEXACT32            reason);

ACIT_VOID
mgmt_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller);

MGMT_STATUS
mgmt_get_link_status_callback(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle,
    PIOCTL_RESPONSE_HEADER  presponse_hdr,
    PACIT_VOID              pdata_buffer);

