/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_cev_h__
#define __be2_gen_cev_h__

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_CQ_Context
//
//   Access CQ context. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// cq_id              - CQ ring ID. 
// p_ctx              - Pointer to CQ context structure. 
// write_flag         - TRUE to write context, FALSE to read context. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_CQ_Context (
   IN void*       function_context,
   IN u32         cq_id,
   IN CQ_CONTEXT* p_ctx,
   IN boolean     write_flag )
{
    u32 i;

    // Write the CEV RAM base address
    BE_GEN_CSR_WRITE( function_context, cev.ctx_address, BE_GEN_FIELD_OFFSET(CEV_MEMORY_MAP, cqt[cq_id]) );

    // Strobe the data. The address register auto-increments.
    for (i = 0; i < sizeof(CQ_CONTEXT)/4; i++)
    {
        if (write_flag)
        {
            BE_GEN_CSR_WRITE( function_context, cev.ctx_data, p_ctx->dw[i] );
        }
        else
        {
            p_ctx->dw[i] = BE_GEN_CSR_READ( function_context, cev.ctx_data );
        }
    }
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_EQ_Context
//
//   Access EQ context. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// eq_id              - EQ ring ID. 
// p_ctx              - Pointer to EQ context structure. 
// write_flag         - TRUE to write context, FALSE to read context. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_EQ_Context (
   IN void*       function_context,
   IN u32         eq_id,
   IN EQ_CONTEXT* p_ctx,
   IN boolean     write_flag )
{
    u32 i;

    // Write the CEV RAM base address
    BE_GEN_CSR_WRITE( function_context, cev.ctx_address, BE_GEN_FIELD_OFFSET(CEV_MEMORY_MAP, eqt[eq_id]) );

    // Strobe the data. The address register auto-increments.
    for (i = 0; i < sizeof(EQ_CONTEXT)/4; i++)
    {
        if (write_flag)
        {
            BE_GEN_CSR_WRITE( function_context, cev.ctx_data, p_ctx->dw[i] );
        }
        else
        {
            p_ctx->dw[i] = BE_GEN_CSR_READ( function_context, cev.ctx_data );
        }
    }
    return;
}


//---------------------------------------------------------------------------------------
// Function: BE2_Gen_CEV_Phase_Reset
//
//   Causes CEV to reset the current phase for all EQs. This should be called after
//   reprogramming the EQ delay multiplier or phase value. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_CEV_Phase_Reset (
   IN void* function_context )
{
    CEV_MAX_SUST_INT_DLY_CNT_CSR delay;

    delay.dw = BE_GEN_CSR_READ( function_context, cev.cev_max_sust_int_dly_cnt_csr);

    BE_GEN_ASSERT (delay.MAX_CSR_STROLL == 95); // The EQ delay to multiplier function assumes this.
    
    delay.PHASE_RESET = 1;

    BE_GEN_CSR_WRITE (function_context, cev.cev_max_sust_int_dly_cnt_csr, delay.dw);
}

#endif  // __be2_gen_cev_h__

