/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_init_h__
#define __be2_gen_init_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"
#include "be2_gen_misc.h"

#include "be2_config.h"


#define EP_DESC_BANK_BASE0 (0x1100)
#define EP_CTXT_BANK_BASE0 (0x1500)
#define EP_DESC_BANK_BASE1 (0x1900)
#define EP_CTXT_BANK_BASE1 (0x1d00)

#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
#define ISCSI_RX_EP_DESC_BANK_BASE0 (0x1080)
#define ISCSI_RX_EP_DESC_BANK_BASE1 (0x1880)
#endif

#define BE2_GEN_LOAD_FIRMWARE_IROM(FunctionCtx, ArrayName, EpName) \
    BE_GEN_CSR_WRITE (FunctionCtx, EpName .ep_minirom.irom_csr0, BE_GEN_SWAP32(ArrayName[0]));

#define BE2_GEN_LOAD_FIRMWARE_INST(FunctionCtx, ArrayName, StartOffs, EpName) \
    BE2_Gen_Load_EP_Firmware_Base(FunctionCtx, (u32 *)ArrayName, \
        ((sizeof (ArrayName)) / (sizeof (u32))), \
        StartOffs, \
        BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, EpName .ep.ep_instr_ram_addr), \
        BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, EpName .ep.ep_instr_ram_rwdata));

#define BE2_GEN_LOAD_FIRMWARE_DATA(FunctionCtx, ArrayName, StartOffs, EpName) \
    BE2_Gen_Load_EP_Firmware_Base(FunctionCtx, (u32 *)ArrayName, \
        ((sizeof (ArrayName)) / (sizeof (u32))), \
        StartOffs, \
        BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, EpName .ep.ep_data_ram_addr), \
        BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, EpName .ep.ep_data_ram_rwdata));

#define BE_ContextCount(field) (BE_GEN_NUMBER_OF_FIELD(CTX_MEMORY_MAP,field))

#define BE_Log2(n)                      \
  ((u32)(n) == (u32)(1 <<  0) ?  0 :    \
  ((u32)(n) == (u32)(1 <<  1) ?  1 :    \
  ((u32)(n) == (u32)(1 <<  2) ?  2 :    \
  ((u32)(n) == (u32)(1 <<  3) ?  3 :    \
  ((u32)(n) == (u32)(1 <<  4) ?  4 :    \
  ((u32)(n) == (u32)(1 <<  5) ?  5 :    \
  ((u32)(n) == (u32)(1 <<  6) ?  6 :    \
  ((u32)(n) == (u32)(1 <<  7) ?  7 :    \
  ((u32)(n) == (u32)(1 <<  8) ?  8 :    \
  ((u32)(n) == (u32)(1 <<  9) ?  9 :    \
  ((u32)(n) == (u32)(1 << 10) ? 10 :    \
  ((u32)(n) == (u32)(1 << 11) ? 11 :    \
  ((u32)(n) == (u32)(1 << 12) ? 12 :    \
  ((u32)(n) == (u32)(1 << 13) ? 13 :    \
  ((u32)(n) == (u32)(1 << 14) ? 14 :    \
  ((u32)(n) == (u32)(1 << 15) ? 15 :    \
  ((u32)(n) == (u32)(1 << 16) ? 16 :    \
  ((u32)(n) == (u32)(1 << 17) ? 17 :    \
  ((u32)(n) == (u32)(1 << 18) ? 18 :    \
  ((u32)(n) == (u32)(1 << 19) ? 19 :    \
  ((u32)(n) == (u32)(1 << 20) ? 20 :    \
  ((u32)(n) == (u32)(1 << 21) ? 21 :    \
  ((u32)(n) == (u32)(1 << 22) ? 22 :    \
  ((u32)(n) == (u32)(1 << 23) ? 23 :    \
  ((u32)(n) == (u32)(1 << 24) ? 24 :    \
  ((u32)(n) == (u32)(1 << 25) ? 25 :    \
  ((u32)(n) == (u32)(1 << 26) ? 26 :    \
  ((u32)(n) == (u32)(1 << 27) ? 27 :    \
  ((u32)(n) == (u32)(1 << 28) ? 28 :    \
  ((u32)(n) == (u32)(1 << 29) ? 29 :    \
  ((u32)(n) == (u32)(1 << 30) ? 30 :    \
  ((u32)(n) == (u32)(1 << 31) ? 31 :    \
  (0/0)))))))))))))))))))))))))))))))))

/* QoS defines */
// Set CBS=64KB, CIR=1000Mbps
// BUG 19283 - Clamp mimimium to 1 Gbps as we reduce the time slice to 520 us
#define BE2_QoS_TIMER_VALUE_64K_100Mbps (4096)
// Set CBS=18KB, CIR=100Mbps
#define BE2_Qos_TIMER_VALUE_18K_100Mbps (11250)

/* Excess QoS defines */
#define DISABLE_EXCESS_QoS_TXULP1 (0x280000)
#if (BE_CONFIG==7)
#define ENABLE_EXCESS_QoS_TXULP1  (0x4E20)
#else
#define ENABLE_EXCESS_QoS_TXULP1  (0x7D0)
#endif
#define DISABLE_EXCESS_QoS_TXULP0 (0x280000)
#define ENABLE_EXCESS_QoS_TXULP0  (0x7D0)


//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Load_TPRE_Classification
//
//   Loads BladeEngine TPRE classification firmware.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Load_TPRE_Classification (
   IN void* function_context )
{
    TPRE_COPROCESSOR_CR tpre_coprocessor;
    TPRE_TPRE_COPROCESSOR2_CR tpre_coprocessor2;
    TPRE_CLASS_CONTROL_CR tpre_class_ctrl;
    TPRE_CLASS_CONST_RWDATA_CSR tpre_class_const_data;
    TPRE_CLASS_JUMP_MAP_1_CR jump_map_1;
    TPRE_CLASS_JUMP_MAP_2_CR jump_map_2;
#if (BE_CONFIG==0)
    TPRE_CLASS_PGA_MAP_CR tpre_pga_map_cr;
#endif

    u32 dword_length;
    u32 *data_array;
    u32 index;
    u32 is_a0 = (0 == BE2_Gen_Chip_Rev(function_context)) ? 1 : 0;

    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TPRE_CLASS " in TPRE classification IRAM\n");

    /* Enable write to the classification engine RAM */
    tpre_coprocessor.dw = BE_GEN_CSR_READ( function_context, tpre.coprocessor_control );
    tpre_coprocessor.class_iram_wr = 1;
    tpre_coprocessor.class_iaddr = 0;
    tpre_coprocessor.class_const_addr = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.coprocessor_control, tpre_coprocessor.dw );

    /* Write the classification microcode */
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32 *)tpre_class_TIE;
        dword_length = sizeof(tpre_class_TIE)/sizeof(u32);
    } else {
        data_array = (u32 *)tpre_class_A1TIE;
        dword_length = sizeof(tpre_class_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32 *)tpre_class_PERF;
        dword_length = sizeof(tpre_class_PERF)/sizeof(u32);
    } else {
        data_array = (u32 *)tpre_class_A1PERF;
        dword_length = sizeof(tpre_class_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datal, data_array[index] );
        index++;
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datau, data_array[index] );
        index++;
    }

#if (BE_CONFIG==0)
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TPRE_LRO_CLASS " in TPRE classification IRAM\n");
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32*)tpre_lroclass_TIE;
        dword_length = sizeof(tpre_lroclass_TIE)/sizeof(u32);
    } else {
        data_array = (u32*)tpre_lroclass_A1TIE;
        dword_length = sizeof(tpre_lroclass_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32*)tpre_lroclass_PERF;
        dword_length = sizeof(tpre_lroclass_PERF)/sizeof(u32);
    } else {
        data_array = (u32*)tpre_lroclass_A1PERF;
        dword_length = sizeof(tpre_lroclass_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datal, data_array[index] );
        index++;
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datau, data_array[index] );
        index++;
    }
#elif (BE_CONFIG==7)
#if ENABLE_PRIORITY_PAUSE
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32 *)tpre_fcoeclass_TIE;
        dword_length = sizeof(tpre_fcoeclass_TIE)/sizeof(u32);
    }
    else {
        data_array = (u32 *)tpre_fcoeclass_A1TIE;
        dword_length = sizeof(tpre_fcoeclass_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32 *)tpre_fcoeclass_PERF;
        dword_length = sizeof(tpre_fcoeclass_PERF)/sizeof(u32);
    }
    else {
        data_array = (u32 *)tpre_fcoeclass_A1PERF;
        dword_length = sizeof(tpre_fcoeclass_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datal, data_array[index] );
        index++;
        BE_GEN_CSR_WRITE( function_context, tpre.class_iram_datau, data_array[index] );
        index++;
    }
#endif /* ENABLE_PRIORITY_PAUSE */
#endif

    /* Enable write to the classifiaction engine constants */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TPRE_CLASS_IM " in TPRE classification constants\n");
    tpre_coprocessor.class_iram_wr = 0;
    tpre_coprocessor.class_iaddr = 0;
    tpre_coprocessor.class_const_wr = 1;

    /* Write the classification constants array */
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32 *)tpre_class_immeds_TIE;
        dword_length = sizeof(tpre_class_immeds_TIE)/sizeof(u32);
    } else {
        data_array = (u32 *)tpre_class_immeds_A1TIE;
        dword_length = sizeof(tpre_class_immeds_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32 *)tpre_class_immeds_PERF;
        dword_length = sizeof(tpre_class_immeds_PERF)/sizeof(u32);
    } else {
        data_array = (u32 *)tpre_class_immeds_A1PERF;
        dword_length = sizeof(tpre_class_immeds_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        tpre_coprocessor.class_const_addr = index;
        BE_GEN_CSR_WRITE( function_context, tpre.coprocessor_control, tpre_coprocessor.dw );
        tpre_class_const_data.class_const_data = data_array[index] ;
        BE_GEN_CSR_WRITE( function_context, tpre.class_const_data, tpre_class_const_data.dw );
        index ++;
    }
    /* Done with classification engine constants */
    tpre_coprocessor.class_const_wr = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.coprocessor_control, tpre_coprocessor.dw );

    /* Write the start addresses */
    tpre_class_ctrl.dw = 0;
    tpre_class_ctrl.dsc_start_addr = 0x8;  // Was 0x5
    tpre_class_ctrl.dsc_end_addr = 0x13;   // Was 0x10
    /* start microcode execution from the start of CTX */
    tpre_class_ctrl.ctx_start_addr = 0x0;
    BE_GEN_CSR_WRITE( function_context, tpre.class_control, tpre_class_ctrl.dw );

    /* program the end address for the packet header write in prefetch buffer */
    tpre_coprocessor2.dw = 0x0;
    tpre_coprocessor2.halign_end_addr = 0x12;
    tpre_coprocessor2.c = 0x1;
    tpre_coprocessor2.d = 0x1;
    tpre_coprocessor2.r = 0x1;
    tpre_coprocessor2.g = 0x0; /* We set this to zero to avoid some race conditions */
    /* According to Randy we can have upto 62 ack_credits which is 31 imm_ack
     * desc
     */
    tpre_coprocessor2.ack_credits = 0x3e;
    BE_GEN_CSR_WRITE( function_context, tpre.coprocessor_control2, tpre_coprocessor2.dw );

    jump_map_1.p0_vector = 0;
    jump_map_1.p1_vector = 0;
    jump_map_1.p2_vector = 0;
    jump_map_1.p3_vector = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.class_jump_1, jump_map_1.dw );

    jump_map_2.p4_vector = 0;
#if ( (BE_CONFIG==0) || (BE_CONFIG==9) )
#if !defined(BE_XT_PERF)
    if (is_a0) {
        jump_map_2.p5_vector = sizeof(tpre_class_TIE)/(2*sizeof(u32));
    } else {
        jump_map_2.p5_vector = sizeof(tpre_class_A1TIE)/(2*sizeof(u32));
    }
#else
    if (is_a0) {
        jump_map_2.p5_vector = sizeof(tpre_class_PERF)/(2*sizeof(u32));
    } else {
        jump_map_2.p5_vector = sizeof(tpre_class_A1PERF)/(2*sizeof(u32));
    }
#endif
#elif (BE_CONFIG==7)
    jump_map_2.p5_vector = 0;
#if ENABLE_PRIORITY_PAUSE
#if !defined(BE_XT_PERF)
    if (is_a0) {
        jump_map_2.p6_vector = sizeof(tpre_class_TIE)/(2*sizeof(u32));
    } else {
        jump_map_2.p6_vector = sizeof(tpre_class_A1TIE)/(2*sizeof(u32));
    }
#else
    if (is_a0) {
        jump_map_2.p6_vector = sizeof(tpre_class_PERF)/(2*sizeof(u32));
    } else {
        jump_map_2.p6_vector = sizeof(tpre_class_A1PERF)/(2*sizeof(u32));
    }
#endif
#else /* ENABLE_PRIORITY_PAUSE */
    jump_map_2.p6_vector = 0;
#endif /* ENABLE_PRIORITY_PAUSE */
#else
    jump_map_2.p5_vector = 0;
    jump_map_2.p6_vector = 0;
#endif
    jump_map_2.p7_vector = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.class_jump_2, jump_map_2.dw );

    /* Enable the classification engine */
    tpre_coprocessor.class_en = 1;
    tpre_coprocessor.class_iram_wr = 0;
    tpre_coprocessor.class_iaddr = 0;
    tpre_coprocessor.class_const_wr = 0;
    tpre_coprocessor.class_const_addr = 0;
    tpre_coprocessor.class_queues =
        ((1 << TPRE_IQ_RCV_ULP_0) |
#if (BE_CONFIG==0)
         (1 << TPRE_IQ_LRO) |
#elif (BE_CONFIG==7)
#if ENABLE_PRIORITY_PAUSE
         (1 << TPRE_IQ_FCoE) |
#endif /* ENABLE_PRIORITY_PAUSE */
#endif
         (1 << TPRE_IQ_RCV_ULP_1) );

    tpre_coprocessor.halign_queues =
        ((1 << TPRE_IQ_RCV_ULP_0) |
#if (BE_CONFIG==0)
     (1 << TPRE_IQ_LRO) |
#endif
         (1 << TPRE_IQ_RCV_ULP_1));

    BE_GEN_CSR_WRITE( function_context, tpre.coprocessor_control, tpre_coprocessor.dw );

#if (BE_CONFIG==0)
    tpre_pga_map_cr.dw = 0;
    tpre_pga_map_cr.pga0_config = 11;  // LRO remaps QCR bit 11 as is_singleton predicate
    tpre_pga_map_cr.pga1_config = 12;  // LRO remaps QCR bit 12 as lro_context_match predicate
    tpre_pga_map_cr.pga2_config = 13;  // LRO remaps QCR bit 13 as is_in_sequence predicate
    BE_GEN_CSR_WRITE( function_context, tpre.class_pga_mapping, tpre_pga_map_cr.dw );
#endif

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Load_TXULP_Classification
//
//   Loads BladeEngine TPRE classification firmware.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// txulp_number       - Indicates which TXULP to initialize (0, 1, or 2).
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Load_TXULP_Classification (
   IN void* function_context,
   IN u8    txulp_number )
{
    TXULP_COPROCESSOR_CR coprocessor_ctrl;
    TXULP_CLASS_CONST_RWDATA_CSR txulp_class_const_data;
#if (BE_CONFIG == 7 && ENABLE_PRIORITY_PAUSE)
    TXULP_QOS_CBS_PBS_FLOW_NUM_CSR qos_cbs_pbs_flow;
    TXULP_QOS_CBS_CSR qos_cbs;
    TXULP_QOS_PBS_CSR qos_pbs;
    //ashwinm - temporarily commenting unused variables
    //TXULP_QOS_EBP_FLOW_NUM_CSR qos_ebp_flow;
    //TXULP_QOS_EBP_CSR qos_ebp;
    //TXULP_QOS_WEIGHTS_CSR qos_weights;
#endif
    TXULP_QOS_TIMER_INIT_CSR qos_timer_init;
    GLOB_QOS_CBS_PBS_FLOW_NUM_CSR global_qos_cbs_pbs_flow;
    GLOB_QOS_CBS_CSR global_qos_cbs;
    GLOB_QOS_PBS_CSR global_qos_pbs;
    //GLOB_QOS_EBP_FLOW_NUM_CSR global_qos_ebp_flow;
    //GLOB_QOS_EBP_CSR global_qos_ebp;
    GLOB_QOS_TIMER_INIT_CSR global_qos_timer_init;
    //GLOB_QOS_WEIGHTS_0_CSR global_qos_weights_0;
    //GLOB_QOS_WEIGHTS_1_CSR global_qos_weights_1;
    u32 bucket;
    u32 is_asic;

    u32 dword_length;
    u32 *data_array;
    u32 index;
    u32 is_a0 = (0 == BE2_Gen_Chip_Rev(function_context)) ? 1 : 0;

    if (txulp_number)
    {
        BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TXULP_CLASS " in TXULP1 classification IRAM\n");
    }
    else
    {
        BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TXULP_CLASS " in TXULP0 classification IRAM\n");
    }

    /* Enable write to the classification engine RAM */
    coprocessor_ctrl.dw = 0;
    coprocessor_ctrl.class_iram_wr = 1;
    coprocessor_ctrl.class_iaddr = 0;
    coprocessor_ctrl.class_const_addr = 0;
    BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, coprocessor_control, coprocessor_ctrl.dw );

    /* Write the classification microcode */
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32 *)txulp_class_TIE;
        dword_length = sizeof(txulp_class_TIE)/sizeof(u32);
    } else {
        data_array = (u32 *)txulp_class_A1TIE;
        dword_length = sizeof(txulp_class_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32 *)txulp_class_PERF;
        dword_length = sizeof(txulp_class_PERF)/sizeof(u32);
    } else {
        data_array = (u32 *)txulp_class_A1PERF;
        dword_length = sizeof(txulp_class_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, class_iram_datal, data_array[index] );
        index++;
        BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, class_iram_datau, data_array[index] );
        index++;
    }

    /* Enable write to the classifiaction engine constants */
    if (txulp_number)
    {
        BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TXULP_CLASS_IM " in TXULP1 classification constants\n");
    }
    else
    {
        BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TXULP_CLASS_IM " in TXULP0 classification constants\n");
    }
    coprocessor_ctrl.class_iram_wr = 0;
    coprocessor_ctrl.class_iaddr = 0;
    coprocessor_ctrl.class_const_wr = 1;

    /* Write the classification constants array */
#if !defined(BE_XT_PERF)
    if (is_a0) {
        data_array = (u32 *)txulp_class_immeds_TIE;
        dword_length = sizeof(txulp_class_immeds_TIE)/sizeof(u32);
    } else {
        data_array = (u32 *)txulp_class_immeds_A1TIE;
        dword_length = sizeof(txulp_class_immeds_A1TIE)/sizeof(u32);
    }
#else
    if (is_a0) {
        data_array = (u32 *)txulp_class_immeds_PERF;
        dword_length = sizeof(txulp_class_immeds_PERF)/sizeof(u32);
    } else {
        data_array = (u32 *)txulp_class_immeds_A1PERF;
        dword_length = sizeof(txulp_class_immeds_A1PERF)/sizeof(u32);
    }
#endif
    index = 0;
    while ( index < dword_length )
    {
        coprocessor_ctrl.class_const_addr = index;
        BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, coprocessor_control, coprocessor_ctrl.dw );
        txulp_class_const_data.class_const_data = data_array[index] ;
        BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, class_const_data, txulp_class_const_data.dw );
        index ++;
    }
    /* Done with classification engine constants */
    coprocessor_ctrl.class_const_wr = 0;
    BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, coprocessor_control, coprocessor_ctrl.dw );

    is_asic = BE2_Gen_Is_ASIC(function_context);

#if (BE_CONFIG == 7 && ENABLE_PRIORITY_PAUSE)
    // Local QOS for config 7
    // We need to set the QOS timer interval to 48us
    // Care should be taken by the driver to disable
    // LSO in case the CBS is smaller than the lso size
    qos_timer_init.qos_timer_value = 375; // 375 * 32 buckets = 12000 * 4 ns = 48us
#else
    /* Setup the local QoS timer values.  We need Committed Burst Size (CBS) of 64KB, Committed
       Information Rate (CIR) of 100Mbps for each flow id.
       Time to transmit 64KB @ 100Mbps = 5.24288 ms
       t(clock) = 40960, since t(clock) * 32 (buckets) *  4nsec = 5.24288 ms
    */
    // the minimum supported rate.  CBS must be >= mtu.
    qos_timer_init.qos_timer_value =
            BE2_QoS_TIMER_VALUE_64K_100Mbps;
#endif

    if (!is_asic) {
        qos_timer_init.qos_timer_value /= 100; // FPGA is clocked 100 times slower
    }
    BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, qos_tim_dat, qos_timer_init.dw );

    /* QOS */
#if (BE_CONFIG == 7 && ENABLE_PRIORITY_PAUSE)
    if (txulp_number == 1) {
        for (bucket = 0; bucket < 7; bucket++) {
            qos_cbs_pbs_flow.cbs_pbs_flow_num = bucket;
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, cbr_pbr_num, qos_cbs_pbs_flow.dw );

            qos_cbs.cbs_data = 6000; // 6000 = 1 Gbps
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, cbr_dat, qos_cbs.dw );

            qos_cbs_pbs_flow.cbs_pbs_flow_num = bucket;
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, cbr_pbr_num, qos_cbs_pbs_flow.dw );

            qos_pbs.pbs_data = 1;
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, pbr_dat, qos_pbs.dw );

        }
    }
#endif /* (BE_CONFIG == 7 && ENABLE_PRIORITY_PAUSE) */

    /* ETX */
    /* By default we will assign 100 Mbps for each port for ETX traffic i.e. for buckets 32, 33 */
    /* Global parameters are only programmed into TXULP 0. */
    /* Global flow labels for ethernet are hardcoded to 32 (port 0) and 33 (port 1) regardless
       of the setting in ethernet context. */
    if (txulp_number == 0) {

        for (bucket = 32; bucket < 34; bucket++) {
            global_qos_cbs_pbs_flow.glob_cbs_pbs_flow_num = bucket;
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, glob_cbr_pbr_num, global_qos_cbs_pbs_flow.dw );

            if (is_asic) {
                /* 0xA21FE80 bytes needs to be transmitted in 136ms to get a bandwidth of 10 Gbps */
                global_qos_cbs.glob_cbs_data = 0xA21FE80;
            } else {
                /* 0x19F0A0 bytes needs to be transmitted in 136ms to get a bandwidth of 100 Mbps */
                global_qos_cbs.glob_cbs_data = 0x19F0A0;
            }
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, glob_cbr_dat, global_qos_cbs.dw );


            global_qos_cbs_pbs_flow.glob_cbs_pbs_flow_num = bucket;
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, glob_cbr_pbr_num, global_qos_cbs_pbs_flow.dw );

            global_qos_pbs.glob_pbs_data = 1; // Value of 1 disables excess
            BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, glob_pbr_dat, global_qos_pbs.dw );
        }


        if (is_asic) {
            global_qos_timer_init.glob_qos_timer_value = 1000000; // 1000000 * 34 buckets = 34000000 * 4 ns = 136ms
        } else {
            global_qos_timer_init.glob_qos_timer_value = 10000; // 10000 * 34 buckets = 340000 * 400 ns = 136ms
        }
        BE_GEN_TXULP_CSR_WRITE(txulp_number, function_context, glob_qos_tim_dat, global_qos_timer_init.dw);

    }

    /* End QOS */

    /* Enable the classification engine */
    coprocessor_ctrl.dw = 0;
#if (BE_CONFIG == 6)
    coprocessor_ctrl.class_en = 0;
#elif (BE_CONFIG == 7)
    // Need to enable classification on TXULP1
    // in order to use QOS
    if (txulp_number == 1) {
        coprocessor_ctrl.class_en = 1;
    } else {
        coprocessor_ctrl.class_en = 1;
    }
#else
    coprocessor_ctrl.class_en = 1;
#endif

    BE_GEN_TXULP_CSR_WRITE( txulp_number, function_context, coprocessor_control, coprocessor_ctrl.dw );
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE_Gen_Set_QoS
//
//   Set QOS parameters for given flow ID.
//
//  flow_id  - Flow ID to apply the QoS to
//  txulp_id - TxULP ID to which the flow_id belongs to
//  cir      - Committed Information Rate in units of 10 Mbps
//
// return -- none
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Set_QoS(u8 flow_id, u8 txulp_id, u32 cir)
{

    TXULP_QOS_CBS_PBS_FLOW_NUM_CSR qos_cbs_pbs_flow;
    TXULP_QOS_CBS_CSR qos_cbs;
    TXULP_QOS_PBS_CSR pbs_csr;

    qos_cbs_pbs_flow.cbs_pbs_flow_num = flow_id;
#if (BE_CONFIG == 7 && ENABLE_PRIORITY_PAUSE)
    // 10 Mbps = 60
    qos_cbs.cbs_data = 60 * cir;
    if (qos_cbs.cbs_data < 1536) {
        // Need to set to atleast mtu to prevent lockups
        qos_cbs.cbs_data = 1536;
    }
#else
    /* Compute CBS data value in BYTES:
       cir is in 10 Mbps units, i.e. 250 -> 2.5Gbps
       (timer count) * 32 (number of flow ids) * 4 (in nanosec) * cir (in 10Mbps)/ (8 (bits))
       = (timer count) * 16 * cir / 100
    */
    // Clamp min rate to 1Gbps
    // BUG 19283 - Clamp mimimium to 1 Gbps as we reduce the time slice to 520 us
    if (cir < 100) {
        cir = 100;
    }
    qos_cbs.cbs_data = (BE2_QoS_TIMER_VALUE_64K_100Mbps * 16 * cir) / 100;
#endif

    // Write CBS data
    BE_GEN_TXULP_CSR_WRITE( txulp_id, NULL, cbr_pbr_num, qos_cbs_pbs_flow.dw );
    BE_GEN_TXULP_CSR_WRITE( txulp_id, NULL, cbr_dat, qos_cbs.dw );

    // Write PBS Data (Disable PBS in BE2 - Bugzilla BUG 10663)
    BE_GEN_TXULP_CSR_WRITE( txulp_id, NULL, cbr_pbr_num, qos_cbs_pbs_flow.dw );
    pbs_csr.pbs_data = 1;
    BE_GEN_TXULP_CSR_WRITE( txulp_id, NULL, pbr_dat, pbs_csr.dw );

} // end of BE2_Gen_Set_QoS

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Load_Ctx_Base
//
//   Private function to the GEN API to write descriptor or context bank data for a given
//   EP into chip context memory.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// data_array         - Pointer to source data.
// dword_length       - Number of valid dwords in data_array.
// byte_offset        - Starting byte offset in chip context memory.
// swap_data          - TRUE to byte-swap the data written, FALSE to leave as-is.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Load_Ctx_Base (
    void* function_context,
    u32 *data_array,
    u32 dword_length,
    u32 byte_offset,
    u8  swap_data )
{
    u32 i = 0;
    u32 align_dword_length = (dword_length / 4) * 4;
    u32 offset_dword_length = dword_length - align_dword_length;

    // Write multiples of 16 bytes.
    for (i = 0; i < align_dword_length; i += 4)
    {
        if (swap_data) {
            u32 data[4];

            data[0] = BE_GEN_SWAP32 (data_array[i+0]);
            data[1] = BE_GEN_SWAP32 (data_array[i+1]);
            data[2] = BE_GEN_SWAP32 (data_array[i+2]);
            data[3] = BE_GEN_SWAP32 (data_array[i+3]);
            BE_GEN_WRITE_CTX (function_context, byte_offset, data, 16);
        } else {
            BE_GEN_WRITE_CTX (function_context, byte_offset, &data_array[i], 16);
        }
        byte_offset += 16;
    }

    // Write the last 4, 8, or 12 bytes
    if (offset_dword_length > 0) {
         if (swap_data) {
            u32 data[4];
            data[0] = BE_GEN_SWAP32 (data_array[i+0]);
            data[1] = (offset_dword_length > 1 ? BE_GEN_SWAP32 (data_array[i+1]) : 0);
            data[2] = (offset_dword_length > 2 ? BE_GEN_SWAP32 (data_array[i+2]) : 0);
            data[3] = (offset_dword_length > 3 ? BE_GEN_SWAP32 (data_array[i+3]) : 0);
            BE_GEN_WRITE_CTX (function_context, byte_offset, data, offset_dword_length * 4);
        } else {
            BE_GEN_WRITE_CTX (function_context, byte_offset, &data_array[i], offset_dword_length * 4);
        }
    }
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Load_EP_Firmware_Base
//
//   Private function to the GEN API to load the EP IRAM or DRAM.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// data_array         - Pointer to the source data.
// dword_length       - Number of valid dwords in data_array.
// starting_address   - Starting dword address.
// addr_register      - Byte address to the address register.
// data_register      - Byte address to the data register.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Load_EP_Firmware_Base (
    void *function_context,
    u32 *data_array,
    u32 dword_length,
    u32 starting_address,
    u32 addr_register,
    u32 data_register )
{
    u32 i;
    /* Set the starting address in the address register */
    BE_GEN_CSR_WRITE_ADDR( function_context, addr_register, starting_address );

    /* Write out the data. Note the address register is expected to auto-increment. */
    for (i=0; i < dword_length; i++) {
        BE_GEN_CSR_WRITE_ADDR( function_context, data_register, BE_GEN_SWAP32(data_array[i]) );
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Init_EP_Ram_Base
//
//   Private function to the GEN API to initialize (0) the EP IRAM or DRAM.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// dword_length       - Number of dwords to write.
// starting_address   - Starting dword address.
// addr_register      - Byte address to the address register.
// data_register      - Byte address to the data register.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Init_EP_Ram_Base (
    void *function_context,
    u32 dword_length,
    u32 starting_address,
    u32 addr_register,
    u32 data_register )
{
    u32 i;
    /* Set the starting address in the address register */
    BE_GEN_CSR_WRITE_ADDR( function_context, addr_register, starting_address );

    /* Write out the data. Note the address register is expected to auto-increment. */
    for (i=0; i < dword_length; i++) {
        BE_GEN_CSR_WRITE_ADDR( function_context, data_register, 0 );
    }

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Load_EP_Firmware
//
//   Loads BladeEngine embedded processor firmware. This function depends on the BE_CONFIG
//   literal value to determine proper firmware to load.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Load_EP_Firmware (
   IN void* function_context )
{
#if !defined(SA_LINUXU)
    u32 is_a0 = (0 == BE2_Gen_Chip_Rev(function_context)) ? 1 : 0;

    /* TPOST */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TPOST " in TPOST EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_TPOST, tpost);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_TPOST, 0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_TPOST, 0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_TPOST, EP_DESC_BANK_BASE0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_TPOST, EP_CTXT_BANK_BASE0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_TPOST, EP_DESC_BANK_BASE1, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_TPOST, EP_CTXT_BANK_BASE1, tpost);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_TPOST, tpost);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_TPOST, 0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_TPOST, 0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_TPOST, EP_DESC_BANK_BASE0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_TPOST, EP_CTXT_BANK_BASE0, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_TPOST, EP_DESC_BANK_BASE1, tpost);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_TPOST, EP_CTXT_BANK_BASE1, tpost);
    }

    /* TPRE */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_TPRE " in TPRE EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_TPRE, tpre);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_TPRE, 0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_TPRE, 0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_TPRE, EP_DESC_BANK_BASE0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_TPRE, EP_CTXT_BANK_BASE0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_TPRE, EP_DESC_BANK_BASE1, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_TPRE, EP_CTXT_BANK_BASE1, tpre);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_TPRE, tpre);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_TPRE, 0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_TPRE, 0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_TPRE, EP_DESC_BANK_BASE0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_TPRE, EP_CTXT_BANK_BASE0, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_TPRE, EP_DESC_BANK_BASE1, tpre);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_TPRE, EP_CTXT_BANK_BASE1, tpre);
    }

    /* TXULP0 */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_ULP0_TX " in TXULP0 EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_ULP0_TX, txulp0);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_ULP0_TX, 0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_ULP0_TX, 0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_TX, EP_DESC_BANK_BASE0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP0_TX, EP_CTXT_BANK_BASE0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_TX, EP_DESC_BANK_BASE1, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP0_TX, EP_CTXT_BANK_BASE1, txulp0);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_ULP0_TX, txulp0);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_ULP0_TX, 0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_ULP0_TX, 0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_TX, EP_DESC_BANK_BASE0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP0_TX, EP_CTXT_BANK_BASE0, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_TX, EP_DESC_BANK_BASE1, txulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP0_TX, EP_CTXT_BANK_BASE1, txulp0);
    }

    /* RXULP0 */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_ULP0_RX " in RXULP0 EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_ULP0_RX, rxulp0);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_ULP0_RX, 0, rxulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_ULP0_RX, 0, rxulp0);
#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_RX, ISCSI_RX_EP_DESC_BANK_BASE0, rxulp0);
#else
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_RX, EP_DESC_BANK_BASE0, rxulp0);
#endif
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP0_RX, EP_CTXT_BANK_BASE0, rxulp0);
#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_RX, ISCSI_RX_EP_DESC_BANK_BASE1, rxulp0);
#else
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP0_RX, EP_DESC_BANK_BASE1, rxulp0);
#endif
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP0_RX, EP_CTXT_BANK_BASE1, rxulp0);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_ULP0_RX, rxulp0);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_ULP0_RX, 0, rxulp0);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_ULP0_RX, 0, rxulp0);
#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_RX, ISCSI_RX_EP_DESC_BANK_BASE0, rxulp0);
#else
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_RX, EP_DESC_BANK_BASE0, rxulp0);
#endif
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP0_RX, EP_CTXT_BANK_BASE0, rxulp0);
#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_RX, ISCSI_RX_EP_DESC_BANK_BASE1, rxulp0);
#else
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP0_RX, EP_DESC_BANK_BASE1, rxulp0);
#endif
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP0_RX, EP_CTXT_BANK_BASE1, rxulp0);
    }

    /* TXULP1 */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_ULP1_TX " in TXULP1 EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_ULP1_TX, txulp1);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_ULP1_TX, 0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_ULP1_TX, 0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP1_TX, EP_DESC_BANK_BASE0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP1_TX, EP_CTXT_BANK_BASE0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP1_TX, EP_DESC_BANK_BASE1, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP1_TX, EP_CTXT_BANK_BASE1, txulp1);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_ULP1_TX, txulp1);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_ULP1_TX, 0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_ULP1_TX, 0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP1_TX, EP_DESC_BANK_BASE0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP1_TX, EP_CTXT_BANK_BASE0, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP1_TX, EP_DESC_BANK_BASE1, txulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP1_TX, EP_CTXT_BANK_BASE1, txulp1);
    }

    /* RXULP1 */
    BE_GEN_DEBUG_PRINT ("BE GEN: loading " NAME_STRING_ULP1_RX " in RXULP1 EP\n");
    if (is_a0)
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A0_IROM_ULP1_RX, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A0_IRAM_ULP1_RX, 0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DRAM_ULP1_RX, 0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP1_RX, EP_DESC_BANK_BASE0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP1_RX, EP_CTXT_BANK_BASE0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_DESC_ULP1_RX, EP_DESC_BANK_BASE1, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A0_CTXT_ULP1_RX, EP_CTXT_BANK_BASE1, rxulp1);
    }
    else
    {
        BE2_GEN_LOAD_FIRMWARE_IROM  (function_context, A1_IROM_ULP1_RX, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_INST  (function_context, A1_IRAM_ULP1_RX, 0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DRAM_ULP1_RX, 0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP1_RX, EP_DESC_BANK_BASE0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP1_RX, EP_CTXT_BANK_BASE0, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_DESC_ULP1_RX, EP_DESC_BANK_BASE1, rxulp1);
        BE2_GEN_LOAD_FIRMWARE_DATA  (function_context, A1_CTXT_ULP1_RX, EP_CTXT_BANK_BASE1, rxulp1);
    }

    /* TPRE Classification */
    BE2_Gen_Load_TPRE_Classification (function_context);

    /* TXULP Classification */
    BE2_Gen_Load_TXULP_Classification (function_context, 0);
    BE2_Gen_Load_TXULP_Classification (function_context, 1);

#endif
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Wakeup_EPs
//
//   Brings embedded processors out of reset.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Wakeup_EPs (
   IN void* function_context )
{
    EP_CONTROL_CSR ep_control;

    /* TPOST */
    ep_control.dw = BE_GEN_CSR_READ( function_context, tpost.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_control, ep_control.dw );

    /* TPRE */
    ep_control.dw = BE_GEN_CSR_READ( function_context, tpre.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_control, ep_control.dw );

    /* TXULP0 */
    ep_control.dw = BE_GEN_CSR_READ( function_context, txulp0.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, txulp0.ep.ep_control, ep_control.dw );

    /* RXULP0 */
    ep_control.dw = BE_GEN_CSR_READ( function_context, rxulp0.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, rxulp0.ep.ep_control, ep_control.dw );

    /* TXULP1 */
    ep_control.dw = BE_GEN_CSR_READ( function_context, txulp1.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, txulp1.ep.ep_control, ep_control.dw );

    /* RXULP1 */
    ep_control.dw = BE_GEN_CSR_READ( function_context, rxulp1.ep.ep_control );
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, rxulp1.ep.ep_control, ep_control.dw );

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_RAM_Init
//
//   Performs BladeEngine memory initialization.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_RAM_Init (
   IN void* function_context )
{
#if !defined(SIMULATION)
    {
        u64 dt_value = 0xffffffffffffffffULL;
        int index ;
        BE_GEN_DEBUG_PRINT ("BE GEN: pre-initialize DT RAM\n");
        for (index = 0; index < 512; index++) {
            BE2_Gen_Access_DT (
                function_context,
                index,
                &dt_value,
                TRUE );
        }
    }


    {
        u32 data[4] = {0, 0, 0, 0};
        u32 i;
#if (BE_CONFIG==7)
        u32 cookie;
        BE_GEN_DEBUG_PRINT ("BE GEN: scrub CONFIG 7 context memory\n");
        BE2_Gen_Turn_Online (function_context, 0, ENUM_CTX_SRC);
        BE_GEN_CTX_RD(function_context, BE_FIELD_OFFSET(CONFIG7_CTX_MEMORY_MAP,armfw_config_info.fc_cr_cookie) ,
                                                                                        &cookie, sizeof(u32));
        BE_GEN_DEBUG_PRINT ("BE GEN: FC_CR:%x\n", cookie);
        for (i = 0; i < sizeof(CTX_MEMORY_MAP) + 32768; i += sizeof(data)) {
            BE_GEN_CTX_WR(function_context, i, (u8 *)&data[0], sizeof(data));
        }
        BE_GEN_CTX_WR(function_context, BE_FIELD_OFFSET(CONFIG7_CTX_MEMORY_MAP,armfw_config_info.fc_cr_cookie) ,
                                                                                        &cookie, sizeof(u32));
#else
        BE_GEN_DEBUG_PRINT ("BE GEN: scrub context memory\n");
        BE2_Gen_Turn_Online (function_context, 0, ENUM_CTX_SRC);
        for (i = 0; i < sizeof(CTX_MEMORY_MAP) + 32768; i += sizeof(data)) {
            BE_GEN_CTX_WR(function_context, i, (u8 *)&data[0], sizeof(data));
        }
#endif
    }

    {
        VLAN_TABLE_ENTRY vte = {0, 0, 0, 0, 0, 0, 0, 0};
        int index;

        BE_GEN_DEBUG_PRINT ("BE GEN: reset RXF CAMs\n");
        BE_GEN_CSR_WRITE (function_context, rxf.cam_soft_reset, 0x0000001f);

        BE_GEN_DEBUG_PRINT ("BE GEN: zero RXF VLAN table\n");
        for (index = 0; index < BE_GEN_NUMBER_OF_FIELD(VLAN_TABLE, TAG); index++) {
            BE2_Gen_Access_VLAN_Table (function_context, index, &vte.dw[0], TRUE);
        }
    }
#if (BE_CONFIG != 6)
    /* Bug 10700*/
    /* Initialize the RXF CAM  */
    {
        u32 index;
        OFFLOAD_CXN_ENTRY data;
        memset ( &data, 0 , sizeof(OFFLOAD_CXN_ENTRY ) );
        for ( index = 0; index < 1024; index++ ) {
            BE2_Gen_Access_Rxf_Table(function_context, ENUM_OFFLOADED_CXN_TABLE_INDEX,
                    BE_FIELD_OFFSET(OFFLOADED_CXN_TABLE, cxn[index])/sizeof(u32) ,
                    sizeof(OFFLOAD_CXN_ENTRY)/sizeof(u32) , (u32 *)&data, TRUE);
        }
    }
#endif /*BE_CONFIG != 6*/

#if (BE_CONFIG==7)
    /* Initialize the FCoE CAM */
    BE_GEN_DEBUG_PRINT ("BE GEN: Initialize FCoE CAM\n");
    {
        u32 index;
        FCOE_TABLE_ENTRY fcoe_table_entry;;
        memset ( &fcoe_table_entry, 0 , sizeof(FCOE_TABLE_ENTRY ) );
        for ( index = 0; index < 64; index++ )
        {
            BE2_Gen_Access_Rxf_Table(function_context, ENUM_FCOE_TABLE_INDEX,
                            index * 4 , sizeof(FCOE_TABLE_ENTRY)/sizeof(u32) ,
                                            (u32 *)&fcoe_table_entry, TRUE);
        }
    }
#endif

#endif
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Registers_Init
//
//   Performs BladeEngine register initialization. This function depends on the BE_CONFIG
//   literal value to determine proper values for ULP registers.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Registers_Init (
   IN void* function_context )
{
    /* Butler registers */
    BTLR_CTRL0 btlr_ctrl0;
    BTLR_CTRL1 btlr_ctrl1;
    BTLR_CTRL2 btlr_ctrl2;

    /* General EP control register */
    EP_CONTROL_CSR ep_control;

    /* EP Input queue weight register */
    EP_INPUT_QUE_WEIGHT_CSR ep_iq_weight;

    /* TPOST PSEL weight register 2 */
    /* TPOST_PROTSEL_CTL2_CR psel_ctl2_weight; */

    /* EP input queue fetch control registers */
    EP_QCTRL0_CSR ep_qctrl0;
    EP_QCTRL1_CSR ep_qctrl1;
    EP_QCTRL2_CSR ep_qctrl2;
    EP_QCTRL3_CSR ep_qctrl3;

    /* EP output queue credit registers */
    EP_QCREDITA_CSR ep_qcreditA;
    EP_QCREDITB_CSR ep_qcreditB;
    EP_QCREDITC_CSR ep_qcreditC;
    EP_QCREDITD_CSR ep_qcreditD;

#if 0
    /* EP IPC overflow registers */
    EP_IPC_OV_CSR ipc_ov_csr;
#endif

    /* UC and PTC control registers */
    UC_CONTROL_CSR uc_control;
    PTC_CONTROL_CSR ptc_control;

    TXP_CONTROL_CSR txp_ctrl;

    ERX_CFG_CSR erx_cfg;

    RXF_RXCTL_CSR rxctl;
    CEV_CTL_CSR cev_ctl;

    WDMA_UC_RELEASE_CSR wdma_uc_rel;
    WDMA_CTRL_CSR wdma_ctrl;
    RDMA_CTRL_CSR rdma_ctrl;
    NDMA_CTRL_CSR ndma_ctrl;

    RDMA_WRR_ENTRY_CSR wrr_entry;

    CTX_WRR_ENTRY_CSR wrr_csr;
    TPOST_PROTSEL_CTL3_CR tpost_protsel_ctl3;

    u32 is_asic;
    u32 chip_rev;

    /* TPRE registers */
    TPRE_DBCRED_CFG_RWDATA_CSR tpre_dbcred;
    TIM_1MS_TIMESTAMP_CR tim_1ms_timestamp_cr;
#if (BE_CONFIG==0)
    TIM_3_2US_TIMESTAMP_CR tim_3_2us_timestamp_cr;
#endif
    TPRE_FW_TIMER_SCALE_CR tim_tcp_timer_scale;
    TIM_TIMESTAMP_EXPIRE_CR tim_timestamp_expire;



    /************************************************************************
     * GENERAL SECTION
     ************************************************************************/
    {
    BE_GEN_DEBUG_PRINT ("BE GEN: Initializing general registers\n");

    is_asic = BE2_Gen_Is_ASIC(function_context);
    chip_rev = BE2_Gen_Chip_Rev(function_context);
    if (TRUE == is_asic) {
        switch (chip_rev)
        {
        case 0:
            BE_GEN_DEBUG_PRINT ("BE GEN: A0 asic detected\n");
            break;
        case 1:
            BE_GEN_DEBUG_PRINT ("BE GEN: A1 asic detected\n");
            break;
        case 2:
            BE_GEN_DEBUG_PRINT ("BE GEN: A2 asic detected\n");
            break;
        default:
#ifdef BE_ARM_FIRMWARE
            BE_GEN_DEBUG_PRINT ("BE GEN: asic rev %d detected\n", chip_rev);
#else
            BE_GEN_DEBUG_PRINT ("BE GEN: asic detected\n");
#endif
        }
    } else {
        switch (chip_rev)
        {
        case 0:
            BE_GEN_DEBUG_PRINT ("BE GEN: A0 fpga detected\n");
            break;
        case 1:
            BE_GEN_DEBUG_PRINT ("BE GEN: A1 fpga detected\n");
            break;
        case 2:
            BE_GEN_DEBUG_PRINT ("BE GEN: A2 fpga detected\n");
            break;
        default:
#ifdef BE_ARM_FIRMWARE
            BE_GEN_DEBUG_PRINT ("BE GEN: fpga rev %d detected\n", chip_rev);
#else
            BE_GEN_DEBUG_PRINT ("BE GEN: fpga detected\n");
#endif
        }
    }

    if (!is_asic) {
        PCIE_FCP_TIMER_BEACON_EN_PCICFG_OVERRIDE pcie_fcp_timer_beacon_en_pcicfg_override;
        pcie_fcp_timer_beacon_en_pcicfg_override.dw = BE_GEN_PCI0_READ ( function_context, pcie_fcp_timer_beacon_en_pcicfg_override );
        pcie_fcp_timer_beacon_en_pcicfg_override.mode_electrical_idle_infer_fc_detect = 0;
        BE_GEN_PCI0_WRITE ( function_context, pcie_fcp_timer_beacon_en_pcicfg_override, pcie_fcp_timer_beacon_en_pcicfg_override.dw );
    }

    cev_ctl.dw = BE_GEN_CSR_READ( function_context, cev.ctl );
    cev_ctl.Swap = DMA_SWAP_BYTE;
#if (BE_CONFIG == 6)
    cev_ctl.QueueOvEn = 1;
#endif
    BE_GEN_CSR_WRITE( function_context, cev.ctl, cev_ctl.dw );

    /* Set the base context memory address for UC */
    BE_GEN_CSR_WRITE( function_context, uc.uc_cache_start_addr, BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, unified_cache) );

    /* Setup UC control register */
    uc_control.dw = 0;
    uc_control.en_uc = 1;
    uc_control.rdma_swap_cntl = DMA_SWAP_BYTE;
    uc_control.wdma_swap_cntl = DMA_SWAP_BYTE;
    uc_control.uc_rdma_throttle = 3;
    uc_control.uc_wdma_throttle = 3;
    uc_control.uc_ptc_throttle  = 3;
    BE_GEN_CSR_WRITE( function_context, uc.control, uc_control.dw );

    /* Set the base context memory address for PTC */
    BE_GEN_CSR_WRITE( function_context, ptc.ptc_cache_start_addr, BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, page_table_cache) );

    /* Setup the PTC control register */
    ptc_control.dw = 0;
    ptc_control.en_ptc = 1;
    ptc_control.rdma_swap_cntl = DMA_SWAP_WORD_BYTE;
    if (0 == chip_rev) {
        /* Part of workaround for bug 9461: limit PTC to RDMA requests */
        ptc_control.ptc_rdma_throttle = 0x1;
    }
    else {
        /* Default number of PTC to RDMA requests */
        ptc_control.ptc_rdma_throttle = 0x7;
    }
    BE_GEN_CSR_WRITE( function_context, ptc.control, ptc_control.dw );

    if (0 == chip_rev) {
        /* Part of workaround for bug 9461: lower weight of PTC to RDMA requests */
#ifdef SIMULATION
        // Sim environment can't do CSR reads. Set defaults for
        // those values we are not changing.
        wrr_entry.dw     = 0;
        wrr_entry.mpu_wt = 8;  // Fix for Bug 10721
        wrr_entry.uc_wt  = 8;  // Fix for Bug 10721
        wrr_entry.erx_wt = 8;  // Fix for Bug 10721
#else
        wrr_entry.dw = BE_GEN_CSR_READ(function_context, rdma.rdma_wrr);
#endif
        wrr_entry.ptc_wt = 1;
        BE_GEN_CSR_WRITE(function_context, rdma.rdma_wrr, wrr_entry.dw);
    }

    {
        UC_WEIGHT_ARB_CSR uc_weight8_9;
        UC_WEIGHT_ARB_CSR uc_weight10_11;
        PQ_CONTROL_CSR pq_control_csr;

        /* CSR writes to enable UC UEs (see bug #s referenced by the bits set) */

        /* Register 8_9 */
        uc_weight8_9.dw = 0x3f010001; // TODO use names
        BE_GEN_CSR_WRITE( function_context, uc.weight8_9, uc_weight8_9.dw );

        /* Register 10. Make queue depth between UC and PTC 6.
        * Do not enable UC CFIFO parity error at this juncture (it will be
        * enabled after the CFIFO is initialized during RAM init code).
        */
        uc_weight10_11.dw = 0x783d0001; // TODO use names
        BE_GEN_CSR_WRITE( function_context, uc.weight10_11, uc_weight10_11.dw );

        pq_control_csr.dw = 0;
        pq_control_csr.en_count = 1;
        BE_GEN_CSR_WRITE( function_context, uc.rxulp0_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rxulp0_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rxulp1_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rxulp1_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rxulp2_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rxulp2_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp0_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp0_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp1_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp1_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp2_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.txulp2_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.mpu_req_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.mpu_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.cev_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.erx_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rdma_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.wdma_rel_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.prefifo_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.ptc_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.rdma_pq, pq_control_csr.dw );
        BE_GEN_CSR_WRITE( function_context, uc.wdma_pq, pq_control_csr.dw );

    }


    // initialize ptc table
    //
    {
        PTC_UE_ENABLE ptc_ue_enable = { 0 };
        PTC_UE_SOURCE ptc_ue_source = { 0 };
        PTC_DT_INDEX_CSR ptc_dt_index = { 0 };
        PTC_PHYS_ADDR_LO dt_pa_lo;
        PTC_PHYS_ADDR_HI dt_pa_hi;
        int i;

        ptc_ue_enable.dw = BE_GEN_CSR_READ(function_context, ptc.ue_enable);

        ptc_ue_enable.dt_parity_error = 0;
        BE_GEN_CSR_WRITE(function_context, ptc.ue_enable, ptc_ue_enable.dw);

        for (i=0; i<512; i++) {
            ptc_dt_index.idx = i;
            dt_pa_lo.dw = 0xffffffff;
            dt_pa_hi.dw = 0xffffffff;
            BE_GEN_CSR_WRITE(function_context, ptc.dt_index, ptc_dt_index.dw);
            BE_GEN_CSR_WRITE(function_context, ptc.dt_pa_lo, dt_pa_lo.dw);
            BE_GEN_CSR_WRITE(function_context, ptc.dt_pa_hi, dt_pa_hi.dw);
        }

        // clear ue source
        ptc_ue_source.dw = BE_GEN_CSR_READ(function_context, ptc.ue_source);
        ptc_ue_source.dt_parity_error = 1;
        ptc_ue_source.tag_ram_perror = 1;
        ptc_ue_source.fifo_uflow_error = 1;
        ptc_ue_source.fifo_oflow_error = 1;
        BE_GEN_CSR_WRITE(function_context, ptc.ue_source, ptc_ue_source.dw);

        // reenable ptc dt parity error
        ptc_ue_enable.dt_parity_error = 1;
        BE_GEN_CSR_WRITE(function_context, ptc.ue_enable, ptc_ue_enable.dw);
    }


    /* Enable TXP to forward management frames to port 0 */
    txp_ctrl.dw = BE_GEN_CSR_READ( function_context, txp.txp_control );
    if (0 == chip_rev) {
        /* Throttle NDMA queue depth to 1 for A0 to workaround
         * iSCSI sequence number jump issues caused by hardware.
         * Two related bugs: SE-AUS bug #11916, SE-SJC bug #4950.
         * Note: this has a serious negative impact on transmit
         * performance
         */
        txp_ctrl.ndma_limit = 0x1;
    }
    else
    {
        txp_ctrl.ndma_limit = 0x8;
    }

    BE_GEN_CSR_WRITE( function_context, txp.txp_control, txp_ctrl.dw );

#if ( (BE_CONFIG==0) || (BE_CONFIG==3) || (BE_CONFIG==9) )
    /* Initialize the partial iSCSI CRC base address */
    BE_GEN_CSR_WRITE( function_context, txp.txp_crc_start_addr,
                  (BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, txp_iscsi_crc_array) - ISCSI_CID_START * 8) / 16);
    BE_GEN_CSR_WRITE( function_context, wdma.wdma_crc_start_addr,
                  (BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, wdma_iscsi_crc_array) - ISCSI_CID_START * 8) / 16 );
    BE_GEN_CSR_WRITE( function_context, wdma.wdma_dif_start_addr,
                  BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, icd) / 16 );
#endif

    /* CAUTION! Do not change these values without the consent
     * of the firmware folks as the functionality is commonly
     * used by ISCSI and TOE firmware.
     */
    wdma_uc_rel.dw = 0;
    wdma_uc_rel.i = 0;
    wdma_uc_rel.vl = 0;
    wdma_uc_rel.req_type = UC_REQ_TYPE_NORMAL;
    wdma_uc_rel.drop = 1;
    wdma_uc_rel.req_opcode = UC_REQ_OPCODE_ACCESS_MINUS;
    wdma_uc_rel.wr = 0;
    wdma_uc_rel.cqe = 0;
    wdma_uc_rel.stall = STALLARC_NO_STALL;
    wdma_uc_rel.pci_func = 0;
    wdma_uc_rel.p = 0;
    BE_GEN_CSR_WRITE( function_context, wdma.wdma_uc_release_status, wdma_uc_rel.dw );

    /* Stall on errors */
    wdma_ctrl.dw = 0;
    wdma_ctrl.stall_on_error = 1;
    BE_GEN_CSR_WRITE( function_context, wdma.wdma_ctrl, wdma_ctrl.dw );

    ndma_ctrl.dw = 0;
    ndma_ctrl.stall_on_error = 1;
    BE_GEN_CSR_WRITE( function_context, ndma.ndma_ctrl, ndma_ctrl.dw );

    rdma_ctrl.dw = 0;
    rdma_ctrl.stall_on_error = 1;
    BE_GEN_CSR_WRITE( function_context, rdma.rdma_ctrl, rdma_ctrl.dw );

#if (BE_CONFIG==2)
    /* For target when 2 chutes are in use, we need more non-payload TXPBs */
    BE_GEN_CSR_WRITE( function_context, txpb.payload_bufs, 176);
#else
    /* Change default payload_bufs to 192 */
    BE_GEN_CSR_WRITE( function_context, txpb.payload_bufs, 192);
#endif


    /* CEV butler */
    btlr_ctrl0.dw = 0;
    BE_GEN_CSR_WRITE( function_context, cev.btlr1.btlr_ctrl0, btlr_ctrl0.dw );

    btlr_ctrl1.dw = 0;
    btlr_ctrl1.btlr_ring = 1;
    btlr_ctrl1.btlr_pd0_promiscuous = TRUE;
    BE_GEN_CSR_WRITE( function_context, cev.btlr1.btlr_ctrl1, btlr_ctrl1.dw );

    btlr_ctrl2.dw = 0;
    btlr_ctrl2.btlr_cid_start    = 0;
    btlr_ctrl2.btlr_cid_end      = 0xc07f;
    BE_GEN_CSR_WRITE( function_context, cev.btlr1.btlr_ctrl2, btlr_ctrl2.dw );

    /* Setup ERX */
    erx_cfg.dw              = BE_GEN_CSR_READ( function_context, erx.cfg );
    erx_cfg.fetch_wm        = ERX_WM_HALF;
    erx_cfg.intr_wm         = ERX_WM_EMPTY;
    erx_cfg.rdma_swap_cntl  = DMA_SWAP_BYTE;
    erx_cfg.oord            = 1;
    erx_cfg.btlr_ignore_partial_fid = 1;
    erx_cfg.btlr_ignore_pdid = 1;
    erx_cfg.btlr_ignore_fid = 0;
    erx_cfg.btlr_pd0_promiscuous = 1;
    /* BUG 20381 - turn off the suck bit so HP sees good performance with vlan tagging on on the IXIA RFC2544 test */
    erx_cfg.rss_flush_first_dma_echo = 0;
    BE_GEN_CSR_WRITE( function_context, erx.cfg, erx_cfg.dw );

#ifndef SC_SIMULATION
    /* Setup RXF */
    rxctl.dw = BE_GEN_CSR_READ(function_context, rxf.rxctl);
    rxctl.enp0fivetuple = 1;
    rxctl.enp1fivetuple = 1;
    rxctl.enp0 = 1;
    rxctl.enp1 = 1;
    BE_GEN_CSR_WRITE(function_context, rxf.rxctl, rxctl.dw);

    /* Disable all interface vectors */
    BE_GEN_CSR_WRITE(function_context, rxf.port0_broadcast_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port0_broadcast_vector_hi, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_broadcast_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_broadcast_vector_hi, 0);

    BE_GEN_CSR_WRITE(function_context, rxf.port0_promiscuous_vlan_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port0_promiscuous_vlan_vector_hi, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_promiscuous_vlan_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_promiscuous_vlan_vector_hi, 0);

    BE_GEN_CSR_WRITE(function_context, rxf.port0_untagged_vlan_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port0_untagged_vlan_vector_hi, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_untagged_vlan_vector_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_untagged_vlan_vector_hi, 0);

    BE_GEN_CSR_WRITE(function_context, rxf.port0_interface_en_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port0_interface_en_hi, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_interface_en_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.port1_interface_en_hi, 0);

    BE_GEN_CSR_WRITE(function_context, rxf.p0_pass_l3_l4_err_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.p0_pass_l3_l4_err_hi, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.p1_pass_l3_l4_err_lo, 0);
    BE_GEN_CSR_WRITE(function_context, rxf.p1_pass_l3_l4_err_hi, 0);
#endif

    /* Configure the EQ delay module global values.
     */
    {
        CEV_MAX_SUST_INT_DLY_CNT_CSR delay;

        delay.dw = 0;
        delay.ENABLE_LKY_BUCKET = 1;
        delay.BUCKET_CREDIT = 0x2;
        delay.PHASE_RESET = 0;
        delay.MAX_CSR_STROLL = 0x5F; // stroll's all 192 EQs (setting = (Num EQs/2)-1)
        BE_GEN_CSR_WRITE (function_context, cev.cev_max_sust_int_dly_cnt_csr, delay.dw);
    }

    /* Configure CQ DMA coalescing watermarks.  Each CQ is assigned to watermark 0-3.
     * Watermark 0 never coalesces.  Watermark 1-3 coalesce when the number of pending
     * CQ entries in the ring is >= the given watermark value.
     */
    {
        CEV_CSR_COALESCWM_CFG wm;

        wm.dw = 0;
        wm.coalesc_wm0 = 0xFF; // no coalescing
        wm.coalesc_wm1 = 2;    // coalesce when >= 2 CQ entries pending
        wm.coalesc_wm2 = 4;
        wm.coalesc_wm3 = 16;

        BE_GEN_CSR_WRITE (function_context, cev.coalesc_wm, wm.dw);
    }

    /* Configure the Context Memory Weights for the client arbiter.  Increase
     * the relative priority for tpost and tpre since they must handle more
     * traffic than the RX/TX ULPs.
     */
    wrr_csr.weight0 = 0x10; // rdma
    wrr_csr.weight1 = 0x10; // wdma
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_rdma_wdma, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // rxulp1
    wrr_csr.weight1 = 0x10; // rxulp0
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_rxulp1_rxulp0, wrr_csr.dw);

    wrr_csr.weight0 = 0x20; // tpre (higher priority)
    wrr_csr.weight1 = 0x10; // rxulp2
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_tpre_rxulp2, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // txulp1
    wrr_csr.weight1 = 0x10; // txulp0
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_txulp1_txulp0, wrr_csr.dw);

    wrr_csr.weight0 = 0x20; // tpost (higher priority)
    wrr_csr.weight1 = 0x10; // etx
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_tpost_etx, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // ndma
    wrr_csr.weight1 = 0x10; // txp
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_ndma_tx, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // erx
    wrr_csr.weight1 = 0x10; // uc
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_erx_uc, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // mpu
    wrr_csr.weight1 = 0x10; // ptc
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_mpu_ptc, wrr_csr.dw);

    wrr_csr.weight0 = 0x10; // cev
    wrr_csr.weight1 = 0x10; // txulp2
    BE_GEN_CSR_WRITE(function_context, ctx.wrr_cev_txulp2, wrr_csr.dw);


    }
    /************************************************************************
     * END GENERAL SECTION
     ************************************************************************/

    /************************************************************************
     * MPU REGISTERS SECTION
     ************************************************************************/
    {
    BE_GEN_DEBUG_PRINT ("BE GEN: Initializing MPU registers\n");

    /* MPU butler configuration 0 register */
    btlr_ctrl0.dw                       = 0;
    btlr_ctrl0.btlr_ignore_fid          = 0;        // Ignore function ID
    btlr_ctrl0.btlr_ignore_partial_fid  = 0;        // Ignore FID bits 7:3
    btlr_ctrl0.btlr_ignore_pdid         = 1;        // Ignore protection domain ID
    btlr_ctrl0.btlr_indx_mask           = 8;        // Encoded max allowable size of ring (2^indx_mask)
    btlr_ctrl0.btlr_cid_mask            = 3;        // Encoded max allowable size of CID value in DB (2^cid_mask)
    btlr_ctrl0.btlr_opaque_mask         = 0;        // Opaque data bits copied from the left-most part of the DB.
    BE_GEN_CSR_WRITE( function_context, mpu.btlr.btlr_ctrl0, btlr_ctrl0.dw );

    /* MPU butler configuration 1 register */
    btlr_ctrl1.dw                           = 0;
    btlr_ctrl1.btlr_base_offset             = BE_GEN_TO_U16 (BE_GEN_FIELD_OFFSET( CTX_MEMORY_MAP, mcc[-MCC_BID_START]) / 16);
    btlr_ctrl1.btlr_be                      = 0xFF;     // Byte-enables for butler management
    btlr_ctrl1.btlr_check_numpost           = 1;        // Enables checking num-posted counts on doorbell ring
    btlr_ctrl1.btlr_report_update_cid_oor   = 0;        // Enables OOR error reporting
    btlr_ctrl1.btlr_pd0_promiscuous         = 0;        // Enables PD0 to write doorbells in all pages
    btlr_ctrl1.btlr_ring                    = 1;        // iSCSI TX rings == 0, all others == 1
    btlr_ctrl1.btlr_stride                  = 0;        // 2^stride * 16 = sizeof(CTX)
    BE_GEN_CSR_WRITE( function_context, mpu.btlr.btlr_ctrl1, btlr_ctrl1.dw );

    /* MPU butler configuration 2 register */
    btlr_ctrl2.dw               = 0;
    btlr_ctrl2.btlr_cid_start   = MCC_BID_START;
    btlr_ctrl2.btlr_cid_end     = MCC_BID_START + MCC_ID_NUM - 1;
    BE_GEN_CSR_WRITE( function_context, mpu.btlr.btlr_ctrl2, btlr_ctrl2.dw );

    /* MPU general control register */
    ep_control.dw = 0;
    ep_control.m0_RxPbuf = 0;
    ep_control.m1_RxPbuf = 0;
    ep_control.m2_RxPbuf = 1;
    ep_control.CPU_reset = 0;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_control, ep_control.dw );

    /* MPU input queue weight registers */
    ep_iq_weight.dw      = 0;
    ep_iq_weight.iq0     = 0xF;
    ep_iq_weight.iq1     = 1;
    ep_iq_weight.iq2     = 1;
    ep_iq_weight.iq3     = 1;
    ep_iq_weight.iq4     = 1;
    ep_iq_weight.iq5     = 1;
    ep_iq_weight.iq6     = 1;
    ep_iq_weight.iq7     = 1;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_input_que_weight, ep_iq_weight.dw );

    /* MPU output queue credit registers */
    ep_qcreditA.dw      = 0;
    ep_qcreditA.oq0     = MPU_EP_QCREDITA_CSR_OQ0;
    ep_qcreditA.oq1     = MPU_EP_QCREDITA_CSR_OQ1;
    ep_qcreditA.oq2     = MPU_EP_QCREDITA_CSR_OQ2;
    ep_qcreditA.oq3     = MPU_EP_QCREDITA_CSR_OQ3;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_qcreditA, ep_qcreditA.dw );

    ep_qcreditB.dw      = 0;
    ep_qcreditB.oq4     = MPU_EP_QCREDITA_CSR_OQ4;
    ep_qcreditB.oq5     = MPU_EP_QCREDITA_CSR_OQ5;
    ep_qcreditB.oq6     = MPU_EP_QCREDITA_CSR_OQ6;
    ep_qcreditB.oq7     = MPU_EP_QCREDITA_CSR_OQ7;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_qcreditB, ep_qcreditB.dw );

    ep_qcreditC.dw      = 0;
    ep_qcreditC.oq8     = MPU_EP_QCREDITA_CSR_OQ8;
    ep_qcreditC.oq9     = MPU_EP_QCREDITA_CSR_OQ9;
    ep_qcreditC.oq10    = MPU_EP_QCREDITA_CSR_OQ10;
    ep_qcreditC.oq11    = MPU_EP_QCREDITA_CSR_OQ11;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_qcreditC, ep_qcreditC.dw );

    ep_qcreditD.dw      = 0;
    ep_qcreditD.oq12    = MPU_EP_QCREDITA_CSR_OQ12;
    BE_GEN_CSR_WRITE( function_context, mpu.ep.ep_qcreditD, ep_qcreditD.dw );

    }
    /************************************************************************
     * END MPU REGISTERS SECTION
     ************************************************************************/


    /************************************************************************
     * BEGIN TPOST REGISTERS SECTION
     ************************************************************************/
    {
        BE_GEN_DEBUG_PRINT ("BE GEN: Initializing TPOST registers\n");

#if 0
        /* Program preassigned context memory area for ipc overflow */
        ipc_ov_csr.dw = 0;
        ipc_ov_csr.base_addr = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, ovflow[0]);
        ipc_ov_csr.amount    = BE_GEN_TO_U16(BE_GEN_SIZEOF_FIELD(CTX_MEMORY_MAP, ovflow[0]) / 16);
        BE_GEN_CSR_WRITE(function_context, tpost.ep.ep_ipc_ov, ipc_ov_csr.dw);
#endif

        /* Output queue credit registers */
        ep_qcreditA.dw      = 0;
        ep_qcreditA.oq0     = TPOST_EP_QCREDITA_CSR_OQ0;
        ep_qcreditA.oq1     = TPOST_EP_QCREDITA_CSR_OQ1;
        ep_qcreditA.oq2     = TPOST_EP_QCREDITA_CSR_OQ2;
        ep_qcreditA.oq3     = TPOST_EP_QCREDITA_CSR_OQ3;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_qcreditA, ep_qcreditA.dw );

        ep_qcreditB.dw      = 0;
        ep_qcreditB.oq4     = TPOST_EP_QCREDITA_CSR_OQ4;
        ep_qcreditB.oq5     = TPOST_EP_QCREDITA_CSR_OQ5;
        ep_qcreditB.oq6     = TPOST_EP_QCREDITA_CSR_OQ6;
        ep_qcreditB.oq7     = TPOST_EP_QCREDITA_CSR_OQ7;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_qcreditB, ep_qcreditB.dw );

        ep_qcreditC.dw      = 0;
        ep_qcreditC.oq8     = TPOST_EP_QCREDITA_CSR_OQ8;
        ep_qcreditC.oq9     = TPOST_EP_QCREDITA_CSR_OQ9;
        ep_qcreditC.oq10    = TPOST_EP_QCREDITA_CSR_OQ10;
        ep_qcreditC.oq11    = TPOST_EP_QCREDITA_CSR_OQ11;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_qcreditC, ep_qcreditC.dw );

        ep_qcreditD.dw      = 0;
        ep_qcreditD.oq12    = TPOST_EP_QCREDITA_CSR_OQ12;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_qcreditD, ep_qcreditD.dw );

        /* Input queue weight register */
        ep_iq_weight.dw      = 0;
        ep_iq_weight.iq0     = 0xF;
        ep_iq_weight.iq1     = 8;
        ep_iq_weight.iq2     = 1;
        ep_iq_weight.iq3     = 1;
        ep_iq_weight.iq4     = 1;
        ep_iq_weight.iq5     = 0xf;
        ep_iq_weight.iq6     = 0xf; /* SIQ1 needs to be at highest priority to break B2B */
        ep_iq_weight.iq7     = 1;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.ep_input_que_weight, ep_iq_weight.dw );

#if 0 // TODO Need to set all PSEL Arbiter weights
        /* Give ACK queue the highest priority in PSEL */
        psel_ctl2_weight.dw = 0;
//    psel_ctl2_weight.tpost_ps_weight_acku0p0 = 0xf;
//    psel_ctl2_weight.tpost_ps_weight_acku0p1 = 0xf;
//    psel_ctl2_weight.tpost_ps_weight_acku1p0 = 0xf;
//    psel_ctl2_weight.tpost_ps_weight_acku1p1 = 0xf;
//    psel_ctl2_weight.tpost_ps_weight_acku2p0 = 0xf;
//    psel_ctl2_weight.tpost_ps_weight_acku2p1 = 0xf;
        psel_ctl2_weight.tpost_ps_weight_timp0 = 1;
        psel_ctl2_weight.tpost_ps_weight_timp1 = 1;
        BE_GEN_CSR_WRITE(function_context, tpost.tpost_protsel_ctl2, psel_ctl2_weight.dw);
#endif

        /* set highest priority to ACK Q */
        tpost_protsel_ctl3.dw = 0;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute0port0 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack0port0 = 3;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute0port1 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack0port1 = 3;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute1port0 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack1port0 = 3;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute1port1 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack1port1 = 3;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute2port0 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack2port0 = 3;
        tpost_protsel_ctl3.tpost_ps_weight_req_chute2port1 = 1;
        tpost_protsel_ctl3.tpost_ps_weight_req_ack2port1 = 3;

        BE_GEN_CSR_WRITE(function_context, tpost.tpost_protsel_ctl3, tpost_protsel_ctl3.dw);

        /* Input queue configuration */

#if (BE_CONFIG==6)
        /* DALTPOST IQ1 Protocol Selector */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, dal_tx_te_cache) / 16;
        ep_qctrl0.length1    = 4;           // length * 16 = bytes to fetch : 64 byte fetch = 8 TEs
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 1;           // 2^stride * 16 = sizeof(data) = 32 byte stride
        ep_qctrl1.field_len1 = 11;          // CID is 11 bits
        ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
        ep_qctrl1.dst1       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = FALSE;       // Enables back to back
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl1, ep_qctrl1.dw );
#else /* BE_CONFIG != 0 */
        /* STD TPOST IQ1 Protocol Selector */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, template_pointers) / 16;
        ep_qctrl0.length1    = 2;           // length * 16 = bytes to fetch : 32 byte fetch = 4 pointers
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 1;           // 2^stride * 16 = sizeof(data) = 32 byte stride
        ep_qctrl1.field_len1 = 3;           // Bit length of field
        ep_qctrl1.field_loc1 = 7;           // Starting bit position of field
        ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = TRUE;        // Enables back to back
        ep_qctrl1.b2b_group  = 1;           // BUG 9349
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl1, ep_qctrl1.dw );

        ep_qctrl2.dw = 0;
        ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
        ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl2, ep_qctrl2.dw );

        ep_qctrl3.dw = 0;
        ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
        ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
        ep_qctrl3.field_loc2 = 0;           // Starting bit position of field: CID starts at bit 0
        ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl3.src2       = 0;           // Source memory port
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; // FIXME TODO @@@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL].ep_qctrl3, ep_qctrl3.dw );
#endif /* BE_CONFIG==6 */

        /* IQ0 IPC */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = 0;
        ep_qctrl0.length1    = 0;           // Second fetch is disabled
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_IPC].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
        ep_qctrl1.field_len1 = 0;           // Bit length of field
        ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
        ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = TRUE;        // Enables back to back
        ep_qctrl1.b2b_group  = 1;           // BUG 9349
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_IPC].ep_qctrl1, ep_qctrl1.dw );

        ep_qctrl2.dw = 0;
        ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
        ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_IPC].ep_qctrl2, ep_qctrl2.dw );

        ep_qctrl3.dw = 0;
        ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
        ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
        ep_qctrl3.field_loc2 = 32;          // Starting bit position of field: CID starts at bit 32
        ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl3.src2       = 0;           // Source memory port
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) - 44)/4; //-  BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_IPC].ep_qctrl3, ep_qctrl3.dw );

        /* IQ2 DMA response */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = 0;
        ep_qctrl0.length1    = 0;           // Second fetch is disabled
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_DMARESP].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
        ep_qctrl1.field_len1 = 0;           // Bit length of field
        ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
        ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = TRUE;        // Enables back to back
        ep_qctrl1.b2b_group  = 1;           // BUG 9349
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_DMARESP].ep_qctrl1, ep_qctrl1.dw );

        ep_qctrl2.dw = 0;
        ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
        ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_DMARESP].ep_qctrl2, ep_qctrl2.dw );

        ep_qctrl3.dw = 0;
        ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
        ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
        ep_qctrl3.field_loc2 = 48;          // Starting bit position of field: CID starts at bit 48
        ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl3.src2       = 0;           // Source memory port
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_DMARESP].ep_qctrl3, ep_qctrl3.dw );

#if (BE_CONFIG==7)
        /* STD FCOE IQ3 Protocol Selector 2 */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, fc_icd) / 16;
        ep_qctrl0.length1    = sizeof(FC_ICD_CONTEXT) / 16;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL2].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 2;           // 2^stride * 16 = sizeof(data)
        ep_qctrl1.field_len1 = 12;          // Bit length of field: ICD is 12 bits
        ep_qctrl1.field_loc1 = 32;          // Starting bit position of field: ICD starts at bit 32
        ep_qctrl1.dst1       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = 0;        // Disables back to back
        ep_qctrl1.b2b_group  = 0;           // BUG 9349
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL2].ep_qctrl1, ep_qctrl1.dw );

        ep_qctrl2.dw = 0;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL2].ep_qctrl2, ep_qctrl2.dw );

        ep_qctrl3.dw = 0;
         // Workaround for bug 9349
         // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
#else

        /* IQ 3 PSEL2 - Workaround for bug 9349 */
        ep_qctrl3.dw = 0;
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_PSEL2].ep_qctrl3, ep_qctrl3.dw );
#endif

        /* IQ 5 TXP - Workaround for bug 9349 */
        ep_qctrl3.dw = 0;
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_TXPP].ep_qctrl3, ep_qctrl3.dw );


        /* IQ6 SIQ1 */
        ep_qctrl0.dw = 0;
        ep_qctrl0.offset1    = 0;
        ep_qctrl0.length1    = 0;           // Second fetch is disabled
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_SIQ1].ep_qctrl0, ep_qctrl0.dw );

        ep_qctrl1.dw = 0;
        ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
        ep_qctrl1.field_len1 = 0;           // Bit length of field
        ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
        ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl1.src1       = 0;           // Source memory port
        ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
        ep_qctrl1.b2b        = TRUE;        // Enables back to back
        ep_qctrl1.b2b_group  = 1;           // BUG 9349
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_SIQ1].ep_qctrl1, ep_qctrl1.dw );

        ep_qctrl2.dw = 0;
        ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
        ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_SIQ1].ep_qctrl2, ep_qctrl2.dw );

        ep_qctrl3.dw = 0;
        ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
        ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
        ep_qctrl3.field_loc2 = 0;          // Starting bit position of field: CID starts at bit 0
        ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
        ep_qctrl3.src2       = 0;           // Source memory port
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_SIQ1].ep_qctrl3, ep_qctrl3.dw );

        /* IQ 7 SIQ2 - Workaround for bug 9349 */
        ep_qctrl3.dw = 0;
        // Always fetch start and len in 4 byte quantities
        ep_qctrl3.af_start   = 44/4; //FIXME TODO @@@ BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr) / 4;
        ep_qctrl3.af_len = (sizeof(TCP_CONTEXT) -  44)/4; //BE_GEN_FIELD_OFFSET(TCP_CONTEXT, delack_tmr)) / 4;
        BE_GEN_CSR_WRITE( function_context, tpost.ep.input_queue[TPOST_IQ_SIQ2].ep_qctrl3, ep_qctrl3.dw );


#if !defined(SIMULATION)
        if (is_asic) {
            int i;

            TPOST_JOURNAL_WB_HO_PHY_ADDR_CR tpost_journal_wb_ho_phy_addr = { 0 };
            TPOST_JOURNAL_WB_XLAT_TABLE_ADDR_CR tpost_journal_wb_xlat_table_addr = { 0 };

            BE_GEN_CSR_WRITE(function_context, tpost.tpost_journal_wb_ho_phy_addr, tpost_journal_wb_ho_phy_addr.dw);

            for (i=0;i<512; i++) {
                tpost_journal_wb_xlat_table_addr.tpost_journal_addr_xlat_table_addr = i;
                BE_GEN_CSR_WRITE(function_context, tpost.tpost_journal_wb_xlat_table_addr, tpost_journal_wb_xlat_table_addr.dw);
                BE_GEN_CSR_WRITE(function_context, tpost.tpost_journal_wb_xlat_table_rwdata, tpost_journal_wb_ho_phy_addr.dw);
            }
        }
#endif

    }
    /************************************************************************
     * END TPOST REGISTERS SECTION
     ************************************************************************/


    /************************************************************************
     * BEGIN TPRE REGISTERS SECTION
     ************************************************************************/
    {
    BE_GEN_DEBUG_PRINT ("BE GEN: Initializing TPRE registers\n");

#if 0
    /* Program preassigned context memory area for ipc overflow */
    ipc_ov_csr.dw = 0;
    ipc_ov_csr.base_addr = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, ovflow[1]);
    ipc_ov_csr.amount    = BE_GEN_TO_U16(BE_GEN_SIZEOF_FIELD(CTX_MEMORY_MAP, ovflow[1]) / 16);
    BE_GEN_CSR_WRITE(function_context, tpre.ep.ep_ipc_ov, ipc_ov_csr.dw);
#endif

    /* Indicate memory interface 2 is the RxPb/PMEM interface */
    ep_control.dw = BE_GEN_CSR_READ( function_context, tpre.ep.ep_control );
    ep_control.m2_RxPbuf = 1 ;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_control, ep_control.dw );

    /* Output queue credit registers */
    ep_qcreditA.dw      = 0;
    ep_qcreditA.oq0     = TPRE_EP_QCREDITA_CSR_OQ0;
    ep_qcreditA.oq1     = 2; //Workaround for vinq_underflow - Prash TPRE_EP_QCREDITA_CSR_OQ1;
    ep_qcreditA.oq2     = TPRE_EP_QCREDITA_CSR_OQ2;
    ep_qcreditA.oq3     = TPRE_EP_QCREDITA_CSR_OQ3;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_qcreditA, ep_qcreditA.dw );

    ep_qcreditB.dw      = 0;
    ep_qcreditB.oq4     = TPRE_EP_QCREDITA_CSR_OQ4;
    ep_qcreditB.oq5     = TPRE_EP_QCREDITA_CSR_OQ5;
    ep_qcreditB.oq6     = TPRE_EP_QCREDITA_CSR_OQ6;
    ep_qcreditB.oq7     = TPRE_EP_QCREDITA_CSR_OQ7;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_qcreditB, ep_qcreditB.dw );

    ep_qcreditC.dw      = 0;
    ep_qcreditC.oq8     = TPRE_EP_QCREDITA_CSR_OQ8;
    ep_qcreditC.oq9     = TPRE_EP_QCREDITA_CSR_OQ9;
    ep_qcreditC.oq10    = TPRE_EP_QCREDITA_CSR_OQ10;
    ep_qcreditC.oq11    = TPRE_EP_QCREDITA_CSR_OQ11;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_qcreditC, ep_qcreditC.dw );

    ep_qcreditD.dw      = 0;
    ep_qcreditD.oq12    = TPRE_EP_QCREDITA_CSR_OQ12;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_qcreditD, ep_qcreditD.dw );

    /* Input queue weight register */
    ep_iq_weight.dw      = 0;
    ep_iq_weight.iq0     = 0xF;
    ep_iq_weight.iq1     = 1;
    ep_iq_weight.iq2     = 1;
    ep_iq_weight.iq3     = 1;
    ep_iq_weight.iq4     = 1;
    ep_iq_weight.iq5     = 1;
    ep_iq_weight.iq6     = 1;
    ep_iq_weight.iq7     = 0xF;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.ep_input_que_weight, ep_iq_weight.dw );

    /* Input queue configuration */

    /* IQ from RXPP */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;
    ep_qctrl0.length1    = 16;          // length * 8 = bytes to fetch
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_0].ep_qctrl0, ep_qctrl0.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_1].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 9;           // Bit length of field
    ep_qctrl1.field_loc1 = 32;          // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 2;           // Source memory port: pbuf
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = TRUE;        // Enables back to back
    ep_qctrl1.b2b_group  = 2;           // Use B2B group ID of 2
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_0].ep_qctrl1, ep_qctrl1.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_1].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;    // CTX TOE start addr
    ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;      // CTX Len is multiples of 16
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_0].ep_qctrl2, ep_qctrl2.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_1].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data) = 128 = TCP ctxt size
    ep_qctrl3.field_len2 = 10;          // Bit length of field: CID is 10 bits
    ep_qctrl3.field_loc2 = 3;          // Starting bit position of field: CID starts at bit 3
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port
    // Always fetch start and len in 4 byte quantities
    ep_qctrl3.af_start   = 0;  // Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len = 11; // always fetch till tpost timer area

    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_0].ep_qctrl3, ep_qctrl3.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_RCV_ULP_1].ep_qctrl3, ep_qctrl3.dw );

    /* IQ0 IPC */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;
    ep_qctrl0.length1    = 0;           // Second fetch is disabled
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_IPC].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 0;           // Bit length of field
    ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 0;           // Source memory port
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = TRUE;        // Enables back to back
    ep_qctrl1.b2b_group  = 2;           // Use B2B group ID of 2
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_IPC].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
    ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_IPC].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
    ep_qctrl3.field_len2 = 16;          // Bit length of field: CID is 16 bits
    ep_qctrl3.field_loc2 = 32;          // Starting bit position of field: CID starts at bit 32
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port
    // Always fetch start and len in 4 byte quantities
    ep_qctrl3.af_start   = 0;  // Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len = 11; // always fetch till tpost timer area
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_IPC].ep_qctrl3, ep_qctrl3.dw );

    /* IQ4 Timer */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;
    ep_qctrl0.length1    = 0;           // Second fetch is disabled
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_TIMER].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 0;           // Bit length of field
    ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 0;           // Source memory port
    ep_qctrl1.lb_cfg     = 0;           // Single entrys (not chained)
    ep_qctrl1.b2b        = TRUE;        // Enables back to back
    ep_qctrl1.b2b_group  = 2;           // Use B2B group ID of 2
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_TIMER].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
    ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_TIMER].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data)
    ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
    ep_qctrl3.field_loc2 = 0;           // Starting bit position of field: CID starts at bit 0
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port
    // Always fetch start and len in 4 byte quantities
    ep_qctrl3.af_start   = 0;  // Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len = 11; // always fetch till tpost timer area
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_TIMER].ep_qctrl3, ep_qctrl3.dw );

#if (BE_CONFIG==0)    // LRO support only defined in CONFIG0 for now

    /*
     * IQ5 LRO (Large Receive Offload) -
     *
     *   TPRE IQ 5 is used for LRO.  An RXPP_TCP_D is delivered as the
     *   primary descriptor from RXF.  From this we perform two
     *   fetches:
     *
     *   The first fetch is for the first PBUF located in PMEM.  This
     *   allows FW to examine the packet headers.  This is indexed by
     *   the "pbuf_head" field of the RXPP_TCP_D which is 9 bits wide.
     *   The headers are loaded into the descriptor bank (following the
     *   RXPP_TCP_D).
     *
     *   The second fetch is the associated LRO_TCP_CONTEXT located in
     *   context memory which is indexed by the lowest 5 bits of the
     *   RXPP_TCP_D "sid" field.  This is loaded into the context
     *   bank.  Note that while HW produces a 6-bit "hash" we only use
     *   5 since we only have room for 32 LRO contexts in the Config0
     *   context memory map.
     */

    ep_qctrl0.dw = 0;                   // Start with a clean slate
    ep_qctrl0.offset1    = 0;           // PMEM is 0-origin
    ep_qctrl0.length1    = (128 / 8);   // length * 8 = fetch 128 bytes (PMEM granularity is 8 bytes)
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;                   // Start with a clean slate
    ep_qctrl1.stride1    = 0;           // PMEM fetches are special, don't use a stride
    ep_qctrl1.field_len1 = 9;           // Bit length of field
    ep_qctrl1.field_loc1 = 32;          // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 2;           // Source memory port: pbuf
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = TRUE;        // Enables back to back on secondary field fetches
    ep_qctrl1.b2b_group  = 5;           // Use queue number as B2B group ID
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;                   // Start with a clean slate
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, lroctx) / 16;
    ep_qctrl2.length2    = sizeof(LRO_TCP_CONTEXT) / 16;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;                   // Start with a clean slate
    ep_qctrl3.stride2    = BE_Log2(sizeof(LRO_TCP_CONTEXT)/16); // 2^stride * 16 = sizeof(data)
    ep_qctrl3.field_len2 = BE_Log2(BE_ContextCount(lroctx));    // LRO Context number bit count
    ep_qctrl3.field_loc2 = 3;           // Starting bit position for LRO CID
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc, 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port (0 = CTXT)
    ep_qctrl3.af_start   = 0;           // Start of AF area
    ep_qctrl3.af_len     = 11;          // Always-fetch first 44 bytes of LRO_TCP_CONTEXT
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl3, ep_qctrl3.dw );
#endif // (BE_CONFIG==0)

#if (BE_CONFIG == 7)
    /* IQ5 for FCoE */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;			// Second fetch is disabled.
    ep_qctrl0.length1    = 0;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 0;           // Bit length of field
    ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 0;           // Source memory port: pbuf
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = FALSE;       // Enables back to back
	ep_qctrl1.b2b_group  = 1;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = 0;
    ep_qctrl2.length2    = 16;          // length * 8 = bytes to fetch
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl3.field_len2 = 9;          	// Bit length of field
    ep_qctrl3.field_loc2 = 32;          // Starting bit position of field
    ep_qctrl3.dst2       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 2;           // Source memory port: pbuf
    // Always fetch start and len in 4 byte quantities
    ep_qctrl3.af_start   = 0;  			// Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len     = 11; 			// always fetch must be same in all queues

    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl3, ep_qctrl3.dw );
#endif

#if (BE_CONFIG == 7)
    /* IQ6 for FCoE */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;
    ep_qctrl0.length1    = 16;            // length * 8 = bytes to fetch
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_FCoE].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 9;           // Bit length of field
    ep_qctrl1.field_loc1 = 32;          // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 2;           // Source memory port: pbuf
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = FALSE;       // Enables back to back
	ep_qctrl1.b2b_group  = 1;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_FCoE].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, fc_ctx) / 16;    // CTX FC Initiator start addr
    ep_qctrl2.length2    = sizeof(FC_CONTEXT) / 16 - 1;  // CTX Len is multiples of 16 (only need 48 bytes for cfg 7)
														 // the remainder of the FC ctxt is reserved
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_FCoE].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 2;           // 2^stride * 16 = sizeof(data) = 64 = FC Initiator ctxt size
    ep_qctrl3.field_len2 = 6;          	// Bit length of field: CID is 10 bits
    ep_qctrl3.field_loc2 = 3;          	// Starting bit position of field: CID starts at bit 3
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port
    // Always fetch start and len in 4 byte quantities
    ep_qctrl3.af_start   = 0;  			// Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len     = 11; 			// always fetch must be same in all queues

    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_FCoE].ep_qctrl3, ep_qctrl3.dw );

    TPRE_VINQ_ULP_LIMITS_CR tpre_vinq_ulp_limits;
    /* Increasing the number of RXPP/TPRE FCoE descriptors from the default (0x80) to (0xb0)
     * in order that we do not drop frames  */
    tpre_vinq_ulp_limits.dw         = 0x100b0050; 			
    BE_GEN_CSR_WRITE( function_context, tpre.vinq_ulp_limits, tpre_vinq_ulp_limits.dw );
    
#endif

#if BE_CONFIG==6

    // DAL Configuration

    ep_qctrl0.dw = 0;                   // Start with a clean slate
    ep_qctrl0.offset1    = 0;           // PMEM is 0-origin
    ep_qctrl0.length1    = (128 / 8);   // length * 8 = fetch 128 bytes (PMEM granularity is 8 bytes)

    // BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_8021au].ep_qctrl0, ep_qctrl0.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl0, ep_qctrl0.dw );

    ep_qctrl1.dw = 0;                   // Start with a clean slate
    ep_qctrl1.stride1    = 0;           // PMEM fetches are special, don't use a stride
    ep_qctrl1.field_len1 = 9;           // Bit length of field
    ep_qctrl1.field_loc1 = 32;          // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 2;           // Source memory pbuf
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = FALSE;       // Enables back to back on secondary field fetches

    // BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_8021au].ep_qctrl1, ep_qctrl1.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl1, ep_qctrl1.dw );

    ep_qctrl2.dw = 0;                   // Start with a clean slate
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, dal_tpre_context) / 16;
    ep_qctrl2.length2    = sizeof(DAL_TPRE_CONTEXT) / 32;
    // BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_8021au].ep_qctrl2, ep_qctrl2.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl2, ep_qctrl2.dw );

    ep_qctrl3.dw = 0;                   // Start with a clean slate
    ep_qctrl3.stride2    = BE_Log2(sizeof(DAL_TPRE_CONTEXT)/16); // 2^stride * 16 = sizeof(data)
    ep_qctrl3.field_len2 = BE_Log2(BE_ContextCount(dal_tpre_context));    // MACID Context number bit count
    ep_qctrl3.field_loc2 = 3;          // Starting bit position for MACID
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc, 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port (0 = CTXT)
    ep_qctrl3.af_start   = 0;           // Start of AF area
    ep_qctrl3.af_len     = 11;          // Always-fetch first 44 bytes of LRO_TCP_CONTEXT
    // BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_8021au].ep_qctrl3, ep_qctrl3.dw );
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_LRO].ep_qctrl3, ep_qctrl3.dw );

    // BE_GEN_CSR_WRITE( function_context, rxf.dot1au_type, 0x0800);

#endif
    /* IQ7 SIQ */
    ep_qctrl0.dw = 0;
    ep_qctrl0.offset1    = 0;
    ep_qctrl0.length1    = 0;           // Second fetch is disabled
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_SIQ].ep_qctrl0, ep_qctrl0.dw);

    ep_qctrl1.dw = 0;
    ep_qctrl1.stride1    = 0;           // 2^stride * 16 = sizeof(data)
    ep_qctrl1.field_len1 = 0;           // Bit length of field
    ep_qctrl1.field_loc1 = 0;           // Starting bit position of field
    ep_qctrl1.dst1       = 0;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl1.src1       = 0;           // Source memory port
    ep_qctrl1.lb_cfg     = 0;           // Last bit config: single entries (not chained)
    ep_qctrl1.b2b        = TRUE;        // Enables back to back
    ep_qctrl1.b2b_group  = 2;           // Use B2B group ID of 2
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_SIQ].ep_qctrl1, ep_qctrl1.dw);

    ep_qctrl2.dw = 0;
    ep_qctrl2.offset2    = BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp) / 16;
    ep_qctrl2.length2    = sizeof(TCP_CONTEXT) / 16;
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_SIQ].ep_qctrl2, ep_qctrl2.dw);

    ep_qctrl3.dw = 0;
    ep_qctrl3.stride2    = 3;           // 2^stride * 16 = sizeof(data) = 128 (TCP ctxt size)
    ep_qctrl3.field_len2 = 11;          // Bit length of field: CID is 11 bits
    ep_qctrl3.field_loc2 = 0;           // Starting bit position of field: CID starts at bit 0
    ep_qctrl3.dst2       = 1;           // Destination bank (0 = desc | 1 = ctxt)
    ep_qctrl3.src2       = 0;           // Source memory port
    ep_qctrl3.af_start   = 0;  // Start of TX section of TCP_CONTEXT
    ep_qctrl3.af_len = 11; // always fetch till tpost timer area
    BE_GEN_CSR_WRITE( function_context, tpre.ep.input_queue[TPRE_IQ_SIQ].ep_qctrl3, ep_qctrl3.dw);


    /*
     * For iSCSI, use pd_id of 1 for now and disable dbcredit logic by setting the cost to zero
     * and credits to a non-zero value.
     *
     * ^^^^^^ THE COMMENT IS INCORRECT. But more than that, this is just a hack. I
     * moved the TOE-related programming into the BE2_Gen_Toerx_Registers_Init function.
     * That approach is likely the same thing we want to do for the iSCSI-related default
     * queue programming as it is BE_CONFIG-specific.
     * --Vahid
     */
    tpre_dbcred.dw = 0;
    tpre_dbcred.dbcred_data = 0;
    tpre_dbcred.dbcred_pdid = 23;
    tpre_dbcred.dbcred_cost_cred_n = 1;
    tpre_dbcred.dbcred_wr_en = 1;
    BE_GEN_CSR_WRITE( function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw );

    tpre_dbcred.dw = 0;
    tpre_dbcred.dbcred_data = 0xf;
    tpre_dbcred.dbcred_pdid = 23;
    tpre_dbcred.dbcred_cost_cred_n = 0;
    tpre_dbcred.dbcred_wr_en = 1;
    BE_GEN_CSR_WRITE( function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw );

#if (BE_CONFIG == 7)
#if ENABLE_PRIORITY_PAUSE
	/*
	 * For FCoE, program the number of credits and cost per frame. The constants
	 * FC_DBCRED_COST, FC_DBCRED_CREDITS, and FC_PORT_PDID_BASE are defined in
	 * fw/def/exp/fc_enums.sg.
	 */
	tpre_dbcred.dw = 0;
	tpre_dbcred.dbcred_data        = FC_DBCRED_COST;
	tpre_dbcred.dbcred_pdid        = FC_PORT_PDID_BASE + 0;
	tpre_dbcred.dbcred_cost_cred_n = 1;
	tpre_dbcred.dbcred_wr_en       = 1;
	BE_GEN_CSR_WRITE(function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw);

	tpre_dbcred.dw = 0;
	tpre_dbcred.dbcred_data        = FC_DBCRED_CREDITS;
	tpre_dbcred.dbcred_pdid        = FC_PORT_PDID_BASE + 0;
	tpre_dbcred.dbcred_cost_cred_n = 0;
	tpre_dbcred.dbcred_wr_en       = 1;
	BE_GEN_CSR_WRITE(function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw);

	tpre_dbcred.dw = 0;
	tpre_dbcred.dbcred_data        = FC_DBCRED_COST;
	tpre_dbcred.dbcred_pdid        = FC_PORT_PDID_BASE + 1;
	tpre_dbcred.dbcred_cost_cred_n = 1;
	tpre_dbcred.dbcred_wr_en       = 1;
	BE_GEN_CSR_WRITE(function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw);

	tpre_dbcred.dw = 0;
	tpre_dbcred.dbcred_data        = FC_DBCRED_CREDITS;
	tpre_dbcred.dbcred_pdid        = FC_PORT_PDID_BASE + 1;
	tpre_dbcred.dbcred_cost_cred_n = 0;
	tpre_dbcred.dbcred_wr_en       = 1;
	BE_GEN_CSR_WRITE(function_context, tpre.dbcred_cfg_data, tpre_dbcred.dw);
#endif
#endif
    /* TCP Timers configuration */

    /* Timer Scale Configuration
     * Timer 7 - Push Timer - 8ms Granularity
     * Timer 6 - Delayed ACK Timer - 8 ms Granularity
     * Timer 5 - FIN_WAIT_2 Timer - 4096 ms Granularity
     * Timer 4 - KA Timer - 256 ms Granularity
     * Timer 3 - Retransmit Timer - 256 ms granularity
     * Timer 2 - Tx Idle Timer - 128 ms Granularity
     * Timer 1 - KA Interval Timer - 128 ms Granularity
     * Timer 0 - Persist Timer - 128 ms granularity
     */
    tim_tcp_timer_scale.dw = 0;
    tim_tcp_timer_scale.timer7 = TIM_GRAN_8MS;
    tim_tcp_timer_scale.timer6 = TIM_GRAN_8MS;
    tim_tcp_timer_scale.timer5 = TIM_GRAN_4KMS;
    tim_tcp_timer_scale.timer4 = TIM_GRAN_256MS;
    tim_tcp_timer_scale.timer3 = TIM_GRAN_64MS;
    tim_tcp_timer_scale.timer2 = TIM_GRAN_128MS;
    tim_tcp_timer_scale.timer1 = TIM_GRAN_128MS;
    tim_tcp_timer_scale.timer0 = TIM_GRAN_128MS;
    BE_GEN_CSR_WRITE( function_context, tpre.tcp_timer_scale, tim_tcp_timer_scale.dw );


    /* Timer Expiration Configuration
     * Timer 7 - PUSH timer - Notify TPRE on expiration
     * Timer 6 - Delayed ACK Timer - Notify TPRE on expiration
     * Timer 5 - FIN_WAIT_2 Timer - Notify TPRE on expiration
     * Timer 4 - KA Timer - Notify TPOST and TPRE on expiration
     * Timer 3 - Retransmit Timer - Notify TPRE on expiration
     * Timer 2 - Tx Idle Timer - Notify TPOST on expiration
     * Timer 1 - KA Interval Timer - Notify TPOST on expiration
     * Timer 0 - Persist Timer - Notify TPOST on expiration
    */
    tim_timestamp_expire.dw = 0;
    /*
     * Disable timer block for config6 & config7.  TODO_arvindp: Find out if there's a
     * better way to do it than setting the steers to 0.
     */
#if (BE_CONFIG != 6)
    tim_timestamp_expire.tim_tcp_tpoststeer =
        TIM_EXPIRE_EN0 | TIM_EXPIRE_EN1 | TIM_EXPIRE_EN2 | TIM_EXPIRE_EN4;
    tim_timestamp_expire.tim_tcp_tpresteer =
        TIM_EXPIRE_EN0 | TIM_EXPIRE_EN2 | TIM_EXPIRE_EN3 | TIM_EXPIRE_EN4 | TIM_EXPIRE_EN5 |
        TIM_EXPIRE_EN6 | TIM_EXPIRE_EN7;
#endif /* BE_CONFIG != 6 */
    if (TRUE == is_asic) {
        /* For 10ms accuracy on chip */
        tim_timestamp_expire.tim_tcp_interval = 0x0131;
    } else {
        tim_timestamp_expire.tim_tcp_interval = 0x007d;
    }
    BE_GEN_CSR_WRITE( function_context, tpre.timestamp_expire, tim_timestamp_expire.dw );

    /* 1 ms timestamp control */
    tim_1ms_timestamp_cr.dw = 0;
    if (TRUE == is_asic) {
        /* Timer integration and divide parameters */
        tim_1ms_timestamp_cr.tim_1ms_ts_divide = 0x3a;
        tim_1ms_timestamp_cr.tim_1ms_ts_integ = 0x10;
        BE_GEN_CSR_WRITE( function_context, tpre.timestamp_1ms_control, tim_1ms_timestamp_cr.dw );
    } else {
        /* Timer integration and divide parameters */
        tim_1ms_timestamp_cr.tim_1ms_ts_divide = 0x0;
        tim_1ms_timestamp_cr.tim_1ms_ts_integ = 0x1a;
        BE_GEN_CSR_WRITE( function_context, tpre.timestamp_1ms_control, tim_1ms_timestamp_cr.dw );
    }

    /* Set the TCP base address. The hardware looks at the TCP context to get the timer
     * values. By design, a 0 offset value refers to 16-byte offset 0x3 of the context.
     * The timer values are at 16-byte offset of 0x2, which means we need a -1 offset value.
     * Note: TPRE_TCP_CM_ADDR_CR.tcp_state_rxoffst value is not used!
     */
    BE_GEN_CSR_WRITE( function_context, tpre.tcp_cm_addr, (u32)(BE_GEN_TO_U16(BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, tcp)/16) - 1) );

#if (BE_CONFIG==0)  // LRO support only defined in CONFIG0 for now

    /*
     * Setup LRO Timer Configuration -
     */

    /* 3.2us timestamp control */

    tim_3_2us_timestamp_cr.dw = 0;

    if (TRUE == is_asic) {   // Magic numbers for ASIC and FPGA from Randy + Vijay
      tim_3_2us_timestamp_cr.tim_3_2us_ts_integ         = 512;
      tim_3_2us_timestamp_cr.tim_3_2us_ts_divide        = 125;
      tim_3_2us_timestamp_cr.tim_lro_tpre_timerpopdelay = 16;
    } else {
      tim_3_2us_timestamp_cr.tim_3_2us_ts_integ         = 0x800;
      tim_3_2us_timestamp_cr.tim_3_2us_ts_divide        = 0x200;
      tim_3_2us_timestamp_cr.tim_lro_tpre_timerpopdelay = 0;
      // Saved for debug purposes.
      //tim_3_2us_timestamp_cr.tim_3_2us_ts_integ         = 128;
      //tim_3_2us_timestamp_cr.tim_3_2us_ts_divide        = 5;
      //tim_3_2us_timestamp_cr.tim_lro_tpre_timerpopdelay = 0;  // Still need this number
    }

    BE_GEN_CSR_WRITE( function_context, tpre.timestamp_3_2us_control, tim_3_2us_timestamp_cr.dw );

    /* LRO Timer Scales and Enables */

    BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale0, 0x49249249 );  // Set up a time scale
    BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale1, 0x92492492 );  // ..of 3.2us for LRO Timers 0-31
    BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale2, 0x24924924 );  // ..(value = 3'b001)

    switch (BE_ContextCount(lroctx)) {
    default:
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timestamp_expire0,
            (1 << BE_ContextCount(lroctx)) - 1);
      break;
    case 64:
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale3, 0x49249249 );  // Set up a time scale
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale4, 0x92492492 );  // ..of 3.2us for LRO Timers 32-63
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timer_scale5, 0x24924924 );  // ..(value = 3'b001)
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timestamp_expire1, 0xFFFFFFFF );
    case 32:
      BE_GEN_CSR_WRITE( function_context, tpre.lro_timestamp_expire0, 0xFFFFFFFF );
    }

#endif                              // (BE_CONFIG==0)

    /* Virtual Queue CrossBar configuration for CONFIG 0 - NEEDS TO GO INTO CONFIG INIT */
    /* RXULP0 output to RXULP 0 input and FCOE input */
    /* RXULP1 output to RXULP 1 input and LRO input */
    {
        TPRE_OQ_CTL_CR oq_ctl ;
        oq_ctl.dw = 0 ;

        oq_ctl.voq_ulp0_wrr = 1;
        oq_ctl.voq_ulp1_wrr = 1;

        oq_ctl.ack_hold_thresh = 6;

        oq_ctl.voq_fcmap_u0 = 0x11;
        oq_ctl.voq_fcmap_u1 = 0xa;

        BE_GEN_CSR_WRITE( function_context, tpre.oq_control, oq_ctl.dw );
    }

    }
    /************************************************************************
     * END TPRE REGISTERS SECTION
     ************************************************************************/

    /* Execute configuration-specific ULP register initialization (these
     * functions are declared within a BE configuration header file in
     * fw/config directory).
     */
    BE2_Gen_Txulp0_Registers_Init(function_context);
    BE2_Gen_Txulp1_Registers_Init(function_context);
    BE2_Gen_Rxulp0_Registers_Init(function_context);
    BE2_Gen_Rxulp1_Registers_Init(function_context);

    /*************************************************************************
     * BEGIN WCON config programming section
     *************************************************************************/
    {
    RXULP_WCON_CFG1_CR wcon_cfg1;
    RXULP_WCON_CFG2_CR wcon_cfg2;

    /* WCON config, command base is the same as cid */
    wcon_cfg1.dw = 0;
    wcon_cfg1.BMbaseAddr      = (BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, wcon_bm_cc[0]) / 4);
    wcon_cfg1.BMstride        = WCON_BM_STRIDE_4BYTES;
    wcon_cfg1.LenViolPolicy   = SGL_VIOL_IS_PE;
    wcon_cfg1.pyldSwapMode    = DMA_SWAP_NONE;
    wcon_cfg1.complSwapMode   = DMA_SWAP_WORD;
    wcon_cfg1.UCreqLim        = 0;
    BE_GEN_CSR_WRITE( function_context, wcon.wcon_cfg1, wcon_cfg1.dw );

    /* WCON config, page mode commands are teh same as block mode commands for TOE */
    wcon_cfg2.dw = 0;
#if (BE_CONFIG==0)
    wcon_cfg2.PMbaseAddr      = ((BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, wcon_pm_cc[0]) - (CFG0_ISCSI_ICDS*8)) / 8);
#elif ( (BE_CONFIG == 3) || (BE_CONFIG == 9) )
    wcon_cfg2.PMbaseAddr      = ((BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, wcon_pm_cc[0]) - (CFG3_ISCSI_ICDS*8)) / 8);
#else
    wcon_cfg2.PMbaseAddr      = (BE_GEN_FIELD_OFFSET(CTX_MEMORY_MAP, wcon_pm_cc[0]) / 8);
#endif
    wcon_cfg2.PMstride        = WCON_PM_STRIDE_8BYTES;
    BE_GEN_CSR_WRITE( function_context, wcon.wcon_cfg2, wcon_cfg2.dw );
    }

    /*************************************************************************
     * END WCON config programming section
     *************************************************************************/

    BE_GEN_DEBUG_PRINT ("BE GEN: End registers initialization\n");
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Init_PHYs
//
//   Performs initialization of the PHY registers.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Init_PHYs (
   IN void* function_context )
{
    u8 port;
#ifdef BE_ARM_FIRMWARE
    extern int g_nemo_enabled;
#endif

    BE_GEN_DEBUG_PRINT ("BE GEN: Begin initializing PHY registers\n");

    for (port = 0; port < 2; port++) {
        u32 xgmac_data;
#if !defined(SIMULATION) && !defined(BE_EMULATION)
        u16 phy_reg_data;

        // Clear PCS TX faults
        phy_reg_data = 0xFFFF;
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 4, &phy_reg_data, TRUE);

        // Clear PCS RX faults
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 6, &phy_reg_data, TRUE);

        // Clear PCS link status, link fault, and sync
        phy_reg_data = 0xC080;
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 2, &phy_reg_data, TRUE);

        // Reset them all to 0
        phy_reg_data = 0;
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 4, &phy_reg_data, TRUE);
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 6, &phy_reg_data, TRUE);
        BE2_Gen_Access_Internal_PHY_Reg (function_context, port, 30, 2, &phy_reg_data, TRUE);

#endif // !defined(SIMULATION)

#ifdef BE_ARM_FIRMWARE
        if (g_nemo_enabled) {
            // Do not write to the MAC registers
            BE_GEN_DEBUG_PRINT ("BE GEN: Skip PHY%d registers\n", port);
            continue;
        }
#endif
        // Initialize the MAC registers with magic fairy dust. See the AXGMAC PDF
        // documentation for register decoding values.
        xgmac_data = 0;
        BE2_Gen_Access_XGMAC_Table (function_context, port, 0, &xgmac_data, TRUE);
        xgmac_data = 0xf00000a5;
        BE2_Gen_Access_XGMAC_Table (function_context, port, 1, &xgmac_data, TRUE);
        // Bug 10277
        // Set IPG to maximum. Also disable automatic remote fault generation
        // when a receive fault is detected.
        //
        // update: 12/15/2008 - VB
        // set ipg extension to 0 to get back performance
        // we gave trying to avoid the crc errors in our unusual lab
        // config case of back-back network connections.
        //
        //xgmac_data = 0x040d0000;
        xgmac_data = 0x040d000d;
        BE2_Gen_Access_XGMAC_Table (function_context, port, 2, &xgmac_data, TRUE);
        xgmac_data = 0x0418fbe7;
        BE2_Gen_Access_XGMAC_Table (function_context, port, 3, &xgmac_data, TRUE);

        // Program the max frame size
        xgmac_data = (BE2_GEN_TX_MAX_FRAME_SIZE / 4 << 16) | BE2_GEN_RX_MAX_FRAME_SIZE;
        BE2_Gen_Access_XGMAC_Table (function_context, port, 8, &xgmac_data, TRUE);

        // Set the MDC clock divider
        if (BE2_Gen_Is_ASIC (function_context)) {
            xgmac_data = 0x3e;
        } else {
            xgmac_data = 0x6;
        }
        BE2_Gen_Access_XGMAC_Table (function_context, port, 18, &xgmac_data, TRUE);
    }

    BE_GEN_DEBUG_PRINT ("BE GEN: End initializing PHY registers\n");
    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Enable_UEs
//
//   Unmasks and enables unrecoverable errors sources.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Enable_UEs (
   IN void* function_context )
{
    union {
        struct {
            MPU_UE_ENABLE ue_enable;
            MPU_UE_ENABLE1 ue_enable1;
        } mpu;

        struct {
            struct {
                JELLM_UE_ENABLE_CSR uee;
            } jellm;
            TPOST_UE_ENABLE_CR tpost_ue_enable;
            TPOST_PROTSEL_UE_ENABLE_CR tpost_protsel_ue_enable;
            TPOST_UE_ENABLE_1_CR tpost_ue_enable_1;
        } tpost;

        struct {
            struct {
                JELLM_UE_ENABLE_CSR uee;
            } jellm;
            TPRE_UE1_SOURCE_SR tpre_ue1_enable;
            TPRE_UE2_SOURCE_SR tpre_ue2_enable;
        } tpre;

        struct {
            CEV_UE_ENABLE_CSR ue_enable;
        } cev;

        struct {
            RXULP_UE_ENABLE ue_enable;
            RXULP_UE_ENABLE1 ue_enable1;
            RXULP_VQUEUE_UE_ENABLE1 vqueue_ue_enable1;
        } rxulp0;

        struct {
            RXULP_UE_ENABLE ue_enable;
            RXULP_UE_ENABLE1 ue_enable1;
            RXULP_VQUEUE_UE_ENABLE1 vqueue_ue_enable1;
        } rxulp1;

        struct {
            WCON_UE_ENABLE_CSR ue_enable;
        } wcon;

        struct {
            TXULP_EP_AND_IPC_UE_ENABLE ep_and_ipc_ue_enable;
            TXULP_CONSEL_ARB_UE_ENABLE consel_arb_ue_enable;
            TXULP_CONSEL_SEL_UE_ENABLE consel_sel_ue_enable;
            TXULP_SCHEDULER_UE_ENABLE_0 scheduler_ue_enable0;
            TXULP_SCHEDULER_UE_ENABLE_1 scheduler_ue_enable1;
            TXULP_RING_MANAGER_UE_ENABLE_0 ring_man_ue_enable_0;
            TXULP_RING_MANAGER_UE_ENABLE_1 ring_man_ue_enable_1;
            TXULP_UC_MERGE_UE_ENABLE uc_merge_ue_enable;
            GLOBAL_QOS_UE_ENABLE qos_ue_enable;
        } txulp0;

        struct {
            TXULP_EP_AND_IPC_UE_ENABLE ep_and_ipc_ue_enable;
            TXULP_CONSEL_ARB_UE_ENABLE consel_arb_ue_enable;
            TXULP_CONSEL_SEL_UE_ENABLE consel_sel_ue_enable;
            TXULP_SCHEDULER_UE_ENABLE_0 scheduler_ue_enable0;
            TXULP_SCHEDULER_UE_ENABLE_1 scheduler_ue_enable1;
            TXULP_RING_MANAGER_UE_ENABLE_0 ring_man_ue_enable_0;
            TXULP_RING_MANAGER_UE_ENABLE_1 ring_man_ue_enable_1;
            TXULP_UC_MERGE_UE_ENABLE uc_merge_ue_enable;
            GLOBAL_QOS_UE_ENABLE qos_ue_enable;
        } txulp1;

        struct {
            UC_RE_ENABLE_CSR re_enable;
            UC_UE_ENABLE_CSR ue_enable;
        } uc;

        struct {
            PTC_UE_ENABLE ue_enable;
        } ptc;

        struct {
            TXP_UE_ENABLE0_CSR txp_ue_enable0;
            TXP_UE_ENABLE1_CSR txp_ue_enable1;
            TXP_UE_ENABLE2_CSR txp_ue_enable2;
            TXP_UE_ENABLE3_CSR txp_ue_enable3;
        } txp;

        struct {
            NDMA_UE_ENABLE_CSR ue_enable;
        } ndma;

        struct {
            RDMA_UE_ENABLE_CSR ue_enable;
        } rdma;

        struct {
            WDMA_UE_ENABLE_CSR ue_enable;
        } wdma;

        struct {
            TXPB_UE_ENABLE ue_enable;
        } txpb;

        struct {
            CTX_UE_ENABLE_CSR ue_enable;
        } ctx;

        struct {
            DBUF_UE_ENABLE_CSR ue_enable;
        } dbuf;

        struct {
            PMEM_UE_ENABLE ue_enable;
        } pmem;

        struct {
            RXF_UE_ENABLE ue_enable;
        } rxf;

        struct {
            ERX_UE_ENABLE ue_enable;
        } erx;
    } enables;

    PCICFG_UE_STATUS_LOW_MASK_CSR ue_status_low_mask = { 0 };
    PCICFG_UE_STATUS_HI_MASK_CSR ue_status_hi_mask = { 0 };
    u32 is_asic;
    u32 chip_rev;

    is_asic = BE2_Gen_Is_ASIC(function_context);
    chip_rev = BE2_Gen_Chip_Rev(NULL);

    /* Mask all UEs to begin with (last step will be to unmask) */
    BE_GEN_PCI0_WRITE(function_context, ue_status_low_mask, 0xFFFFFFFF);
    BE_GEN_PCI0_WRITE(function_context, ue_status_hi_mask, 0xFFFFFFFF);

    /* Disable RXF UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if (is_asic && (0 == chip_rev)) {
        // asic A0 RXF UEs
        enables.rxf.ue_enable.mac_perr_ue_en = 0;
        enables.rxf.ue_enable.vlan_perr_ue_en = 0;
        enables.rxf.ue_enable.conn_perr_ue_en = 0;
    } else if (is_asic && (0 != chip_rev)) {
        // All other asic RXF UEs
        /* Bug 13265: A1 ECO (Bug 10026) not working on few boards */
        enables.rxf.ue_enable.conn_perr_ue_en = 0;
        /* Bug 27520: This UE should never occur given the speeds of clk_core_d3 and
         * clk_mgmt ,but the way this UE is synced into clk_core domain is incorrect.
         */
        enables.rxf.ue_enable.mgmt_tx_syncfifo_overflow_ue_en = 0;
    } else if (is_asic == FALSE) {
        // FPGA RXF UEs
        enables.rxf.ue_enable.tsv0_syncfifo_overflow_ue_en = 0;
        enables.rxf.ue_enable.tsv0_syncfifo_underflow_ue_en = 0;
        enables.rxf.ue_enable.tsv1_syncfifo_overflow_ue_en = 0;
        enables.rxf.ue_enable.tsv1_syncfifo_underflow_ue_en = 0;
    }

    /* BUG 15064 - Multiple 64 byte packets followed by 1 large 1514 byte packet
    *  may overflow the RXF fifos if these are true:
    * (1) packets are sent with minimum inter-frame gap. Typically this is only
    *     from a switch that has buffered packets during a flow control pause.
    * (2) 1 vlan tag is enabled (if 2 VLAN tags are present the minimum size packet is
    *     large enough to avoid the problem).
    * (3) packets are TCP and the connection CAM lookup occurs.
    * Masking the UE will cause some packets to be dropped at this fifo until a
    * large packet arrives to resync it.
    */
    enables.rxf.ue_enable.mac1_syncfifo_overflow_ue_en = 0;
    enables.rxf.ue_enable.mac0_syncfifo_overflow_ue_en = 0;

    BE_GEN_CSR_WRITE(function_context, rxf.ue_source, enables.rxf.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rxf.ue_enable, enables.rxf.ue_enable.dw);


    /* Disable MPU UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    enables.mpu.ue_enable.arm_ddr2 = 0;
    if (is_asic && (0 == chip_rev))
    {
        // A0 MPU UEs
        enables.mpu.ue_enable.rr_ue_nomatch = 0;    //workaround for BUG 8424
        enables.mpu.ue_enable.sram_access_during_init_bank_x_ctx = 0;
        enables.mpu.ue_enable.sram_access_during_init_bank_x_desc = 0;
        enables.mpu.ue_enable.sram_access_during_init_bank_y_ctx = 0;
        enables.mpu.ue_enable.sram_access_during_init_bank_y_desc = 0;
        enables.mpu.ue_enable.sram_access_during_init_shadow_ram = 0;
    }
    else if(FALSE == is_asic)
    {
        // Parity UEs on FPGA not hooked-up
        enables.mpu.ue_enable.ctx_parity_ue = 0;
        enables.mpu.ue_enable.pmem_parity_ue = 0;
        enables.mpu.ue_enable.txpb_parity_ue = 0;
    }
    BE_GEN_CSR_WRITE(function_context, mpu.ue_source, enables.mpu.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, mpu.ue_enable, enables.mpu.ue_enable.dw);

    BE_GEN_CSR_WRITE(function_context, mpu.ue_source1, enables.mpu.ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, mpu.ue_enable1, enables.mpu.ue_enable1.dw);

    /* Disable TPOST UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.tpost.tpost_ue_enable.tpost_ctx_parity_error_enable = 0;
        enables.tpost.tpost_ue_enable.tpost_txpb_parity_error_enable = 0;
    }
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_ue_source, enables.tpost.tpost_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_ue_enable, enables.tpost.tpost_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.jellm.ues, enables.tpost.jellm.uee.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.jellm.uee, enables.tpost.jellm.uee.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_protsel_ue_source, enables.tpost.tpost_protsel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_protsel_ue_enable, enables.tpost.tpost_protsel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_ue_source_1, enables.tpost.tpost_ue_enable_1.dw);
    BE_GEN_CSR_WRITE(function_context, tpost.tpost_ue_enable_1, enables.tpost.tpost_ue_enable_1.dw);

    /* Disable TPRE UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.tpre.tpre_ue1_enable.ctx_bus_perr = 0;
        enables.tpre.tpre_ue1_enable.pmem_bus_perr = 0;
    }

    //Bug 10378 and 9914 are pushed to BE3
#if ( (BE_CONFIG==0)  || (BE_CONFIG==7) || (BE_CONFIG==9) ) // BUG 10378
    enables.tpre.tpre_ue1_enable.class_iram_perr = 0;
#endif
#if ( (BE_CONFIG==0)  || (BE_CONFIG==7) || (BE_CONFIG==9) ) // BUG #9914 
    enables.tpre.tpre_ue1_enable.vinq_wr_create_same_time = 0;
#endif
#if (BE_CONFIG==7)
#if ENABLE_PRIORITY_PAUSE
    /* Disable DBCRED module UE, since we are going to use the module's ALU to do signed
     * arithmetic. This causes "overflows" that we have to ignore.
     */
    enables.tpre.tpre_ue2_enable.dbcred_ue = 0;
#endif
#endif
    BE_GEN_CSR_WRITE(function_context, tpre.jellm.ues, enables.tpre.jellm.uee.dw);
    BE_GEN_CSR_WRITE(function_context, tpre.jellm.uee, enables.tpre.jellm.uee.dw);
    BE_GEN_CSR_WRITE(function_context, tpre.tpre_ue1_source, enables.tpre.tpre_ue1_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpre.tpre_ue1_enable, enables.tpre.tpre_ue1_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpre.tpre_ue2_source, enables.tpre.tpre_ue2_enable.dw);
    BE_GEN_CSR_WRITE(function_context, tpre.tpre_ue2_enable, enables.tpre.tpre_ue2_enable.dw);

    /* Disable CEV UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     *
     * cev.ue_enable.TooManyPops : disabled as it triggers a false UE on signed
     *      extension/comparison (bug 9387).
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.cev.ue_enable.ctx_perr = 0;
    }
    enables.cev.ue_enable.TooManyPops = 0;
    BE_GEN_CSR_WRITE(function_context, cev.ue_cause, enables.cev.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, cev.ue_enable, enables.cev.ue_enable.dw);

    /* Disable RXULP0 UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.rxulp0.ue_enable.ctx_parity_ue = 0;
        enables.rxulp0.ue_enable.pmem_parity_ue = 0;
    }
    BE_GEN_CSR_WRITE(function_context, rxulp0.ue_source, enables.rxulp0.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp0.ue_enable, enables.rxulp0.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp0.ue_source1, enables.rxulp0.ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp0.ue_enable1, enables.rxulp0.ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp0.vqueue_ue_source1, enables.rxulp0.vqueue_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp0.vqueue_ue_enable1, enables.rxulp0.vqueue_ue_enable1.dw);

    /* Disable RXULP1 UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.rxulp1.ue_enable.ctx_parity_ue = 0;
        enables.rxulp1.ue_enable.pmem_parity_ue = 0;
    }
    BE_GEN_CSR_WRITE(function_context, rxulp1.ue_source, enables.rxulp1.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp1.ue_enable, enables.rxulp1.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp1.ue_source1, enables.rxulp1.ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp1.ue_enable1, enables.rxulp1.ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp1.vqueue_ue_source1, enables.rxulp1.vqueue_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, rxulp1.vqueue_ue_enable1, enables.rxulp1.vqueue_ue_enable1.dw);

    /* Disable WCON UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.wcon.ue_enable.wcon_ctx_perr = 0;
    }
    enables.wcon.ue_enable.wcon_d_parse = 0;
    BE_GEN_CSR_WRITE(function_context, wcon.ue_source, enables.wcon.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, wcon.ue_enable, enables.wcon.ue_enable.dw);

    /* Disable TXULP0 UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.txulp0.scheduler_ue_enable1.ctx_parity_ue_enable = 0;
        enables.txulp0.scheduler_ue_enable1.txpb_parity_ue_enable = 0;
    }

    /* Disable the below UC request error UEs so that the errors get sent to firmware */
    enables.txulp0.scheduler_ue_enable1.tcpsch_uc_wrb_req_err_enable = 0;
    enables.txulp0.scheduler_ue_enable1.tcpsch_uc_rel_err_enable = 0;
    enables.txulp0.scheduler_ue_enable1.tcpsch_uc_rel_recieved_err_enable = 0;

    BE_GEN_CSR_WRITE(function_context, txulp0.ep_and_ipc_ue_source, enables.txulp0.ep_and_ipc_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.ep_and_ipc_ue_enable, enables.txulp0.ep_and_ipc_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.consel_arb_ue_source, enables.txulp0.consel_arb_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.consel_arb_ue_enable, enables.txulp0.consel_arb_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.consel_sel_ue_source, enables.txulp0.consel_sel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.consel_sel_ue_enable, enables.txulp0.consel_sel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.scheduler_ue_source0, enables.txulp0.scheduler_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.scheduler_ue_enable0, enables.txulp0.scheduler_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.scheduler_ue_source1, enables.txulp0.scheduler_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.scheduler_ue_enable1, enables.txulp0.scheduler_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.ring_man_ue_source_0, enables.txulp0.ring_man_ue_enable_0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.ring_man_ue_enable_0, enables.txulp0.ring_man_ue_enable_0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.ring_man_ue_source_1, enables.txulp0.ring_man_ue_enable_1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.ring_man_ue_enable_1, enables.txulp0.ring_man_ue_enable_1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.uc_merge_ue_source, enables.txulp0.uc_merge_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.uc_merge_ue_enable, enables.txulp0.uc_merge_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.qos_ue_source, enables.txulp0.qos_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp0.qos_ue_enable, enables.txulp0.qos_ue_enable.dw);

    /* Disable TXULP1 UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if (FALSE == is_asic) {
        enables.txulp1.scheduler_ue_enable1.ctx_parity_ue_enable = 0;
        enables.txulp1.scheduler_ue_enable1.txpb_parity_ue_enable = 0;
    }

    /* Disable the below UC request error UEs so that the errors get sent to firmware */
    enables.txulp1.scheduler_ue_enable1.tcpsch_uc_wrb_req_err_enable = 0;
    enables.txulp1.scheduler_ue_enable1.tcpsch_uc_rel_err_enable = 0;
    enables.txulp1.scheduler_ue_enable1.tcpsch_uc_rel_recieved_err_enable = 0;

    BE_GEN_CSR_WRITE(function_context, txulp1.ep_and_ipc_ue_source, enables.txulp1.ep_and_ipc_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.ep_and_ipc_ue_enable, enables.txulp1.ep_and_ipc_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.consel_arb_ue_source, enables.txulp1.consel_arb_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.consel_arb_ue_enable, enables.txulp1.consel_arb_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.consel_sel_ue_source, enables.txulp1.consel_sel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.consel_sel_ue_enable, enables.txulp1.consel_sel_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.scheduler_ue_source0, enables.txulp1.scheduler_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.scheduler_ue_enable0, enables.txulp1.scheduler_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.scheduler_ue_source1, enables.txulp1.scheduler_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.scheduler_ue_enable1, enables.txulp1.scheduler_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.ring_man_ue_source_0, enables.txulp1.ring_man_ue_enable_0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.ring_man_ue_enable_0, enables.txulp1.ring_man_ue_enable_0.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.ring_man_ue_source_1, enables.txulp1.ring_man_ue_enable_1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.ring_man_ue_enable_1, enables.txulp1.ring_man_ue_enable_1.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.uc_merge_ue_source, enables.txulp1.uc_merge_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.uc_merge_ue_enable, enables.txulp1.uc_merge_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.qos_ue_source, enables.txulp1.qos_ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txulp1.qos_ue_enable, enables.txulp1.qos_ue_enable.dw);

    /* Disable UC UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    BE_GEN_CSR_WRITE(function_context, uc.ue_cause, 0);
    BE_GEN_CSR_WRITE(function_context, uc.ue_enable, enables.uc.ue_enable.dw);

    /* Disable PTC UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    BE_GEN_CSR_WRITE(function_context, ptc.ue_source, enables.ptc.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, ptc.ue_enable, enables.ptc.ue_enable.dw);

    /* Disable TXP UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.txp.txp_ue_enable2.ctx_parity_ue = 0;
        enables.txp.txp_ue_enable2.txpb_parity_ue0 = 0;
        enables.txp.txp_ue_enable2.txpb_parity_ue1 = 0;
    }
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_cause0, enables.txp.txp_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_enable0, enables.txp.txp_ue_enable0.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_cause1, enables.txp.txp_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_enable1, enables.txp.txp_ue_enable1.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_cause2, enables.txp.txp_ue_enable2.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_enable2, enables.txp.txp_ue_enable2.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_cause3, enables.txp.txp_ue_enable3.dw);
    BE_GEN_CSR_WRITE(function_context, txp.txp_ue_enable3, enables.txp.txp_ue_enable3.dw);

    /* Disable NDMA UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.ndma.ue_enable.sram_access_during_init_en = 0;
    }
    BE_GEN_CSR_WRITE(function_context, ndma.ue_source, enables.ndma.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, ndma.ue_enable, enables.ndma.ue_enable.dw);

    /* Disable RDMA UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.rdma.ue_enable.sram_access_during_init_en = 0;
    }
    BE_GEN_CSR_WRITE(function_context, rdma.ue_source, enables.rdma.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, rdma.ue_enable, enables.rdma.ue_enable.dw);

    /* Disable WDMA UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.wdma.ue_enable.wdma_ctx_bus_ue_enable = 0;
        enables.wdma.ue_enable.wdma_pmem_bus_ue_enable = 0;
        enables.wdma.ue_enable.wdma_txpb_bus_ue_enable = 0;
    }
    BE_GEN_CSR_WRITE(function_context, wdma.ue_source, enables.wdma.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, wdma.ue_enable, enables.wdma.ue_enable.dw);

    /* Disable TXPB UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.txpb.ue_enable.parityerr_mpu_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_ndma_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_tpost_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_txp2_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_txp_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_txulp0_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_txulp1_ue_enable = 0;
        enables.txpb.ue_enable.parityerr_wdma_ue_enable = 0;
    }
    BE_GEN_CSR_WRITE(function_context, txpb.ue_cause, enables.txpb.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, txpb.ue_enable, enables.txpb.ue_enable.dw);

    /* Disable CTX UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    BE_GEN_CSR_WRITE(function_context, ctx.ue_source, enables.ctx.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, ctx.ue_enable, enables.ctx.ue_enable.dw);

    /* Disable DBUF UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    BE_GEN_CSR_WRITE(function_context, dbuf.ue_source, enables.dbuf.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, dbuf.ue_enable, enables.dbuf.ue_enable.dw);

    /* Disable PMEM UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    if(FALSE == is_asic) {
        enables.pmem.ue_enable.rxulp0_read_bus_ue_en = 0;
        enables.pmem.ue_enable.rxulp1_read_bus_ue_en = 0;
        enables.pmem.ue_enable.mpu_read_bus_ue_en = 0;
        enables.pmem.ue_enable.tpre_read_bus_ue_en = 0;
        enables.pmem.ue_enable.wdma_read_bus_ue_en = 0;
        enables.pmem.ue_enable.rxpp_write_bus_ue_en = 0;
        enables.pmem.ue_enable.rxpp_fcoe_write_bus_ue_en = 0;
    } else {
        /* BUG 9945: disable PMEM FCOE parity error */
        enables.pmem.ue_enable.rxpp_fcoe_write_bus_ue_en = 0;
    }
    /* False trigger of PMEM perr (bug 9810) */
    enables.pmem.ue_enable.head_a_perr_ue_en = 0;
    BE_GEN_CSR_WRITE(function_context, pmem.ue_source, enables.pmem.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, pmem.ue_enable, enables.pmem.ue_enable.dw);

    /* Disable ERX UEs.
     * Every disabled UE must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    memset(&enables, -1, sizeof(enables));
    BE_GEN_CSR_WRITE(function_context, erx.ue_source, enables.erx.ue_enable.dw);
    BE_GEN_CSR_WRITE(function_context, erx.ue_enable, enables.erx.ue_enable.dw);


    /* Clear any previous UEs (defensive step) */
    BE_GEN_PCI0_WRITE(function_context, ue_status_low, 0);
    BE_GEN_PCI0_WRITE(function_context, ue_status_hi, 0);


    /* Mask UE sources as appropriate.
     * Every masked UE source must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    ue_status_low_mask.dw = 0;
    ue_status_low_mask.host_gpio_ue_mask = 1; // this ue is used only for external trigger


    /* Mask UE sources as appropriate.
     * Every masked UE source must have a valid explanation in this source.
     * If available, bugzilla numbers should be noted as well.
     */
    ue_status_low_mask.dw = 0;
    ue_status_low_mask.host_gpio_ue_mask = 1; // this ue is used only for external trigger
    if (is_asic) {
        // BUG 9918: Disable NDMA and RDMA UEs on the asic until the bug is resolved.
        ue_status_low_mask.ndma_ue_mask = 1;
        ue_status_low_mask.rdma_ue_mask = 1;
    }
    BE_GEN_PCI0_WRITE(function_context, ue_status_low_mask, ue_status_low_mask.dw);
    BE_GEN_PCI0_WRITE(function_context, ue_status_hi_mask, ue_status_hi_mask.dw);

    return;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Init_Management_MAC
//
//   Initializes management MAC registers and enables the management MAC.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type -
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Init_Management_MAC (
   IN void* function_context )
{
    RXF_RXCTL_CSR rxctl;
    u32 data;

    data = 0x00000700;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x01, 1, &data, TRUE);
    data = 0x00000000;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x01, 1, &data, TRUE);
    data = 0x00000e46;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x00, 1, &data, TRUE);
    data = 0x0000dead;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x15, 1, &data, TRUE);
    data = 0xbabef00d;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x16, 1, &data, TRUE);
    data = 0x00008100;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x19, 1, &data, TRUE);
    data = 0x0000000d;
    BE2_Gen_Access_Rxf_Table(function_context, MGMT_MAC_REGISTER_TABLE_INDEX, 0x32, 1, &data, TRUE);

    rxctl.dw = BE_GEN_CSR_READ(function_context, rxf.rxctl);
    rxctl.enmgmtmac = 1;
    BE_GEN_CSR_WRITE(function_context, rxf.rxctl, rxctl.dw);

    return;
}

#endif  // __be2_gen_init_h__

