/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <scsi/scsi_host.h>
#include "scsi/scsi_cmnd.h"
#include "be2iscsi_common.h"
#include "be2iscsi_debug.h"
#include<linux/pci.h>
#include "be2iscsi_osm.h"
#include "acit_iface.h"


extern PACIT_IOCTL_REQUEST
se_build_internal_ioctl(
	se_hba		*phba,
	ACIT_UEXACT32	page_code,
	ACIT_UEXACT32	data_length,
	ACIT_UEXACT32	timeout,
	ACIT_UEXACT8	domain,
	ACIT_UEXACT8	type);
/*++
Function:
	se_get_free_bt_slot

Parameter:
	phba:     [in]  pointer to hba struct
	pchannel: [out] pointer to channel no
	ptarget:  [out] pointer to target no

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INSUFFICIENT_RESOURCE - run out of free slots

Context:
	Any

Description:
	This routine searches driver internal structures and gets
	a free slot for a target entry.

--*/

ACIT_STATUS
se_get_free_bt_slot(
	se_hba		*phba,
	UEXACT16	*pchannel,
	UEXACT16	*ptarget
)
{
	ACIT_U32 i = 0;

	i = find_next_zero_bit((unsigned long *)phba->bt_map,
		MAX_TARGETS_SUPPORTED, 0);

	if (i > MAX_TARGETS_SUPPORTED)
		return ACIT_STATUS_INSUFFICIENT_RESOURCE;

#ifdef BUS_FIRST_MAPPING
	/* BUS First Algorithm, fill tgts one for buses-->2nd tgt in 1st Bus */
	*pchannel = (ACIT_U16) (i % MAX_CHANNEL);
	*ptarget = (ACIT_U16) (i / MAX_CHANNEL);
#else
	/* Tgt First Algorithm, fill tgts in one bus-->then 2nd Bus */
	*ptarget = (ACIT_U16) (i % MAX_ID_PER_CHANNEL);
	*pchannel = (ACIT_U16) (i / MAX_ID_PER_CHANNEL);
#endif

	if (i < MAX_TARGETS_SUPPORTED)
		set_bit(i, (unsigned long *)phba->bt_map);
	else
		OSM_LOG_CONFIG(phba, KERN_ERR,"ERROR: BIT in BTmap is %d???\n",i);

	return ACIT_STATUS_SUCCESS;

}

/*++
Function:
	se_return_to_bt_slot

Parameter:
	phba:     [in]  pointer to hba struct
	channel:  [in]  channel/bus number
	target:   [in]  target id

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INVALID_PARAMETER - hba ptr is NULL (or)
		target id (or) channel num is incorrect

Context:
	Any

Description:
	This routine searches driver internal structures and gets
	a free slot for a target entry.

--*/

ACIT_STATUS
se_return_to_bt_slot(
	se_hba		*phba,
	UEXACT16	channel,
	UEXACT16	target
)
{

	int i = 0;

#ifdef BUS_FIRST_MAPPING
	i = (target * MAX_CHANNEL)+ channel;
#else
	i = (channel * MAX_ID_PER_CHANNEL)+target;
#endif

	if (i < MAX_TARGETS_SUPPORTED)
	{
		clear_bit(i , (unsigned long *)phba->bt_map);
		return ACIT_STATUS_SUCCESS;
	}
	else
	{
		OSM_LOG_CONFIG(phba, KERN_ERR,"ERROR: BIT in BTmap is %d???\n",i);
		return ACIT_STATUS_INVALID_PARAMETER;
	}
}

ACIT_STATUS
se_get_session_eto(
	se_hba			*phba,
	PACIT_TARGET_ENTRY	ptarget_entry,
	se_target	*ptarget)
{
	PACIT_IOCTL_REQUEST				pacit_ioctl_req = NULL;
	IOCTL_HEADER					*pioctl_hdr;
	PIOCTL_ISCSI_INI_SESSION_GET_CONFIGURATION	pioctl_data;
	ACIT_STATUS                     status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_ISCSI_INI_SESSION_GET_CONFIGURATION) +
		(phba->max_luns_per_target * sizeof(MGMT_LUN_MAPPING));
	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		down(&phba->ioctl_lock);

        pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_SESSION_GET_CONFIGURATION, data_length, 20,
		ptarget_entry->domain,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}
	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);
	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* setup the session id in request header */
	pioctl_data->params.request.session_handle = ptarget_entry->session_handle;



	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);
	ptarget->extended_timeout = pioctl_data->params.response.extended_timeout;


	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

	status = (unsigned char)pioctl_hdr->response.status;
error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		up(&phba->ioctl_lock);

	return status;
}


ACIT_STATUS
se_get_session_conf(
	se_hba			*phba,
	PACIT_TARGET_ENTRY	ptarget_entry,
	PMGMT_BTL_MAPPING	ptarget_btl_mapping
)
{
	PACIT_IOCTL_REQUEST		pacit_ioctl_req = NULL;
	IOCTL_HEADER			*pioctl_hdr;
	PIOCTL_ISCSI_INI_SESSION_GET_CONFIGURATION	pioctl_data;
	PMGMT_BTL_MAPPING		pbtl_mapping = NULL;
	ACIT_STATUS			status = ACIT_STATUS_ERROR;
	ACIT_UEXACT32			session_handle;

	int data_length = sizeof(IOCTL_ISCSI_INI_SESSION_GET_CONFIGURATION) +
		(phba->max_luns_per_target * sizeof(MGMT_LUN_MAPPING));

	session_handle = ptarget_entry->session_handle;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		down(&phba->ioctl_lock);

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_SESSION_GET_CONFIGURATION, data_length, 20,
			ptarget_entry->domain,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);
	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* setup the session id in request header */
	pioctl_data->params.request.session_handle = session_handle;

	pbtl_mapping = &pioctl_data->params.response.btl_mapping;

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	data_length = sizeof(MGMT_BTL_MAPPING);

	if (pbtl_mapping->lun_count) {

		if (pbtl_mapping->lun_count > phba->max_luns_per_target) {
			pbtl_mapping->lun_count = phba->max_luns_per_target;
		}

		data_length += (pbtl_mapping->lun_count - 1) *
			sizeof(MGMT_LUN_MAPPING);
	}

	memcpy(ptarget_btl_mapping, pbtl_mapping, data_length);

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

	status = (unsigned char)pioctl_hdr->response.status;

#if defined(__VMKERNEL_MODULE__)
	OSM_LOG_CONFIG(phba, KERN_INFO,"get_session_conf: status = %x os_bus = %x os_target = %x\n",
		status,
		0,
		pbtl_mapping->os_target);
#else
	OSM_LOG_CONFIG(phba, KERN_INFO,"get_session_conf: status = %x os_bus = %x os_target = %x\n",
		status,
		pbtl_mapping->os_bus,
		pbtl_mapping->os_target);
#endif

error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		up(&phba->ioctl_lock);

	return status;
}



ACIT_STATUS
se_get_cntlr_info(
	se_hba			*phba,
	PMGMT_HBA_ATTRIBUTES	pattr_buf
)
{
	PACIT_IOCTL_REQUEST		pacit_ioctl_req = NULL;
	IOCTL_HEADER			*pioctl_hdr;
	PIOCTL_ISCSI_INI_CFG_GET_CNTL_ATTRIBUTES	pioctl_data;
	PMGMT_HBA_ATTRIBUTES		pmgmt_hba_attr = NULL;
	ACIT_STATUS			status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_ISCSI_INI_CFG_GET_CNTL_ATTRIBUTES)*2;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_GET_CNTL_ATTRIBUTES, data_length, 20,
		0x0,1);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	pmgmt_hba_attr = &pioctl_data->cntl_attributes_info.hba_attribs;

	data_length = sizeof(MGMT_HBA_ATTRIBUTES);

	memcpy(pattr_buf,pmgmt_hba_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}

ACIT_STATUS
se_get_port_name(
	se_hba			*phba,
	uint8_t			*pport_name
)
{
	PACIT_IOCTL_REQUEST		pacit_ioctl_req = NULL;
	IOCTL_HEADER			*pioctl_hdr;
	PIOCTL_COMMON_GET_PORT_NAME	pioctl_data;
	ACIT_STATUS			status = ACIT_STATUS_ERROR;
	PIOCTL_COMMON_GET_PORT_NAME_PARAMS_OUT pioctl_portname;

	int data_length = sizeof(IOCTL_COMMON_GET_PORT_NAME);

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_GET_PORT_NAME, data_length, 20,
		0x0,1);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	pioctl_portname = &pioctl_data->params.response;

	data_length = sizeof(IOCTL_COMMON_GET_PORT_NAME);

	/* assumption: be2 is always either starting from 0 or 2 */
	if (PCI_FUNC(phba->pcidev->devfn) % 2)
		*pport_name = pioctl_portname->port1_name;
	else
		*pport_name = pioctl_portname->port0_name;

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}

ACIT_STATUS
se_get_boot_target_info(
	se_hba			*phba,
	PGET_BOOT_TARGET_OUT	pattr_buf
)
{
	PACIT_IOCTL_REQUEST		pacit_ioctl_req = NULL;
	IOCTL_HEADER			*pioctl_hdr;
	PIOCTL_ISCSI_INI_BOOT_GET_BOOT_TARGET pioctl_data;
	PGET_BOOT_TARGET_OUT 		pget_boottgt_out = NULL;
	ACIT_STATUS			status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_ISCSI_INI_BOOT_GET_BOOT_TARGET)*2;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_BOOT_GET_BOOT_TARGET, data_length, 20,
		0x0,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	pget_boottgt_out  = &pioctl_data->response;

	data_length = sizeof(GET_BOOT_TARGET_OUT);

	memcpy(pattr_buf,pget_boottgt_out, data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}


ACIT_STATUS
se_get_default_gateway(
	se_hba			*phba,
	PGET_DEFAULT_ROUTE_PARAMS	pattr_buf
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_COMMON_ISCSI_NTWK_GET_DEFAULT_GATEWAY	pioctl_data;
	PGET_DEFAULT_ROUTE_PARAMS		proute_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_COMMON_ISCSI_NTWK_GET_DEFAULT_GATEWAY)*2;

	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_ISCSI_NTWK_GET_DEFAULT_GATEWAY, data_length, 20,
		0x0,SUBSYSTEM_COMMON_ISCSI_NTWK_GET_DEFAULT_GATEWAY);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;
	proute_attr = &pioctl_data->params;

	proute_attr->request.ip_type = IP_TYPE_V4;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	data_length = sizeof(GET_DEFAULT_ROUTE_PARAMS);

	memcpy(pattr_buf,proute_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}

ACIT_STATUS
se_get_all_if_id(
	se_hba			*phba,
	PGET_ALL_IF_ID_PARAMS	pattr_buf
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_COMMON_ISCSI_NTWK_GET_ALL_IF_ID	pioctl_data;
	PGET_ALL_IF_ID_PARAMS			pifid_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_COMMON_ISCSI_NTWK_GET_ALL_IF_ID)*2;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}
	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_ISCSI_NTWK_GET_ALL_IF_ID, data_length, 20,
		0x0, SUBSYSTEM_COMMON_ISCSI_NTWK_GET_ALL_IF_ID);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	pifid_attr = &pioctl_data->params;
	data_length = sizeof(GET_ALL_IF_ID_PARAMS);

	memcpy(pattr_buf,pifid_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}


ACIT_STATUS
se_get_nw_info(
	se_hba			*phba,
	PGET_IF_INFO_PARAMS	pattr_buf
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_COMMON_ISCSI_NTWK_GET_IF_INFO	pioctl_data;
	PGET_ALL_IF_ID_PARAMS			pifid_attr = NULL;
	PGET_IF_INFO_PARAMS			pifinfo_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_COMMON_ISCSI_NTWK_GET_IF_INFO)*2;

	pifid_attr = (PGET_ALL_IF_ID_PARAMS ) kmalloc(sizeof(GET_ALL_IF_ID_PARAMS),GFP_KERNEL);
	status = se_get_all_if_id(phba,pifid_attr);

	if (status) {
		kfree(pifid_attr);
		return status;
	}

	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	data_length = sizeof(IOCTL_COMMON_ISCSI_NTWK_GET_IF_INFO)*2;

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_ISCSI_NTWK_GET_IF_INFO, data_length, 20,
		0x0, SUBSYSTEM_COMMON_ISCSI_NTWK_GET_IF_INFO);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pifinfo_attr = &pioctl_data->params;

	pifinfo_attr->request.interface_hndl = pifid_attr->response.if_hndl_list[0];
	pifinfo_attr->request.ip_type = IP_TYPE_V4;

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	data_length = sizeof(GET_IF_INFO_PARAMS);

	memcpy(pattr_buf,pifinfo_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	kfree(pifid_attr);
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}

ACIT_STATUS
se_get_nic_info(
	se_hba			*phba,
	PNIC_CONFIG_PARAMS	pattr_buf
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_COMMON_ISCSI_NTWK_NIC_CONFIG	pioctl_data;
	PNIC_CONFIG_PARAMS			pnic_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_COMMON_ISCSI_NTWK_NIC_CONFIG)*100;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_COMMON_ISCSI_NTWK_GET_NIC_CONFIG, data_length, 20,
		0x0,2);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	pnic_attr = &pioctl_data->params;

	data_length = sizeof(NIC_CONFIG_PARAMS);

	memcpy(pattr_buf,pnic_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}

	return status;
}



ACIT_STATUS
se_get_iqn_info(
	se_hba			*phba,
	PISCSI_HBA_NAME_PARAMS	pattr_buf
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_ISCSI_INI_CFG_HBA_NAME		pioctl_data;
	PISCSI_HBA_NAME_PARAMS			piqn_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_ISCSI_INI_CFG_HBA_NAME)*2;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_CFG_GET_HBA_NAME, data_length, 20,
		0x0,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	piqn_attr = &pioctl_data->params;

	data_length = sizeof(ISCSI_HBA_NAME_PARAMS);

	memcpy(pattr_buf,piqn_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
ACIT_STATUS
se_get_portal_group_tag(
	se_hba			*phba,
	se_target		*ptarget,
	unsigned		*portal_group_tag
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_ISCSI_INI_DISCOVERY_GET_ALL_DISCOVERED_TARGET_ID	pioctl_data;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;
	int					data_length;
	PGET_ALL_DISCOVERED_TARGET_ID_PARAMS	pget_all_tgt_id;
	PMGMT_TARGET_ID_PORTAL_COUNT		plist;
	unsigned				target_count;
	unsigned				i;
	unsigned				j;
	unsigned				k;
	unsigned				port_group_count;
	unsigned				actual_size;
	char					buf_name[255];
	char					buf_port[255];
	char					buf_ip[255];

	PIOCTL_ISCSI_INI_DISCOVERY_REPORT_A_TARGET	pioctl_targ_data;
	PREPORT_A_TARGET_PARAMS			preport_a_target;
	PMGMT_TARGET_INFO			ptinfo = NULL;
	PMGMT_PORTAL				pmgmt_portal;

	BEGIN;


	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	/*
	 * --------------------------------------------------------------------
	 * We will do it once to find out the real size.
	 */
	data_length = 
		sizeof(IOCTL_ISCSI_INI_DISCOVERY_GET_ALL_DISCOVERED_TARGET_ID);

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_DISCOVERY_GET_ALL_DISCOVERED_TARGET_ID,
		data_length, 20, 0x0, 0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	pget_all_tgt_id = &pioctl_data->params;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	status = (unsigned char)pioctl_hdr->response.status;

	/*
	 * Now we know the real size!!!
	 */
	actual_size = pioctl_hdr->response.actual_response_length;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/*
	 * --------------------------------------------------------------------
	 * Do it again now that we know the real size
	 */
	data_length = sizeof(IOCTL_HEADER) + actual_size;

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_DISCOVERY_GET_ALL_DISCOVERED_TARGET_ID,
		data_length, 20, 0x0, 0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	pget_all_tgt_id = &pioctl_data->params;

	init_completion(&phba->ioctl_completion);

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	status = (unsigned char)pioctl_hdr->response.status;

	target_count = pget_all_tgt_id->response.target_count;

	data_length = sizeof(MGMT_TARGET_ID_PORTAL_COUNT) * target_count;

	plist = (PMGMT_TARGET_ID_PORTAL_COUNT)kmalloc(data_length, GFP_KERNEL);

	memcpy(plist, pget_all_tgt_id->response.list, data_length);

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	status = ACIT_STATUS_ERROR;

	/*
	 * --------------------------------------------------------------------
	 * Loop through targets
	 */

	for (i = 0 ; i < target_count ; i++) {

		/*
		 * -----------------------------------------------------------
		 * We will do it once to find out the real size.
		 */
		data_length = sizeof(IOCTL_ISCSI_INI_DISCOVERY_REPORT_A_TARGET);

		pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(
			phba, OPCODE_ISCSI_INI_DISCOVERY_REPORT_A_TARGET, 
			data_length, 20, 0x0, 0);

		if (!pacit_ioctl_req) {
			OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
			status = ACIT_STATUS_ERROR;
			goto error;
		}

		pioctl_targ_data = phba->ioctl_buf.virtual_address;
		memset(pioctl_targ_data, 0, data_length);

		pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

		preport_a_target = &pioctl_targ_data->params;

		preport_a_target->request.target_id = plist[i].target_id;

		init_completion(&phba->ioctl_completion);

		/* currently acit_start_ioctl returns only success */
		status = acitapi_start_ioctl(phba, &phba->controller_key,
			pacit_ioctl_req);

		wait_for_completion(&phba->ioctl_completion);

		actual_size = pioctl_hdr->response.actual_response_length;

		/*
		 * Now we know the real size!!!
		 */

		se_pci_free_consistent(phba->pcidev,
			(size_t)phba->ioctl_buf.size_in_bytes,
			(void *)phba->ioctl_buf.virtual_address,
			phba->ioctl_buf.bus_address.u.a64.address);


		/*
		 * -----------------------------------------------------------
		 * Now do it again with the real size
		 */
		data_length = sizeof(IOCTL_HEADER) + actual_size;

		pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(
			phba, OPCODE_ISCSI_INI_DISCOVERY_REPORT_A_TARGET, 
			data_length, 20, 0x0, 0);

		if (!pacit_ioctl_req) {
			OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
			status = ACIT_STATUS_ERROR;
			goto error;
		}

		pioctl_targ_data = phba->ioctl_buf.virtual_address;
		memset(pioctl_targ_data, 0, data_length);

		pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

		preport_a_target = &pioctl_targ_data->params;

		preport_a_target->request.target_id = plist[i].target_id;

		init_completion(&phba->ioctl_completion);

		/* currently acit_start_ioctl returns only success */
		status = acitapi_start_ioctl(phba, &phba->controller_key,
			pacit_ioctl_req);

		wait_for_completion(&phba->ioctl_completion);

		ptinfo = &preport_a_target->response.target_info;

		OSM_LOG_CONFIG(phba, KERN_INFO,"target_id = 0x%x target_name = %s"
				" target_alias = %s pg_count = 0x%x\n",
			 	ptinfo->target_id,ptinfo->target_name,
				ptinfo->target_alias,ptinfo->pg_count);

		status = ACIT_STATUS_ERROR;

		for (j = 0 ; j < ptinfo->pg_count ; j++) {

			OSM_LOG_CONFIG(phba, KERN_INFO,"   tag = 0x%x - "
				"portal_count = 0x%x\n",
				ptinfo->pg_list[j].pg_tag,
				ptinfo->pg_list[j].portal_count);

			for (k = 0 ; k < ptinfo->pg_list[j].portal_count ; 
				k++) {
				pmgmt_portal = &(ptinfo->pg_list[j].
					portal_list[k].portal);

				OSM_LOG_CONFIG(phba, KERN_INFO,"   %d.%d.%d.%d - %d\n",
					pmgmt_portal->ip_address.IPAddress[0],
					pmgmt_portal->ip_address.IPAddress[1],
					pmgmt_portal->ip_address.IPAddress[2],
					pmgmt_portal->ip_address.IPAddress[3],
					pmgmt_portal->port_number);

				sprintf(buf_ip, "%d.%d.%d.%d",
					pmgmt_portal->ip_address.IPAddress[0],
					pmgmt_portal->ip_address.IPAddress[1],
					pmgmt_portal->ip_address.IPAddress[2],
					pmgmt_portal->ip_address.IPAddress[3]);

				if (strcmp(buf_ip, ptarget->buf_ip) != 0) {
					OSM_LOG_CONFIG(phba, KERN_INFO,"   \"%s\" and \"%s\" "
						"don't match\n",
						buf_ip, ptarget->buf_ip);
					continue;
				}

				sprintf(buf_port, "%u", 
					pmgmt_portal->port_number);

				if (strcmp(buf_port, ptarget->buf_port) != 0) {
					OSM_LOG_CONFIG(phba, KERN_INFO,"   \"%s\" and \"%s\" "
						"don't match\n",
						buf_port, ptarget->buf_port);
					continue;
				}

				if (strcmp(ptinfo->target_name, 
					ptarget->buf_name) != 0) {

					OSM_LOG_CONFIG(phba, KERN_INFO,"   \"%s\" and \"%s\" "
						"don't match\n",
						ptinfo->target_name,
						ptarget->buf_name);

					continue;
				}

				*portal_group_tag = ptinfo->pg_list[j].pg_tag;

				OSM_LOG_CONFIG(phba, KERN_INFO,"   Found match!! %d\n",
					*portal_group_tag);

				status = ACIT_STATUS_SUCCESS;

				break;
			}

			if (status == ACIT_STATUS_SUCCESS) {
				break;
			}
		}

		se_pci_free_consistent(phba->pcidev,
			(size_t)phba->ioctl_buf.size_in_bytes,
			(void *)phba->ioctl_buf.virtual_address,
			phba->ioctl_buf.bus_address.u.a64.address);

		if (status == ACIT_STATUS_SUCCESS) {
			break;
		}
	}

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}

	END;

	return status;
}
#endif
#endif

ACIT_STATUS
se_get_allsess_info(
	se_hba				*phba,
	PGET_ALL_SESSION_ID_PARAMS	pattr_buf,
	unsigned int			data_length
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_ISCSI_INI_SESSION_GET_ALL_SESSION_ID	pioctl_data;
	PGET_ALL_SESSION_ID_PARAMS		piqn_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_SESSION_GET_ALL_SESSION_ID , data_length, 20,
		0x0,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	piqn_attr = (PGET_ALL_SESSION_ID_PARAMS)pacit_ioctl_req->pdata_buffer;
	piqn_attr->request.session_flags = MGMT_SESSION_FLAG_DATA_SESSION|MGMT_SESSION_FLAG_LIST_SESSION;

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	piqn_attr = &pioctl_data->params;

	memcpy(pattr_buf,piqn_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	if (4 == status)
		status = pioctl_hdr->response.actual_response_length;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}
	return status;
}


ACIT_STATUS
se_get_asession_info(
	se_hba			*phba,
	PGET_A_SESSION_PARAMS	pattr_buf,
	UEXACT32		session_handle
)
{
	PACIT_IOCTL_REQUEST			pacit_ioctl_req = NULL;
	IOCTL_HEADER				*pioctl_hdr;
	PIOCTL_ISCSI_INI_SESSION_GET_A_SESSION	pioctl_data;
	PGET_A_SESSION_PARAMS			piqn_attr = NULL;
	ACIT_STATUS				status = ACIT_STATUS_ERROR;

	int data_length = sizeof(IOCTL_ISCSI_INI_SESSION_GET_A_SESSION)*2;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		down(&phba->ioctl_lock);
	}

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_SESSION_GET_A_SESSION, data_length, 20,
		0x0,0);

	if (!pacit_ioctl_req) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"No request\n");
		status = ACIT_STATUS_ERROR;
		goto error;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);
	pioctl_hdr = &pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);
	pioctl_data->params.request.session_handle = session_handle;

	/* currently acit_start_ioctl returns only success */
	status = acitapi_start_ioctl(phba, &phba->controller_key,
			pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	piqn_attr = &pioctl_data->params;

	data_length = sizeof(GET_A_SESSION_PARAMS);

	memcpy(pattr_buf,piqn_attr , data_length);

	status = (unsigned char)pioctl_hdr->response.status;

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;


error:
	if (!(phba->state & SE_STATE_LOGIN_PENDING)) {
		up(&phba->ioctl_lock);
	}

	return status;
}



ACIT_STATUS
se_get_btl_info(
	se_hba *phba,
	PACIT_TARGET_ENTRY ptarget_entry,
	PMGMT_BTL_MAPPING ptarget_btl_mapping)
{
	ACIT_STATUS	status = MGMT_STATUS_FAILED;
	int		i = 0;

	status = se_get_session_conf(phba, ptarget_entry, ptarget_btl_mapping);

	if (status != MGMT_STATUS_SUCCESS ||
		ptarget_btl_mapping->os_bus == 0xffff ||
		ptarget_btl_mapping->os_target == 0xffff) {

		/* Update the BT inuse with the latest one */
		se_get_bt_inuse(phba, ptarget_entry->domain);

		/* assign a bt if the ioctl returns invalid values */
		status = se_get_free_bt_slot(phba,
			&(ptarget_btl_mapping)->os_bus,
			&(ptarget_btl_mapping)->os_target);
	}
	else
	{
		/* BTmap is from Firmware --> Target is Persistent reserve it */
#ifdef BUS_FIRST_MAPPING
		i = (ptarget_btl_mapping->os_target * MAX_CHANNEL) +
			ptarget_btl_mapping->os_bus;
#else
#if defined(__VMKERNEL_MODULE__)
		i = ptarget_btl_mapping->os_target;
#else
		i = (ptarget_btl_mapping->os_bus * MAX_ID_PER_CHANNEL) +
			ptarget_btl_mapping->os_target;
#endif
#endif
		if (i < MAX_TARGETS_SUPPORTED)
			set_bit(i , (unsigned long *)phba->bt_map);
		else
			OSM_LOG_CONFIG(phba, KERN_ERR,"ERROR:BIT in BTmap is %d???\n",i);

#if defined(__VMKERNEL_MODULE__)
		OSM_LOG_CONFIG(phba, KERN_INFO,"get_btl_info: bit=%x, bus=%x,tgt=%x\n ",
			i,
			0,
			ptarget_btl_mapping->os_target);
#else
		OSM_LOG_CONFIG(phba, KERN_INFO,"get_btl_info: bit=%x, bus=%x,tgt=%x\n ",
			i,
			ptarget_btl_mapping->os_bus,
			ptarget_btl_mapping->os_target);
#endif
	}

#if defined(__VMKERNEL_MODULE__)
	OSM_LOG_CONFIG(phba, KERN_INFO,"target = %x bus = %x custom_mapping = %d\n",
		ptarget_btl_mapping->os_target,
		0,
		ptarget_btl_mapping->dev_addr_mode);
#else
	OSM_LOG_CONFIG(phba, KERN_INFO,"target = %x bus = %x custom_mapping = %d\n",
		ptarget_btl_mapping->os_target,
		ptarget_btl_mapping->os_bus,
		ptarget_btl_mapping->dev_addr_mode);
#endif

	if (phba->top_scsi_id < ptarget_btl_mapping->os_target)
		phba->top_scsi_id = ptarget_btl_mapping->os_target;

#if !defined(__VMKERNEL_MODULE__)
	if (phba->top_scsi_channel < ptarget_btl_mapping->os_bus)
		phba->top_scsi_channel = ptarget_btl_mapping->os_bus;
#endif

	return status;
}


unsigned int
se_get_bt_inuse(
	se_hba		*phba,
	ACIT_UEXACT8	domain
)
{
	PACIT_IOCTL_REQUEST	pacit_ioctl_req = NULL;
	IOCTL_HEADER		ioctl_hdr;
	PIOCTL_ISCSI_INI_BTL_MAPPING_QUERY_OS_BT_INUSE	pioctl_data;
	PQUERY_OS_BUS_TARGET_INUSE_PARAMS_OUT		pbt_in_use;
	ACIT_STATUS		status = ACIT_STATUS_SUCCESS;

	int data_length = sizeof(
		IOCTL_ISCSI_INI_BTL_MAPPING_QUERY_OS_BT_INUSE) -
		sizeof(IOCTL_HEADER) +
		((phba->max_iscsi_targets - 1) * sizeof(MGMT_OS_BT));

	/* preparing the request and sending it is an atomic operation */
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		down(&phba->ioctl_lock);

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_BTL_MAPPING_QUERY_OS_BT_INUSE, data_length,
		20, domain,0);

	if (!pacit_ioctl_req) {
		status = ACIT_STATUS_ERROR;
		goto no_req;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);

	ioctl_hdr = pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

#if 0
	/* COMMENT THIS OUT TO WORK AROUND THE FW/ACIT BUG */
	if (ioctl_hdr.response.status != MGMT_STATUS_SUCCESS) {
		status = ACIT_STATUS_ERROR;
		goto ioctl_failed;
	}
#endif

	pbt_in_use = &pioctl_data->response;

	data_length = sizeof(QUERY_OS_BUS_TARGET_INUSE_PARAMS_OUT) +
		(pbt_in_use->os_bt_count * sizeof(MGMT_OS_BT));

	OSM_LOG_CONFIG(phba, KERN_INFO,"IOCTL_FAILED,returned error code =%d\n",
		ioctl_hdr.response.status);

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

no_req:
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		 up(&phba->ioctl_lock);

	return status;
}

unsigned int
se_set_session_conf(
	se_hba			*phba,
	se_target		*ptarget,
	PACIT_TARGET_ENTRY	ptarget_entry
)
{
	PACIT_IOCTL_REQUEST		pacit_ioctl_req = NULL;
	IOCTL_HEADER			ioctl_hdr;
	PIOCTL_ISCSI_INI_SESSION_SET_CONFIGURATION	pioctl_data;
	ACIT_STATUS			status = ACIT_STATUS_SUCCESS;
	ACIT_UEXACT32			session_handle;
	int lun_length = 0;
	int data_length = sizeof(IOCTL_ISCSI_INI_SESSION_SET_CONFIGURATION);

	session_handle = ptarget_entry->session_handle;

	OSM_LOG_CONFIG(phba, KERN_INFO,"lun_count is %d\n",
		ptarget->pbtl_mapping->lun_count);

	if (ptarget->pbtl_mapping->lun_count)
	{
		lun_length = sizeof(MGMT_LUN_MAPPING) *
			(ptarget->pbtl_mapping->lun_count - 1);
	}

	data_length += lun_length;

	/* preparing the request and sending it is an atomic operation */
	/* if login is pending then ioctl_lock is already taken */
	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		 down(&phba->ioctl_lock);

	pacit_ioctl_req = (PACIT_IOCTL_REQUEST)se_build_internal_ioctl(phba,
		OPCODE_ISCSI_INI_SESSION_SET_CONFIGURATION, data_length, 20,
		ptarget_entry->domain,0);

	if (!pacit_ioctl_req) {
		return ACIT_STATUS_ERROR;
	}

	pioctl_data = phba->ioctl_buf.virtual_address;
	memset(pioctl_data, 0, data_length);
	ioctl_hdr = pacit_ioctl_req->ioctl_hdr;

	init_completion(&phba->ioctl_completion);

	 /* setup the session id in request header */
	pioctl_data->params.request.session_handle = session_handle;

	memcpy(&pioctl_data->params.request.btl_mapping,
		ptarget->pbtl_mapping,
		sizeof(MGMT_BTL_MAPPING) + lun_length
	);

	pioctl_data->params.request.extended_timeout = -1;

	pioctl_data->params.request.param_flags =
		((ENUM_SET_SESSCONFIG_PARAM_FLAG_FULL) &
			(~SET_SESSCONFIG_PARAM_FLAG_ETO));

	status = acitapi_start_ioctl(phba, &phba->controller_key,
		pacit_ioctl_req);

	wait_for_completion(&phba->ioctl_completion);

	if (ioctl_hdr.response.status != MGMT_STATUS_SUCCESS) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"Set_session_conf ioctl failed\n");
		status = ACIT_STATUS_ERROR;
	}

	se_pci_free_consistent(phba->pcidev, (size_t)phba->ioctl_buf.size_in_bytes,
		(void *)phba->ioctl_buf.virtual_address,
		phba->ioctl_buf.bus_address.u.a64.address);

	/* this is how you can test if internal ioctl is in progress */
	phba->ioctl_buf.size_in_bytes = 0;

	if (!(phba->state & SE_STATE_LOGIN_PENDING))
		 up(&phba->ioctl_lock);

	/* MBM FIXME: Remove it, for testing only */
	se_get_bt_inuse(phba, ptarget_entry->domain);

	return status;
}
