/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#include <linux/version.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/pci.h>
#if LINUX_VERSION_CODE != KERNEL_VERSION(2,6,5)
#include "scsi/scsi.h"
#else
#include "../drivers/scsi/scsi.h"
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include "scsi/scsi_cmnd.h"
#include "scsi/scsi_device.h"
#endif
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/reboot.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "uerr.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <asm/delay.h>
#include <linux/genhd.h>
#include <linux/spinlock.h>
#include <scsi/scsi_host.h>
#include <linux/proc_fs.h>
#include <linux/debugfs.h>
#include <linux/seq_file.h>
#include <linux/ctype.h>
#include "be2iscsi_common.h"
#include "version.h"
#include "be2iscsi_osm.h"
#include "be2iscsi_debug.h"
#include "acit_iface.h"
#include <linux/slab.h>
#include "iscsi_drvr_stats.h"


struct proc_dir_entry *be_proc_root_dir;
ACIT_UEXACT8 be_proc_bdir_present;
static DEFINE_MUTEX(be_pgbl_mutex);
static int
be_get_stats(char *buffer, char **start, off_t offset, int length, int *eof, void *data);
static int
be_set_loglvl(struct file *file, const char *buffer, unsigned long count, void *data);
static int
be_get_loglvl(char *buffer, char **start, off_t offset, int length, int *eof, void *data);

extern ACIT_STATUS
acit_process_stat_ioctl(
        PACIT_CONTROLLER_KEY    pcontroller_key,
        PDRVR_STATS_REQUEST    pstats_hdr);
static void se_qdetails(char *str,int *qid, int *options);

#define DRIVER_STATS    "driver_stats"
/* OSM Logging related Functions */
#define OSM_LOG_LEVEL	"log_level"
extern int log_level;

#ifndef __VMKLNX__

static DEFINE_MUTEX(se_pgbl_mutex);
ACIT_UEXACT8 se_proc_bdir_present;
struct dentry	*se_gbl_pdir;

static int se_single_open_seq(struct inode *inode, struct file *file);
static int se_cid_read(struct seq_file *pseq,void *v);
static int se_icd_read(struct seq_file *pseq,void *v);
static int se_icdfind_read(struct seq_file *pseq,void *v);
static int se_wrb_read(struct seq_file *pseq,void *v);
static int se_eq_read(struct seq_file *pseq,void *v);
static int se_eqfind_read(struct seq_file *pseq,void *v);
static int se_cq_read(struct seq_file *pseq,void *v);
static int se_cqfind_read(struct seq_file *pseq,void *v);
static int se_etofind_read(struct seq_file *pseq,void *v);
static int se_intr_read(struct seq_file *pseq,void *v);
static int se_tmf_read(struct seq_file *pseq,void *v);
static int se_invl_read(struct seq_file *pseq,void *v);
static int se_iosqd_read(struct seq_file *pseq,void *v);
static int se_help_read(struct seq_file *pseq,void *v);
static int dbg_log_activate(struct seq_file *pseq,void *v);
static int se_qwrite_read(struct seq_file *pseq,void *v);
static ssize_t se_qwrite_proc(struct file *filep,const char __user *buff,
		size_t len,	loff_t *off);


#define Q_WLEN			(256)
#define EQ_QTYPE		"eq"
#define CQ_QTYPE		"cq"
#define CID_QTYPE		"cid"
#define ICD_QTYPE		"icd"
#define WRB_QTYPE		"wrb"
#define TMF_QTYPE		"tmfwalk"
#define ICDFIND_QTYPE	"findicd"
#define IOSQUEUED_QTYPE "iosqueue"
#define ACIT_LOG		"dbg_log"


/* PROC Files */
#define SE_QWRITE		"debug_data" 
#define SE_EQ_READ		"eq_data" /* Done */
#define SE_CQ_READ		"cq_data" /* Done */
#define SE_CID_READ		"cid_data" /* Done */
#define SE_ICD_READ		"icd_data" /* Done */
#define SE_WRB_READ		"wrb_data" /* Done */
#define SE_ICDFIND_READ	"findicd_data" /* Done */
#define SE_EQFIND_READ	"findeq_data" /* Done */
#define SE_CQFIND_READ	"findcq_data" /* Done */
#define SE_ETOFIND_READ	"etofind_data" /* Done */
#define SE_INTR_READ	"intr_data" /* Done */
#define SE_TMF_READ		"tmfwalk_data" /* Done */
#define SE_INVL_READ	"invalid_data" /* Done */
#define SE_IOSQD_READ	"iosqd_data"
#define SE_HELP_READ	"dbg_help_proc"
#define SE_LOG_OP		"dbg_log_status" /* Enable/Disable Debugging operation */


#define SE_SEQ_FOPS_RW_SET(__wfunc) 		\
		.seq_fops = { 						\
			.owner	= THIS_MODULE, 			\
			.open	= se_single_open_seq, 	\
			.read	= seq_read,  			\
			.write	= __wfunc,   			\
			.llseek	= seq_lseek, 			\
			.release= single_release, 		\
		} 

#define SE_PDE_PARAMS_SET(pde,Name,Mode,	\
		Parent,Data,Sq_Data) 						\
		do{ 								\
			memcpy(pde.pname,Name,strlen(Name) + 1);	\
				(pde.mode) = (Mode); 		\
				(pde.parent) = (Parent);	\
				(pde.data) = (Data);		\
				(pde.seq_data) = (Sq_Data); \
		}while(0);


static struct se_seq_data se_log_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = dbg_log_activate,
};

static struct se_seq_data dbg_help_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_help_read,
};


static struct se_seq_data se_iosqd_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_iosqd_read,
};


static struct se_seq_data se_invl_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_invl_read,
};


static struct se_seq_data se_tmf_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_tmf_read,
};

static struct se_seq_data se_intr_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_intr_read,
};

static struct se_seq_data se_etofind_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_etofind_read,
};

static struct se_seq_data se_cqfind_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_cqfind_read,
};

static struct se_seq_data se_cq_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_cq_read,
};

static struct se_seq_data se_eqfind_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_eqfind_read,
};

static struct se_seq_data se_eq_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_eq_read,
};

static struct se_seq_data se_wrb_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_wrb_read,
};

static struct se_seq_data se_icdfind_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_icdfind_read,
};

static struct se_seq_data se_icd_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_icd_read,
};

static struct se_seq_data se_cid_seq = {
	SE_SEQ_FOPS_RW_SET(NULL),
	.show = se_cid_read,
};

static struct se_seq_data se_qwrite_seq = {
	SE_SEQ_FOPS_RW_SET(se_qwrite_proc),
	.show = se_qwrite_read,
};

extern void
acit_process_proc_iosqd(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_SEXACT32           cid,
		struct seq_file			*pseq);

extern void
acit_process_proc_invl(
		PACIT_CONTROLLER_KEY    pcontroller,
		struct seq_file			*pseq);

extern void
acit_process_proc_tmf(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_SEXACT32           cid,
		struct seq_file			*pseq);

extern void
acit_process_proc_intr(
		PACIT_CONTROLLER_KEY    pcontroller,
		struct seq_file			*pseq);

extern void
acit_process_proc_etofind(
		PACIT_CONTROLLER_KEY    pcontroller,
		struct seq_file			*pseq);

extern void
acit_process_proc_cqfind(
		PACIT_CONTROLLER_KEY    pcontroller,
		struct seq_file			*pseq);

extern void
acit_process_proc_cq(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_UEXACT32           qid,
		ACIT_UEXACT32           options,
		struct seq_file			*pseq);

extern void
acit_process_proc_eqfind(
		PACIT_CONTROLLER_KEY    pcontroller,
		struct seq_file			*pseq);

extern void
acit_process_proc_eq(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_UEXACT32           qid,
		ACIT_UEXACT32           options,
		struct seq_file			*pseq);

extern void
acit_process_proc_wrb(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_UEXACT32           qid,
		ACIT_UEXACT32           options,
		struct seq_file			*pseq);

extern void
acit_process_proc_icdfind(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_UEXACT32           qid,
		ACIT_UEXACT32           options,
		struct seq_file			*pseq);


extern void
acit_process_proc_icd(
		PACIT_CONTROLLER_KEY	pcontroller,
		ACIT_UEXACT32			icd_idx,
		struct seq_file			*pseq);

extern void
acit_process_proc_cid(
		PACIT_CONTROLLER_KEY    pcontroller,
		ACIT_UEXACT32           qid,
		ACIT_UEXACT32           options,
		struct seq_file			*pseq);

#endif

#ifdef BE2ISCSI_DEBUG
unsigned int alloc_count = 0;
unsigned int free_count = 0;

extern struct list_head hba_chain;

#if 0
#undef kmalloc
#undef kfree

void *
dmalloc(unsigned int size, unsigned int flags, char *fn)
{
	void *ptr = kmalloc(size, flags);

	alloc_count++;

	return ptr;	
}

void
dfree(void *ptr, char *fn)
{
	free_count++;

	kfree(ptr);
}
#endif

void clear_device_stats(void)
{
	struct list_head	*hba_entry;
	se_hba			*phba;
	se_target		*ptarget;
	se_device		*pdevice;
	unsigned		hba_index;
	unsigned		target_index;
	unsigned		lun_index;

	hba_index = 0;
	target_index = 0;
	lun_index = 0;

	list_for_each(hba_entry, &hba_chain) {

		phba = list_entry(hba_entry, se_hba, hba_queue);

		hba_index++;

		list_for_each_entry(ptarget, &phba->target_list, sibling_list) {

			target_index++;

			list_for_each_entry(pdevice, &ptarget->lun_list,
					sibling_list) {

				pdevice->num_errors = 0;

				lun_index++;
			}
		}
	}
}
#endif

struct info_str {
        char *buffer;
        int length;
        off_t offset;
        int pos;
};

static void
copy_mem_info(struct info_str *info, char *data, int len)
{
        if (info->pos + len > info->offset + info->length) {
                len = info->offset + info->length - info->pos;
        }

        if (info->pos + len < info->offset) {
                info->pos += len;
                return;
        }

        if (info->pos < info->offset) {
                off_t partial;

                partial = info->offset - info->pos;
                data += partial;
                info->pos += partial;
                len  -= partial;
        }

        if (len > 0) {
                memcpy(info->buffer, data, len);
                info->pos += len;
                info->buffer += len;
        }
}

static int
copy_info(struct info_str *info, char *fmt, ...)
{
        va_list args;
        char buf[256];
        int len;
	
        va_start(args, fmt);
        len = vsprintf(buf, fmt, args);
        va_end(args);

        copy_mem_info(info, buf, len);
        return (len);
}
unsigned int g_pending_ioctls = 0;

int get_acit_stats(se_hba *phba, PDRVR_STATS_REQUEST prequest, void * buffer)
{
	int status = 0;
	void *pbuffer = kmalloc((sizeof(DRVR_STATS_REQUEST) + prequest->buffer_len),GFP_KERNEL);
	if(pbuffer){
                memset(pbuffer, 0, (sizeof(DRVR_STATS_REQUEST)+ prequest->buffer_len));
                memcpy(pbuffer, prequest, sizeof(DRVR_STATS_REQUEST));

		status = acit_process_stat_ioctl(&phba->controller_key, pbuffer);
		if(status != ACIT_STATUS_SUCCESS) {
			printk("acit failed stat ioctl\n");
			return -1;
		}
		memcpy(buffer, (char *)pbuffer + sizeof(DRVR_STATS_REQUEST),prequest->buffer_len);
		kfree(pbuffer);
	}
	return 0;
}


void dump_osm_params(se_hba *phba, struct info_str *info)
{
}

void dump_acit_params(se_hba *phba, struct info_str *info)
{
	int idx = 0;
	DRVR_STATS_REQUEST request;
	ACIT_PARAMETERS acit_params;

	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));
	request.ioctl_code = IOCTL_GET_ACIT_PARAMETERS;
	request.buffer_len = sizeof(ACIT_PARAMETERS);
	
	if(get_acit_stats(phba, &request, &acit_params) == 0){
		idx = phba->hba_unique_id;
		phba->num_cq = acit_params.num_cq;
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_MaxTargets", acit_params.targets_per_controller);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_ConnectionsPerSession", acit_params.connections_per_session);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_IosPerController", acit_params.ios_per_controller);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_ICDsPerController", acit_params.icds_per_controller);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_LogoutResources", acit_params.logouts_per_controller);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_LogoutTimeoutTicks", acit_params.logout_timeout_ticks);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TicksPerSec", acit_params.ticks_per_second);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TMFResources", acit_params.tmfs_per_ctrl);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TMFTimeoutTicks", acit_params.tmf_timeout_ticks);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_ETO", acit_params.eto);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_LDTO", acit_params.ldto);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DebugFlags", acit_params.dbg_flags);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TMFsPerCxn", acit_params.num_tmfs_per_cxn);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_WRBsPerCxn", acit_params.num_cmd_per_cxn);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_MaxCxns", acit_params.num_cxn);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_MCCWRbs", acit_params.ioctls_per_ctrl);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumEQE", acit_params.num_eq_entries);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumCQE", acit_params.num_cq_entries);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_EQTimer", acit_params.eq_timer);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumMCCWrbPages", acit_params.num_mcc_pages);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumMCCCqPages", acit_params.num_mcc_cq_pages);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumiSCSICQPages", acit_params.num_cq_pages);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumEQPages", acit_params.num_eq_pages);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumDefaultPDUBufferPages", acit_params.num_async_pdu_buf_pages);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumDefaultPDUSGLPages", acit_params.num_async_pdu_buf_sgl_pages);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DefaultPDUBufferSize", acit_params.default_pdu_buf_sz);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumiSCSISGE", acit_params.num_sge);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TicksAllowedForInvalidation", acit_params.invalidation_ticks);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DefaultTicksAllowedForTgtIOs", acit_params.tgt_default_ticks);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_TimeoutFactor", acit_params.timeout_factor);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_IORetriesBeforeERL0", acit_params.io_retries_b4_erl0);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_IORetriesBeforeERR", acit_params.io_retries_b4_err);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_HardwareHealthCheckFreq", acit_params.health_check_freq);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumAsyncPDUs", acit_params.num_async_pdus);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_MaxCQEToProcessInIsr", acit_params.max_cqe_per_isr);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_IORetryAllowed", acit_params.io_retries);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICEnabled", acit_params.enable_dic);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICPolicy", acit_params.dic.policy);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICMinTimer", acit_params.dic.min_timer_val);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICMaxTimer", acit_params.dic.max_timer_val);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICTimerVariation", acit_params.dic.timer_increments);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICMinInterruptRate", acit_params.dic.min_interrupt_rate);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICMaxInterruptRate", acit_params.dic.max_interrupt_rate);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICIoThresholdForDIC", acit_params.dic.io_threshold_for_dic);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICIoResolutionForDIC", acit_params.dic.io_resolution_for_dic);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICInterruptResolution", acit_params.dic.interrupt_resolution);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICTargetThresholdForLoad", acit_params.dic.tgt_threshold_for_load);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICInterrruptVariation", acit_params.dic.interrupt_dec_steps);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_DICIoSafetyValue", acit_params.dic.io_safety_valve);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_SyncCrashDumpIOCompletion", acit_params.crash_dump.sync_completion);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_NumCQ", acit_params.num_cq);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_ChipResetDuringCrashDump", acit_params.crash_dump.dump_rst);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_AcitInterruptRedirection", acit_params.interrupt_redirection);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_OffloadFailureThreshold", acit_params.offload_failure_threshold);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_UERecovery", acit_params.ue_recovery);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Param_MaxUERecoveries", acit_params.max_ue_recoveries);
	}

}

void dump_osm_counters(se_hba *phba, struct info_str *info)
{
}

void dump_osm_eh_counters(se_hba *phba, struct info_str *info)
{
}

void dump_acit_fw_config(se_hba *phba, struct info_str *info)
{
	int idx = 0;
        DRVR_STATS_REQUEST request;
        ACIT_FW_CONFIG acit_fw_config;

        memset(&request, 0, sizeof(DRVR_STATS_REQUEST));
        request.ioctl_code = IOCTL_GET_ACIT_FW_CONFIG;
        request.buffer_len = sizeof(ACIT_FW_CONFIG);

	if(get_acit_stats(phba, &request, &acit_fw_config) == 0){
		idx = phba->hba_unique_id;

		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_iSCSICidStart", acit_fw_config.iscsi_cid_start);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_PhysicalPort", acit_fw_config.phys_port);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_ISROffset", acit_fw_config.isr_offset);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_iSCSIIcdStart", acit_fw_config.iscsi_icd_start);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_iSCSICidCount", acit_fw_config.iscsi_cid_count);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_iSCSIIcdCount", acit_fw_config.iscsi_icd_count);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_FunctionMode", acit_fw_config.function_mode);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_PCIFunction", acit_fw_config.pci_function);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWConfig_ConfigNumber", acit_fw_config.be_config_number);
	}
	
	return;

}
void dump_acit_fw_addi_attribs(se_hba *phba, struct info_str *info)
{
	int idx = 0;
        DRVR_STATS_REQUEST request;
	ACIT_FW_ADDI_ATTRIBS acit_fw_addi_attribs;

        memset(&request, 0, sizeof(DRVR_STATS_REQUEST));
        request.ioctl_code = IOCTL_GET_ACIT_FW_ADDI_ATTRIBS;
        request.buffer_len = sizeof(ACIT_FW_ADDI_ATTRIBS);

	if(get_acit_stats(phba, &request, &acit_fw_addi_attribs) == 0){
	
		idx = phba->hba_unique_id;
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWAttrib_IPLFileNo", acit_fw_addi_attribs.ipl_file_number);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWAttrib_IPLVersion", acit_fw_addi_attribs.ipl_file_version);
		copy_info(info,"C%d%-50s = %-d\n",idx, " FWAttrib_InterruptRedirection", acit_fw_addi_attribs.interrupt_redirection);
	}
	return;
}
void dump_acit_counters(se_hba *phba, struct info_str *info)
{
	int idx = 0;
        DRVR_STATS_REQUEST request;
        ACIT_COUNTERS  acit_counters;


	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));
	request.ioctl_code = IOCTL_GET_ACIT_COUNTERS;
	request.buffer_len = sizeof(ACIT_COUNTERS);

	if(get_acit_stats(phba, &request, &acit_counters) == 0){
		idx = phba->hba_unique_id;
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_ACITCurrentTickCount", acit_counters.current_tick_count);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_IOHandlesAvailable", acit_counters.io_handles_available);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_TMFHandlesAvailable", acit_counters.tmf_handles_available);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_LogoutHandlesAvailable", acit_counters.logout_handles_available);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_NOPHandlesAvailable", acit_counters.nop_handles_available);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_InvalidationHandlesAvailable", acit_counters.inv_handles_available);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_IOsPending", acit_counters.pending_ios);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_IOsQueued", acit_counters.queued_ios);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_TMFsPending", acit_counters.pending_tmfs);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_TargetsQueuedForTMF", acit_counters.tmf_queued_targets);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_TMFsQueued", acit_counters.queued_tmfs);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_NOPsPending", acit_counters.pending_nops);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_LogoutsPending", acit_counters.pending_logouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_IOCTLsPending", acit_counters.pending_ioctls);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_IOsInInvalidation", acit_counters.num_io_invs);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_TargetsInRecovery", acit_counters.num_targets_in_recovery);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_DelayedIOCTLs", acit_counters.delayed_ioctls);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Counters_CurrentLoggedInTargets", acit_counters.current_num_targets);

		g_pending_ioctls = acit_counters.pending_ioctls;
	}
    return;
}

void dump_acit_ctrl_stats(se_hba *phba, struct info_str *info)
{
	int idx = 0;
        DRVR_STATS_REQUEST request;
        ACIT_CONTROLLER_STATS  acit_ctrl_stats;

	unsigned int pi;

	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));

	request.ioctl_code = IOCTL_GET_ACIT_CONTROLLER_STATS;
	request.buffer_len = sizeof(ACIT_CONTROLLER_STATS);

	if(get_acit_stats(phba, &request, &acit_ctrl_stats) == 0){
		idx = phba->hba_unique_id;
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_Success", acit_ctrl_stats.io_status[0]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_Error", acit_ctrl_stats.io_status[1]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_TransportError", acit_ctrl_stats.io_status[2]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_CRCError", acit_ctrl_stats.io_status[3]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_TargetDeleted", acit_ctrl_stats.io_status[4]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_Aborted", acit_ctrl_stats.io_status[5]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_IOTimeout", acit_ctrl_stats.io_status[6]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_LinkDownTimeout", acit_ctrl_stats.io_status[7]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_TargetTimeout", acit_ctrl_stats.io_status[8]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AcitIOStatus_TargetError", acit_ctrl_stats.io_status[9]);

		copy_info(info,"C%d%-50s = %-lu\n",idx, " Stats_IOsSubmitted", acit_ctrl_stats.ios_submitted);
		copy_info(info,"C%d%-50s = %-lu\n",idx, " Stats_IOsCompleted", acit_ctrl_stats.ios_completed);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_MaxIOQueueDepth", acit_ctrl_stats.max_queue_depth);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DataSessionLogins", acit_ctrl_stats.data_session_logins);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DriverSessionLogouts", acit_ctrl_stats.data_session_drvr_logouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_FirmwareSessionLogouts", acit_ctrl_stats.data_session_armfw_logouts);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DefaultPDUHeaderPosted", acit_ctrl_stats.async_hdr_posted);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DefaultPDUBufferPosted", acit_ctrl_stats.async_data_posted);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_TxULPDoorbellRung", acit_ctrl_stats.txulp_db);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_MaxIOsQueuedDuringFlush", acit_ctrl_stats.max_num_queued);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_MaxIOsFlushedDuringFlush", acit_ctrl_stats.max_num_flushed);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_LastIOSubmissionTick", acit_ctrl_stats.last_io_tick);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_MaxCQEProcessedPerISR", acit_ctrl_stats.max_cqe_per_isr);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_MaxEQEProcessedPerISR", acit_ctrl_stats.max_eqe_per_isr);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DrvrMessagesPosted", acit_ctrl_stats.drvr_msg_sent);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DrvrMessagesReceived", acit_ctrl_stats.drvr_msg_recvd);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_DrvrMessagesMultiple Received", acit_ctrl_stats.drvr_msg_multiple_recvd);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_AsyncSCSIEvent", acit_ctrl_stats.async_scsi_recvd);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_SCSICmdUntagged", acit_ctrl_stats.untagged_cmd);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_TargetBusy", acit_ctrl_stats.target_busy);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_TargetReady", acit_ctrl_stats.target_ready);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ControllerBusy", acit_ctrl_stats.controller_busy);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ControllerReady", acit_ctrl_stats.controller_ready);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_IOCTLsSubmitted", acit_ctrl_stats.ioctls_submitted);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_IOCTLsCompleted", acit_ctrl_stats.ioctls_completed);

		for(pi = 0; pi < g_pending_ioctls; pi++)
		{
			char arr_name[64] = {0};
			sprintf(arr_name, " Stats_PendingIOCTL[%d]", pi);
			copy_info(info,"C%d%-50s = %-d\n",idx, arr_name, acit_ctrl_stats.pending_ioctl_info[pi]);
		}

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ReopenAllSessions", acit_ctrl_stats.reopen_all_sessions);

		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_NewTarget", acit_ctrl_stats.async_event_new_target);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_NewiSCSICxn", acit_ctrl_stats.async_event_new_iscsi_cxn);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_NewTCPCxn", acit_ctrl_stats.async_event_new_tcp_cxn);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_LinkDown", acit_ctrl_stats.async_event_link_down);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_LinkUp", acit_ctrl_stats.async_event_link_up);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ArmFwAsync_iSCSICos",acit_ctrl_stats.async_event_iscsi_cos);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ARMFwAsync_Unknown", acit_ctrl_stats.async_event_unknown);
		copy_info(info,"C%d%-50s = %-d\n",idx, " Stats_ARMFwAsync_Ignored", acit_ctrl_stats.async_event_ignored);
	}
	


}

void dump_acit_int_stats(se_hba *phba, struct info_str *info, int pidx)
{
	int idx = 0;
	DRVR_STATS_REQUEST request;
	ACIT_CONTROLLER_INT_STATS acit_int_stats;
	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));

	request.ioctl_code = IOCTL_GET_ACIT_INTERRUPT_STATS;
	request.buffer_len = sizeof(ACIT_CONTROLLER_INT_STATS);
	request.parameter1 = 0;

	if(get_acit_stats(phba, &request, &acit_int_stats) == 0){
                idx = phba->hba_unique_id;

		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_EQ Major", acit_int_stats.eq_maj_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQ", acit_int_stats.cq_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_DefaultPDU", acit_int_stats.dft_pdu_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_DrvrMessages", acit_int_stats.drvr_msg_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CommandCompletion", acit_int_stats.iscsi_cmd_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQ_Processed", acit_int_stats.iscsi_cq_processed);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_LastInterruptCompletionTick", acit_int_stats.last_isr_tick);


		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQEProcessedGT64", acit_int_stats.cqe_gt64);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQEProcessedGT32", acit_int_stats.cqe_gt32);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQEProcessedGT16", acit_int_stats.cqe_gt16);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_CQEProcessedLT16", acit_int_stats.cqe_lt16);

		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_UnknownInterrupts", acit_int_stats.unknown_ints);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_UnknownEQE", acit_int_stats.unknown_eqes);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_UnknownCQE", acit_int_stats.unknown_cqes);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_UnknownCmdCompletion", acit_int_stats.unknown_cmds);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_UnknownCmdError", acit_int_stats.unknown_cmd_errors);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, "Stats_InterruptStats_DelayedEQEDma", acit_int_stats.delayed_eq_dma);


    }

    return;
}

void dump_acit_dic_stats(se_hba *phba, struct info_str *info, int  pidx)
{
	int idx = 0;

	DRVR_STATS_REQUEST request;
	ACIT_CONTROLLER_DIC_COUNTERS acit_dic_counters;
	unsigned int eqi;

	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));

	request.ioctl_code = IOCTL_GET_ACIT_DIC_COUNTERS;
	request.buffer_len = sizeof(ACIT_CONTROLLER_DIC_COUNTERS);
	request.parameter1 = 0;

	if(get_acit_stats(phba, &request, &acit_dic_counters) == 0){
		idx = phba->hba_unique_id;

		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_Enabled", acit_dic_counters.enabled);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_ReconfigeInProgress", acit_dic_counters.reconfigure_in_progress);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_InterruptCount", acit_dic_counters.interrupts);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_MaxInterruptsPerSec", acit_dic_counters.max_interrupts_per_sec);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_IOsPerSec", acit_dic_counters.ios_per_sec);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_NumEQProcessed", acit_dic_counters.num_eq_processed);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_LastIOsPerSec", acit_dic_counters.last_ios_per_sec);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_MaxIOsPerSec", acit_dic_counters.max_ios_per_sec);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_MaxOverload", acit_dic_counters.max_overload);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_LastInterruptCount", acit_dic_counters.last_interrupt_count);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_InterruptPerSec", acit_dic_counters.interrupts_per_sec);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_CurrentEQTimer", acit_dic_counters.eq_timer);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_CurrentInterruptRate", acit_dic_counters.interrupt_rate);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_CurrentLoadFactor", acit_dic_counters.current_load_factor);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_CurrentTargetLoad", acit_dic_counters.current_target_load);
		copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx,"DIC_SafetyValveTriggers", acit_dic_counters.safety_valve_trigger);

		for(eqi = 0; eqi < 32; eqi++)
		{
			char arr_name[64] = {0};
			sprintf(arr_name, "DIC_EQRunsAtIndex[%d]", eqi);
			copy_info(info,"C%d MSIX[%d]%-42s = %-d\n",idx,pidx, arr_name, acit_dic_counters.eq_runs[eqi]);
		}

	}

	return;
}



void dump_acit_eh_stats(se_hba *phba, struct info_str *info)
{
	int idx = 0;
	DRVR_STATS_REQUEST request;
	unsigned int cqe;
	ACIT_CONTROLLER_EH_STATS acit_eh_stats;

	memset(&request, 0, sizeof(DRVR_STATS_REQUEST));
	request.ioctl_code = IOCTL_GET_ACIT_EH_STATS;
	request.buffer_len = sizeof(ACIT_CONTROLLER_EH_STATS);

	if(get_acit_stats(phba, &request, &acit_eh_stats) == 0){
		idx = phba->hba_unique_id;
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TargetRecoveries", acit_eh_stats.target_recoveries);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOTimeouts", acit_eh_stats.io_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFTimeouts", acit_eh_stats.tmf_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFInternalTimeouts", acit_eh_stats.tmf_internal_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFErrors", acit_eh_stats.tmf_errors);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LinkDownTimeouts", acit_eh_stats.ld_timeouts);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IORetryCount", acit_eh_stats.io_retry_count);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFsIssued", acit_eh_stats.tmfs_issued);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_OsmTmfs", acit_eh_stats.osm_tmfs);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_OsmAborts", acit_eh_stats.osm_aborts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_OsmLunResets", acit_eh_stats.osm_lun_resets);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_OsmTargetResets", acit_eh_stats.osm_target_resets);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionFailureCount", acit_eh_stats.ssn_failure_count);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionLogoutNormal", acit_eh_stats.ssn_logout_normals);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionLogoutOthers", acit_eh_stats.ssn_logout_others);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionLastFailureType", acit_eh_stats.last_ssn_failure_type);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionDigestErrors", acit_eh_stats.ssn_digest_errors);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionCxnTimeouts", acit_eh_stats.ssn_cxn_timeout_errors);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_SessionFormatErrors", acit_eh_stats.ssn_format_errors);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationStatusSuccess", acit_eh_stats.inv_sts_success);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationStatusIoctlError", acit_eh_stats.inv_sts_ioctl_err);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationStatusIoctlTimeout", acit_eh_stats.inv_sts_ioctl_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationStatusNotFound", acit_eh_stats.inv_sts_not_found);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationStatusError", acit_eh_stats.inv_sts_err);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOBusyDuringLunReset", acit_eh_stats.io_busy_during_lun_rst_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOQueuedDuringLunReset", acit_eh_stats.io_queued_during_lun_rst_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOResetPendingDuringLunReset", acit_eh_stats.io_rstpend_during_lun_rst_timeout);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOBusyDuringTargetReset", acit_eh_stats.io_busy_during_tgt_rst_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOQueuedDuringTargetReset", acit_eh_stats.io_queued_during_tgt_rst_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_IOResetPendingDuringTargetReset", acit_eh_stats.io_rstpend_during_tgt_rst_timeout);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_Success", acit_eh_stats.tmf_err_stats[0]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_Error", acit_eh_stats.tmf_err_stats[1]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_NotSupported", acit_eh_stats.tmf_err_stats[2]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_FunctionRejected", acit_eh_stats.tmf_err_stats[3]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_InvalidTask", acit_eh_stats.tmf_err_stats[4]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_InProgress", acit_eh_stats.tmf_err_stats[5]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_Timeout", acit_eh_stats.tmf_err_stats[6]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_LinkDownTimeout", acit_eh_stats.tmf_err_stats[7]);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TMFStatus_TargetTimeout", acit_eh_stats.tmf_err_stats[8]);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationDeferredTotal", acit_eh_stats.invs_deferred);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationDeferredNoMCCWrb", acit_eh_stats.invs_deferred_no_ioctls);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvalidationDeferredNoInvH", acit_eh_stats.invs_deferred_no_invh);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_CQErrorsInvalid", acit_eh_stats.cq_errors_invalid);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_OffloadErrors", acit_eh_stats.offload_failures);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_CommandWindowWaitTimeouts", acit_eh_stats.ping_wait_timeouts);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_ETOExpiry", acit_eh_stats.eto_expiry);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_CxnInvalidations", acit_eh_stats.cxn_invalidations);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_CxnUploads", acit_eh_stats.cxn_uploads);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_IoRetriesExceededLimit", acit_eh_stats.ioretry_exceeded);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_RejectReceived", acit_eh_stats.rejects);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_AysncLogout", acit_eh_stats.async_logouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_AsyncDropCxn", acit_eh_stats.async_reopens);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_AsyncParamRenego", acit_eh_stats.async_param_renego);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_TargetErrorInIO", acit_eh_stats.io_status_errors);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_RecoveryReason_UnknownDefaultPDU", acit_eh_stats.unknown_defpdu);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_ReopenSessionIdError", acit_eh_stats.reopen_mismatch);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LoginFailure_ARMFW", acit_eh_stats.login_failure_arm_fw);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LoginFailure_Driver", acit_eh_stats.login_failure_driver);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LogoutFailure_IOBusy", acit_eh_stats.logout_failure_io_busy);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LogoutFailure_Recovery", acit_eh_stats.logout_failure_recovery);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_LogoutFailure_ARMFW", acit_eh_stats.logout_failure_armfw_session);

		for(cqe = 0; cqe < 40; cqe++)
		{
			char arr_name[64] = {0};
			sprintf(arr_name, " EHStat_CQError[%d]", cqe);
			copy_info(info,"C%d%-50s = %-d\n",idx, arr_name, acit_eh_stats.cq_errors[cqe]);
		}

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvTimeOverrun", acit_eh_stats.inv_over_time_limit);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_InvMaxTicksOverLimit", acit_eh_stats.max_inv_ticks_over_limit);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfTimeoutAborts", acit_eh_stats.tmf_abort_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfTimeoutLunRst", acit_eh_stats.tmf_lunrst_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfTimeoutTgtRst", acit_eh_stats.tmf_tgtrst_timeouts);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfInternalTimeoutAborts", acit_eh_stats.tmf_abort_internal_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfInternalTimeoutLunRst", acit_eh_stats.tmf_lunrst_internal_timeouts);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_TmfInternalTimeoutTgtRst", acit_eh_stats.tmf_tgtrst_internal_timeouts);

		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_UELow", acit_eh_stats.ue_low);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_UEHigh", acit_eh_stats.ue_hi);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_FWLine", acit_eh_stats.ue_fw_line);
		copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_UEDetected", acit_eh_stats.ue_detected);
        copy_info(info,"C%d%-50s = %-d\n",idx, " EHStat_UERecoveries", acit_eh_stats.ue_recoveries);
	}
}

void dump_adapter_info(se_hba *phba, struct info_str *info)
{
	int pidx;
	dump_osm_params(phba,info);
	dump_acit_params(phba,info);
	dump_osm_counters(phba,info);
	dump_osm_eh_counters(phba,info);
	dump_acit_fw_config(phba,info);
	dump_acit_fw_addi_attribs(phba,info);
	dump_acit_counters(phba,info);
	dump_acit_ctrl_stats(phba,info);

	for(pidx=0; pidx < phba->num_cq; pidx++) {
		dump_acit_int_stats(phba,info,pidx);
		dump_acit_dic_stats(phba,info,pidx);
	}

	dump_acit_eh_stats(phba,info);


}

static int
be_get_stats(char *buffer, char **start, off_t offset, int length, int *eof, void *data)
{

	int retval = -EINVAL;
	struct info_str info;
	se_hba  *phba = NULL;
	if(data){
		phba = data;

		if (start)
			*start = buffer;

		info.buffer     = buffer;
		info.length     = length;
		info.offset     = offset;
		info.pos        = 0;

		dump_adapter_info(phba,&info);
		copy_info(&info, "\n\0");

		retval = info.pos > info.offset ? info.pos - info.offset : 0;
	}else{
		printk("data is NULL in be_get_stats\n");	
	}
	return retval;
}

static int
be_set_loglvl(struct file *file, const char *buffer, unsigned long count, void *data)
{

	char buf[64];
	se_hba  *phba = NULL;

	if(data){
		phba = data;

		if (count >= sizeof(buf)) {
			return -EINVAL;
		}
		memcpy(buf, buffer, count);
		buf[count] = '\0';

		log_level = simple_strtoul(buf, NULL, 16);
		printk("setting log_level to 0x%x \n",log_level);
		UPDATE_LOG_LEVEL(log_level);	
	}

	return (count);

}


static int
be_get_loglvl(char *buffer, char **start, off_t offset, int length, int *eof, void *data)
{
	
	int retval = -EINVAL;
	struct info_str info;
	se_hba  *phba = NULL;
	if(data){
		phba = data;

		if (start)
			*start = buffer;

		info.buffer     = buffer;
		info.length     = length;
		info.offset     = offset;
		info.pos        = 0;
		copy_info(&info, "Current log level is 0x%0x\n",log_level);
		copy_info(&info,"\n");
		copy_info(&info, "OSM Log Level Setting \n"
				 "----------------------\n"
				 "echo <log-level-value> > log_level\n"
				 "Available Log levels:\n"
				 "LOG_INIT_UNLOAD\t0x01\n"
				 "LOG_IO\t\t0x02\n"
				 "LOG_EH\t\t0x04\n"
				 "LOG_CONFIG\t0x08\n"
				 "LOG_IOCTL\t0x10\n");
		 copy_info(&info,"\n\0");

		retval = info.pos > info.offset ? info.pos - info.offset : 0;
	}else{
		printk("data is NULL in be_get_loglvl\n");	
	}
	*eof = 1;
	return retval;	



}

int se_proc_create(se_hba *phba)
{

	char dname[20] = {0};
	struct proc_dir_entry   *ent;
	struct Scsi_Host *shost = phba->host;
	struct scsi_host_template *sht = shost->hostt;
	if(sht == NULL)
		return -1;
	
	mutex_lock(&be_pgbl_mutex);
	if(!be_proc_bdir_present++){
		be_proc_root_dir = proc_mkdir(sht->name,NULL);
		if (be_proc_root_dir == NULL){
			printk("be_proc_root_dir is NULL\n");
			return -ENOTDIR;
		}
		/*OSM log level*/
		ent = create_proc_entry(OSM_LOG_LEVEL, S_IFREG|S_IRUGO,be_proc_root_dir);
		if(ent) {
			ent->read_proc = be_get_loglvl;
			ent->write_proc = be_set_loglvl;
			ent->data = phba;
		} else {
			OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Creating PROC Entry for %s failed\n",OSM_LOG_LEVEL);
			return -ENOMEM;
		}
	}
	mutex_unlock(&be_pgbl_mutex);


	sprintf(dname,"%s%02d",sht->name,phba->hba_unique_id);
	phba->phba_proc_dir =  proc_mkdir(dname,be_proc_root_dir);
	if (phba->phba_proc_dir == NULL){
		printk("phba_dir is NULL\n");
		return -ENOTDIR;
	}

	ent = create_proc_entry(DRIVER_STATS, S_IFREG|S_IRUGO, phba->phba_proc_dir);
	if (ent){
		ent->read_proc = be_get_stats;
		ent->data = phba;
	} else {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Creating PROC Entry for %s failed\n",DRIVER_STATS);
		return -ENOMEM;
	}


	return 0;
}

void se_proc_destroy(se_hba *phba)
{
	char dname[20] = {0};
	struct Scsi_Host *shost = phba->host;
        struct scsi_host_template *sht = shost->hostt;
	if(sht == NULL)
		return ;
	
	remove_proc_entry(DRIVER_STATS,phba->phba_proc_dir);

	sprintf(dname,"%s%02d",sht->name,phba->hba_unique_id);
	remove_proc_entry(dname,be_proc_root_dir);

	mutex_lock(&be_pgbl_mutex);
        if(!--be_proc_bdir_present && be_proc_root_dir){
		remove_proc_entry(OSM_LOG_LEVEL, be_proc_root_dir);
		remove_proc_entry(sht->name, NULL);
	}
        mutex_unlock(&be_pgbl_mutex);
}

#if defined(__VMKERNEL_MODULE__) || defined(LINUX_2_4)
ACIT_SEXACT32 		
se_proc_info(
		char	*buffer, 
		char	**start, 
		off_t	offset,
		int	length, 
		int	inode, 
		int	inout
	    )
{
#ifdef BE2ISCSI_DEBUG
	int			len;
	unsigned		i;
	u64			avg;
	struct list_head	*hba_entry;
	struct list_head	*target_entry;
	se_hba			*phba;
	se_target		*ptarget;
	se_device		*pdevice;
	unsigned		hba_index;
	unsigned		target_index;
	unsigned		lun_index;
	PACIT_TARGET_ENTRY	ptarget_entry; 

	len = 0;

	if (inout) {
		if (strncmp(buffer, "clear", 5) == 0) {
			printk("clearing\n");

			hba_index = 0;
			target_index = 0;
			lun_index = 0;

			clear_device_stats();

			return length;
		}
		if (strncmp(buffer, "help", 4) == 0) {
			printk("clear             - clear statistics\n");
		}
		printk("unknown: ");
		for (i = 0 ; i < length ; i++)
		{
			printk("%c", *(buffer + i));
		}
		printk("\n");
		
		return -EINVAL;
	}

	hba_index = 0;

        list_for_each(hba_entry, &hba_chain) {

		target_index = 0;

                phba = list_entry(hba_entry, se_hba, hba_queue);

		hba_index++;

		len += sprintf(buffer + len, "HBA%d - ", hba_index);

		list_for_each_entry(ptarget, &phba->target_list, sibling_list) {

			lun_index = 0;

			target_index++;

			list_for_each_entry(pdevice, &ptarget->lun_list,
				sibling_list) {

				len += sprintf(buffer + len,
					"%2d|%c ",
						pdevice->num_ios_outstanding,
						(ptarget->state & SE_TARGET_BUSY?'B':' '));
				lun_index++;
			}
		}

		len += sprintf(buffer + len, "\n");

		ptarget_entry = &phba->pacit_target_table->target_entry[0];

		for (i = 0; i < phba->pacit_target_table->total_entries; i++)
		{
			if ((ptarget_entry[i].acit_tflags & ACIT_TFLAG_IN_USE)
				== 0)
			{
				continue;
			}

			len += sprintf(buffer + len, "  %d: phba=0x%x | "
				"ptarget_key=0x%x | ",
				i, phba, ptarget_entry[i].ptarget_key);

			if (ptarget_entry[i].ptarget_key != NULL)
			{
				len += sprintf(buffer + len, "osm_key %x\n",
					ptarget_entry[i].ptarget_key->osm_key);
			}
			else
			{
				len += sprintf(buffer + len, "osm_key NULL\n");
			}
		}
	}

	if (offset > len)
		return 0;

	*start = buffer + offset;
	len -= offset;

	if (len > length)
		len = length;

	return len;
#else
	return 0;
#endif
}
#endif

static void se_qdetails(char *str,int *qid, int *options)
{
	char *start,*end;

	start = str;

	while(isspace(*start))
		start++;

	*qid = simple_strtol(start,&end,0);

	while(isspace(*end))
		end++;

	*options = simple_strtol(end,&end,0);
}

#ifndef __VMKLNX__
static int se_iosqd_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	
	acit_process_proc_iosqd(
			&phba->controller_key,
			pse_proc->iosqd_qinfo.qid,
			pseq);
	
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;
}

static int se_invl_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_invl(
			&phba->controller_key,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_tmf_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
		
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	
	acit_process_proc_tmf(
			&phba->controller_key,
			pse_proc->tmf_qinfo.qid,
			pseq);
	
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_intr_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_intr(
			&phba->controller_key,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_etofind_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_etofind(
			&phba->controller_key,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_cqfind_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_cqfind(
			&phba->controller_key,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_cq_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

		
	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);

	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_cq(
			&phba->controller_key,
			pse_proc->cq_qinfo.qid,
			pse_proc->cq_qinfo.options,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_eqfind_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;
	
	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);

	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_eqfind(
			&phba->controller_key,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_eq_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
		
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_eq(
			&phba->controller_key,
			pse_proc->eq_qinfo.qid,
			pse_proc->eq_qinfo.options,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_wrb_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
		
	/* Check for WRB_IDX and CID Valid */
	if( pse_proc->wrb_qinfo.qid < 0 ||
		pse_proc->wrb_qinfo.options < 0){
		printk(KERN_ERR"Enter Proper WRB_IDX and CID\n");
		return -EINVAL;
	}

	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_wrb(
			&phba->controller_key,
			pse_proc->wrb_qinfo.qid,
			pse_proc->wrb_qinfo.options,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;

}

static int se_icdfind_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);

	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_icdfind(
			&phba->controller_key,
			pse_proc->icdfind_qinfo.qid,
			pse_proc->icdfind_qinfo.options,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;
}

static int se_icd_read(struct seq_file *pseq,void *v)
{
	ACIT_U32_64			flags = 0;
	struct se_pentry	*pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	acit_process_proc_icd(
			&phba->controller_key,
			pse_proc->icd_qinfo.qid,
			pseq);
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

	return 0;
}

#define DISP_CUR_CID	0x0C
#define DISP_VAL_CID	0x0A
#define DISP_ALL_CID	0x0F

static int se_cid_read(struct seq_file *pseq,void *v)
{
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	ACIT_DBG_DUMP_CHECK(pse_proc,pseq);
	/* Check the options passed */
	if((pse_proc->cid_qinfo.options != DISP_CUR_CID) &&
	   (pse_proc->cid_qinfo.options != DISP_VAL_CID) &&
	   (pse_proc->cid_qinfo.options != DISP_ALL_CID)){

		printk(KERN_ERR"Flag Entered is not Proper\n");
		return -EINVAL;
	}

	/* Check the CID passed */
	if( pse_proc->cid_qinfo.options == DISP_CUR_CID &&
			pse_proc->cid_qinfo.qid < 0 ){
		printk(KERN_ERR"CID Entered is not proper\n");
		return -EINVAL;
	}

	acit_process_proc_cid(
			&phba->controller_key,
			pse_proc->cid_qinfo.qid,
			pse_proc->cid_qinfo.options,
			pseq);

	return 0;
}


static int dbg_log_activate(struct seq_file *pseq,void *v)
{
	struct se_pentry *pse_proc;
	se_hba *phba = (se_hba *)pseq->private;

	pse_proc = phba->pse_pentry;

	seq_printf(pseq,"The Log option is : ");
	if(pse_proc->log_info.qid)
		seq_printf(pseq," Enabled\n");
	else
		seq_printf(pseq," Disabled\n");

	return 0;
}

static int se_qwrite_read(struct seq_file *pseq,void *v)
{
	seq_printf(pseq,"No Read Option for this file\n");
	seq_printf(pseq,"On how to pass parameters Read Help file\n");
	
	return 0;
}

static struct dentry *
se_dentry_crea(struct se_pdetails *pde_dtl)
{

	struct dentry *pde = NULL;
	
	/* Check if the debugfs creation is for DIR */
	if(pde_dtl->mode & S_IFDIR)
		return (debugfs_create_dir(pde_dtl->pname,
					pde_dtl->parent));

	pde = debugfs_create_file(pde_dtl->pname,
			pde_dtl->mode,
			pde_dtl->parent,
			pde_dtl->data,
			&pde_dtl->seq_data->seq_fops);

	return pde;
}

static ssize_t se_qwrite_proc(struct file *filep,
		const char __user *buff,size_t len,
		loff_t *off)
{
	char *start;
	int qid,options;
	char tmp_buff[Q_WLEN];
	
	se_hba *phba;
	struct se_pentry *pse_proc;

	if(len> Q_WLEN) {
		printk(" Passed Args Length(%ld) > Args Buff length(%d)\n",
				len,Q_WLEN);
		return -ENOSPC;
	}

	if(copy_from_user(tmp_buff,buff,len))
		return -EFAULT;

	tmp_buff[len] = '\0';


	/* Iterate till the Q details
	 * are are encountered.
	 */
	start = tmp_buff;
	while(isspace(*start) ||
			isalpha(*start))
		start++;

	se_qdetails(start,&qid,&options);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	phba = (se_hba *)filep->f_dentry->d_inode->i_private;
#else
	phba = (se_hba *)filep->f_dentry->d_inode->u.generic_ip;
#endif
	if(!phba){
		printk(KERN_ERR" phba is NULL\n");
		return -EFAULT;
	}
	
	pse_proc = phba->pse_pentry;

	/* Iterate till the first char
	 * is encountered
	 */
	start = tmp_buff;
	while(isspace(*start))
		start++;

	if(!strnicmp(start,CID_QTYPE,strlen(CID_QTYPE))){
		pse_proc->cid_qinfo.qid = qid;
		pse_proc->cid_qinfo.options = options;

	}
	else if(!strnicmp(start,ICD_QTYPE,strlen(ICD_QTYPE))){
		pse_proc->icd_qinfo.qid = qid;
		pse_proc->icd_qinfo.options = options;

	}
	else if(!strnicmp(start,ICDFIND_QTYPE,strlen(ICDFIND_QTYPE))){
		pse_proc->icdfind_qinfo.qid = qid;
		pse_proc->icdfind_qinfo.options = options;
	}
	else if(!strnicmp(start,WRB_QTYPE,strlen(WRB_QTYPE))){
		pse_proc->wrb_qinfo.qid = qid;
		pse_proc->wrb_qinfo.options = options;
	}
	else if(!strnicmp(start,EQ_QTYPE,strlen(EQ_QTYPE))){
		pse_proc->eq_qinfo.qid = qid;
		pse_proc->eq_qinfo.options = options;
	}
	else if(!strnicmp(start,CQ_QTYPE,strlen(CQ_QTYPE))){
		pse_proc->cq_qinfo.qid = qid;
		pse_proc->cq_qinfo.options = options;
	}
	else if(!strnicmp(start,TMF_QTYPE,strlen(TMF_QTYPE))){
		pse_proc->tmf_qinfo.qid = qid;
	}
	else if(!strnicmp(start,IOSQUEUED_QTYPE,strlen(IOSQUEUED_QTYPE))){
		pse_proc->iosqd_qinfo.qid = qid;
	}
	else if(!strnicmp(start,ACIT_LOG,strlen(ACIT_LOG))){
		pse_proc->log_info.qid = qid;
	}
	else
		return -EINVAL;

	return len;
}

static int se_single_open_seq(struct inode *inode, struct file *file)
{
	struct se_seq_data *pseq_data = NULL;

	pseq_data = container_of(inode->i_fop,struct se_seq_data,seq_fops);	

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
	return single_open(file,pseq_data->show,inode->i_private);
#else
	return single_open(file,pseq_data->show,inode->u.generic_ip);
#endif

}


static int se_padd_entries(se_hba *phba)
{
	struct dentry	*pde;
	struct se_pdetails pde_details;
	struct se_pentry *pse_proc;

	pse_proc = phba->pse_pentry;

	/* debugfs entry to set Q Details */
	SE_PDE_PARAMS_SET(pde_details,SE_QWRITE,\
			S_IWUGO,pse_proc->dbg_fdir,phba,\
			&se_qwrite_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde){
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_QWRITE);
		return -ENOMEM;
	}
	pse_proc->qdata_pde = pde;

	/* debugfs entry to set the CID details */
	SE_PDE_PARAMS_SET(pde_details,SE_CID_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_cid_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_CID_READ);
		return -ENOMEM;
	}
	pse_proc->cid_pde = pde;

	/* debugfs entry to set the ICD details */
	SE_PDE_PARAMS_SET(pde_details,SE_ICD_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_icd_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_ICD_READ);
		return -ENOMEM;
	}
	pse_proc->icd_pde = pde;
	
	/* debugfs entry to set the ICD FIND details */
	SE_PDE_PARAMS_SET(pde_details,SE_ICDFIND_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_icdfind_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_ICDFIND_READ);
		return -ENOMEM;
	}
	pse_proc->icdfind_pde = pde;

	/* debugfs entry to set the WRB details */
	SE_PDE_PARAMS_SET(pde_details,SE_WRB_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_wrb_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_WRB_READ);
		return -ENOMEM;
	}
	pse_proc->wrb_pde = pde;

	/* debugfs entry to set the EQ details */
	SE_PDE_PARAMS_SET(pde_details,SE_EQ_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_eq_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_EQ_READ);
		return -ENOMEM;
	}
	pse_proc->eq_pde = pde;

	/* debugfs entry to set the EQFIND details */
	SE_PDE_PARAMS_SET(pde_details,SE_EQFIND_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_eqfind_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_EQFIND_READ);
		return -ENOMEM;
	}
	pse_proc->eqfind_pde = pde;

	/* debugfs entry to set the CQ details */
	SE_PDE_PARAMS_SET(pde_details,SE_CQ_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_cq_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_CQ_READ);
		return -ENOMEM;
	}
	pse_proc->cq_pde = pde;

	/* debugfs entry to set the CQFIND details */
	SE_PDE_PARAMS_SET(pde_details,SE_CQFIND_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_cqfind_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_CQFIND_READ);
		return -ENOMEM;
	}
	pse_proc->cqfind_pde = pde;

	/* debugfs entry to set the ETOFIND details */
	SE_PDE_PARAMS_SET(pde_details,SE_ETOFIND_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_etofind_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_ETOFIND_READ);
		return -ENOMEM;
	}
	pse_proc->etofind_pde = pde;

	/* debugfs entry to set the INTR details */
	SE_PDE_PARAMS_SET(pde_details,SE_INTR_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_intr_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_INTR_READ);
		return -ENOMEM;
	}
	pse_proc->intr_pde = pde;


	/* debugfs entry to set the INV details */
	SE_PDE_PARAMS_SET(pde_details,SE_INVL_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_invl_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_INVL_READ);
		return -ENOMEM;
	}
	pse_proc->invl_pde = pde;

	/* debugfs entry to set the TMF details */
	SE_PDE_PARAMS_SET(pde_details,SE_TMF_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_tmf_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_TMF_READ);
		return -ENOMEM;
	}
	pse_proc->tmf_pde = pde;

	/* debugfs entry to set the IOS QD details */
	SE_PDE_PARAMS_SET(pde_details,SE_IOSQD_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_iosqd_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_IOSQD_READ);
		return -ENOMEM;
	}
	pse_proc->iosqd_pde = pde;

	/* debugfs entry to enable/disable logging */
	SE_PDE_PARAMS_SET(pde_details,SE_LOG_OP,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&se_log_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_LOG_OP);
		return -ENOMEM;
	}
	pse_proc->dbg_log_pde = pde;

	/* debugfs entry to set the Debugging Help details */
	SE_PDE_PARAMS_SET(pde_details,SE_HELP_READ,\
			S_IRUGO,pse_proc->dbg_fdir,phba,\
			&dbg_help_seq);

	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugs_create failed : %s\n",
				SE_HELP_READ);
		return -ENOMEM;
	}
	pse_proc->dbg_help_pde = pde;

	return 0;
}


ACIT_SEXACT32
se_proc_add_hostdir(se_hba *phba)
{

	char dname[20] = {0};
	ACIT_SEXACT32 rval = 0;
	struct dentry *pde;
	struct se_pentry *pse_proc;
	struct se_pdetails pde_details;
	struct Scsi_Host *shost = phba->host;
	struct scsi_host_template *sht = shost->hostt;

	mutex_lock(&se_pgbl_mutex);

	if(!se_proc_bdir_present++){

		/* Creating the Base DIR */
		SE_PDE_PARAMS_SET(pde_details,sht->proc_name,\
				S_IFDIR,NULL,NULL,NULL);
		se_gbl_pdir = se_dentry_crea(&pde_details);

		if(!se_gbl_pdir) {
			OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugfs_mkdir failed : %s\n",
					sht->proc_name);
			mutex_unlock(&se_pgbl_mutex);
			return -ENOMEM;
		}
	}

	mutex_unlock(&se_pgbl_mutex);

	pse_proc = (struct se_pentry *)
		kzalloc(sizeof(struct se_pentry),GFP_KERNEL);
	if(!pse_proc){
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Memory Allocation Failed\n");
		return -ENOMEM;
	}

	/* function/DIR name combination of
	 * Bus_Num + Slot + Function  
	 */
	sprintf(dname,"%s_%02x%02x%02x",sht->proc_name,
			 phba->pcidev->bus->number,
			 PCI_SLOT(phba->pcidev->devfn),
			 PCI_FUNC(phba->pcidev->devfn));

	/* create function/host specific debug DIR */
	SE_PDE_PARAMS_SET(pde_details,dname,S_IFDIR,\
			se_gbl_pdir,NULL,NULL);
	pde = se_dentry_crea(&pde_details);
	if(!pde) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"debugfs_mkdir failed : %s\n",dname);
		kfree(pse_proc);
		return -ENOMEM;
	}

	pse_proc->dbg_fdir= pde;
	phba->pse_pentry = pse_proc;

	rval = se_padd_entries(phba);

	return (rval ? rval : 0);
}


void se_prm_entries(se_hba *phba)
{
	struct se_pentry *pse_proc;

	pse_proc = phba->pse_pentry;
	
	/* Delete the IOS QD debugfs entry */
	if(pse_proc->iosqd_pde) {
		debugfs_remove(pse_proc->iosqd_pde);
		pse_proc->iosqd_pde = NULL;
	}

	/* Delete the TMF debugfs entry */
	if(pse_proc->tmf_pde) {
		debugfs_remove(pse_proc->tmf_pde);
		pse_proc->tmf_pde = NULL;
	}

	/* Delete the INVL debugfs entry */
	if(pse_proc->invl_pde) {
		debugfs_remove(pse_proc->invl_pde);
		pse_proc->invl_pde = NULL;
	}

	/* Delete the INTR debugfs entry */
	if(pse_proc->intr_pde) {
		debugfs_remove(pse_proc->intr_pde);
		pse_proc->intr_pde = NULL;
	}

	
	/* Delete the ETOFIND debugfs entry */
	if(pse_proc->etofind_pde) {
		debugfs_remove(pse_proc->etofind_pde);
		pse_proc->etofind_pde = NULL;
	}

	/* Delete the CQFIND debugfs entry */
	if(pse_proc->cqfind_pde) {
		debugfs_remove(pse_proc->cqfind_pde);
		pse_proc->cqfind_pde = NULL;
	}

	/* Delete the CQ debugfs entry */
	if(pse_proc->cq_pde) {
		debugfs_remove(pse_proc->cq_pde);
		pse_proc->cq_pde = NULL;
	}
	
	/* Delete the EQFIND debugfs entry */
	if(pse_proc->eqfind_pde) {
		debugfs_remove(pse_proc->eqfind_pde);
		pse_proc->eqfind_pde = NULL;
	}

	/* Delete the EQ debugfs entry */
	if(pse_proc->eq_pde) {
		debugfs_remove(pse_proc->eq_pde);
		pse_proc->eq_pde = NULL;
	}

	/* Delete the WRB debugfs entry */
	if(pse_proc->wrb_pde) {
		debugfs_remove(pse_proc->wrb_pde);
		pse_proc->wrb_pde = NULL;
	}
	
	/* Delete the ICDFIND debugfs entry */
	if(pse_proc->icdfind_pde) {
		debugfs_remove(pse_proc->icdfind_pde);
		pse_proc->icdfind_pde = NULL;
	}

	/* Delete the ICD debugfs entry */
	if(pse_proc->icd_pde) {
		debugfs_remove(pse_proc->icd_pde);
		pse_proc->icd_pde = NULL;
	}

	/* Delete the CID debugfs entry */
	if(pse_proc->cid_pde) {
		debugfs_remove(pse_proc->cid_pde);
		pse_proc->cid_pde = NULL;
	}

	/* Delete the Q Info debugfs entry */
	if(pse_proc->qdata_pde) {
		debugfs_remove(pse_proc->qdata_pde);
		pse_proc->qdata_pde = NULL;
	}

	/* Delete the Log debugfs entry */
	if(pse_proc->dbg_log_pde) {
		debugfs_remove(pse_proc->dbg_log_pde);
		pse_proc->dbg_log_pde = NULL;
	}
	
	/* Delete the Help debugfs entry */
	if(pse_proc->dbg_help_pde){
		debugfs_remove(pse_proc->dbg_help_pde);
		pse_proc->dbg_help_pde = NULL;
	}

}

void  se_proc_rm_hostdir(se_hba *phba)
{
	struct se_pentry *pse_proc = phba->pse_pentry;

	se_prm_entries(phba);

	/* function/DIR name */
	if(pse_proc->dbg_fdir)
		debugfs_remove(pse_proc->dbg_fdir);

	if(pse_proc)
		kfree(pse_proc);

	mutex_lock(&se_pgbl_mutex);

	if(!--se_proc_bdir_present && se_gbl_pdir)
		debugfs_remove(se_gbl_pdir);

	mutex_unlock(&se_pgbl_mutex);

}

static int se_help_read(struct seq_file *pseq,void *v)
{
	seq_printf(pseq,"   help													\n"
		"   Shows this help.                                                    \n"
		"                                                                       \n"
		"Log Enable/Disble                                                          \n"
		"-----------------                                                          \n"
		" Enable/Disable dumping the logs											\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"dbg_log 0/1\" > debug_data									\n"
		"       0 -  Disable 1 - Enable 											\n"
		" 	                                                                        \n"
		" Read the Log Status                                                       \n"
		" 	    cat dbg_log_status                                                  \n"
		"CID                                                                        \n"
		"---                                                                        \n"
		" Prints the Information about the passed CID								\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"cid <cid_number> <flag>\" > debug_data						\n"
		"       cid   : The CID info to be dumped                                   \n"
		"       flag  : 															\n"
		"			0xC - Dump the Passed CID Number                            	\n"
		"           0xA - All Active CID, the CID Number passed is ignored.         \n"
		"           0xF - All CID, the CID Number passed is ignored.                \n"
		"                                                                           \n"
		"           Eg: echo \"cid 13 0xC\"  > debug_data -- Dump CID 13			\n"
		"           Eg: echo \"cid 0 0xA\" > debug_data   -- Dump all ACTIVE CID	\n"
		"           Eg: echo \"cid 0 0xF\" > debug_data   -- Dump all CID			\n"
		"                                                                           \n"
		" ==> DUMP THE CID INFO                                                     \n"
		"       cat cid_data														\n"
		"																			\n"
		"																			\n"
		"ICD                                                                        \n"
		"---                                                                        \n"
		" Prints the Information about the passed ICD								\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"icd <icd_number> 0\" > debug_data							\n"
		"       icd   : The ICD info to be dumped                                   \n"
		"																			\n"
		"           Eg: echo \"icd 13 0  \"  > debug_data -- Dump ICD 13			\n"
		"                                                                           \n"
		" ==> DUMP THE ICD INFO                                                     \n"
		"       cat icd_data														\n"
		"																			\n"
		"																			\n"
		"FINDICD                                                                    \n"
		"-------                                                                    \n"
		" Find all the ICD's pending with the Hardware								\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"findicd <cid_number> <count>\" > debug_data					\n"
		"       cid   : The CID info to be dumped                                   \n"
		"																			\n"
		" Eg: 																		\n"
		"   echo \"findicd 13 0 \"  > debug_data 									\n"
		"	-- Dump All PENDING ICD's of CID 13										\n"
		"																			\n"
		"   echo \"findicd -1 0 \"  > debug_data 									\n"
		"	-- Dump All PENDING ICD's 												\n"
		"                                                                           \n"
		"   echo \"findicd -1 -1\"  > debug_data 									\n"
		"	-- Dump All Last COMPLETED ICD's 										\n"
		"                                                                           \n"
		"   echo \"findicd  9 -1\"  > debug_data 									\n"
		"	-- Dump All Last COMPLETED ICD's on CID 9								\n"
		"                                                                           \n"
		"   echo \"findicd  9 5\"  > debug_data 									\n"
		"	-- Dump Last 5 COMPLETED ICD's on CID 9									\n"
		"                                                                           \n"
		" ==> DUMP THE ICD INFO                                                     \n"
		"       cat icd_data														\n"
		"																			\n"
		"WRB                                                                        \n"
		"---                                                                        \n"
		" Prints the Information about the passed WRB								\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"wrb <wrb_idx> <cid>\" > debug_data							\n"
		"		wrb_idx	: The WRB_IDX info to be dumped								\n"
		"		cid		: The CID to which the WRB_IDX belongs						\n"
		"																			\n"
		"   Eg: echo \"wrb 13 5  \"  > debug_data 									\n"
		"	-- Dump WRB_IDX 13 of CID 5												\n"
		"                                                                           \n"
		" ==> DUMP THE WRB INFO                                                     \n"
		"       cat wrb_data														\n"
		"																			\n"
		"EQ/CQ																		\n"
		"--																			\n"
		" Prints the Information about the passed EQ/CQ								\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"eq/cq <eq/cq_id> <count>\" > debug_data						\n"
		"		eq/cq_id	: The EQ/CQ_ID to be dumped								\n"
		"		count		: No of Enteries to be dumped							\n"
		"																			\n"
		"   Eg: echo \"eq 13 5  \"  > debug_data									\n"
		"	-- Dump EQ_ID 13 NEXT 5 Enteries										\n"
		"																			\n"
		"   Eg: echo \"cq 13 5  \"  > debug_data									\n"
		"	-- Dump CQ_ID 13 NEXT 5 Enteries										\n"
		"                                                                           \n"
		"   Eg: echo \"eq 13 -5  \"  > debug_data 									\n"
		"	-- Dump EQ_ID 13 LAST 5 Enteries										\n"
		"                                                                           \n"
		"   Eg: echo \"cq 13 -5  \"  > debug_data 									\n"
		"	-- Dump CQ_ID 13 LAST 5 Enteries										\n"
		"                                                                           \n"
		"   Eg: echo \"eq -1 5  \"  > debug_data 									\n"
		"	-- Dump All EQ_ID's  NEXT 5 Enteries									\n"
		"                                                                           \n"
		"   Eg: echo \"cq -1 5  \"  > debug_data 									\n"
		"	-- Dump All CQ_ID's  NEXT 5 Enteries									\n"
		"                                                                           \n"
		" ==> DUMP THE EQ/CQ_ID INFO                                                \n"
		"       cat eq_data															\n"
		"       cat cq_data															\n"
		"																			\n"
		"																			\n"
		"EQ/CQ_FIND																	\n"
		"----------																	\n"
		" Prints all  EQ_ID's or CQ_ID's											\n"
		" ==> DUMP THE EQ_ID INFO													\n"
		"       cat findeq_data														\n"
		"																			\n"
		" ==> DUMP THE CQ_ID INFO													\n"
		"       cat findcq_data														\n"
		"																			\n"
		"ETOFIND                                                                    \n"
		"-------                                                                    \n"
		" Prints All the Target WorkSpace that have ETO Expired						\n"
		"                                                                           \n"
		" ==> DUMP THE ETO FIND														\n"
		"       cat etofind_data													\n"
		"																			\n"
		"																			\n"
		"INTERRUPT 																	\n"
		"---------                                                                  \n"
		" Prints Interrupt Statistics												\n"
		"                                                                           \n"
		" ==> DUMP THE Interrupt Stats 												\n"
		"       cat intr_data														\n"
		"																			\n"
		"																			\n"
		"INVALIDATION 																\n"
		"---------                                                                  \n"
		" Dump the List of ICD invalidations that are pending 						\n"
		"                                                                           \n"
		" ==> DUMP THE Invalidation Info											\n"
		"       cat invalid_data													\n"
		"																			\n"
		"TMFWALK																	\n"
		"---------                                                                  \n"
		" Dump all the TMF that are queued and waiting to be submitted to HW		\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"tmfwalk	<cid> 0\" > debug_data								\n"
		"		cid : The CID specific TMF will only be dumped.						\n"
		"                                                                           \n"
		"   Eg: echo \"tmfwalk 13 0  \"  > debug_data 								\n"
		"	-- Dump TMF's for the CID 13 \n"
		"                                                                           \n"
		"   Eg: echo \"tmfwalk -1 0  \"  > debug_data 								\n"
		"	-- Dump All the queued TMF's 											\n"
		"                                                                           \n"
		" ==> DUMP The TMF Info														\n"
		"       cat tmfwalk_data													\n"
		"																			\n"
		"IOSQUEUED																	\n"
		"---------                                                                  \n"
		" Dump the IO's queued in the driver internal queue,waiting to be submitted	\n"
		" to the Hardware															\n"
		" ==> PASS THE PARAMTERS                                                    \n"
		"       echo \"iosqueue <cid> 0\" > debug_data								\n"
		"		cid : The CID specific IO's Q will only be dumped.					\n"
		"                                                                           \n"
		"   Eg: echo \"iosqueue 13 0  \"  > debug_data 								\n"
		"	-- Dump IO's queued for the CID 13										\n"
		"                                                                           \n"
		"   Eg: echo \"iosqueue -1 0  \"  > debug_data 								\n"
		"	-- Dump All the queued IO's 											\n"
		"                                                                           \n"
		" ==> DUMP The IO's Queued Info												\n"
		"       cat iosqd_data														\n"
		"																			\n"
		"																			\n");

	return 0;
}

#endif /* __VMKLNX__ */
