/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */


/*++

    Filename:

        be2iscsi_debug.h

    Purpose:

        This file contains macro definitions for supporting debug


--*/

/*
 *
 *  Common Debug code for driver
 *
 */
#ifndef _BE2ISCSIDEBUG_H_
#define _BE2ISCSIDEBUG_H_
/*  Assertions and KDB codes  */
#if defined(BE2ISCSI_DEBUG) && !defined(_IA64_)
#define KDBENTER() panic()
#define ASSERT(x) if(!(x)) __asm__ volatile ("int $3");
#else
#define KDBENTER() panic()
#define ASSERT(x)
#endif

#define BEGIN          /**/
#define END            /**/
#define ERROR_END      /**/

#define YES 1
#define NO  0

#define BE2ISCSI_STABLE_CODE NO

#if BE2ISCSI_STABLE_CODE
 #define RETRY_CODE NO
#else
 #define RETRY_CODE YES 
#endif

/* Check if debug_info is enabled to dump the data */
#define ACIT_DBG_DUMP_CHECK(pde,pseq)\
	do{						\
		if(!(pde->log_info.qid)) {						\
			seq_printf(pseq,"Enable Logging Support\n");\
			return 0;									\
		}												\
	}while(0);											\


/* OSM Logging Switch */
extern unsigned int log_init_unload;
extern unsigned int log_io;
extern unsigned int log_eh;
extern unsigned int log_config;
extern unsigned int log_ioctl;

#define LOG_INIT_UNLOAD_MASK	0x01
#define LOG_IO_MASK		0x02
#define LOG_EH_MASK		0x04
#define LOG_CONFIG_MASK		0x08
#define LOG_IOCTL_MASK		0x10

#define UPDATE_LOG_LEVEL( level)\
	log_init_unload = (level & LOG_INIT_UNLOAD_MASK)? 1 : 0;\
	log_io = (level & LOG_IO_MASK)? 1 : 0;\
	log_eh = (level & LOG_EH_MASK)? 1 : 0;\
	log_config = (level & LOG_CONFIG_MASK)? 1 : 0;\
	log_ioctl = (level & LOG_IOCTL_MASK)? 1 : 0

#define GET_LOGLEVEL()\
	((log_init_unload == 1)?LOG_INIT_UNLOAD_MASK : 0x00)|\
	((log_io)?LOG_IO_MASK : 0x00)|\
	((log_eh)?LOG_EH_MASK : 0x00)|\
	((log_config)? LOG_CONFIG_MASK : 0x00)|\
	((log_ioctl)? LOG_IOCTL_MASK : 0x00)

#define PRINT_LOG(phba, level, fmt, args...)\
			if(phba)\
                                dev_printk(level , &((phba)->pcidev)->dev, \
                                fmt, ##args); \
                        else\
                                printk(level BE2ISCSI_DRIVER fmt,\
                                         ##args);\

/* To be used during driver load and unload */
#define OSM_LOG_INIT_UNLOAD(phba, level, fmt, args...)	\
	do{\
		if(unlikely(log_init_unload)) {\
			PRINT_LOG(phba, level, fmt, ##args)\
		}\
	}while(0)

#define OSM_LOG_IO(phba, level, fmt, args...)		\
	do{											\
		if(unlikely(log_io)) {					\
			PRINT_LOG(phba, level, fmt, ##args)\
		}										\
												\
	}while(0);								\

/* Used during the Error Handle Path */
#define OSM_LOG_EH(phba, level, fmt, args...)	\
	do{										\
		if(unlikely(log_eh)) {				\
			PRINT_LOG(phba, level, fmt, ##args)\
		}									\
	}while(0);								\

/* Used during the ioctl Path */
#define OSM_LOG_IOCTL(phba, level, fmt, args...)	\
	do{										\
		if(unlikely(log_ioctl)) {				\
			PRINT_LOG(phba, level, fmt, ##args)\
		}									\
	}while(0);								\

/* Used in the configuration/Control Path */
#define OSM_LOG_CONFIG(phba, level, fmt, args...)	\
	do{										\
		if(unlikely(log_config)) {			\
			PRINT_LOG(phba, level, fmt, ##args)\
		}									\
	}while(0);								\

#define PROC_RETURN_FROM_FUNC(p,pg,off,start,eof,len) \
	do {										\
		len = ( p - pg) - off; 					\
		if(len < 0) len = 0; 					\
			*eof = ( len <= count) ? 1 : 0;		\
		*start = pg + off;   					\
		return len; 							\
	}while(0)


#if 0
#ifdef BE2ISCSI_DEBUG
#undef kmalloc
#undef kfree

void *dmalloc(unsigned int size, unsigned int flags, char *fn);
void dfree(void *ptr, char *fn);
#define kmalloc(size, flag) dmalloc(size, flag, __FUNCTION__)
#define kfree(ptr) dfree(ptr, __FUNCTION__)
extern unsigned int alloc_count, free_count;
#endif
#endif
#endif /* _BE2ISCSIDEBUG_H_ */
