/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
// Autogenerated file. Do not edit!
// srcgen version: 0146
#ifndef __beclib_ll_bmap_h__
#define __beclib_ll_bmap_h__
#include "setypes.h"

#undef SG_PACK
#if defined(SG_PRAGMA_PACK)
   #pragma pack(push, 1)
   #define SG_PACK
#elif defined(SG_ATTRIBUTE_PACK)
   #define SG_PACK __attribute__ ((packed))
#else
   #define SG_PACK
#endif

#ifndef SG_C_ASSERT
   #define SG_C_ASSERT(_name_, _condition_)
#endif

#ifndef SG_ASSERT
   #define SG_ASSERT(_condition_)
#endif

#ifndef SG_CALL
   #define SG_CALL
#endif

#ifndef SG_STATIC_INLINE
   #define SG_STATIC_INLINE static inline
#endif

//---------------------------------------------------------------------------------------
// Function: be_initialize_library
//
//   Call to initialize the static library. This initializes the debug traces.
//
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_initialize_library (
   void
);

//---------------------------------------------------------------------------------------
// Function: be_mpu_set_poll_timeout
//
//   Call to set the MPU (BladeEngine Management processor) IOCTL-polled mode timeout. 
//
// timeout       - IOCTL-poll timeout
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mpu_set_poll_timeout (
   IN u32 timeout
);

//---------------------------------------------------------------------------------------
// Function: be_mpu_set_post_timeout
//
//   Call to set the MPU (BladeEngine Management processor) POST timeout. 
//
// timeout       - post timeout
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mpu_set_post_timeout (
   IN u32 timeout
);

//---------------------------------------------------------------------------------------
// Function: be_function_and_chip_create_ex
//
//   Create both function and chip object with one call. The lower level API will rarely
//   use more than one function object. 
//
// sa_dev          - Previously created device.
// function_type   - 
// mailbox_sgl     - SGL with length equal to sizeof(MCC_MAILBOX). This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// emulation_sgl   - SGL with length equal to one page . This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// caps            - 
// function_object - On input this is a pointer to an allocated function object. On
//                   output, it is initialized. This is an opaque object. Do not access
//                   the members. 
// chip            - On input this is a pointer to an allocated chip object. On output, it
//                   is initialized. This is an opaque object. Do not access the members. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_and_chip_create_ex (
   IN SA_DEV*                                         sa_dev,
   IN u32                                             function_type,
   IN PSA_SGL                                         mailbox_sgl,
   IN PSA_SGL                                         emulation_sgl,
   IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps,
   OUT PBE_FUNCTION_OBJECT                            function_object,
   OUT PBE_CHIP_OBJECT                                chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_and_chip_create
//
//   Create both function and chip object with one call. The lower level API will rarely
//   use more than one function object. 
//
// sa_dev          - Previously created device.
// function_type   - 
// mailbox_sgl     - SGL with length equal to sizeof(MCC_MAILBOX). This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// emulation_sgl   - SGL with length equal to one page . This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// function_object - On input this is a pointer to an allocated function object. On
//                   output, it is initialized. This is an opaque object. Do not access
//                   the members. 
// chip            - On input this is a pointer to an allocated chip object. On output, it
//                   is initialized. This is an opaque object. Do not access the members. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_and_chip_create (
   IN SA_DEV*              sa_dev,
   IN u32                  function_type,
   IN PSA_SGL              mailbox_sgl,
   IN PSA_SGL              emulation_sgl,
   OUT PBE_FUNCTION_OBJECT function_object,
   OUT PBE_CHIP_OBJECT     chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_object_create
//
//   Create function object . 
//
// sa_dev          - Previously created device.
// function_type   - 
// mailbox_sgl     - SGL with length equal to sizeof(MCC_MAILBOX). This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// function_object - On input this is a pointer to an allocated function object. On
//                   output, it is initialized. This is an opaque object. Do not access
//                   the members. 
// chip            - On input this is a pointer to an allocated chip object. On output, it
//                   is initialized. This is an opaque object. Do not access the members. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_object_create (
   IN SA_DEV*              sa_dev,
   IN u32                  function_type,
   IN PSA_SGL              mailbox_sgl,
   OUT PBE_FUNCTION_OBJECT function_object,
   OUT PBE_CHIP_OBJECT     chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_object_create_ex
//
//   Create function object . 
//
// sa_dev          - Previously created device.
// function_type   - 
// mailbox_sgl     - SGL with length equal to sizeof(MCC_MAILBOX). This must be physically
//                   contiguous. The starting address must be aligned to 16 byte boundary.
// caps            - The capabilities supported by driver on input, and supported by
//                   firmware on output. 
// function_object - On input this is a pointer to an allocated function object. On
//                   output, it is initialized. This is an opaque object. Do not access
//                   the members. 
// chip            - On input this is a pointer to an allocated chip object. On output, it
//                   is initialized. This is an opaque object. Do not access the members. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_object_create_ex (
   IN SA_DEV*                                         sa_dev,
   IN u32                                             function_type,
   IN PSA_SGL                                         mailbox_sgl,
   IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps,
   OUT PBE_FUNCTION_OBJECT                            function_object,
   OUT PBE_CHIP_OBJECT                                chip
);

//---------------------------------------------------------------------------------------
// Function: be_chip_object_create
//
//   Create chip object . 
//
// chip          - On input this is a pointer to an allocated chip object. On output, it
//                 is initialized. This is an opaque object. Do not access the members. 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_chip_object_create (
   OUT PBE_CHIP_OBJECT chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_and_chip_destroy
//
//   This destroys the function and chip object created by the be_function_and_chip_create
//   function. It assumes that all sub-objects created for this function have already been
//   destroyed -- it asserts if this is not true. This function is called by
//   be_function_cleanup -- Use be_function_cleanup to automatically cleanup all
//   sub-objects. 
//
// function_object - Previously created function object.
// chip            - Previously created chip object.
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_and_chip_destroy (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PBE_CHIP_OBJECT     chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_object_destroy
//
//   This destroys the function object created by the be_function_object_create function.
//   It assumes that all sub-objects created for this function have already been destroyed
//   -- it asserts if this is not true. This function is called by be_function_cleanup --
//   Use be_function_cleanup to automatically cleanup all sub-objects. 
//
// function_object - Previously created function object.
// chip            - Previously created chip object.
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_object_destroy (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PBE_CHIP_OBJECT     chip
);

//---------------------------------------------------------------------------------------
// Function: be_chip_object_destroy
//
//   This destroys the chip object created by the be_chip_object_create function. It
//   assumes that all sub-objects created for this chip have already been destroyed -- it
//   asserts if this is not true. This function is called by be_function_cleanup -- Use
//   be_function_cleanup to automatically cleanup all sub-objects. 
//
// chip          - Previously created chip object.
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_chip_object_destroy (
   IN PBE_CHIP_OBJECT chip
);

//---------------------------------------------------------------------------------------
// Function: be_function_cleanup
//
//   This function destroys all objects that reference the function object including the
//   following.  - all offloaded connections (if offloaded thru beclib API)  - all queues
//   (mcc, etx, erx, iscsi/toe wrbq/defq, toe rq, cq, eq)  - iscsi ooo buffers  - software
//   jell buffers (during emulation)  - template header buffers (if host based)  - WDMA
//   zero buffer (during emulation)  - vlan tags  - multicast addresses  - rss
//   configuration  - wake on lan entries (ACPI)  - chip object  - function_object  All
//   objects are destroyed with synchronous ioctls that poll for their completions. This
//   function is optional, all objects can be explicitly cleaned up. 
//
// function_object - Previously created function object
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_cleanup (
   PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_reset_ex
//
//   Issues a function level reset on the specified function. 
//
// function_object    - 
// domain             - 
// callback           - 
// callback_context   - 
// queue_context      - 
// return return_type - 
//---------------------------------------------------------------------------------------
BESTATUS
be_function_reset_ex (
   IN BE_FUNCTION_OBJECT*       function_object,
   u32                          domain,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_get_function_number
//
//   Returns the PCI function number.
//
// function_object        - Previously created function object
// return function_number - 
//---------------------------------------------------------------------------------------
u32
be_function_get_function_number (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_get_function_type
//
//   Returns the function type.
//
// function_object      - 
// return function_type - 
//---------------------------------------------------------------------------------------
u32
be_function_get_function_type (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_is_iscsi
//
//   Returns TRUE for the ISCSI function.
//
// function_object - 
// return is_iscsi - 
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
be_function_is_iscsi (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_is_networking
//
//   Returns TRUE for the networking function.
//
// function_object      - 
// return is_networking - 
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
be_function_is_networking (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_copy_stats
//
//   Copies the function object stats into the memory supplied by the caller.
//
// function_object    - The previously created function object.
// stats              - Pointer to a stat struct. The stats are copied.
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
be_function_copy_stats (
   IN PBE_FUNCTION_OBJECT     function_object,
   OUT PBECLIB_FUNCTION_STATS stats
);

//---------------------------------------------------------------------------------------
// Function: be_function_iscsi_ooo_buffers_post
//
//   Posts out-of-order buffers to BladeEngine.
//
// function_object - 
// page_count      - 
// sgl             - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_iscsi_ooo_buffers_post (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 page_count,
   IN PSA_SGL             sgl
);

//---------------------------------------------------------------------------------------
// Function: be_function_iscsi_ooo_buffers_remove
//
//   Removes the out-of-order buffers from the chip. The caller may free the memory when
//   this function completes. All iSCSI connections must be closed first. This function is
//   called as part of be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_iscsi_ooo_buffers_remove (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_template_header_memory
//
//   Posts buffers for template headers. Post 1 page per 32 connections. 
//
// function_object - 
// sgl             - 
// response        - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_template_header_memory (
   IN PBE_FUNCTION_OBJECT                                 function_object,
   IN PSA_SGL                                             sgl,
   OUT IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS_RESPONSE* response
);

//---------------------------------------------------------------------------------------
// Function: be_function_remove_template_header_memory
//
//   Removes buffers for template headers. Do not remove the buffers until all connections
//   have been torn down. This function is called as part of be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_remove_template_header_memory (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_page_tables
//
//   Posts a list of page tables used to organize unified table. Each ring or iSCSI
//   fragment uses these page tables to allow discontiguous physical pages. These pages
//   are used in both NIC mode and for offloaded connections.  The required number of page
//   tables is: [table col="2" delimit=":"] PCI Function : Required Page Tables : ISCSI: 
//   - 1 per 8 connections  - 1 per 2MB of fragments  - 1 for EQ (emulation only, normally
//   posted by ARM)  - 1 for JELL (emulation only, normally posted by ARM) : NET:  - 1 per
//   8 connections  - 1 for JELL (emulation only, normally posted by ARM)  - 1 for EQ
//   (emulation only, normally posted by ARM) : [/table]  The page tables for connections
//   may be added dynamically, before creating the corresponding WRB and RQ queues.
//   However, both functions require the first page table (for connections 0-7) be posted
//   immediately. This page table will contain other device rings, including CQs and
//   DEFQs. Further, the fragment, EQ, and JELL tables are required immediately.  NOTE:
//   This function will not be required if page tables are all in DDR. 
//
// function_object - 
// sgl             - SGL with at the "num_pages" physical pages. The SGL offset should be
//                   zero, meaning that all pages are page aligned. For emulation, the SGL
//                   must contain a valid virtual address. 
// num_pages       - The number of page to post from the beginning of the SGL. This allows
//                   the SGL to be larger than the number of posted pages. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_page_tables (
   IN BE_FUNCTION_OBJECT* function_object,
   IN PSA_SGL             sgl,
   IN u32                 num_pages
);

//---------------------------------------------------------------------------------------
// Function: be_function_remove_page_tables
//
//   Removes a maximum of "num" page tables from the chip. Page tables are only removed if
//   they are not used. The actual_num parameter gives the actual number of pages removed
//   which may be less than the requested number. The pa_list_optional may be NULL.
//   Otherwise, it should point to an array where the physical addresses for the removed
//   pages are stored. These pages may be freed after this function completes. This
//   function is called as part of be_function_cleanup. 
//
// function_object  - 
// num              - 
// actual_num       - 
// pa_list_optional - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_remove_page_tables (
   BE_FUNCTION_OBJECT* function_object,
   u32                 num,
   u32*                actual_num,
   u64*                pa_list_optional
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_jell
//
//   Currently, the JELL uses up to 8k, 256 byte entries for a total of 2MB. Fewer pages
//   may be posted.  NOTE: This function is ONLY used during emulation. In normal
//   operation the arm will allocate the JELL in DDR. 
//
// function_object - 
// sgl             - The SGL does not require a valid virtual address.
// num_pages       - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_jell (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 num_pages
);

//---------------------------------------------------------------------------------------
// Function: be_function_remove_jell
//
//   Removes all JELL pages from the chip. This function is called as part of
//   be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_remove_jell (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_zero_buffer
//
//   The physical address must point to 1 byte. That byte must be zero. NOTE: This
//   function is ONLY used during emulation. In normal operation the arm will allocate the
//   buffer in DDR. 
//
// function_object  - 
// physical_address - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_zero_buffer (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u64                 physical_address
);

//---------------------------------------------------------------------------------------
// Function: be_function_remove_zero_buffer
//
//   Removes the zero buffer from the chip. This function is called as part of
//   be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_remove_zero_buffer (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_nop
//
//   Issues a NOP command to the MCC ring. The command is completed with a successful
//   status. This can be used to pend operations until a short time in the future, since
//   the callback function will be invoked upon command completion. The returned status
//   will be BE_PENDING if the command was issued successfully. 
//
// function_object    - 
// callback           - Callback function invoked when the NOP completes. 
// callback_context   - Passed to the callback function. 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_nop (
   IN PBE_FUNCTION_OBJECT   function_object,
   IN MCC_WRB_CQE_CALLBACK  callback,
   IN PVOID                 callback_context,
   IN PBE_NOP_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_get_fw_version
//
//   Retrieves the firmware version on the adpater. If the callback is NULL this call
//   executes synchronously. If the callback is not NULL, the returned status will be
//   BE_PENDING if the command was issued successfully. 
//
// function_object    - 
// fw_version         - Pointer to response buffer if callback is NULL. 
// callback           - Callback function invoked when the IOCTL completes. 
// callback_context   - Passed to the callback function. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_get_fw_version (
   IN PBE_FUNCTION_OBJECT                            function_object,
   OUT IOCTL_COMMON_GET_FW_VERSION_RESPONSE_PAYLOAD* fw_version,
   IN MCC_WRB_CQE_CALLBACK                           callback,
   IN PVOID                                          callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_configure_arm_prints
//
//   This routine can be used to change the debug print level of the BladeEngine
//   management processor. This is only applicable in debug release builds. 
//
// function_object - 
// debug_level     - Debug print level. The lower the value the more debug info will be
//                   printed.
// idle_print      - Set to print debug info in the management firmware idle periods. 
// old_debug_level - Previous debug print level. 
// old_idle_print  - Previous idle print setting. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_configure_arm_prints (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 debug_level,
   IN boolean             idle_print,
   OUT u32*               old_debug_level,
   OUT boolean*           old_idle_print
);

//---------------------------------------------------------------------------------------
// Function: be_function_manage_FAT_log
//
//   This routine can be used to manage the BladeEngine Fault Analysis Tool (FAT) log.
//   This includes querying the FAT log size, retrieving the FAT log, and clearing the FAT
//   log. The log data can be anaylzed by BladeEngine support tools to diagnose faults.
//   Only host domains (domain 0) may issue this request. 
//
// function_object  - 
// sgl              - The SGL representing the FAT log buffer landing space. The SGL
//                    should be page aligned. It does not require a virtual address. 
// num_pages        - The number of pages in the SGL. A value of zero (0) implies no FAT
//                    log data transfer will take place. The num_pages is limited to 27
//                    pages. 
// page_offset      - The page_offset specifies the starting page offset when retrieving
//                    the FAT log. For example, a value of 0 requests data starting at
//                    byte offset 0 of the FAT log. Likewise, a value of 5 requests data
//                    starting at byte offset 20480 of the FAT log. A caller may choose to
//                    issue multiple queries when the FAT log buffer size is larger than
//                    the number of pages specified. 
// clear_log        - Set to clear the log. 
// output           - 
// callback         - 
// callback_context - 
// queue_context    - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_manage_FAT_log (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN PSA_SGL                      sgl,
   IN u32                          num_pages,
   IN u32                          page_offset,
   IN boolean                      clear_log,
   OUT FAT_RESPONSE*               output,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_MANAGE_FAT_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_measure_dma_bandwidth
//
//   This routine performs a diagnostic DMA bandwidth test. 
//
// function_object  - 
// ioctl            - The IOCTL request should be filled in. The data is copied into the
//                    WRB. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_measure_dma_bandwidth (
   IN PBE_FUNCTION_OBJECT                       function_object,
   IN OUT IOCTL_LOWLEVEL_MEASURE_DMA_BANDWIDTH* ioctl,
   IN MCC_WRB_CQE_CALLBACK                      callback,
   IN PVOID                                     callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_pe_pages
//
//   Posts buffers for protection entries. 
//
// function_object - 
// sgl             - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_pe_pages (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_te_pages
//
//   Posts buffers for translation entries. 
//
// function_object - 
// sgl             - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_te_pages (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl
);

//---------------------------------------------------------------------------------------
// Function: be_function_free_pe_pages
//
//   Remove buffers for protection entries. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_free_pe_pages (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_free_te_pages
//
//   Remove buffers for translation entries. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_free_te_pages (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_query_fw_config
//
//   Queries the firmware configuration currently loaded. This configuration includes
//   information regarding the EP processor configuration for various Upper Layer
//   Protocols. 
//
// function_object - 
// config          - The configuration parameters currently loaded by firmware. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_query_fw_config (
   IN PBE_FUNCTION_OBJECT function_object,
   OUT BE_FW_CFG*         config
);

//---------------------------------------------------------------------------------------
// Function: be_function_internal_driver_capabilities
//
//   Sets the driver capabilities and returns the capabilities that the driver should use.
//
// function_object - 
// caps            - The capabilities supported by driver on input, and supported by
//                   firmware on output. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_internal_driver_capabilities (
   IN PBE_FUNCTION_OBJECT                          function_object,
   OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps
);

//---------------------------------------------------------------------------------------
// Function: be_function_is_asic
//
//   Returns TRUE if the driver is currently running on the ASIC, rather than the FPGA. 
//
// function_object - 
// return is_asic  - TRUE, if the driver is currently running on the ASIC.
//---------------------------------------------------------------------------------------
boolean
be_function_is_asic (
   BE_FUNCTION_OBJECT* function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_chip_revision
//
//   Returns the overall revision number of the adapter. 
//
// function_object      - 
// return chip_revision - the overall revision number of the adapter
//---------------------------------------------------------------------------------------
u32
be_function_chip_revision (
   BE_FUNCTION_OBJECT* function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_passthru_ioctl
//
//   This routine issues an embedded IOCTL in pass-through mode. 
//
// function_object  - 
// payload          - The embedded payload for the MCC_WRB structure. The input buffer is
//                    overwritten with response data. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_passthru_ioctl (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN OUT MCC_WRB_PAYLOAD* payload,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_passthru_nonembedded_ioctl
//
//   This routine issues an embedded IOCTL in pass-through mode. 
//
// function_object  - 
// ioctl_va         - Virtual address to the IOCTL. 
// ioctl_pa         - Physical address to the IOCTL. 
// ioctl_len        - Length of the IOCTL. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_passthru_nonembedded_ioctl (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN IOCTL_HEADER*        ioctl_va,
   IN u64                  ioctl_pa,
   IN u32                  ioctl_len,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_passthru_nonembedded_ioctl_with_queue_context
//
//   This routine issues an embedded IOCTL in pass-through mode. 
//
// function_object  - 
// ioctl_va         - Virtual address to the IOCTL. 
// ioctl_pa         - Physical address to the IOCTL. 
// ioctl_len        - Length of the IOCTL. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// queue_context    - Queue context 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_passthru_nonembedded_ioctl_with_queue_context (
   IN PBE_FUNCTION_OBJECT               function_object,
   IN IOCTL_HEADER*                     ioctl_va,
   IN u64                               ioctl_pa,
   IN u32                               ioctl_len,
   IN MCC_WRB_CQE_CALLBACK              callback,
   IN PVOID                             callback_context,
   IN PBE_GET_TEMPERATURE_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_dcn_query
//
//   This routine issues an embedded IOCTL in pass-through mode. 
//
// function_object  - 
// ioctl_va         - Virtual address to the IOCTL. 
// ioctl_pa         - Physical address to the IOCTL. 
// ioctl_len        - Length of the IOCTL. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// queue_context    - Queue context 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_dcn_query (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN IOCTL_HEADER*               ioctl_va,
   IN u64                         ioctl_pa,
   IN u32                         ioctl_len,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_QUERY_DCN_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_get_privileges
//
//   This function returns the privilege-categories for this function 
//
// function_object  - Previously created function object.
// privilege_mask   - Privilege mask for the function.
// callback         - 
// callback_context - 
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_get_privileges (
   IN BE_FUNCTION_OBJECT*             function_object,
   OUT u32*                           privilege_mask,
   IN MCC_WRB_CQE_CALLBACK            callback,
   IN PVOID                           callback_context,
   IN PBE_FUNCTION_PRIVILEGES_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_set_privileges
//
//   This function sets the privilege-categories for a proxy domain. 
//
// function_object  - Previously created function object.
// domain           - Proxy domain for which this function sets the privileges.
// privilege_mask   - Privilege mask for the function.
// callback         - 
// callback_context - 
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_set_privileges (
   IN BE_FUNCTION_OBJECT*             function_object,
   IN u32                             domain,
   IN u32                             privilege_mask,
   IN MCC_WRB_CQE_CALLBACK            callback,
   IN PVOID                           callback_context,
   IN PBE_FUNCTION_PRIVILEGES_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_set_emi_canceller
//
//   This functions sets the 10G Base T EMI Canceller State. 
//
// function_object    - Previously created function object.
// port_number        - PHY (either 0 or 1)
// canceller_state    - EMI canceller state (either 0 or 1)
// return return_type - 
//---------------------------------------------------------------------------------------
BESTATUS
be_function_set_emi_canceller (
   IN BE_FUNCTION_OBJECT* function_object,
   IN u32                 port_number,
   IN u32                 canceller_state
);

//---------------------------------------------------------------------------------------
// Function: be_function_get_phy_details
//
//   This function may be issued by VF and PF drivers to determine the PHY type and
//   settings for the port assigned to their function. 
//
// function_object    - Previously created function object.
// phy_details        - PHY details, type, speeds
// return return_type - 
//---------------------------------------------------------------------------------------
BESTATUS
be_function_get_phy_details (
   IN BE_FUNCTION_OBJECT*             function_object,
   OUT PGET_PHY_DETAILS_PARAMS_OUT_V1 phy_details
);

//---------------------------------------------------------------------------------------
// Function: be_eq_modify_delay
//
//   Changes the EQ delay for a group of EQs.
//
// function_object           - 
// num_eq                    - The number of EQs in the eq_array to adjust. This also is
//                             the number of delay values in the eq_delay_array. 
// eq_array                  - Array of BE_EQ_OBJECT pointers to adjust. 
// eq_delay_multiplier_array - Array of "num_eq" timer delay multipliers. These are in
//                             arbitrary units. Use be_eq_delay_to_multiplier to convert
//                             from a microsecond delay to a multiplier. 
// eq_phase_array            - Array of "num_eq" phase values from 0 to 3. Each setting
//                             represents a 25 percent phase shift. 
// 
//  The phase is
//                             measured relative to other EQs to allow the driver to
//                             synchronize interrupts or stagger interrupts. 
// 
//  The phase
//                             only comes into play when the EQs are operating with zero
//                             credits in their leaky bucket. During this condition, the
//                             interrupt rate is exactly (1,000,000 /
//                             delay_in_microseconds) interrupts/sec. Phase is only
//                             meaningful, among EQs with the same delay setting. 
// callback                  - 
// callback_context          - 
// queue_context             - Optional. Pointer to a previously allocated struct. If the
//                             MCC WRB ring is full, this structure is used to queue the
//                             operation. It will be posted to the MCC ring when space
//                             becomes available. All queued commands will be posted to
//                             the ring in the order they are received. It is always valid
//                             to pass a pointer to a generic BE_GENERIC_QUEUE_CONTEXT.
//                             However, the specific context structs are generally smaller
//                             than the generic struct. 
// return pend_status        - BE_SUCCESS (0) on success. BE_PENDING (postive value) if
//                             the IOCTL completion is pending. Negative error code on
//                             failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_modify_delay (
   IN PBE_FUNCTION_OBJECT               function_object,
   IN u32                               num_eq,
   IN BE_EQ_OBJECT**                    eq_array,
   IN u32*                              eq_delay_multiplier_array,
   IN u32*                              eq_phase_array,
   IN MCC_WRB_CQE_CALLBACK              callback,
   IN PVOID                             callback_context,
   IN PBE_EQ_MODIFY_DELAY_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eq_create
//
//   Creates an eq_object. This is a synchronous IOCTL.
//
// function_object  - 
// sgl              - No virtual address required.
// eqe_size         - EQ entry size in bytes. Either 4 or 16.
// num_entries      - Power of 2, fom 256 to 4k.
// delay_multiplier - EQ delay in arbitrary units. Use be_eq_delay_to_multiplier to
//                    convert a delay from microseconds to a multiplier. A value of 0 is
//                    no delay. The delay may be modified during runtime using
//                    be_eq_modify_delay. 
// eq_object        - Created EQ object.
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 eqe_size,
   IN u32                 num_entries,
   IN u32                 delay_multiplier,
   OUT BE_EQ_OBJECT*      eq_object
);

//---------------------------------------------------------------------------------------
// Function: be_eq_create_async
//
//   Creates an eq_object. This is a synchronous IOCTL.
//
// function_object  - 
// sgl              - No virtual address required.
// eqe_size         - EQ entry size in bytes. Either 4 or 16.
// num_entries      - Power of 2, fom 256 to 4k.
// delay_multiplier - EQ delay in arbitrary units. Use be_eq_delay_to_multiplier to
//                    convert a delay from microseconds to a multiplier. A value of 0 is
//                    no delay. The delay may be modified during runtime using
//                    be_eq_modify_delay. 
// eq_object        - Created EQ object.
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_create_async (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN u32                       eqe_size,
   IN u32                       num_entries,
   IN u32                       delay_multiplier,
   OUT BE_EQ_OBJECT*            eq_object,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eq_fake_create
//
//   Creates an eq_object that does not have an associated memory ring. The object is used
//   to generate an interrupt from software. The EQ will interrupt the host every time the
//   EQ is rearmed, after the EQ delay expires. The EQ should always be rearmed with a
//   num_popped count of 0 in the doorbell register. The be_function_enable_interrupts
//   function will cause an immediate interrupt from this EQ. This is a synchronous IOCTL.
//
// function_object  - 
// delay_multiplier - EQ delay in arbitrary units. Use be_eq_delay_to_multiplier to
//                    convert a delay from microseconds to a multiplier. A value of 0 is
//                    no delay. The delay may be modified during runtime using
//                    be_eq_modify_delay. 
// eq_object        - Created EQ object.
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_fake_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 delay_multiplier,
   OUT BE_EQ_OBJECT*      eq_object
);

//---------------------------------------------------------------------------------------
// Function: be_eq_destroy
//
//   Destroys an eq_object. This function is called as part of be_function_cleanup.
//
// eq            - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_destroy (
   IN BE_EQ_OBJECT* eq
);

//---------------------------------------------------------------------------------------
// Function: be_eq_destroy_async
//
//   Destroys an eq_object. This function is called as part of be_function_cleanup.
//
// eq               - 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eq_destroy_async (
   IN BE_EQ_OBJECT*             eq,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eq_get_id
//
//   Return the allocated ID that must be used to ring the doorbell for the EQ.
//
// eq        - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_eq_get_id (
   IN BE_EQ_OBJECT* eq
);

//---------------------------------------------------------------------------------------
// Function: be_eq_delay_to_multiplier
//
//
// function_object   - 
// microsecond_delay - The delay value in microseconds.
// return multiplier - The multiplier value required for the be_eq_modify_delay function.
//---------------------------------------------------------------------------------------
u32
be_eq_delay_to_multiplier (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 microsecond_delay
);

//---------------------------------------------------------------------------------------
// Function: be_eq_multiplier_to_delay
//
//
// function_object - 
// multiplier      - The multiplier value required for the be_eq_modify_delay function.
// return delay    - The delay value in microseconds.
//---------------------------------------------------------------------------------------
u32
be_eq_multiplier_to_delay (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 multiplier
);

//---------------------------------------------------------------------------------------
// Function: be_cq_create
//
//   Creates a CQ. This uses a synchronous IOCTL.
//
// function_object     - 
// sgl                 - No virtual address required, except for MCC CQ.
// length              - Length is bytes of queue. Number of entries is power of 2, from
//                       256 to 1k. 
// solicited_eventable - If TRUE, only CQEs with solicited event bit cause EQE write. 
// no_delay            - TRUE means force interrupt immediately without waiting for any EQ
//                       delay. 
// cqe_dma_coalescing  - Each CQ supports coalescing the DMA of CQs into larger transfers.
//                       The maximum DMA transfer is up to 4 CQEs in a single DMA. When
//                       the CQ ring contains 'N' entries that the driver has not
//                       consumed, it will start coalescing future DMAs. The future DMA
//                       will not occur until the driver rings the CQ doorbell to reduce
//                       the outstanding number of CQ entries to less than 'N' -- or when
//                       CEV has buffered 4 CQEs to transfer. 
// 
//  The value 'N' is
//                       configurable by this setting using the enumeration
//                       BE_CQ_DMA_COALESCING. Using a small value (i.e. 2) will allow
//                       more DMA coalescing. However, the driver must be careful about
//                       increased latencies. 
// eq_object           - Optional, if set this is an eventable CQ. 
// cq_object           - Created CQ object. 
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN boolean             solicited_eventable,
   IN boolean             no_delay,
   IN u32                 cqe_dma_coalescing,
   IN PBE_EQ_OBJECT       eq_object,
   OUT PBE_CQ_OBJECT      cq_object
);

#if defined(SG_REVERSE_ENDIAN)
   
   #if defined(SG_BMAP_UNION)
      typedef struct _CQ_CREATE_PARAMS {
         union {
            struct { 
               u32 solicited_eventable;
               u32 no_delay;
               u32 cqe_dma_coalescing;
            } SG_PACK ; // unnamed struct
            u32 dw[3]; // dword union
         }; // unnamed union
      } SG_PACK CQ_CREATE_PARAMS, *PCQ_CREATE_PARAMS;

      SG_C_ASSERT(__sizeof__CQ_CREATE_PARAMS, sizeof (CQ_CREATE_PARAMS) == 12);

   #else
      typedef struct _CQ_CREATE_PARAMS {
         u32 solicited_eventable;
         u32 no_delay;
         u32 cqe_dma_coalescing;
      } SG_PACK CQ_CREATE_PARAMS, *PCQ_CREATE_PARAMS;

      typedef union _CQ_CREATE_PARAMS_UNION {
         CQ_CREATE_PARAMS bits;
         u32 dw[3]; // dword union
      } SG_PACK CQ_CREATE_PARAMS_UNION, *PCQ_CREATE_PARAMS_UNION;

      SG_C_ASSERT(__sizeof__CQ_CREATE_PARAMS_UNION, sizeof (CQ_CREATE_PARAMS_UNION) == 12);
   #endif


#else
   
   #if defined(SG_BMAP_UNION)
      typedef struct _CQ_CREATE_PARAMS {
         union {
            struct { 
               u32 solicited_eventable;
               u32 no_delay;
               u32 cqe_dma_coalescing;
            } SG_PACK ; // unnamed struct
            u32 dw[3]; // dword union
         }; // unnamed union
      } SG_PACK CQ_CREATE_PARAMS, *PCQ_CREATE_PARAMS;

      SG_C_ASSERT(__sizeof__CQ_CREATE_PARAMS, sizeof (CQ_CREATE_PARAMS) == 12);

   #else
      typedef struct _CQ_CREATE_PARAMS {
         u32 solicited_eventable;
         u32 no_delay;
         u32 cqe_dma_coalescing;
      } SG_PACK CQ_CREATE_PARAMS, *PCQ_CREATE_PARAMS;

      typedef union _CQ_CREATE_PARAMS_UNION {
         CQ_CREATE_PARAMS bits;
         u32 dw[3]; // dword union
      } SG_PACK CQ_CREATE_PARAMS_UNION, *PCQ_CREATE_PARAMS_UNION;

      SG_C_ASSERT(__sizeof__CQ_CREATE_PARAMS_UNION, sizeof (CQ_CREATE_PARAMS_UNION) == 12);
   #endif


#endif
//---------------------------------------------------------------------------------------
// Function: be_cq_create_ex
//
//   Creates a CQ. This uses a synchronous IOCTL.
//
// function_object  - 
// sgl              - No virtual address required, except for MCC CQ.
// length           - Length is bytes of queue. Number of entries is power of 2, from 256
//                    to 1k. 
// param            - Parameter structure.
// eq_object        - Optional, if set this is an eventable CQ. 
// cq_object        - Created CQ object. 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_create_ex (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN u32                       length,
   IN CQ_CREATE_PARAMS*         param,
   IN PBE_EQ_OBJECT             eq_object,
   OUT PBE_CQ_OBJECT            cq_object,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_cq_destroy
//
//   Destroys a CQ. All clients that reference this CQ must be destroyed first. This
//   function is called as part of be_function_cleanup. 
//
// cq            - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_destroy (
   IN BE_CQ_OBJECT* cq
);

//---------------------------------------------------------------------------------------
// Function: be_cq_destroy_async
//
//   Destroys a CQ. All clients that reference this CQ must be destroyed first. This
//   function is called as part of be_function_cleanup. 
//
// cq               - 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                    ring is full, this structure is used to queue the operation. It will
//                    be posted to the MCC ring when space becomes available. All queued
//                    commands will be posted to the ring in the order they are received.
//                    It is always valid to pass a pointer to a generic
//                    BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs are
//                    generally smaller than the generic struct. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cq_destroy_async (
   IN BE_CQ_OBJECT*             cq,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_cq_get_id
//
//   Returns the allocated CQ id used for ringing the doorbell. 
//
// cq        - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_cq_get_id (
   IN BE_CQ_OBJECT* cq
);

//---------------------------------------------------------------------------------------
// Function: be_mcc_ring_create_ext
//
//   This creates a MCC ring and switches from mailbox mode to ring mode. The number of
//   outstanding MCC commands is the minimum of the number of entries in the ring and the
//   number of context structures (i.e. num_context_entries). BECLIB uses the context
//   entries to track outstanding MCC WRBs since they may complete out-of-order. You may
//   provide zero context entries, which will limit BECLIB to one MCC WRB at a time. 
//
// function_object     - 
// sgl                 - A virtual address is required. 
// length              - length in bytes 
// context_array       - Array of context structs. Each outstanding MCC WRB requires a
//                       context entry. BECLIB has 1 by default. The context array allows
//                       you to post more -- this must be non-pageable memory, but it is
//                       not used for DMA. Most likely the number of context structs will
//                       match the size of the MCC ring. Since commands complete
//                       out-of-order, this lets beclib fully utilize the ring. However,
//                       the size may be either lower or higher than the ring size. 
// num_context_entries - number of context structs 
// cq                  - Associated completion queue created with an SGL with a mapped
//                       virtual address 
// mcc                 - On input this is a pointer to an allocated MCC_OBJECT> On ouput
//                       this is a created MCC object. This is an opaque object. Do not
//                       reference any members. 
// flags               - Flags indicating the async events to subscribe to. 
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mcc_ring_create_ext (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN PBE_MCC_WRB_CONTEXT context_array,
   IN u32                 num_context_entries,
   IN PBE_CQ_OBJECT       cq,
   OUT PBE_MCC_OBJECT     mcc,
   IN u32                 flags
);

//---------------------------------------------------------------------------------------
// Function: be_mcc_ring_create
//
//   This creates a MCC ring and switches from mailbox mode to ring mode. The number of
//   outstanding MCC commands is the minimum of the number of entries in the ring and the
//   number of context structures (i.e. num_context_entries). BECLIB uses the context
//   entries to track outstanding MCC WRBs since they may complete out-of-order. You may
//   provide zero context entries, which will limit BECLIB to one MCC WRB at a time. 
//
// function_object     - 
// sgl                 - A virtual address is required. 
// length              - length in bytes 
// context_array       - Array of context structs. Each outstanding MCC WRB requires a
//                       context entry. BECLIB has 1 by default. The context array allows
//                       you to post more -- this must be non-pageable memory, but it is
//                       not used for DMA. Most likely the number of context structs will
//                       match the size of the MCC ring. Since commands complete
//                       out-of-order, this lets beclib fully utilize the ring. However,
//                       the size may be either lower or higher than the ring size. 
// num_context_entries - number of context structs 
// cq                  - Associated completion queue created with an SGL with a mapped
//                       virtual address 
// mcc                 - On input this is a pointer to an allocated MCC_OBJECT> On ouput
//                       this is a created MCC object. This is an opaque object. Do not
//                       reference any members. 
// return status       - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mcc_ring_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length,
   IN PBE_MCC_WRB_CONTEXT context_array,
   IN u32                 num_context_entries,
   IN PBE_CQ_OBJECT       cq,
   OUT PBE_MCC_OBJECT     mcc
);

//---------------------------------------------------------------------------------------
// Function: be_mcc_ring_destroy
//
//   Cleans up MCC ring and switches to mailbox mode. This function is called as part of
//   be_function_cleanup. 
//
// mcc_object    - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mcc_ring_destroy (
   IN PBE_MCC_OBJECT mcc_object
);

//---------------------------------------------------------------------------------------
// Function: be_mcc_process_cq
//
//   Processes mcc completions. This should be called from the interrupt or DPC to process
//   the MCC cq, when a corresponding event queue entry is received. It can also be called
//   to poll the CQ from a timer routine for non-eventable completions. Keep in mind, only
//   one thread can be inside beclib at any given time. That means this function must be
//   serialized with the posting of MCC WRBs. (This behavior may change. TBD.) 
//
// mcc_object    - 
// rearm         - rearm should be TRUE only if called due to EQE 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mcc_process_cq (
   IN BE_MCC_OBJECT* mcc_object,
   IN boolean        rearm
);

//---------------------------------------------------------------------------------------
// Function: be_mcc_add_async_event_callback
//
//
// mcc_object       - 
// callback         - Function callback invoked when an asynchronous event is received on
//                    the MCC cq. It follows the prototype:
//  <code> typedef void (BECALL
//                    *MCC_ASYNC_EVENT_CALLBACK) (PVOID context, u32 event_code, PVOID
//                    event); </code> 
// 
//  The context parameter is a copy of the
//                    callback_context pointer provided to this function. The event_code
//                    is the code from the common portion of the event entry. The event
//                    parameter is a pointer to the 16 byte async event message. 
// callback_context - Passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_mcc_add_async_event_callback (
   IN BE_MCC_OBJECT*           mcc_object,
   IN MCC_ASYNC_EVENT_CALLBACK callback,
   IN PVOID                    callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_pci_soft_reset_nowait
//
//   This function is called to issue a BladeEngine soft reset. Callers should acquire the
//   soft reset semaphore before calling this function. Additionaly, callers should ensure
//   they cannot be pre-empted while the routine executes. Upon completion of this
//   routine, callers should release the reset semaphore. This routine DOES NOT WAIT for
//   BladeEngine POST to complete. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_pci_soft_reset_nowait (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_pci_soft_reset
//
//   This function is called to issue a BladeEngine soft reset. Callers should acquire the
//   soft reset semaphore before calling this function. Additionaly, callers should ensure
//   they cannot be pre-empted while the routine executes. Upon completion of this
//   routine, callers should release the reset semaphore. This routine implicitly waits
//   for BladeEngine POST to complete. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_pci_soft_reset (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_acquire_reset_semaphore
//
//   Acquires the reset semaphore. If success is returned, the caller owns the reset
//   semaphore. Otherwise the caller does not own the reset semaphore. Callers must
//   acquire the reset semaphore before inducing a BladeEngine runtime reset. Callers must
//   also release the reset semaphore once they are done with a soft reset. Release of the
//   reset semaphore is accomplished with be_release_reset_semaphore. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_acquire_reset_semaphore (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_release_reset_semaphore
//
//   Releases the reset semaphore. Callers must release the reset semaphore once they are
//   done with a soft reset. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_release_reset_semaphore (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_drive_POST
//
//   This function is called to drive BladeEngine POST. The caller should ensure they
//   cannot be pre-empted while this routine executes. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_drive_POST (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_download_management_firmware
//
//   This function is called to download management firmware to DDR. The caller should
//   ensure they cannot be pre-empted while this routine executes. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_download_management_firmware (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_defq_create
//
//   Creates a default PDU ring for the current protection domain. Each queue contains 8
//   byte physical addresses to default PDU buffers. 
//
// function_object       - 
// sgl                   - no virtual address required
// length                - bytes in total ring
// default_buffer_length - bytes in each buffer posted to ring
// cq_object             - cq that will receive completions
// defq                  - 
// return status         - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_defq_create (
   IN PBE_FUNCTION_OBJECT           function_object,
   IN PSA_SGL                       sgl,
   IN u32                           length,
   IN u32                           default_buffer_length,
   IN PBE_CQ_OBJECT                 cq_object,
   OUT PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_defq_destroy
//
//   Destroys the default queue. All connections for the PD should be torn down first.
//   This function is called as part of be_function_cleanup. 
//
// defq          - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_defq_destroy (
   PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_defq_get_id
//
//   Gets the allocated ID for this queue. The id is used to ring the doorbell to post
//   buffers to the default queue. 
//
// defq      - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_iscsi_defq_get_id (
   IN PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_wrb_queue_create
//
//   Creates a WRB ring. Use the "get_cid" function to query the corresponding cid for the
//   ring. 
//
// function_object - 
// sgl             - 
// num_entries     - max = 32k / sizeof(ISCSI_INITIATOR_WRB) 
// wrbq            - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_wrb_queue_create (
   PBE_FUNCTION_OBJECT        function_object,
   PSA_SGL                    sgl,
   u32                        num_entries,
   PBE_ISCSI_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_wrb_queue_destroy
//
//   Destroys the wrb ring. Make sure the connection is torn down before destroying the
//   ring. This function is called as part of be_function_cleanup. 
//
// wrbq          - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_wrb_queue_destroy (
   IN PBE_ISCSI_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_wrbq_get_cid
//
//   Queries the allocated cid for the wrbq. 
//
// wrbq       - 
// return cid - 
//---------------------------------------------------------------------------------------
u32
be_iscsi_wrbq_get_cid (
   IN PBE_ISCSI_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_post_sgl_pages
//
//   Posts memory for iSCSI SGLs to the chip for the given protection domain. The memory
//   may be posted with multiple calls to this function by incrementing the page_offset.
//   The SGL does not require a valid virtual address. The SGL should be page aligned.
//   This must be done before any connections are offloaded. 1 page table must be posted
//   to the chip per 2MB of frags. 
//
// function_object - 
// sgl             - memory is posted starting with page 0 in this sgl
// page_offset     - offset on the chip where the pages are mapped
// num_pages       - number of pages to post from the current sgl
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_post_sgl_pages (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 page_offset,
   IN u32                 num_pages
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_remove_sgl_pages
//
//   Removes all the sgl memory for the given protection domain. This function is called
//   as part of be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_remove_sgl_pages (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_offload_from_tcp_stack
//
//   Use only for emulation. Performs a low-level connection offload with a diagnostic
//   IOCTL. The ioctl includes the entire template header, tcp state, and iscsi context
//   that must be filled in. 
//
// wrbq               - 
// defq               - 
// cq                 - 
// va_for_ioctl       - ISCSI_CONTEXT, TEMPLATE_HEADER, TCP_CONTEXT & OFFLOAD_CXN_ENTRY
//                      must be set 
// pa_for_ioctl       - 
// cxn                - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_offload_from_tcp_stack (
   IN PBE_ISCSI_WRB_QUEUE_OBJECT    wrbq,
   IN PBE_DEFAULT_PDU_QUEUE_OBJECT  defq,
   IN PBE_CQ_OBJECT                 cq,
   IN IOCTL_LOWLEVEL_ISCSI_OFFLOAD* va_for_ioctl,
   IN u64                           pa_for_ioctl,
   OUT PBE_ISCSI_CONNECTION_OBJECT  cxn,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_iscsi_terminate
//
//   Terminates an iSCSI offload created from the be_iscsi_offload_from_tcp_stack
//   function. This function is called as part of be_function_cleanup. 
//
// cxn                - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_iscsi_terminate (
   IN PBE_ISCSI_CONNECTION_OBJECT cxn,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_wrb_queue_create
//
//   Creates a WRB ring. Use the "get_cid" function to query the corresponding cid for the
//   ring. 
//
// function_object - 
// sgl             - 
// num_entries     - 
// wrbq            - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_wrb_queue_create (
   PBE_FUNCTION_OBJECT       function_object,
   PSA_SGL                   sgl,
   u32                       num_entries,
   PBE_FCOE_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_wrb_queue_destroy
//
//   Destroys the wrb ring. Make sure the connection is torn down before destroying the
//   ring. This function is called as part of be_function_cleanup. 
//
// wrbq          - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_wrb_queue_destroy (
   IN PBE_FCOE_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_wrbq_get_cid
//
//   Queries the allocated cid for the wrbq. 
//
// wrbq       - 
// return cid - 
//---------------------------------------------------------------------------------------
u32
be_fcoe_wrbq_get_cid (
   IN PBE_FCOE_WRB_QUEUE_OBJECT wrbq
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_post_sgl_pages
//
//   Posts memory for FCoE SGLs to the chip for the given protection domain. The memory
//   may be posted with multiple calls to this function by incrementing the page_offset.
//   The SGL does not require a valid virtual address. The SGL should be page aligned.
//   This must be done before any connections are offloaded. 1 page table must be posted
//   to the chip per 2MB of frags. 
//
// function_object - 
// sgl             - memory is posted starting with page 0 in this sgl
// page_offset     - offset on the chip where the pages are mapped
// num_pages       - number of pages to post from the current sgl
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_post_sgl_pages (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 page_offset,
   IN u32                 num_pages
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_remove_sgl_pages
//
//   Removes all the sgl memory for the given protection domain. This function is called
//   as part of be_function_cleanup. 
//
// function_object - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_remove_sgl_pages (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_inteface_offload
//
//   Use only for emulation. Performs a low-level interface offload with a diagnostic
//   IOCTL. The ioctl includes the entire template header, tcp state, and iscsi context
//   that must be filled in. 
//
// wrbq               - 
// cq                 - 
// intf               - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_inteface_offload (
   IN PBE_FCOE_WRB_QUEUE_OBJECT  wrbq,
   IN PBE_CQ_OBJECT              cq,
   OUT PBE_FCOE_INTERFACE_OBJECT intf,
   IN MCC_WRB_CQE_CALLBACK       callback,
   IN PVOID                      callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_terminate_offload
//
//   Terminates an iSCSI offload created from the be_iscsi_offload_from_tcp_stack
//   function. This function is called as part of be_function_cleanup. 
//
// intf               - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_terminate_offload (
   IN PBE_FCOE_INTERFACE_OBJECT intf,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_defq_create
//
//   Creates a default PDU ring for the current protection domain. Each queue contains 8
//   byte physical addresses to default PDU buffers. 
//
// function_object       - 
// sgl                   - no virtual address required
// length                - bytes in total ring
// default_buffer_length - bytes in each buffer posted to ring
// cq_object             - cq that will receive completions
// defq                  - 
// return status         - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_defq_create (
   IN PBE_FUNCTION_OBJECT           function_object,
   IN PSA_SGL                       sgl,
   IN u32                           length,
   IN u32                           default_buffer_length,
   IN PBE_CQ_OBJECT                 cq_object,
   OUT PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_defq_destroy
//
//   Destroys the default queue. All connections for the PD should be torn down first.
//   This function is called as part of be_function_cleanup. 
//
// defq          - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_fcoe_defq_destroy (
   PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_fcoe_defq_get_id
//
//   Gets the allocated ID for this queue. The id is used to ring the doorbell to post
//   buffers to the default queue. 
//
// defq      - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_fcoe_defq_get_id (
   IN PBE_DEFAULT_PDU_QUEUE_OBJECT defq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create
//
//   Creates an ethernet send ring. 
//
// function_object - 
// sgl             - no virtual address required
// length_in_bytes - 
// type            - The type of ring to create.
// ulp             - The requested ULP number for the ring. This should be zero based,
//                   i.e. 0,1,2. This must be valid NIC ULP based on the firmware
//                   configuration. Drivers may query the firmware configuration with
//                   be_function_query_firmware_config. All doorbells for this ring must
//                   be sent to this ULP. 
// bound_port      - port number to use for bound port type
// cq_object       - cq object for completions
// eth_sq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN u32                 length_in_bytes,
   IN u32                 type,
   IN u32                 ulp,
   IN u32                 bound_port,
   IN PBE_CQ_OBJECT       cq_object,
   OUT PBE_ETHSQ_OBJECT   eth_sq
);

// This is a set of optional parameters for be_eth_sq_create_ex. Certain ring types
// require additional parameters provided in this struct. 
typedef struct _BE_ETH_SQ_PARAMETERS {
   u32 rsvd0[4];
} SG_PACK BE_ETH_SQ_PARAMETERS, *PBE_ETH_SQ_PARAMETERS;



//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create_ex
//
//   Creates an ethernet send ring - extended version with additional parameters. 
//
// function_object  - 
// sgl              - no virtual address required
// length_in_bytes  - 
// type             - The type of ring to create.
// ulp              - The requested ULP number for the ring. This should be zero based,
//                    i.e. 0,1,2. This must be valid NIC ULP based on the firmware
//                    configuration. Drivers may query the firmware configuration with
//                    be_function_query_firmware_config. All doorbells for this ring must
//                    be sent to this ULP. 
// bound_port       - port number to use for bound port type
// cq_object        - cq object for completions
// ex_parameters    - Additional parameters (that may increase in future revisions). These
//                    parameters are only used for certain ring types -- see
//                    BE_ETH_SQ_PARAMETERS for details. 
// eth_sq           - 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create_ex (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN u32                       length_in_bytes,
   IN u32                       type,
   IN u32                       ulp,
   IN u32                       bound_port,
   IN PBE_CQ_OBJECT             cq_object,
   IN BE_ETH_SQ_PARAMETERS*     ex_parameters,
   OUT PBE_ETHSQ_OBJECT         eth_sq,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_create_v1
//
//   Creates an ethernet send ring - extended version with additional parameters. 
//
// function_object  - 
// sgl              - no virtual address required
// length_in_bytes  - 
// type             - The type of ring to create.
// rxif_object      - 
// cq_object        - cq object for completions
// eth_sq           - 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_create_v1 (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN u32                       length_in_bytes,
   IN u32                       type,
   IN PBE_RXIF_OBJECT           rxif_object,
   IN PBE_CQ_OBJECT             cq_object,
   OUT PBE_ETHSQ_OBJECT         eth_sq,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_destroy
//
//   Destroys an ethernet send ring. The driver must wait for all send completions before
//   destroying the ring. This function is called as part of be_function_cleanup. 
//
// eth_sq        - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_destroy (
   IN PBE_ETHSQ_OBJECT eth_sq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_destroy_async
//
//   Destroys an ethernet send ring. The driver must wait for all send completions before
//   destroying the ring. This function is called as part of be_function_cleanup. 
//
// eth_sq           - 
// callback         - Callback function on completion
// callback_context - Callback context
// queue_context    - 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_sq_destroy_async (
   IN PBE_ETHSQ_OBJECT          eth_sq,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_get_id
//
//   Query the allocated id for the send ring. This ID is used to ring the doorbell. 
//
// eth_sq    - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_eth_sq_get_id (
   IN PBE_ETHSQ_OBJECT eth_sq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_sq_get_completion_id
//
//   Query the id that is stored in each TX completion for this ring. 
//
// eth_sq        - 
// return cqe_id - 
//---------------------------------------------------------------------------------------
u32
be_eth_sq_get_completion_id (
   IN PBE_ETHSQ_OBJECT eth_sq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_set_flow_control
//
//   This function sets the flow control characteristics of BladeEngine. 
//
// function_object - 
// txfc_enable     - 
// rxfc_enable     - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_set_flow_control (
   IN PBE_FUNCTION_OBJECT function_object,
   IN boolean             txfc_enable,
   IN boolean             rxfc_enable
);

//---------------------------------------------------------------------------------------
// Function: be_eth_get_flow_control
//
//   This function gets the flow control characteristics of BladeEngine. 
//
// function_object - 
// txfc_enable     - 
// rxfc_enable     - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_get_flow_control (
   IN PBE_FUNCTION_OBJECT function_object,
   IN boolean*            txfc_enable,
   IN boolean*            rxfc_enable
);

//---------------------------------------------------------------------------------------
// Function: be_eth_set_qos
//
//   This function sets the ethernet transmit Quality of Service (QoS) characteristics of
//   BladeEngine for the domain. All ethernet transmit rings of the domain will evenly
//   share the bandwidth. The exeception to sharing is the host primary (super) ethernet
//   transmit ring as well as the host ethernet forwarding ring for missed offload data. 
//
// function_object - 
// max_bps         - the maximum bits per second in units of 10 Mbps (valid 0-100)
// max_pps         - the maximum packets per second in units of 1 Kpps (0 indicates no
//                   limit)
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_set_qos (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 max_bps,
   IN u32                 max_pps
);


typedef struct _BE_QOS_PARAMETERS {
   u32 version;
   u32 pdid;
   u32 max_bps;
   u32 max_pps;
} SG_PACK BE_QOS_PARAMETERS, *PBE_QOS_PARAMETERS;



//---------------------------------------------------------------------------------------
// Function: be_eth_set_qos_ex
//
//   This function sets the ethernet transmit Quality of Service (QoS) characteristics of
//   BladeEngine for the domain. All ethernet transmit rings of the domain will evenly
//   share the bandwidth. The exeception to sharing is the host primary (super) ethernet
//   transmit ring as well as the host ethernet forwarding ring for missed offload data. 
//
// function_object    - 
// parameters         - 
// callback           - Callback function on completion
// callback_context   - Callback context
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_set_qos_ex (
   IN PBE_FUNCTION_OBJECT              function_object,
   IN BE_QOS_PARAMETERS*               parameters,
   IN MCC_WRB_CQE_CALLBACK             callback,
   IN PVOID                            callback_context,
   IN BE_COMMON_SET_QOS_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_get_qos
//
//   This function retrieves the ethernet transmit Quality of Service (QoS)
//   characteristics for the domain. 
//
// function_object - 
// max_bps         - the maximum bits per second in units of 10 Mbps (valid 0-100)
// max_pps         - the maximum packets per second in units of 1 Kpps (0 indicates no
//                   limit)
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_get_qos (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32*                max_bps,
   IN u32*                max_pps
);

//---------------------------------------------------------------------------------------
// Function: be2_eth_set_frame_size
//
//   This function sets the ethernet maximum frame size. The previous values are returned.
//
// function_object - 
// tx_frame_size   - maximum transmit frame size in bytes
// rx_frame_size   - maximum receive frame size in bytes
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_eth_set_frame_size (
   IN PBE_FUNCTION_OBJECT function_object,
   IN OUT u32*            tx_frame_size,
   IN OUT u32*            rx_frame_size
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_create
//
//   Creates an ethernet receive ring for posting rx buffers. Each buffer posted to the
//   ring should be the size indicated from calling the be_eth_rq_get_frag_size function.
//   Only the host protection domain can receive broadcast/multicast frames. It can choose
//   to receive the CQ entries for these frames on a separate completion queue, or it can
//   use the same CQ. This is determines by the bcmc_cq_object parameter. 
//
// function_object - 
// sgl             - 
// cq_object       - 
// bcmc_cq_object  - optional
// frag_size       - 
// max_frame_size  - 
// rss_queue       - Indicates that an rss queue must be created
// rxif_object     - Associated rxif object for this rq
// eth_rq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN PBE_CQ_OBJECT       cq_object,
   IN PBE_CQ_OBJECT       bcmc_cq_object,
   IN u32                 frag_size,
   IN u32                 max_frame_size,
   IN boolean             rss_queue,
   IN PBE_RXIF_OBJECT     rxif_object,
   OUT PBE_ETHRQ_OBJECT   eth_rq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_create_async
//
//   Creates an ethernet receive ring for posting rx buffers. Each buffer posted to the
//   ring should be the size indicated from calling the be_eth_rq_get_frag_size function.
//   Only the host protection domain can receive broadcast/multicast frames. It can choose
//   to receive the CQ entries for these frames on a separate completion queue, or it can
//   use the same CQ. This is determines by the bcmc_cq_object parameter. 
//
// function_object    - 
// sgl                - 
// cq_object          - 
// bcmc_cq_object     - optional
// frag_size          - 
// max_frame_size     - 
// rss_queue          - Indicates that an rss queue must be created
// rxif_object        - Associated rxif object for this rq
// eth_rq             - 
// queue_context      - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_create_async (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN PBE_CQ_OBJECT             cq_object,
   IN PBE_CQ_OBJECT             bcmc_cq_object,
   IN u32                       frag_size,
   IN u32                       max_frame_size,
   IN boolean                   rss_queue,
   IN PBE_RXIF_OBJECT           rxif_object,
   OUT PBE_ETHRQ_OBJECT         eth_rq,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context
);


typedef struct _BE_ETH_RQ_PARAMETERS {
   u32 version;
   u32 frag_size;
   u32 max_frame_size;
   u32 rss_queue;
   u32 backfill;
   u32 dca;
   u32 fixed_split_location;
   u32 header_dma;
   u32 header_size;
   u32 header_split;
   u32 split_location;
} SG_PACK BE_ETH_RQ_PARAMETERS, *PBE_ETH_RQ_PARAMETERS;



//---------------------------------------------------------------------------------------
// Function: be_eth_rq_create_async_v1
//
//   Creates an ethernet receive ring for posting rx buffers. Each buffer posted to the
//   ring should be the size indicated from calling the be_eth_rq_get_frag_size function.
//   Only the host protection domain can receive broadcast/multicast frames. It can choose
//   to receive the CQ entries for these frames on a separate completion queue, or it can
//   use the same CQ. This is determines by the bcmc_cq_object parameter. 
//
// function_object    - 
// sgl                - 
// rxif_object        - Associated rxif object for this rq
// cq_object          - 
// parameters         - 
// eth_rq             - 
// queue_context      - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_create_async_v1 (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PSA_SGL                   sgl,
   IN PBE_RXIF_OBJECT           rxif_object,
   IN PBE_CQ_OBJECT             cq_object,
   IN BE_ETH_RQ_PARAMETERS*     parameters,
   OUT PBE_ETHRQ_OBJECT         eth_rq,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_destroy
//
//   Destroys an ethernet receive ring. This function is called as part of
//   be_function_cleanup. 
//
// eth_rq        - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_destroy (
   IN PBE_ETHRQ_OBJECT eth_rq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_destroy_options
//
//   Destroys an ethernet receive ring with finer granularity options than the standard
//   be_eth_rq_destroy() API function. 
//
// eth_rq           - 
// flush            - Set to 1 to flush the ring, set to 0 to bypass the flush and force
//                    invalidation. 
// queue_context    - 
// callback         - Callback function on completion
// callback_context - Callback context
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_destroy_options (
   IN PBE_ETHRQ_OBJECT          eth_rq,
   IN boolean                   flush,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_destroy_async
//
//   Same as be_eth_rq_destroy, but this version allows an async callback. 
//
// eth_rq             - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_eth_rq_destroy_async (
   IN PBE_ETHRQ_OBJECT     eth_rq,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_get_id
//
//   Queries the allocated ID for the ethernet receive ring. This should be used for
//   ringing the doorbell to post buffers. 
//
// EthRq     - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_eth_rq_get_id (
   IN PBE_ETHRQ_OBJECT EthRq
);

//---------------------------------------------------------------------------------------
// Function: be_eth_rq_get_rss_cpu_index
//
//   Only valid for RSS rings. This is the CPU id value to place in the RSS Indirection
//   table for this RQ. For example, if the driver assigns this RQ to CPU 1, it should
//   translate each 1 in the indirection table, to this value before programming the
//   indirection table. 
//
// EthRq                - 
// return rss_cpu_index - See function description.
//---------------------------------------------------------------------------------------
u8
be_eth_rq_get_rss_cpu_index (
   IN PBE_ETHRQ_OBJECT EthRq
);

//---------------------------------------------------------------------------------------
// Function: be_lro_enable
//
//   Enables LRO for the corresponding host interface 
//
// function_object - 
// enable          - 
// max_chainlen    - 
// full_chainlen   - 
// max_segcount    - 
// max_chaintime   - 
// rxif_object     - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_lro_enable (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 enable,
   IN u32                 max_chainlen,
   IN u32                 full_chainlen,
   IN u32                 max_segcount,
   IN u32                 max_chaintime,
   IN PBE_RXIF_OBJECT     rxif_object
);

//---------------------------------------------------------------------------------------
// Function: be_lro_rq_create
//
//   Creates an ethernet receive ring for posting rx buffers. Each buffer posted to the
//   ring should be the size indicated from calling the be_eth_rq_get_frag_size function.
//   Only the host protection domain can receive broadcast/multicast frames. It can choose
//   to receive the CQ entries for these frames on a separate completion queue, or it can
//   use the same CQ. This is determines by the bcmc_cq_object parameter. 
//
// function_object - 
// sgl             - 
// cq_object       - 
// page_size       - 
// ring_size       - 
// rxif_object     - 
// lro_rq          - 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_lro_rq_create (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PSA_SGL             sgl,
   IN PBE_CQ_OBJECT       cq_object,
   IN u32                 page_size,
   IN u32                 ring_size,
   IN PBE_RXIF_OBJECT     rxif_object,
   OUT PBE_LRORQ_OBJECT   lro_rq
);

//---------------------------------------------------------------------------------------
// Function: be_lro_rq_destroy
//
//   Destroys an ethernet receive ring with finer granularity options than the standard
//   be_eth_rq_destroy() API function. 
//
// lro_rq           - 
// callback         - Callback function on completion
// callback_context - Callback context
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_lro_rq_destroy (
   IN PBE_LRORQ_OBJECT     lro_rq,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_lro_rq_destroy_async
//
//   Same as be_eth_rq_destroy, but this version allows an async callback. 
//
// eth_rq             - 
// callback           - 
// callback_context   - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_lro_rq_destroy_async (
   IN PBE_LRORQ_OBJECT     eth_rq,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be_lro_rq_get_id
//
//   Queries the allocated ID for the ethernet receive ring. This should be used for
//   ringing the doorbell to post buffers. 
//
// rq        - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_lro_rq_get_id (
   IN PBE_LRORQ_OBJECT rq
);

//---------------------------------------------------------------------------------------
// Function: be_function_config_red
//
//   This function configures global and/or ULP specific Random Early Drop (RED)
//   functionality. 
//
// function_object    - 
// parameters         - The RED parameters. Only parameters for the chutes owned by the
//                      current PCI function are applied.
// callback           - Optional callback function.
// callback_context   - Optional context for callback function.
// queue_context      - Optional context for queueing the IOCTL.
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_config_red (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN PBE_RED_PARAMETERS           parameters,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_CONFIG_RED_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_function_set_tcp_parameters
//
//   This function updates the global tcp parameters per Upper Layer Protocol (ULP) chute.
//   Only the iSCSI PCI function may modify iSCSI protocol chutes. Only the networking PCI
//   function may modify TOE and RDMA chutes. 
//
// function_object    - 
// ulp_mask           - The ULPs to which the parameters will be applied. The storage PCI
//                      function may only change iSCSI ULPs. The networking PCI function
//                      may change TOE or RDMA ULPs. The mask may contain the bitwise OR
//                      of the valid values. 0x1 = ULP 0, 0x2 = ULP 1, etc. 
// parameters         - The new TCP parameters used for all connections on the given ULP
//                      chutes. 
// callback           - 
// callback_context   - 
// queue_context      - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_set_tcp_parameters (
   IN PBE_FUNCTION_OBJECT                  function_object,
   IN u32                                  ulp_mask,
   IN PBE_DRIVER_TCP_PARAMETERS            parameters,
   IN MCC_WRB_CQE_CALLBACK                 callback,
   IN PVOID                                callback_context,
   IN PBE_SET_TCP_PARAMETERS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_fpga_crc_calculation
//
//   calculates crc for fpga 
//
// dst        - Local address is destination for incoming packet 
// src        - Remote address is source for incoming packet 
// da_id      - Destination address id 
// ifId       - Interface Id. 
// srcPort    - Remote port is source port for incoming packet 
// dstPort    - Local port is destination port for incoming packet 
// isIpv6     - ipv6 
// return crc - 
//---------------------------------------------------------------------------------------
u32
be_toe_fpga_crc_calculation (
   IN u32* dst,
   IN u32* src,
   IN u32  da_id,
   IN u32  ifId,
   IN u16  srcPort,
   IN u16  dstPort,
   IN u32  isIpv6
);

//---------------------------------------------------------------------------------------
// Function: be_toe_cxn_offload
//
//   This completes the offload with one ioctl to program the connection context, program
//   the template header, program tcp state, and finally enable the offload. 
//
// sq                      - Required. The send queue to use. 
// rq                      - Optional. The preposted receive buffer queue to use. This
//                           must match the cid for the send queue. 
// dq                      - Required. The default receive buffer queue to use. 
// dq_cq                   - The CQ object to use for default completions. 
// rxif_object             - Handle to the RXIF object required for CAM programming 
// va_for_ioctl            - TEMPLATE_HEADER, TCP_CONTEXT & OFFLOAD_CXN_ENTRY must be set 
// pa_for_ioctl            - Physical address for the IOCTL buffer. 
// rx_delegated_data_bytes - Delegated RX data preceeds rcv_nxt in sequence number space,
//                           and it should be placed at the beginning of the first
//                           preposted buffer. To allow the driver to unconditionally post
//                           all preposted buffers to the ASIC, the delegated data bytes
//                           must be provided during offload. The TOE firmware will skip
//                           this many bytes in the first preposted buffer(s) to leave
//                           space for the driver to copy the delegated data. This is
//                           exactly like receiving default receive data before the
//                           preposted buffer is posted to the ASIC. 
// cxn                     - 
// callback                - 
// callback_context        - 
// queue_context           - Optional. Pointer to a previously allocated struct. If the
//                           MCC WRB ring is full, this structure is used to queue the
//                           operation. It will be posted to the MCC ring when space
//                           becomes available. All queued commands will be posted to the
//                           ring in the order they are received. It is always valid to
//                           pass a pointer to a generic BE_GENERIC_QUEUE_CONTEXT.
//                           However, the specific context structs are generally smaller
//                           than the generic struct. 
// return pend_status      - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the
//                           IOCTL completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_cxn_offload (
   IN PBE_TOE_SQ_OBJECT             sq,
   IN PBE_TOE_RQ_OBJECT             rq,
   IN PBE_TOE_DQ_OBJECT             dq,
   IN PBE_CQ_OBJECT                 dq_cq,
   IN PBE_RXIF_OBJECT               rxif_object,
   IN IOCTL_TOE_CONNECTION_OFFLOAD* va_for_ioctl,
   IN u64                           pa_for_ioctl,
   IN u32                           rx_delegated_data_bytes,
   OUT PBE_TOE_CONNECTION_OBJECT    cxn,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context,
   IN PBE_TOE_OFFLOAD_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_cxn_offload_ex
//
//   This completes the offload with one ioctl to program the connection context, program
//   the template header, program tcp state, and finally enable the offload. 
//
// sq                      - Required. The send queue to use. 
// rq                      - Optional. The preposted receive buffer queue to use. This
//                           must match the cid for the send queue. 
// dq                      - Required. The default receive buffer queue to use. 
// dq_cq                   - The CQ object to use for default completions. 
// rxif_object             - Handle to the RXIF object required for CAM programming 
// va_for_ioctl            - TEMPLATE_HEADER, TCP_CONTEXT & OFFLOAD_CXN_ENTRY must be set 
// pa_for_ioctl            - Physical address for the IOCTL buffer. 
// rx_delegated_data_bytes - Delegated RX data preceeds rcv_nxt in sequence number space,
//                           and it should be placed at the beginning of the first
//                           preposted buffer. To allow the driver to unconditionally post
//                           all preposted buffers to the ASIC, the delegated data bytes
//                           must be provided during offload. The TOE firmware will skip
//                           this many bytes in the first preposted buffer(s) to leave
//                           space for the driver to copy the delegated data. This is
//                           exactly like receiving default receive data before the
//                           preposted buffer is posted to the ASIC. 
// tx_delegated_data_bytes - Delegated TX data has already been transmitted, but not yet
//                           ACKed. The SndMax and SndNxt should be set equal to SndUna
//                           during offload. As the delegated TX data is posted, the ASIC
//                           will increment SndNxt and SndMax. If the connection is
//                           uploaded before transmitting all delegated tx data, the
//                           driver must fix the uploaded SndMax variable. 
// global_flow_id          - 
// local_flow_id           - 
// default_mode            - Sets the default buffer packing mode for this connection. See
//                           TOERX_DEFAULT_MODE defines. (0 = No packing, 1 = MSS packing,
//                           2 = full packing). 
// disable_repeat_wrb      - If repeat is used the first TX fragment of each set of 4 is a
//                           repeat of the last fragment in the previous set of 4. This
//                           allow coalescing of 2 cache lines together. If disabled, this
//                           repeated fragment is not used. 
// cxn                     - 
// callback                - 
// callback_context        - 
// queue_context           - Optional. Pointer to a previously allocated struct. If the
//                           MCC WRB ring is full, this structure is used to queue the
//                           operation. It will be posted to the MCC ring when space
//                           becomes available. All queued commands will be posted to the
//                           ring in the order they are received. It is always valid to
//                           pass a pointer to a generic BE_GENERIC_QUEUE_CONTEXT.
//                           However, the specific context structs are generally smaller
//                           than the generic struct. 
// ka_probe_count          - 
// ka_timeout_delta        - 
// return pend_status      - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the
//                           IOCTL completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_cxn_offload_ex (
   IN PBE_TOE_SQ_OBJECT             sq,
   IN PBE_TOE_RQ_OBJECT             rq,
   IN PBE_TOE_DQ_OBJECT             dq,
   IN PBE_CQ_OBJECT                 dq_cq,
   IN PBE_RXIF_OBJECT               rxif_object,
   IN IOCTL_TOE_CONNECTION_OFFLOAD* va_for_ioctl,
   IN u64                           pa_for_ioctl,
   IN u32                           rx_delegated_data_bytes,
   IN u32                           tx_delegated_data_bytes,
   IN u32                           global_flow_id,
   IN u32                           local_flow_id,
   IN u32                           default_mode,
   IN u32                           disable_repeat_wrb,
   OUT PBE_TOE_CONNECTION_OBJECT    cxn,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context,
   IN PBE_TOE_OFFLOAD_QUEUE_CONTEXT queue_context,
   IN u32                           ka_probe_count,
   IN u32                           ka_timeout_delta
);

//---------------------------------------------------------------------------------------
// Function: be_toe_cxn_upload
//
//   This completely uploads the connection and returns the final tcp state. If the
//   BE_TOE_UPLOAD_FLAG_RESET flag is set, the connection is reset. This function is
//   called as part of be_function_cleanup, resetting the connection. 
//
// cxn                - 
// rxif_object        - Handle to the RXIF object required for CAM programming 
// flags              - 
// rst_sequence       - The TCP sequence number to use with the RST frame, only used when
//                      the BE_TOE_UPLOAD_FLAG_RST_WITH_SEQUENCE flag is set. Normal RSTs
//                      are transmitted using the delegated snd_max sequence number.
//                      During offload, the driver may provide a fake value of snd_max to
//                      avoid offloading during retransmit. The RST must be sent with the
//                      real value of snd_max to make sure it falls within the receiver's
//                      window. 
// tcp_state          - 
// callback           - 
// callback_context   - 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_cxn_upload (
   IN PBE_TOE_CONNECTION_OBJECT    cxn,
   IN PBE_RXIF_OBJECT              rxif_object,
   IN u32                          flags,
   IN u32                          rst_sequence,
   OUT PTCP_CONTEXT                tcp_state,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_TOE_UPLOAD_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_cxn_update
//
//   This updates the template header and tcp_context for the connection while the
//   connection is offloaded. Only select fields in tcp context are updated. 
//
// cxn                - 
// va_for_ioctl       - virtual address for ioctl TEMPLATE_HEADER & TCP_CONTEXT must be
//                      set 
// pa_for_ioctl       - corresponding physical address 
// callback           - 
// callback_context   - 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_cxn_update (
   IN PBE_TOE_CONNECTION_OBJECT    cxn,
   IN IOCTL_TOE_CONNECTION_UPDATE* va_for_ioctl,
   IN u64                          pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_TOE_UPDATE_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_cxn_query
//
//   This queries the template header and tcp_context for the connection while the
//   connection is offloaded. Only select fields in tcp context are updated. 
//
// cxn                - 
// va_for_ioctl       - virtual address for ioctl TEMPLATE_HEADER & TCP_CONTEXT must be
//                      set 
// pa_for_ioctl       - corresponding physical address 
// callback           - 
// callback_context   - 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_cxn_query (
   IN PBE_TOE_CONNECTION_OBJECT   cxn,
   IN IOCTL_TOE_CONNECTION_QUERY* va_for_ioctl,
   IN u64                         pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_TOE_QUERY_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_rq_create
//
//   Creates a TOE receive queue for preposted buffers. Each queue corresponds to a
//   previously created send queue. This queue can be allocated once and used for many
//   different connections. The cid provided to this function should be returned from the
//   be_toe_sq_get_id function. 
//
// function_object    - 
// sgl                - no virtual address required 
// length             - in bytes for ring 
// cid                - 
// cq_object          - 
// toe_rq             - 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_rq_create (
   IN PBE_FUNCTION_OBJECT             function_object,
   IN PSA_SGL                         sgl,
   IN u32                             length,
   IN u32                             cid,
   IN PBE_CQ_OBJECT                   cq_object,
   OUT PBE_TOE_RQ_OBJECT              toe_rq,
   IN MCC_WRB_CQE_CALLBACK            callback,
   IN PVOID                           callback_context,
   IN PBE_TOE_RQ_CREATE_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_rq_destroy
//
//   Destroys the receive queue. The connection must be uploaded first. This function is
//   called as part of be_function_cleanup. 
//
// toe_rq        - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_rq_destroy (
   IN PBE_TOE_RQ_OBJECT toe_rq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_dq_create
//
//   This routine creates a TOE default receive queue. 
//
// function_object       - 
// sgl                   - no virtual address required 
// length                - byte length of queue 
// default_buffer_length - byte length of each buffer on the queue 
// ulp_id                - The ULP assigned to this TOE default queue: 0,1,2 
// toe_dq                - 
// return status         - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_dq_create (
   PBE_FUNCTION_OBJECT function_object,
   PSA_SGL             sgl,
   u32                 length,
   u32                 default_buffer_length,
   u32                 ulp_id,
   PBE_TOE_DQ_OBJECT   toe_dq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_dq_destroy
//
//   Destroys the default queue. All connections should be uploaded first. This function
//   is called as part of be_function_cleanup. 
//
// toe_dq        - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_dq_destroy (
   PBE_TOE_DQ_OBJECT toe_dq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_dq_get_id
//
//   Returns the TOE default queue id allocated during the create function. This id is
//   used for ringing the doorbell to post buffer to the queue. 
//
// toe_dq    - 
// return id - 
//---------------------------------------------------------------------------------------
u32
be_toe_dq_get_id (
   PBE_TOE_DQ_OBJECT toe_dq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_dq_get_buffer_length
//
//   Returns the default_buffer_length passed to the create function. 
//
// toe_dq               - 
// return buffer_length - 
//---------------------------------------------------------------------------------------
u32
be_toe_dq_get_buffer_length (
   IN PBE_TOE_DQ_OBJECT toe_dq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_sq_create
//
//   Creates a TOE send queue. This queue can be allocated once and used for many
//   different connections. 
//
// function_object    - 
// sgl                - no virtual address required 
// length_in_bytes    - 
// cq_object          - 
// tcp_sq             - 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_sq_create (
   IN PBE_FUNCTION_OBJECT             function_object,
   IN PSA_SGL                         sgl,
   IN u32                             length_in_bytes,
   IN PBE_CQ_OBJECT                   cq_object,
   OUT PBE_TOE_SQ_OBJECT              tcp_sq,
   IN MCC_WRB_CQE_CALLBACK            callback,
   IN PVOID                           callback_context,
   IN PBE_TOE_SQ_CREATE_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_toe_sq_destroy
//
//   Destroys the toe send queue. This function is called as part of be_function_cleanup. 
//
// tcp_sq        - 
// return status - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_toe_sq_destroy (
   IN PBE_TOE_SQ_OBJECT tcp_sq
);

//---------------------------------------------------------------------------------------
// Function: be_toe_sq_get_cid
//
//   Returns the TOE send queue cid allocated during the create function. This id is used
//   for ringing the doorbell to post WRBs to the queue. 
//
// sq         - 
// return cid - 
//---------------------------------------------------------------------------------------
u32
be_toe_sq_get_cid (
   IN PBE_TOE_SQ_OBJECT sq
);

//---------------------------------------------------------------------------------------
// Function: be_function_prepare_embedded_ioctl
//
//   Use the BE_FUNCTION_PREPARE_EMBEDDED_IOCTL macro, instead of calling this function
//   directly. 
//
// function_object - 
// wrb             - 
// payload_length  - 
// request_length  - 
// response_length - 
// opcode          - 
// subsystem       - 
// return ioctl    - 
//---------------------------------------------------------------------------------------
PVOID
be_function_prepare_embedded_ioctl (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PMCC_WRB            wrb,
   IN u32                 payload_length,
   IN u32                 request_length,
   IN u32                 response_length,
   IN u32                 opcode,
   IN u32                 subsystem
);

//---------------------------------------------------------------------------------------
// Function: be_function_prepare_nonembedded_ioctl
//
//   Use the BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL macro, instead of calling this function
//   directly. 
//
// function_object - 
// wrb             - 
// ioctl_header_va - 
// ioctl_header_pa - 
// payload_length  - 
// request_length  - 
// response_length - 
// opcode          - 
// subsystem       - 
// version         - 
// return ioctl    - 
//---------------------------------------------------------------------------------------
PVOID
be_function_prepare_nonembedded_ioctl (
   IN PBE_FUNCTION_OBJECT function_object,
   IN PMCC_WRB            wrb,
   IN PVOID               ioctl_header_va,
   IN u64                 ioctl_header_pa,
   IN u32                 payload_length,
   IN u32                 request_length,
   IN u32                 response_length,
   IN u32                 opcode,
   IN u32                 subsystem,
   IN u8                  version
);

//---------------------------------------------------------------------------------------
// Function: be_function_lock_mcc
//
//   Acquires the MCC posting lock. This lock must be held while calling
//   be_function_post_mcc_wrb. If using the be_function_peek_mcc_wrb, the caller should
//   acquire the lock before the peek function call and hold it until after the post
//   function call. Additionally, the "peeked" WRB should not be accessed after this lock
//   is released since another thread may reuse the WRB. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_lock_mcc (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_unlock_mcc
//
//   Releases the MCC posting lock acquired by calling be_function_lock_mcc. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_unlock_mcc (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_peek_mcc_wrb
//
//   Returns a pointer to the next WRB to post, either the mailbox buffer or the next WRB
//   in the ring. When using this function, the caller must ensure that no other thread
//   will post an MCC wrb until after this WRB is posted. 
//
// function_object - 
// return wrb      - Pointer to the next WRB to post, NULL on failure.
//---------------------------------------------------------------------------------------
PMCC_WRB
be_function_peek_mcc_wrb (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_post_mcc_wrb
//
//   Posts a WRB to either mailbox or mcc ring. The wrb pointer may be the pointer
//   returned by the be_function_peek_mcc_wrb function. It not, the wrb is copied into the
//   ring/mailbox. The callback may be NULL, in which case the function will not return
//   until the WRB completes. It will poll the mailbox status or MCC CQ in that case.
//   Other outstanding MCC WRBs may be completed in this context if the callback is NULL. 
//   This function (nor any other synchronous IOCTL function) may be called from an
//   MCC_WRB_CQE_CALLBACK context. Only asynchronous IOCTLs may be posted during the
//   completion of another IOCTL. Otherwise, the software will deadlock. 
//
// funcobj            - 
// wrb                - 
// callback           - 
// callback_context   - 
// optional_ioctl_va  - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_post_mcc_wrb (
   IN PBE_FUNCTION_OBJECT  funcobj,
   IN PMCC_WRB             wrb,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context,
   IN PVOID                optional_ioctl_va
);

//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_create
//
//   Creates a receive interface object. 
//
// function_object    - Previously created function object. 
// domain             - Domain for the command. 
// params             - Function-specific parameters. 
// rxif_object        - Pointer to an opaque object representing the interface 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// queue_context      - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_create (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN u32                       domain,
   IN PBE2_RXF_IF_CREATE_PARAMS params,
   OUT PBE_RXIF_OBJECT          rxif_object,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_destroy
//
//   Deletes a receive interface object. 
//
// rxif_object        - Receive interface object to destroy. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// queue_context      - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_destroy (
   IN PBE_RXIF_OBJECT           rxif_object,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT* queue_context
);

#if defined(SG_REVERSE_ENDIAN)
   // Defines the function argument structure to be2_rxf_vlan_add and be2_rxf_vlan_remove.
   typedef struct _BE2_RXF_VLAN_PARAMS {
      u32 version;
      u16 vlan_tag_outer;
      u16 vlan_tag_inner;
   } SG_PACK BE2_RXF_VLAN_PARAMS, *PBE2_RXF_VLAN_PARAMS;



#else
   // Defines the function argument structure to be2_rxf_vlan_add and be2_rxf_vlan_remove.
   typedef struct _BE2_RXF_VLAN_PARAMS {
      u32 version;
      u16 vlan_tag_outer;
      u16 vlan_tag_inner;
   } SG_PACK BE2_RXF_VLAN_PARAMS, *PBE2_RXF_VLAN_PARAMS;



#endif
//---------------------------------------------------------------------------------------
// Function: be2_rxf_add_vlan
//
//   Adds a VLAN to a receive interface. 
//
// rxif_object        - Previously created RXIF object. 
// params             - Function-specific parameters. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_add_vlan (
   IN OUT PBE_RXIF_OBJECT  rxif_object,
   IN BE2_RXF_VLAN_PARAMS  params,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

//---------------------------------------------------------------------------------------
// Function: be2_rxf_vlan_remove
//
//   Removes a VLAN from a receive interface. 
//
// rxif_object        - Previously created RXIF object. 
// params             - Function-specific parameters. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_vlan_remove (
   IN OUT PBE_RXIF_OBJECT  rxif_object,
   IN BE2_RXF_VLAN_PARAMS  params,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);

#if defined(SG_REVERSE_ENDIAN)
   // Defines the function argument structure to be2_rxf_umac_addr_add,
   // be2_rxf_umac_addr_remove, be2_rxf_mmac_addr_add, and be2_rxf_mmac_addr_remove. 
   typedef struct _BE2_RXF_MAC_ADDR_PARAMS {
      u32 version;
      SA_MAC_ADDRESS mac_address;
      boolean perfect;
   } SG_PACK BE2_RXF_MAC_ADDR_PARAMS, *PBE2_RXF_MAC_ADDR_PARAMS;



#else
   // Defines the function argument structure to be2_rxf_umac_addr_add,
   // be2_rxf_umac_addr_remove, be2_rxf_mmac_addr_add, and be2_rxf_mmac_addr_remove. 
   typedef struct _BE2_RXF_MAC_ADDR_PARAMS {
      u32 version;
      SA_MAC_ADDRESS mac_address;
      boolean perfect;
   } SG_PACK BE2_RXF_MAC_ADDR_PARAMS, *PBE2_RXF_MAC_ADDR_PARAMS;



#endif
//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_add_mac
//
//   Adds a unicast MAC address to a receive interface. 
//
// rxif_object        - Previously created RXIF object. 
// mac                - MAC to add. 
// mac_id             - Handle to the MAC for removing it. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// domain             - Domain for the command. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_add_mac (
   IN PBE_RXIF_OBJECT              rxif_object,
   IN SA_MAC_ADDRESS*              mac,
   OUT u32*                        mac_id,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_IF_ADD_MAC_QUEUE_CONTEXT queue_context,
   IN u8                           domain
);

//---------------------------------------------------------------------------------------
// Function: be2_rxf_if_remove_mac
//
//   Adds a unicast MAC address to a receive interface. 
//
// rxif_object        - Previously created RXIF object. 
// mac_id             - Handle to the MAC for removing it. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback context used in conjunction with the callback
//                      function pointer. 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// domain             - Domain for the command. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be2_rxf_if_remove_mac (
   IN PBE_RXIF_OBJECT              rxif_object,
   IN u32                          mac_id,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_IF_ADD_MAC_QUEUE_CONTEXT queue_context,
   IN u8                           domain
);

//---------------------------------------------------------------------------------------
// Function: be_set_hsw_config
//
//
// function_object    - 
// rxif_object        - 
// vlan_tag           - 
// callback           - 
// callback_context   - 
// queue_context      - 
// return return_type - 
//---------------------------------------------------------------------------------------
BESTATUS
be_set_hsw_config (
   IN BE_FUNCTION_OBJECT*       function_object,
   IN PBE_RXIF_OBJECT           rxif_object,
   IN u8                        domain,
   IN u16                       vlan_tag,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context
);

BESTATUS
be_get_hsw_config (
   IN PBE_FUNCTION_OBJECT    function_object,
   IN PBE_RXIF_OBJECT        rxif_object,
   IN u8                     domain,
   IN MCC_WRB_CQE_CALLBACK   callback,
   IN PVOID                  callback_context,
   IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
   OUT GET_HSW_CONFIG_PARAMS_OUT_V1* response_cpoy
);


//---------------------------------------------------------------------------------------
// Function: be_rxf_mac_address_read_write
//
//   Reads or writes the MAC address. 
//
// function_object    - 
// port1              - Specifies port 0 or 1
// mac1               - Specifies addres 0 or 1 (BE has 2 per port)
// mgmt               - Specifies the management port. If TRUE, the port1 and mac1
//                      parameters are ignored.
// write              - If TRUE, the MAC address is set.
// permanent          - If TRUE, the MAC read is the permanent address
// rxif_object        - RXIF object to modify
// mac_address        - Address to set, or address returned.
// callback           - optional
// callback_context   - optional
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_mac_address_read_write (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN boolean              port1,
   IN boolean              mac1,
   IN boolean              mgmt,
   IN boolean              write,
   IN boolean              permanent,
   IN PBE_RXIF_OBJECT      rxif_object,
   IN OUT PSA_MAC_ADDRESS  mac_address,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context
);


typedef struct _BE_READ_WRITE_MAC_PARAMETERS {
   u32 version;
   u32 pdid;
   u32 write;
   u32 permanent;
} SG_PACK BE_READ_WRITE_MAC_PARAMETERS, *PBE_READ_WRITE_MAC_PARAMETERS;



//---------------------------------------------------------------------------------------
// Function: be_rxf_read_write_pd_mac
//
//   Reads or writes the MAC address. 
//
// function_object    - 
// parameters         - 
// mac_address        - Address to set, or address returned.
// callback           - optional
// callback_context   - optional
// queue_context      - optional.
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_read_write_pd_mac (
   IN PBE_FUNCTION_OBJECT                 function_object,
   IN BE_READ_WRITE_MAC_PARAMETERS*       parameters,
   IN OUT PSA_MAC_ADDRESS                 mac_address,
   IN MCC_WRB_CQE_CALLBACK                callback,
   IN PVOID                               callback_context,
   IN PBE_PD_READ_WRITE_MAC_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_multicast_config
//
//   Writes all multicast addresses for this protection domain. This function is called as
//   part of be_function_cleanup to remove all multicast filters. 
//
// function_object    - 
// rxif_object        - Receive object to use for the multicast addresses. 
// promiscuous        - TRUE, enables promiscuous mode where all multicast packets are
//                      passed up. 
// num                - Number of mac addresses in the mac_table list. There are a maximum
//                      of 32 valid addresses. Set num to 0 to delete multicast settings
//                      associated with the receive object. 
// mac_table          - List of multicast MAC addresses to associate with the receive
//                      object. 
// callback           - Optional callback function pointer. 
// callback_context   - Optional callback function context. 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_multicast_config (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN PBE_RXIF_OBJECT             rxif_object,
   IN boolean                     promiscuous,
   IN u32                         num,
   IN SA_MAC_ADDRESS*             mac_table,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_MULTICAST_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_vlan_config
//
//   Writes all vlan tags for this protection domain. This function is called as part of
//   be_function_cleanup to remove all vlan filters. 
//
// function_object    - 
// rxif_object        - Optional interface object that VLAN tags are applied to. If NULL,
//                      the default (i.e. first) RXF interface is used. For iSCSI, NULL is
//                      always suitable. 
// promiscuous        - TRUE, enables vlan promiscuous mode where all vlan packets are
//                      passed up the stack. Allows supporting >32 vlan tags. 
// untagged           - TRUE, enables untagged mode for the interface. FALSE, disables
//                      untagged mode for the interface. 
// num                - Number of vlan tags in the array.
// vlan_tag_array     - 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_vlan_config (
   IN PBE_FUNCTION_OBJECT    function_object,
   IN PBE_RXIF_OBJECT        rxif_object,
   IN boolean                promiscuous,
   IN boolean                untagged,
   IN u32                    num,
   IN u16*                   vlan_tag_array,
   IN MCC_WRB_CQE_CALLBACK   callback,
   IN PVOID                  callback_context,
   IN PBE_VLAN_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_wake_on_lan_config
//
//   Configures wake-on-lan for the controller. Wake-on-lan is only supported for the host
//   system virtual machine - wake-on-lan is not supported per protection domain. To
//   enable magic packet reception, use null pointer arguments for pattern & pattern mask,
//   set magic_packet_enable to true, and set the port enables to true. To disable magic
//   packet reception, use null pointer arguments for pattern & pattern mask, set
//   magic_packet_enable to true, and set the port enables to false. To remove a pattern,
//   set both enable_port0 and enable_port1 to FALSE, and provide the index you previously
//   used to enable the pattern. This function is not called as part of
//   be_function_cleanup to remove all wake-on-lan patterns, since be_function_cleanup
//   could be called when the system wants to wake up on a magic or interesting packet. 
//
// function_object     - 
// enable_port0        - If TRUE, enable this pattern for port0. 
// enable_port1        - If TRUE, enable this pattern for port1.
// magic_packet_enable - If TRUE, enable magic packet for the enabled ports.
// index               - index 0-3 for the packet to program
// pattern             - 128 byte magic pattern
// pattern_mask        - 128 bit mask indicating which bytes in the pattern should be
//                       matched.
// callback            - optional
// callback_context    - optional
// queue_context       - Optional. Pointer to a previously allocated struct. If the MCC
//                       WRB ring is full, this structure is used to queue the operation.
//                       It will be posted to the MCC ring when space becomes available.
//                       All queued commands will be posted to the ring in the order they
//                       are received. It is always valid to pass a pointer to a generic
//                       BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                       are generally smaller than the generic struct. 
// return pend_status  - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the
//                       IOCTL completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_wake_on_lan_config (
   IN PBE_FUNCTION_OBJECT           function_object,
   IN boolean                       enable_port0,
   IN boolean                       enable_port1,
   IN boolean                       magic_packet_enable,
   IN u32                           index,
   IN PVOID                         pattern,
   IN PVOID                         pattern_mask,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context,
   IN PBE_WAKE_ON_LAN_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_rss_config
//
//   Configures Receive Side Scaling for the host networking function. This function is
//   called as part of be_function_cleanup to disable RSS. 
//
// function_object    - 
// rxif_object        - 
// rss_type           - Use enumeration ENABLE_RSS_ENUM.
// flush              - If 1, generates 1 rss flush completion per CQ per port. All
//                      packets with the old RSS hash will preceed the flush completion.
//                      This allows the driver to flush old packets. 
// hash               - 40 byte hash.
// cpu_table_length   - bytes (power of 2 from 2 to 128)
// cpu_table          - cpu_table_length bytes. Each byte indicates a CPU from 0 to
//                      num_cq-1.
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_rss_config (
   IN PBE_FUNCTION_OBJECT   function_object,
   IN PBE_RXIF_OBJECT       rxif_object,
   IN u32                   rss_type,
   IN boolean               flush,
   IN PVOID                 hash,
   IN u32                   cpu_table_length,
   IN u8*                   cpu_table,
   IN MCC_WRB_CQE_CALLBACK  callback,
   IN PVOID                 callback_context,
   IN PBE_RSS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_link_status
//
//   Queries the link status. 
//
// function_object    - 
// link_status        - returned status
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_link_status (
   IN PBE_FUNCTION_OBJECT           function_object,
   OUT BE_LINK_STATUS*              link_status,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context,
   IN PBE_LINK_STATUS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_query_eth_statistics
//
//   Queries the ethernet statistics. The resulting statistics are stored in the ioctl
//   response buffer. 
//
// function_object    - 
// va_for_ioctl       - Virtual address of the IOCTL buffer. This buffer must be
//                      physically contiguous. 
// pa_for_ioctl       - Physical address of the IOCTL buffer. This buffer must be
//                      physically contiguous. 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_query_eth_statistics (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN IOCTL_ETH_GET_STATISTICS*   va_for_ioctl,
   IN u64                         pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_ETH_STATS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_query_eth_statistics_v1
//
//   Queries the ethernet statistics for BE3. The resulting statistics are stored in the
//   ioctl response buffer. 
//
// function_object    - 
// va_for_ioctl       - Virtual address of the IOCTL_ETH_GET_STATISTICS_V1 buffer. This
//                      buffer must be physically contiguous. 
// pa_for_ioctl       - Physical address of the IOCTL buffer. This buffer must be
//                      physically contiguous. 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_query_eth_statistics_v1 (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN IOCTL_ETH_GET_STATISTICS_V1* va_for_ioctl,
   IN u64                          pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_ETH_STATS_QUEUE_CONTEXT  queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_promiscuous
//
//   Enables or disables promiscuous mode for each MAC port. 
//
// function_object    - 
// enable_port0       - Enable/Disable promiscuous mode on port 0.
// enable_port1       - Enable/Disable promiscuous mode on port 1.
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_promiscuous (
   IN PBE_FUNCTION_OBJECT           function_object,
   IN boolean                       enable_port0,
   IN boolean                       enable_port1,
   IN MCC_WRB_CQE_CALLBACK          callback,
   IN PVOID                         callback_context,
   IN PBE_PROMISCUOUS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_query_tcp_statistics
//
//   Queries the TCP statistics from the firmware. 
//
// function_object    - 
// va_for_ioctl       - Virtual address to IOCTL buffer.
// pa_for_ioctl       - Physical address to IOCTL buffer.
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated BE_QUEUE_CONTEXT
//                      struct. If the MCC WRB ring is full, this structure is used to
//                      queue the operation. It will be posted to the MCC ring when space
//                      becomes available. All queued commands will be posted to the ring
//                      in the order they are received. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_query_tcp_statistics (
   IN PBE_FUNCTION_OBJECT              function_object,
   IN IOCTL_COMMON_TCP_GET_STATISTICS* va_for_ioctl,
   IN u64                              pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK             callback,
   IN PVOID                            callback_context,
   IN PBE_TCP_STATS_QUEUE_CONTEXT      queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_query_toe_statistics
//
//   Queries the TOE statistics from the firmware. 
//
// function_object    - 
// va_for_ioctl       - Virtual address to IOCTL buffer.
// pa_for_ioctl       - Physical address to IOCTL buffer.
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated BE_QUEUE_CONTEXT
//                      struct. If the MCC WRB ring is full, this structure is used to
//                      queue the operation. It will be posted to the MCC ring when space
//                      becomes available. All queued commands will be posted to the ring
//                      in the order they are received. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_query_toe_statistics (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN IOCTL_TOE_GET_STATISTICS*   va_for_ioctl,
   IN u64                         pa_for_ioctl,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_TOE_STATS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_query_eth_network_stats
//
//   Queries the NIC/TOE statistics from the firmware. 
//
// function_object    - 
// query_toe          - 
// stats              - 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated BE_QUEUE_CONTEXT
//                      struct. If the MCC WRB ring is full, this structure is used to
//                      queue the operation. It will be posted to the MCC ring when space
//                      becomes available. All queued commands will be posted to the ring
//                      in the order they are received. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_query_eth_network_stats (
   IN PBE_FUNCTION_OBJECT                 function_object,
   IN u32                                 query_toe,
   OUT PNIC_TOE_STATS                     stats,
   IN MCC_WRB_CQE_CALLBACK                callback,
   IN PVOID                               callback_context,
   IN PBE_ETH_NETWORK_STATS_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_filter_config
//
//   Configures BladeEngine ethernet receive filter settings. 
//
// function_object    - 
// rxif_object        - Previously created RXIF object if receive interface options are
//                      changed. 
// settings           - Pointer to the requested filter settings. The response from
//                      BladeEngine will be placed back in this structure. 
// callback           - optional
// callback_context   - optional
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_filter_config (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN OUT PBE_RXIF_OBJECT          rxif_object,
   IN OUT NTWK_RX_FILTER_SETTINGS* settings,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_RXF_FILTER_QUEUE_CONTEXT queue_context
);

//---------------------------------------------------------------------------------------
// Function: be_rxf_set_link_speed_v1
//
//   Sets the link status for the client's port. Note: This command is used to program the
//   PHY from some vendors, which require a manual setting as opposed to PHYs from others
//   that can auto detect the speed settings. 
//
// function_object    - 
// phy_port           - 
// mac_speed          - 
// return return_type - 
//---------------------------------------------------------------------------------------
BESTATUS
be_rxf_set_link_speed_v1 (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 phy_port,
   IN u16                 mac_speed
);

//---------------------------------------------------------------------------------------
// Function: be_function_enable_interrupts
//
//   Enables interrupts for the given PCI function. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_enable_interrupts (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_disable_interrupts
//
//   Disables interrupts for the given PCI function. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_disable_interrupts (
   IN PBE_FUNCTION_OBJECT function_object
);

//---------------------------------------------------------------------------------------
// Function: be_cev_modify_msi_messages
//
//   Changes the number of MSI messages reported by CEV for a given function.
//
// function_object    - 
// num_msi_messages   - The number of MSI messages reported by CEV. Needs to be 0, 1, 2,
//                      4, or 8. Any other values are not allowed. 
// callback           - 
// callback_context   - 
// queue_context      - Optional. Pointer to a previously allocated struct. If the MCC WRB
//                      ring is full, this structure is used to queue the operation. It
//                      will be posted to the MCC ring when space becomes available. All
//                      queued commands will be posted to the ring in the order they are
//                      received. It is always valid to pass a pointer to a generic
//                      BE_GENERIC_QUEUE_CONTEXT. However, the specific context structs
//                      are generally smaller than the generic struct. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_cev_modify_msi_messages (
   IN PBE_FUNCTION_OBJECT                       function_object,
   IN u32                                       num_msi_messages,
   IN MCC_WRB_CQE_CALLBACK                      callback,
   IN PVOID                                     callback_context,
   IN PBE_CEV_MODIFY_MSI_MESSAGES_QUEUE_CONTEXT queue_context
);


typedef struct _UE_SCRATCHPAD {
   union {
      struct { 
         u32 disable : 12;                        // DWORD 0
         u32 rsvd0 : 2;                           // DWORD 0
         u32 no_soft_ue : 1;                      // DWORD 0
         u32 master : 1;                          // DWORD 0
         u32 master_function : 5;                 // DWORD 0
         u32 rsvd1 : 11;                          // DWORD 0
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK UE_SCRATCHPAD, *PUE_SCRATCHPAD;

SG_C_ASSERT(__sizeof__UE_SCRATCHPAD, sizeof (UE_SCRATCHPAD) == 4);


//---------------------------------------------------------------------------------------
// Function: be_ue_create
//
//   This function initializes the opaque BE_ERROR_OBJECT structure. It requires a valid
//   SA_DEV and a function_type of either UE_DISABLE_NET or UE_DISABLE_STORAGE. The
//   BE_ERROR_OBJECT must be created before any other functions are called. 
//
// sa_dev             - Previously created SA_DEV object for accessing registers.
// errobj             - The caller must allocate memory for the error object. If this
//                      function succeeds, it is initialized.
// return return_type - BE_SUCCESS on successful completion.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_create (
   IN SA_DEV*           sa_dev,
   OUT BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_destroy
//
//   Frees any resources allocated during be_ue_create. 
//
// errobj             - Previously created error object.
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_ue_destroy (
   BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_set_timeouts
//
//   This function overrides the default timeout values for all timeouts. 
//
// errobj                - Previously created error object.
// which_value           - 
// value_in_microseconds - New timeout duration in microseconds.
// return return_type    - BE_SUCCESS on successful completion.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_set_timeouts (
   BE_ERROR_OBJECT* errobj,
   u32              which_value,
   u64              value_in_microseconds
);

//---------------------------------------------------------------------------------------
// Function: be_ue_generate_software_ue
//
//   This function generates a hardware UE based due to a software failure. The
//   optional_code is only used for logging the error in the tiny log. 
//
// errobj             - Previously created error object.
// optional_code      - Optional, opaque code for logging. The listed values are used by
//                      BECLIB
// return return_type - BE_SUCCESS if the PCI function was able to generate the UE.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_generate_software_ue (
   BE_ERROR_OBJECT* errobj,
   u64              optional_code
);

//---------------------------------------------------------------------------------------
// Function: be_ue_check_hardware_ue
//
//   This function polls the hardware for a UE. It returns BE_SUCCESS if NO ue exists, and
//   failure if a UE is currently asserted. 
//
// errobj             - Previously created error object.
// return return_type - BE_SUCCESS if no UE exists.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_check_hardware_ue (
   BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_get_last_ue_status
//
//   This function returns the last cached value for the UE status registers, with the
//   masks applied. This can be used for logging the UE information during recovery. 
//
// errobj             - Previously created error object.
// return return_type - The previous cached value for the UE status registers with the UE
//                      masks applied.
//---------------------------------------------------------------------------------------
u64
be_ue_get_last_ue_status (
   BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_driver_load
//
//   This function must be called at PASSIVE. It checks for a UE and recovers during
//   driver load. 
//
// errobj             - Previously created error object.
// return return_type - BE_SUCCESS if the device is operational.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_driver_load (
   IN BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_recovery
//
//   This function must be run as PASSIVE context to allow sleeping. The
//   be_ue_recovery_start and be_ue_recovery_poll should be used to run the algorithm at
//   DISPATCH (interrupts disabled) context.  This function should be called if the driver
//   detects a UE condition and wants to run the recovery algorithm. The entire algorithm
//   is performed from passive context which may take 30 seconds or more. It returns
//   success if the algorithm succeeds and the device is ready for POST. 
//
// errobj             - Previously created error object.
// crash_dump         - If TRUE, this function is requesting to steal the master
//                      privileges from another device. It will also cause this function
//                      to steal the semaphore after a timeout. Typically, this would only
//                      be used by the crash dump driver that knows it is the only driver
//                      remaining in operation. 
// return return_type - BE_SUCCESS if the algorithm recovers the hardware.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_recovery (
   IN BE_ERROR_OBJECT* errobj,
   IN boolean          crash_dump
);

//---------------------------------------------------------------------------------------
// Function: be_ue_recovery_start
//
//   This function starts the UE recovery process. It should be called if the driver
//   detects a UE condition and wants to run the recovery algorithm. If this function
//   succeeds, the algorithm is running asynchronously. The be_ue_recovery_poll function
//   should be polled to determine the final status of the UE recovery algorithm. 
//
// errobj             - Previously created error object.
// crash_dump         - If TRUE, this function is requesting to steal the master
//                      privileges from another device. It will also cause this function
//                      to steal the semaphore after a timeout. Typically, this would only
//                      be used by the crash dump driver that knows it is the only driver
//                      remaining in operation. 
// return return_type - BE_SUCCESS if the UE recovery algorithm starts running.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_recovery_start (
   IN BE_ERROR_OBJECT* errobj,
   IN boolean          crash_dump
);

//---------------------------------------------------------------------------------------
// Function: be_ue_recovery_poll
//
//   After calling be_ue_recovery_start, the driver polls this function to check the
//   status of the UE recovery algorithm. If this function returns BE_PENDING, the
//   algorithm is still in progress. Eventually the algorithm will either return
//   BE_SUCCESS or a failure code. 
//
// errobj             - Previously created error object.
// return return_type - BE_PENDING while the algorithm is running.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_recovery_poll (
   IN BE_ERROR_OBJECT* errobj
);

//---------------------------------------------------------------------------------------
// Function: be_ue_link_tiny_log
//
//   Optional method to link a tiny log to the UE object. Many registration and recovery
//   events are logged to the tiny log. 
//
// errobj             - Previously created error object.
// tlog               - Previously created tiny log.
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_ue_link_tiny_log (
   BE_ERROR_OBJECT* errobj,
   SA_TLOG*         tlog
);

//---------------------------------------------------------------------------------------
// Function: be_ue_disable_recovery
//
//   Optional method write the "disable" field in the scratchpad register. This only
//   modifies the bit corresponding to this PCI function as defined in the
//   UE_SCRATCHPAD_DISABLE_ENUM. 
//
// errobj             - Previously created error object.
// disable            - If TRUE, set the bit in the disable field corresponding to this
//                      PCI function. If FALSE, clear the bit in the disable field
//                      corresponding to this PCI function. 
// return return_type - BE_SUCCESS if the UE scratchpad is successfully written.
//---------------------------------------------------------------------------------------
BESTATUS
be_ue_disable_recovery (
   BE_ERROR_OBJECT* errobj,
   boolean          disable
);

//---------------------------------------------------------------------------------------
// Function: be_function_abort_ioctls
//
//   When the driver detects a UE (or generates a software UE) this function may be used
//   to abort all pending IOCTLS and fail any future IOCTL requests. The aborted IOCTLs
//   will be completed with an error status code from the context of this call. Before
//   calling this function, the upper driver should quiesce all interrupts and threads
//   that may be processing the MCC cq ring. 
//
// function_object    - 
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_abort_ioctls (
   IN BE_FUNCTION_OBJECT* function_object
);

//---------------------------------------------------------------------------------------
// Function: be_function_link_error_object
//
//   This function links the error object to a function object. The MCC polling routines
//   use the error object for checking for hardware errors. Without an error object, the
//   polling routines may timeout, but they will not immediately detect an error. 
//
// function_object    - Previously created function object.
// errobj             - Previously created error object.
// return return_type - 
//---------------------------------------------------------------------------------------
void
be_function_link_error_object (
   IN BE_FUNCTION_OBJECT* function_object,
   IN BE_ERROR_OBJECT*    errobj
);


#ifdef SG_PRAGMA_PACK
   #pragma pack(pop)
#endif

#endif // __beclib_ll_bmap_h__

