/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   emulation_ll.h

@brief
   
CVS Header Information
   $Header: $
*/
#ifndef __emulation_ll_h__
#define __emulation_ll_h__

#define BE_GEN_STATIC_INLINE STATIC INLINE
#define BE_GEN_ASSERT ASSERT
// #include "be2_gen_id_ranges.h"   // Ring ID ranges and conversions

#define BE2_MAX_FUNCTIONS 8

#define BE_PORT_COUNT 2
#define BE_PMAC_TABLE_COUNT 128
#define BE_HMAC_TABLE_COUNT 64
#define BE_VLAN_TABLE_COUNT 64
#define BE_VLAN_ARRAY_COUNT 2

// Tracks one free page table during emulation.
typedef struct _BE_MCC_EMUL_FREE_PT {
    u64 pa;
    PVOID va;
} BE_MCC_EMUL_FREE_PT, *PBE_MCC_EMUL_FREE_PT;

// MAC filter managment
typedef struct _BE_MCC_EMUL_PMAC_TABLE {
    boolean valid;
    SA_MAC_ADDRESS mac_address;
    u64 interface_map[BE_PORT_COUNT];
} BE_MCC_EMUL_PMAC_TABLE, *PBE_MCC_EMUL_PMAC_TABLE;

// VLAN tag filter management
typedef struct _BE_MCC_EMUL_VLAN_TABLE {
    boolean valid;
    u16 vlan_tag;
    u64 interface_map;
} BE_MCC_EMUL_VLAN_TABLE, *PBE_MCC_EMUL_VLAN_TABLE;

#define BE_HARDWARE_IF_COUNT 34
#define BE_RSS_IF_COUNT 2
#define BE_ULP_IF_COUNT 128

#define BE_IF_BASE(__type__) (BE_ ## __type__ ## _IF_BASE)
#define BE_IF_COUNT(__type__) (BE_ ## __type__ ## _IF_COUNT)
#define BE_IF_END(__type__) (BE_IF_BASE(__type__) + BE_IF_COUNT(__type__))

#define BE_RSS_IF_BASE 0
#define BE_HARDWARE_IF_BASE (BE_RSS_IF_BASE + BE_RSS_IF_COUNT)
#define BE_ULP_IF_BASE (BE_HARDWARE_IF_BASE + BE_HARDWARE_IF_COUNT)
#define BE_IF_TOTAL (BE_ULP_IF_BASE + BE_ULP_IF_COUNT)

#define BE_IF_ID_TO_RELATIVE_ID(__type__, __id__) ((__id__) - BE_IF_BASE(__type__))
#define BE_RELATIVE_ID_TO_IF_ID(__type__, __id__) ((__id__) + BE_IF_BASE(__type__))

#ifndef BE_FIELD_OFFSET
#define BE_FIELD_OFFSET(type, field)   ((long)(unsigned int)&(((type *)0)->field))
#endif
// Interface based filter management
typedef struct _BE_MCC_EMUL_IF {
    u32 mac_id;
    u32 ref_count;
    u32 port_map;
    boolean broadcast_enable;
    boolean promiscuous_vlan_enable;
    boolean untagged_enable;
    boolean rss_enable;
    boolean ulp_if;
    boolean valid;
} BE_MCC_EMUL_IF, *PBE_MCC_EMUL_IF;

#define MAX_EMULATION_LRO_RINGS 64

// Emulation context -- one per chip object.  This info would normally
// be stored in ARM firmware.
typedef struct _BE_MCC_EMUL {

    SA_BV_DECLARE( etx_id_bitvector, 62 )
    SA_BV_DECLARE( lro_id_bitvector, MAX_EMULATION_LRO_RINGS )
    SA_BV_DECLARE( erx_rss_id_bitvector, 8 ) // 8 RSS ERX rings

    struct {
        SA_BV_DECLARE( eq_id_bitvector, 8 )
    } eq_func[BE2_MAX_FUNCTIONS];

    SA_BV_DECLARE( mcc_id_bitvector, MCC_ID_NUM )

    SA_BV_DECLARE( toe_defq_id_bitvector,   32 )
    SA_BV_DECLARE( iscsi_defq_id_bitvector, 32 )

    SA_BV_DECLARE( cq_id_iscsi_bitvector, 64 )
    SA_BV_DECLARE( cq_id_net_bitvector,   448 )
    
    SA_BV_DECLARE( toe_id_bitvector,     TOE_TX_ID_NUM )
    SA_BV_DECLARE( iscsi_id_bitvector,   ISCSI_TX_ID_NUM )

    SA_BV_DECLARE( rdma_id_bitvector,    RDMA_QP_ID_NUM )

    SA_BV_DECLARE (fcoe_id_bitvector,  FC_TX_ID_NUM)

    SA_BV_DECLARE( hardware_if_bitvector, BE_IF_COUNT(HARDWARE) )
    SA_BV_DECLARE( rss_if_bitvector, BE_IF_COUNT(RSS) )
    SA_BV_DECLARE( ulp_if_bitvector, BE_IF_COUNT(ULP) )

    BE_MCC_EMUL_PMAC_TABLE pmac_table[ BE_PMAC_TABLE_COUNT ];
    BE_MCC_EMUL_VLAN_TABLE vlan_table[ BE_PORT_COUNT ][ BE_VLAN_ARRAY_COUNT ][ BE_VLAN_TABLE_COUNT ];
    BE_MCC_EMUL_IF interface_table[ BE_IF_TOTAL ];

    boolean qinq_enable;
    u64 enable_map[2];
    u64 broadcast_map[2];
    u64 vlan_promiscuous_map[2];
    u64 untagged_map[2];

    u32 num_frag_pages_mapped;

    boolean shared_pt_mode;
    u32 num_free_pt[BE2_MAX_FUNCTIONS];
    u32 num_pt[BE2_MAX_FUNCTIONS];

    // START Needs to be defined for MPU
    SA_SGL emulation_sgl;
    u32 host_addr_lo;
    u32 host_addr_hi;
    // END Needs to be defined for MPU
    
    union {
        BE_MCC_EMUL_FREE_PT split[BE2_MAX_FUNCTIONS][256]; // second index is function
        BE_MCC_EMUL_FREE_PT shared[512];
    } free_pt;

    struct 
    {
        u16   ref_count;
        u16   function_number;
        PVOID va;
    } dt[ 512 ];

    struct {
        u8 port0    : 1;
        u8 port1    : 1;
        u8 function : 1;
    } acpi_table[4];

    u8 magic_packet_p0_storage;
    u8 magic_packet_p1_storage;

    u8 magic_packet_p0_net;
    u8 magic_packet_p1_net;

    u32 num_ooo_posted;

    u32 num_template_pages[3]; // one for each type (iscsi, toe, rdma)
    u64 scratch_pa[BE2_MAX_FUNCTIONS];         // one for each function
    PVOID scratch_va[BE2_MAX_FUNCTIONS];       // one for each function
    PVOID template_header_va[SA_NUMBER_OF_FIELD (CTX_MEMORY_MAP, template_pointers)];

} BE_MCC_EMUL, *PBE_MCC_EMUL;


void
be_mccemul_init (
    PBE_MCC_EMUL mccemul);

void
mccemul_write_xgmac (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 port,
    u32                 offset,
    u32                 data);


// ------------------------------------------------------
//  MPU Emulation
// ------------------------------------------------------
void 
_be_mpu_emul_access_mac_table (
        struct _BE_FUNCTION_OBJECT *function_object, 
        u32 index,
        u32 *mac_table_entry,
        boolean write );

boolean 
_be_mpuemul_is_link_up (
        struct _BE_FUNCTION_OBJECT *function_object, 
        u8 port, 
        u8 if_type,
        u8 phy_type);

void 
_be_mpu_emul_erx_ring_init (
        struct _BE_FUNCTION_OBJECT *function_object, 
        u32 qid, 
        u64* phys_addr, 
        u32 pd_reg, 
        u32 cq_id, 
        u32 mtu,
        u32 rss_if);

void 
_be_mpuemul_init (
    struct _BE_FUNCTION_OBJECT* function_object);

void 
_be_mpuemul_read_cev_mem (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 address,
    void*               data,
    u32                 bytes);

void 
_be_mpuemul_write_cev_mem (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 address,
    void*               data,
    u32                 bytes);


BESTATUS
_be_mpuemul_write_ctx_mem (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 address,
    void*               data,
    u32                 bytes);

BESTATUS
_be_mpuemul_write_ctx_ndma (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 address,
    void*               data,
    u32                 bytes);


BESTATUS
_be_mpuemul_write_ctx_mem_nopq (
    struct _BE_FUNCTION_OBJECT* function_object,
    u32                 address,
    void*               data,
    u32                 bytes);


BESTATUS
_be_mpuemul_read_ctx_mem (
    struct _BE_FUNCTION_OBJECT* FunctionObject,
    u32                 address,
    void*               data,
    u32                 bytes);


BESTATUS
_be_mpuemul_read_ctx_mem_nopq (
    struct _BE_FUNCTION_OBJECT* FunctionObject,
    u32                 address,
    void*               data,
    u32                 bytes);

void
_be_mpu_emul_erx_mtu_config(
   IN void* function_context,
   IN u32   ring_id,
   IN u32   mtu );

void
be_ep_read_mem (
    IN  struct _BE_FUNCTION_OBJECT* FunctionObject,
    IN  u32 processorID,
    IN  u32 offset,
    OUT u32 *result);

void
be_ep_write_mem (
    IN  struct _BE_FUNCTION_OBJECT* FunctionObject,
    IN  u32 processorID,
    IN  u32 offset,
    IN  u32 value);

BESTATUS
be_mpuemul_rrmsg (
    struct _BE_FUNCTION_OBJECT* FunctionObject,
    u32    txulp_number,
    u32    cid,
    u32    rrmsg_data0,
    u32    rrmsg_data1);

boolean
be_mccemul_is_emulated (
    struct _BE_FUNCTION_OBJECT* function_object, 
    PMCC_WRB wrb, 
    PVOID optional_ioctl_va);

BESTATUS
be_mccemul_process_wrb (
    IN struct _BE_FUNCTION_OBJECT* function_object, 
    IN PMCC_WRB wrb, 
    IN PVOID optional_ioctl_va);

BESTATUS
mccemul_update_template_header (
    IN struct _BE_FUNCTION_OBJECT* function_object,
    u32 cid,
    MPU_TEMPLATE_HEADER* header);

void
mccemul_write_cam (
    IN struct _BE_FUNCTION_OBJECT* function_object,
    u32 cid,
    MPU_OFFLOAD_CXN_ENTRY* entry);

void
mccemul_read_cam (
    IN struct _BE_FUNCTION_OBJECT* function_object,
    u32 cid,
    MPU_OFFLOAD_CXN_ENTRY* entry);

#endif // __emulation_ll_h__

