/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file
   test_ll.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/test_ll.h,v 1.3 2006/03/24 17:46:25 menz Exp $
*/
#ifndef __test_ll_h__
#define __test_ll_h__

typedef struct _BE_TEST_OBJECTS {
    BE_FUNCTION_OBJECT          func;
    BE_CHIP_OBJECT              chip;
    BE_EQ_OBJECT                eq;
    BE_CQ_OBJECT                cq;
    BE_CQ_OBJECT                cq_array[16];
    BE_MCC_OBJECT               mcc;
    BE_DEFAULT_PDU_QUEUE_OBJECT iscsi_defq;
    BE_ISCSI_WRB_QUEUE_OBJECT   iscsi_wrbq;
    BE_ETHSQ_OBJECT             eth_sq;
    BE_ETHRQ_OBJECT             eth_rq;
    BE_TOE_SQ_OBJECT            toe_sq;
    BE_TOE_RQ_OBJECT            toe_rq;
    BE_TOE_DQ_OBJECT            toe_dq;

    
    BE_MCC_WRB_CONTEXT mcc_context[32];
    SA_MAC_ADDRESS mac[16];
    u32 num_mac_callbacks;
    
} BE_TEST_OBJECTS, *PBE_TEST_OBJECTS;

void
be_read_mac_callback (PVOID context, BESTATUS status, MCC_WRB* optional_wrb)
{
    BE_TEST_OBJECTS *obj = (BE_TEST_OBJECTS*)context;
    obj->num_mac_callbacks++;
    TRACE (DL_INFO, "Callback number: %d status:0x%x wrb:0x%x", obj->num_mac_callbacks, status, optional_wrb);
    ASSERT (status == 0);
}


// Requires RING mode.
STATIC INLINE
BESTATUS
be_read_mac_n_times (
    BE_TEST_OBJECTS *obj,
    u32 num_times)
{
    u32 i=0;
    BESTATUS status = 0;
    SA_MAC_ADDRESS mac;
    
    for (i=0; i<SA_NUMBER_OF(mac.bytes); i++) 
        mac.bytes[i] = 0x10*i;

    obj->num_mac_callbacks = 0;
    SA_ZERO_MEM (obj->mac);

    ASSERT (num_times <= SA_NUMBER_OF(obj->mac));

    be_rxf_mac_address_read_write (&obj->func, 0, 0, 0, TRUE, &mac, NULL, NULL);

    for (i=0; i<num_times; i++) {
        status = be_rxf_mac_address_read_write (&obj->func, 0, 0, 0, FALSE, &obj->mac[i], be_read_mac_callback, obj);
        ASSERT (status >= 0);
    }

    while (obj->num_mac_callbacks < num_times) {

        be_mcc_process_cq (&obj->mcc, FALSE);

        sa_dev_stall (obj->func.sa_dev, 1);

        if ((++i & 0x7fff) == 0) {
            TRACE (DL_WARN, "waiting for mac read callbacks");
        }
    }

    for (i=0; i<num_times; i++) {
        TRACE (DL_INFO, "MAC address[%d]:" SA_MAC_FORMAT, i, SA_MAC_ARGS (obj->mac[i]));
    }

    return status;
}



STATIC INLINE
BESTATUS
be_test_ll (SA_DEV* sa_dev,
         u32 function_type, // BE_FUNCTION_TYPE_ISCSI or BE_FUNCTION_TYPE_NETWORK
         BE_TEST_OBJECTS *obj,
         u32 num_sgl,
         PSA_SGL *sgl // array of sgls, 8 pages each
         )  
{
    BESTATUS status = 0;
    u32 i;
    IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS_RESPONSE response;

    const u32 pt_sgl = 0;
    const u32 trash_sgl = 1;
    const u32 mcc_sgl = 2;
    const u32 cq_sgl = 3;

       
    status =
    be_function_and_chip_create (sa_dev, 
                                 function_type,
                                 sgl[mcc_sgl],
                                 &obj->func,
                                 &obj->chip);
    ASSERT (status == 0);


    status =
    be_function_post_template_header_memory (&obj->func, sgl[trash_sgl], &response);
    

    status = 
    be_function_post_page_tables (&obj->func, sgl[pt_sgl], 8);
    ASSERT (status == 0);

    status = 
    be_function_post_jell (&obj->func, sgl[trash_sgl], 1);
    ASSERT (status == 0);

    status = 
    be_function_post_zero_buffer (&obj->func, 0xbad00000ULL);
    ASSERT (status == 0);

    status = be_eq_create (
                 &obj->func,
                 sgl[trash_sgl],
                 4,                           // eqe size
                 SA_PAGE_SIZE / 4, // num entries
                 0xFFFFFFFF,                  // watermark
                 0xFFFFFFFF,                  // timer
                 &obj->eq);
    ASSERT (status == 0);

    status = be_cq_create (
                &obj->func, // IN PBE_FUNCTION_OBJECT function_object,
                sgl[cq_sgl],     // IN PSA_SGL             sgl,
                4096,       // IN u32                 length,
                FALSE,      // IN boolean             solicited_eventable,
                TRUE,       // IN boolean             no_delay,
                0xFFFFFFFF, // IN u32                 wm_thresh,
                NULL,       // IN PBE_EQ_OBJECT       eq_object,
                &obj->cq    //OUT PBE_CQ_OBJECT      cq_object
                );
    ASSERT (status == 0);    

    status = be_mcc_ring_create (
                 &obj->func,   // IN PBE_FUNCTION_OBJECT function_object,
                 sgl[mcc_sgl],          // IN PSA_SGL             sgl,
                 SA_PAGE_SIZE, // IN u32                 length,
                 obj->mcc_context,         // IN PBE_MCC_WRB_CONTEXT context_array,
                 SA_NUMBER_OF (obj->mcc_context),            // IN u32                 num_context_entries,
                 &obj->cq,     // IN PBE_CQ_OBJECT       cq,
                 &obj->mcc     // OUT PBE_MCC_OBJECT     mcc
                 );
    ASSERT (status == 0);

    status = be_read_mac_n_times (obj, 15);

    for (i=0; i<16; i++) {
        status = be_cq_create (
                    &obj->func, // IN PBE_FUNCTION_OBJECT function_object,
                    sgl[trash_sgl],        // IN PSA_SGL             sgl,
                    4096,       // IN u32                 length,
                    FALSE,      // IN boolean             solicited_eventable,
                    TRUE,       // IN boolean             no_delay,
                    0xFFFFFFFF, // IN u32                 wm_thresh,
                    &obj->eq,   // IN PBE_EQ_OBJECT       eq_object,
                    &obj->cq_array[i]    //OUT PBE_CQ_OBJECT      cq_object
                    );
        ASSERT (status == 0);
    }

    status = be_eth_sq_create (
                 &obj->func,   // IN PBE_FUNCTION_OBJECT function_object,
                 sgl[trash_sgl],          //  IN PSA_SGL             sgl,
                 SA_PAGE_SIZE, // IN u32                 length_in_bytes,
                 BE_ETH_TX_RING_TYPE_STANDARD,
                 1, // legacy default ULP
                 0, // port number
                 &obj->cq,     // IN PBE_CQ_OBJECT       cq_object,
                 &obj->eth_sq  // OUT PBE_ETHSQ_OBJECT   eth_sq
                 );
    ASSERT (status == 0);

    status = 
    be_eth_rq_create (
        &obj->func,  // IN PBE_FUNCTION_OBJECT function_object,
        sgl[trash_sgl],         // IN PSA_SGL             sgl,
        &obj->cq,    // IN PBE_CQ_OBJECT       cq_object,
        &obj->cq_array[0],    // IN PBE_CQ_OBJECT       bcmc_cq_object,
        &obj->eth_rq // OUT PBE_ETHRQ_OBJECT   eth_rq
        );
    ASSERT (status == 0);



    if (function_type == BE_FUNCTION_TYPE_ISCSI) {
        status = be_iscsi_defq_create (
                     &obj->func,      // IN PBE_FUNCTION_OBJECT           function_object,
                     sgl[trash_sgl],             // IN PSA_SGL                       sgl,
                     SA_PAGE_SIZE,    // IN u32                           length,
                     128,             // IN u32                           default_buffer_length,
                     &obj->cq,        // IN PBE_CQ_OBJECT                 cq_object,
                     &obj->iscsi_defq // OUT PBE_DEFAULT_PDU_QUEUE_OBJECT defq
                     );
        ASSERT (status == 0);

        status = be_iscsi_wrb_queue_create (
                     &obj->func,      // PBE_FUNCTION_OBJECT        function_object,
                     sgl[trash_sgl],             // PSA_SGL                    sgl,
                     SA_PAGE_SIZE     / sizeof (ISCSI_INITIATOR_WRB), // u32                        num_entries,
                     &obj->iscsi_wrbq // PBE_ISCSI_WRB_QUEUE_OBJECT wrbq
                     );
        ASSERT (status == 0);

        status = be_iscsi_post_sgl_pages (
                     &obj->func, // IN PBE_FUNCTION_OBJECT function_object,
                     sgl[trash_sgl],        //  IN PSA_SGL             sgl,
                     0,          // IN u32                 page_offset,
                     1           // IN u32                 num_pages
                     );
        ASSERT (status == 0);

        status = be_function_iscsi_ooo_buffers_post (
                     &obj->func, // IN PBE_FUNCTION_OBJECT function_object,
                     1,          // IN u32                 page_count,
                     sgl[trash_sgl]         // IN PSA_SGL             sgl
                     );
        ASSERT (status == 0);

    } else {
        // Networking
        status =  be_toe_sq_create (
                      &obj->func,   // IN PBE_FUNCTION_OBJECT function_object,
                      sgl[trash_sgl],          // IN PSA_SGL             sgl,
                      SA_PAGE_SIZE, // IN u32                 length_in_bytes,
                      &obj->cq,     // IN PBE_CQ_OBJECT       cq_object,
                      &obj->toe_sq  // OUT PBE_TOE_SQ_OBJECT  tcp_sq
                      );
        ASSERT (status == 0);


        status = be_toe_rq_create (
                     &obj->func,                       // IN PBE_FUNCTION_OBJECT function_object,
                     sgl[trash_sgl],                              // IN PSA_SGL             sgl,
                     SA_PAGE_SIZE,                     // IN u32                 length,
                     be_toe_sq_get_cid (&obj->toe_sq), // IN u32                 cid,
                     &obj->cq,                         // IN PBE_CQ_OBJECT       cq_object,
                     &obj->toe_rq                      // OUT PBE_TOE_RQ_OBJECT  toe_rq
                     );
        ASSERT (status == 0);

        status = be_toe_dq_create (
                     &obj->func,   // PBE_FUNCTION_OBJECT function_object,
                     sgl[trash_sgl],          // PSA_SGL             sgl,
                     SA_PAGE_SIZE, // u32                 length,
                     2048,         // u32                 default_buffer_length,
                     0x4,          // default chute for TOE
                     &obj->toe_dq  // PBE_TOE_DQ_OBJECT   toe_dq
                     );
        ASSERT (status == 0);

        {
            u8 cpus[2] = {0};
            u8 hash[16] = {0};
            u32 cq_array[4] = {0};
                
            status = be_rxf_rss_config (
                         &obj->func,     // IN PBE_FUNCTION_OBJECT  function_object,
                         ENRSS_TCP_IPV4, // IN u32                  rss_type,
                         4,              // IN u32                  num_cq,
                         cq_array,       // IN u32* cq_id_array,
                         hash,           // IN PVOID                hash,
                         2,              // IN u32                  cpu_table_length,
                         cpus,           // IN u8* cpu_table,
                         NULL,           // IN MCC_WRB_CQE_CALLBACK callback,
                         NULL            // IN PVOID                callback_context
                         );
            ASSERT (status == 0);
        }

        {
            SA_MAC_ADDRESS mac[4];
            u32 j,k;
            
            for (k=0; k<SA_NUMBER_OF(mac); k++) {
                for (j=0; j<SA_NUMBER_OF(mac[0].bytes); j++) 
                    mac[k].bytes[j] = 0x10*k + j;
            }

            status = be_rxf_multicast_config (
                         &obj->func, // IN PBE_FUNCTION_OBJECT  function_object,
                         FALSE,      // IN boolean              promiscuous,
                         2,          // IN u32                  num,
                         mac,       // IN SA_MAC_ADDRESS* mac_table,
                         NULL,       // IN MCC_WRB_CQE_CALLBACK callback,
                         NULL        //  IN PVOID                callback_context
                         );
            ASSERT (status == 0);
        }

        {
            u16 tags[6] = {1, 2, 3, 4, 5, 6 };
            status = be_rxf_vlan_config (
                         &obj->func, // IN PBE_FUNCTION_OBJECT  function_object,
                         FALSE,      // IN boolean              promiscuous,
                         TRUE,       // IN boolean              untagged
                         2,          // IN u32                  num,
                         tags,       // IN u16* vlan_tag_array,
                         NULL,       // IN MCC_WRB_CQE_CALLBACK callback,
                         NULL        // IN PVOID                callback_context
                         );
            ASSERT (status == 0);

            status = be_rxf_vlan_config (
                         &obj->func, // IN PBE_FUNCTION_OBJECT  function_object,
                         FALSE,      // IN boolean              promiscuous,
                         TRUE,       // IN boolean              untagged
                         6,          // IN u32                  num,
                         tags,       // IN u16* vlan_tag_array,
                         NULL,       // IN MCC_WRB_CQE_CALLBACK callback,
                         NULL        // IN PVOID                callback_context
                         );
            ASSERT (status == 0);

            status = be_rxf_vlan_config (
                         &obj->func, // IN PBE_FUNCTION_OBJECT  function_object,
                         FALSE,      // IN boolean              promiscuous,
                         TRUE,       // IN boolean              untagged
                         3,          // IN u32                  num,
                         tags,       // IN u16* vlan_tag_array,
                         NULL,       // IN MCC_WRB_CQE_CALLBACK callback,
                         NULL        // IN PVOID                callback_context
                         );
            ASSERT (status == 0);
        }
    }


    //-------------------------------------------------------------
    //  And now cleanup....
    //-------------------------------------------------------------
    status = be_function_cleanup (&obj->func);

    BREAKPOINT ();

    return status;
}


STATIC INLINE
void
be_test_hl (
    IN      SA_DEV*                 sa_dev)
{
    #define num_sgl 4

    BESTATUS status = 0;    
    BE_TEST_OBJECTS *obj = NULL;    
    BE_SHARED_MEM_DESCRIPTOR mem[num_sgl];
    PSA_SGL sgl[num_sgl];
    u32 i;

    SA_ZERO_MEM (&mem);

    TRACE (DL_INFO, "Test the LL api cleanup. Start.");    
    
    for (i=0; i<num_sgl; i++) {
        status  = be_shared_memory_allocate (SA_PAGE_SIZE * 8, SA_TAG ("mem_"),  TRUE, NULL, &mem[i]);
        ASSERT (status == 0);
        sgl[i] = mem[i].sgl;
    }

    obj = sa_malloc (sizeof (BE_TEST_OBJECTS), SA_MALLOC_NORMAL, SA_TAG ("obj "));
    ASSERT (obj);

    //be_test_ll (sa_dev, BE_FUNCTION_TYPE_ISCSI, obj, num_sgl, sgl);
    be_test_ll (sa_dev, BE_FUNCTION_TYPE_NETWORK, obj, num_sgl, sgl);

    for (i=0; i<num_sgl; i++) {
        be_shared_memory_free (&mem[i]);
    }
    sa_free (obj);

    TRACE (DL_INFO, "Test the LL api cleanup. Done.");
}

#endif // __test_ll_h__

