/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
@file
   ulpchaos.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/ulpchaos_ll.h,v 1.5 2006/03/23 17:16:49 menz Exp $
*/
#ifndef __ulpchaos_h__
#define __ulpchaos_h__

#ifdef BE_ENABLE_ULP_CHAOS

    INLINE STATIC BESTATUS
    be_function_map_ut_pages (
        PBE_FUNCTION_OBJECT function_object,
        u32 cra,
        u32 num_pages,
        PSA_SGL sgl)
    {
        BESTATUS  status = BE_SUCCESS;  
        MCC_WRB *wrb = NULL;
        IOCTL_LOWLEVEL_MAP_UT_PAGES *ioctl = NULL;

        be_lock_wrb_post (function_object);

        wrb = be_function_peek_mcc_wrb (function_object);
        if (!wrb) {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_MAP_UT_PAGES);

        ioctl->params.request.cra = cra;
        ioctl->params.request.num_pages = (u16)num_pages;

        // Create a page list for the IOCTL.
        be_sgl_to_pa_list (sgl, 
                           ioctl->params.request.pages, 
                           SA_NUMBER_OF (ioctl->params.request.pages));

        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "MCC to map UT pages failed.");
            goto Error;
        }    

    Error:
        be_unlock_wrb_post (function_object);

        return status;
    }
        
    INLINE STATIC BESTATUS
    be_function_unmap_ut_pages (
        PBE_FUNCTION_OBJECT function_object,
        u32 cra,
        u32 num_pages)
    {
        BESTATUS  status = BE_SUCCESS;  
        MCC_WRB *wrb = NULL;
        IOCTL_LOWLEVEL_UNMAP_UT_PAGES *ioctl = NULL;
        CRA_UNION crau;

        be_lock_wrb_post (function_object);

        wrb = be_function_peek_mcc_wrb (function_object);
        if (!wrb) {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }

        // Due to PTC BUG we cannot use tag==0 in PTC.  This makes some CRAs illegal.
        crau.dw = cra;
        ASSERT (crau.ptc_tag.tag != 0);

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_UNMAP_UT_PAGES);

        ioctl->params.request.cra = cra;
        ioctl->params.request.num_pages = (u16)num_pages;

        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "MCC to map UT pages failed.");
            goto Error;
        }    

    Error:
        be_unlock_wrb_post (function_object);

        return status;
    }

    #define ULPCHAOS_STAG (0x20000)

    // This function establishes the ULP Chaos firmware running in the ULP0 processors that
    // are not normally used.  This only works in an fpga config1 that includes these 
    // processors.  Also, the firmware must be previously downloaded in the MPU initialization.
    STATIC INLINE void 
    be_function_ulpchaos_create (
        PBE_FUNCTION_OBJECT function_object,
        PSA_SGL sgl)                            // 2 pages, page aligned with a virtual address
    {
        union { PE_CRA pe; u32 dw; } cra;
        u32 config_ctx[4] = { 0 };
        u32 ctx[4] = { 0x11111111, 0x22222222, 0x33333333, 0x44444444};
        u32 config_data = 0, i = 0;  
        u8 *va = NULL;
        BESTATUS status = 0;

        // # of bytes to initialize in context for read DMA test
        const u32 fixed_extr_wdma_sz = 4096; 

        // Use the RDMA section in ctx that is available
        const u32 host_pa_ctx_loc    = SA_FIELD_OFFSET (CTX_MEMORY_MAP, qp); 
        const u32 wdma_ctx_src_addr  = host_pa_ctx_loc + 0x20;
        
        ASSERT (sgl);
        va = sa_sgl_get_base_va (sgl);

        ASSERT (va);
        ASSERT (sa_sgl_get_page_count (sgl) >= 2);
        ASSERT (sa_sgl_get_byte_count (sgl) >= 2 * SA_PAGE_SIZE);
        ASSERT (sa_sgl_get_offset (sgl) == 0); // SGL must be page aligned

        // Initialize the entire memory
        for (i=0; i<sa_sgl_get_byte_count (sgl); i++) {
            va[i] = 0xAA + i;
        }

        // Create any CRA in an unused portion of the UT.
        cra.dw = 0;
        cra.pe.constant = UC_CRA_PE;
        cra.pe.offset = 0;
        cra.pe.stag = ULPCHAOS_STAG;

        // Map the first page into the UT.
        status = be_function_map_ut_pages (function_object, cra.dw, 1, sgl);
        ASSERT (status == 0);

        // WDMA traffic setup
        for (i=0; i<fixed_extr_wdma_sz/sizeof(ctx); i++) {
            _be_mpuemul_write_ctx_mem (function_object, wdma_ctx_src_addr + i * sizeof (ctx), ctx, sizeof (ctx));
        }

        config_data = (config_data << 1) | 0; // $use_txulp2]
        config_data = (config_data << 1) | 0; // $use_txulp1]
        config_data = (config_data << 1) | 1; // $use_txulp0]
        config_data = (config_data << 1) | 0; // $use_rxulp2]
        config_data = (config_data << 1) | 0; // $use_rxulp1]
        config_data = (config_data << 1) | 1; // $use_rxulp0]
        config_data = (config_data << 1) | 1; // $extr_ipc]
        config_data = (config_data << 1) | 1; // $extr_ctx]
        config_data = (config_data << 1) | 1; // $extr_dbuf]
        config_data = (config_data << 1) | 1; // $extr_rdma]
        config_data = (config_data << 1) | 1; // $extr_wdma]

        // Save the second page address in ctx
        config_ctx[0] = sa_lo (sa_sgl_get_page (sgl, 1));
        config_ctx[1] = sa_hi (sa_sgl_get_page (sgl, 1));
        config_ctx[2] = cra.dw;
        config_ctx[3] = config_data;

        _be_mpuemul_write_ctx_mem (function_object, host_pa_ctx_loc, config_ctx, sizeof (config_ctx));

        TRACE (DL_INFO, 
            "ULP Chaos. PA:0x%x_%08x cra:0x%x config:0x%x", 
            sa_hi (sa_sgl_get_page (sgl, 1)), 
            sa_lo (sa_sgl_get_page (sgl, 1)), 
            cra.dw, 
            config_data);

        // This starts the processors running.  It sets the upper most bit in each semaphore
        // register to indicate that it is waiting for a semaphore write to continue.
        CSR_WRITE_CONST (function_object, rxulp0.ep.ep_control, 0);
        CSR_WRITE_CONST (function_object, txulp0.ep.ep_control, 0);

        //  These writes start the chaos. Normally we don't start it automatically
        // but since we're in the debugging stages we do.
        CSR_WRITE_CONST (function_object, rxulp0.ep.ep_semaphore, 0);
        CSR_WRITE_CONST (function_object, txulp0.ep.ep_semaphore, 0);
    }


    STATIC INLINE void 
    be_function_ulpchaos_destroy (
        PBE_FUNCTION_OBJECT function_object)
    {
        EP_CONTROL_CSR ctrl;
        union { PE_CRA pe; u32 dw; } cra;
        BESTATUS status = 0;

        ctrl.dw = 0;
        ctrl.CPU_reset = 1;

        TRACE (DL_INFO, "Destroy the ULP chaos.");

        // Stop the ULP chaos by halting the processors.  
        // BUG -- Improve ULP chaos stopping.  
        // This probably leaks all kinds of chip resources and generally causes a complete mess.  Plus,
        // it doesn't actually wait for the DMA's to complete.
        CSR_WRITE_CONST (function_object, rxulp0.ep.ep_control, ctrl.dw);
        CSR_WRITE_CONST (function_object, txulp0.ep.ep_control, ctrl.dw);
        
        // Create any CRA in an unused portion of the UT.
        cra.dw = 0;
        cra.pe.constant = UC_CRA_PE;
        cra.pe.offset = 0;
        cra.pe.stag = ULPCHAOS_STAG;

        // UnMap the page in the UT.
        status = be_function_unmap_ut_pages (function_object, cra.dw, 1);
        ASSERT (status == 0);
    }


#endif

#endif // __ulpchaos_h__

