/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

    iscsiapi.c - iSCSI Layer Interface Routines

    Purpose:

        The iSCSI layer interface routines for ACIT and management layer.
        

--*/

#include "acitpch.h"

ACIT_VOID
iscsi_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller);

ACIT_UEXACT32
iscsi_get_ws_sizes(
    PACIT_UEXACT32          pcontroller_ws_size,
    PACIT_UEXACT32          ptarget_ws_size,
    PACIT_UEXACT32          plun_ws_size,
    PACIT_UEXACT32          pio_ws_size,
    PACIT_UEXACT32          ptmf_ws_size)
{
    *pcontroller_ws_size = sizeof(ISCSI_CONTROLLER_WS);
    *ptarget_ws_size = sizeof(ISCSI_TARGET_WS);
    *plun_ws_size = 0;
    *pio_ws_size = 0; //sizeof(ISCSI_IO_WS);
    *ptmf_ws_size = sizeof(ISCSI_TMF_WS);

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_init_mem_descriptors(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PACIT_CONTROLLER_WS     pc = pcontroller;
    PISCSI_CONTROLLER_WS    pi = GET_ISCSI_CONTROLLER_WS(pc);
    PACIT_PARAMETERS        p = &pc->params;

    //const ACIT_UEXACT32 a1 = ACIT_ALIGN_1BYTE;
    const ACIT_UEXACT32 a8 = ACIT_ALIGN_8BYTE;
    const ACIT_UEXACT32 a4k = ACIT_ALIGN_4096BYTE;
    const ACIT_UEXACT32 cached = ACIT_MEM_CATEGORY_CACHED;
    const ACIT_UEXACT32 non_cached = ACIT_MEM_CATEGORY_NON_CACHED;

    // allocate 2 template headers, for data out and nop out PDUs
    ap_set_mem_const (pc, (PUEXACT8)"g_header",   ISCSI_MEM_GLOBAL_HEADER,  ACIT_ISCSI_PDU_HEADER_SIZE * 2,      a4k, non_cached);   

    pi->cxn_mem_index = pc->acit_var_desc_index.range.cxn;
    ap_set_mem_variable (pc, (PUEXACT8)"cxn", pi->cxn_mem_index, p->num_cxn, sizeof(ISCSI_CONNECTION),a8, cached);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_init_controller_ws(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    OSM_MEMSET(piscsi_controller, 0, sizeof(ISCSI_CONTROLLER_WS));

    return ISCSI_STATUS_SUCCESS;
}

ACIT_VOID
iscsi_init_global_templates(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PPDU_DATA_OUT           pdata_out;
    PPDU_NOP_OUT            pnop_out;

    /* 
     * Patch the opcode field of global template headers with write code. The rest
     * will be patched up by the EP
     */
    pdata_out = (PPDU_DATA_OUT)pcontroller->
        mem_descriptors[ISCSI_MEM_GLOBAL_HEADER].virtual_address;

    OSM_MEMSET(pdata_out, 0, ACIT_ISCSI_PDU_HEADER_SIZE);

    pdata_out->opcode = IIOC_SCSI_DATA;

    pnop_out = (PPDU_NOP_OUT)((ACIT_UEXACT8 *)pcontroller->
        mem_descriptors[ISCSI_MEM_GLOBAL_HEADER].virtual_address + 
                                        ACIT_ISCSI_PDU_HEADER_SIZE);

    OSM_MEMSET(pnop_out, 0, ACIT_ISCSI_PDU_HEADER_SIZE);

    pnop_out->ttt = 0xffffffff;
    pnop_out->f_bit = 1;
    pnop_out->i_bit = 0;
}

ACIT_UEXACT32
iscsi_init_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    ACIT_UEXACT32           cxn_pool_size;
    ACIT_UEXACT32           desc_index,i, num_cxns = 0;
    
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    piscsi_controller->max_cxnpool_per_desc = (ACIT_CFG_MAX_CONTIGUOUS_MEM / sizeof(ISCSI_CONNECTION));

    desc_index = piscsi_controller->cxn_mem_index;

    for(i=0; i<pcontroller->acit_var_desc_index.num.cxn; i++)
    {
        piscsi_controller->connections_pool[i] = pcontroller->mem_descriptors[desc_index].virtual_address;
        cxn_pool_size = pcontroller->mem_descriptors[desc_index].size_in_bytes;
        OSM_MEMSET(piscsi_controller->connections_pool[i], 0, cxn_pool_size);

        num_cxns++;
        if(num_cxns >= piscsi_controller->max_cxnpool_per_desc)
        {
            num_cxns++;
            desc_index++;
        }
        piscsi_controller->connection_table[i] = ACIT_NULL;
    }

    OSM_ASSERT (pcontroller->params.num_cxn <= BE_MAX_CXNS);

    iscsi_init_global_templates(pcontroller);

    iscsi_unitt_init_controller(pcontroller);

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_init_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_UEXACT16          pconnection_handle,
    PACIT_UEXACT16          pcid,
    PPHYS_ADDR              ptemplate_address)
{
    PISCSI_TARGET_WS        piscsi_target;    
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    OSM_MEMSET( piscsi_target,
                0,
                sizeof(ISCSI_TARGET_WS));

    piscsi_target->pacit_ws = ptarget;
    //piscsi_target->min_load = 0xFFFFFFFF;

    //The connection resources are reserved for IOs, TMFs and 1 NOPOut
    piscsi_target->max_ios_per_cxn = (ACIT_UEXACT16)(pcontroller->params.num_cmd_per_cxn - 
                                                    (pcontroller->params.num_tmfs_per_cxn + 1));

    // Initialize a connection to send out I/Os later
    pcxn = iscsi_alloc_cxn(pcontroller);

    if(pcxn == ACIT_NULL)
    {
        OSM_ASSERT(0);
        return ISCSI_STATUS_NO_CONNECTIONS;
    }

    // Initialize num_connections once its allocated
    piscsi_target->num_connections = 1;

    piscsi_target->pcxn_list[0] = pcxn;
    pcxn->connection_handle = 0;
    //Zero out the stastistics counters
    pcxn->digest_errors = 0;
    pcxn->format_errors = 0;
    pcxn->iscsi_cmd_pdus = 0;
    pcxn->iscsi_rsp_pdus = 0;
    pcxn->total_bytes_xferred = 0;
    pcxn->total_bytes_read = 0;
    ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_INITIALIZING);

    //
    // This is not valid at this point. However, the offload session
    // IOCTL needs a connection handle. So we're passing a 0.
    // Note that offload connection IOCTL (post leading connection) has
    // to have a valid connection handle.
    //
    *pconnection_handle = (ACIT_UEXACT16)pcxn->connection_handle;
    *pcid = pcxn->cid;
    ISCSI_GET_PDU_TEMPLATE_ADDRESS(pcontroller, ptemplate_address);

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_reinit_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_UEXACT16          pconnection_handle,
    PACIT_UEXACT16          pcid,
    PPHYS_ADDR              ptemplate_address)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i = 0;
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    for (i = 0; i < piscsi_target->num_connections; i++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[i]);
        pcxn = piscsi_target->pcxn_list[i];
        if(pcxn)
        {
            pcxn->status = 0;
            //Zero out the stastistics counters
            pcxn->digest_errors = 0;
            pcxn->format_errors = 0;
            pcxn->iscsi_cmd_pdus = 0;
            pcxn->iscsi_rsp_pdus = 0;
            pcxn->total_bytes_xferred = 0;
            pcxn->total_bytes_read = 0;

            ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_INITIALIZING);
            OSM_ASSERT(!pcxn->num_of_active_cmds);
        }
    }

    piscsi_target->num_connections_invalidated = 0;
    piscsi_target->num_connections_reset = 0;
    piscsi_target->num_connections_uploaded = 0;
    //piscsi_target->num_connections = 0;
    piscsi_target->num_connections_reset_needed = 0;

    //piscsi_target->min_load = 0xFFFFFFFF;

    OSM_ASSERT(piscsi_target->pcxn_list[0]);
    if(piscsi_target->pcxn_list[0])
    {
        *pconnection_handle = (ACIT_UEXACT16)piscsi_target->pcxn_list[0]->connection_handle;
        *pcid = piscsi_target->pcxn_list[0]->cid;
    }
    ISCSI_GET_PDU_TEMPLATE_ADDRESS(pcontroller, ptemplate_address);

    return ISCSI_STATUS_SUCCESS;

}

ACIT_UEXACT32
iscsi_update_session_info(
    PACIT_CONTROLLER_WS     pcontroller,                            
    PACIT_TARGET_WS         ptarget,
    POFFLOAD_SESSION_OUT    psession)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           acit_status = ACIT_STATUS_SUCCESS;

    // only when MGMT has called this function to init the session
    // info can the initialized flag be set
    ptarget->initialized = 1;

    ap_init_eto(pcontroller, ptarget, psession->extended_timeout);

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    piscsi_target->pacit_ws = ptarget;
    /* Vivek 3/5/04 Changing line below to use 
    session_id instead of session_handle.unique_object_id */
    piscsi_target->session_handle = psession->session_handle; 
    piscsi_target->session_status = psession->session_status; 
    /* open/close/in recovery */
    piscsi_target->exp_cmd_sn = psession->expcmdsn;
    piscsi_target->max_cmd_sn = psession->maxcmdsn;
    piscsi_target->cmd_sn = psession->cmdsn;

    piscsi_target->max_connections = (ACIT_UEXACT8)psession->max_connections;

    if(psession->max_connections == 0)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("MaxConnections=0 in offload completion!\n"));
        acit_status = ACIT_STATUS_ERROR;
    }

    piscsi_target->login_options.max_outstanding_r2t = 1;
    piscsi_target->login_options.erl = psession->error_recovery_level;
    piscsi_target->flags = 0;

    iscsi_init_connection_information(ptarget, 
                                      &psession->connection_info,
                                      ACIT_TRUE);

    /* If the target was in recovery, complete the recovery cycle */
    if(ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY))
    {
        //
        // Bug 19481 : Ignore link status here. We want to start ETO
        // if offload fails here or stop ETO and mark target ready
        // if offload succeeds. If the target is not marked ready and
        // link toggles after target reconnected, but before it could
        // offload, the target will marked busy forever. Link up event
        // will not change target state, so do it here. If the link stays
        // down it will trigger another timeout and session recovery.
        //

        ap_complete_target_recovery(pcontroller, ptarget, acit_status, 1);
        pcontroller->counters.num_targets_in_recovery--;
    }

    if(acit_status == ACIT_STATUS_SUCCESS)
    {
        // Reset the offload err counter
        ptarget->counters.offload_err = 0;
    }

    return acit_status;
}

ACIT_UEXACT32
iscsi_update_session_info_from_async(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT32           connection_handle)
{
    PISCSI_TARGET_WS        piscsi_target;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    piscsi_target->session_handle = session_handle; 

    // Assumes no MCS
    piscsi_target->pcxn_list[0]->connection_handle = connection_handle;
    piscsi_target->pcxn_list[0]->piscsi_target_ws = piscsi_target;

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
iscsi_init_hiber_target_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           session_handle,
    ACIT_UEXACT16           connection_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    piscsi_target->session_handle = session_handle; 
    piscsi_target->session_status = 0; 
    piscsi_target->max_connections = 1;
    piscsi_target->pcxn_list[0]->connection_handle = connection_handle;
    piscsi_target->num_connections = 1;

    ptarget->event_flag = 0;
    ptarget->cleanup_flag = 0;
    ptarget->flags = 0;
    ptarget->eto_status = 0;
    ptarget->sub_state = 0;
    ptarget->initialized = 1;

    return ISCSI_STATUS_SUCCESS;
}

PISCSI_CONNECTION
iscsi_get_cxn_from_cxn_handle(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget, 
    ACIT_UEXACT32           connection_handle,
    PACIT_UEXACT32          pis_leading_cxn)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i;
    PISCSI_CONNECTION       pcxn = ACIT_NULL;

    OSM_ASSERT(pis_leading_cxn);

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    *pis_leading_cxn = ACIT_FALSE;

    for(i=0; i<MAX_CXNS_PER_SESSION; i++)
    {
        if(piscsi_target->pcxn_list[i] &&
            piscsi_target->pcxn_list[i]->connection_handle == connection_handle)
        {
            pcxn = piscsi_target->pcxn_list[i];
            break;
        }
    }

    if(pcxn && i == 0)
        *pis_leading_cxn = 1;

    return pcxn;
}

ACIT_UEXACT32
iscsi_process_connection_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           connection_handle,
    ACIT_UEXACT32           session_handle)
{
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           is_leading_cxn = 0;

    pcxn = iscsi_get_cxn_from_cxn_handle(pcontroller, 
                                         ptarget, 
                                         connection_handle,
                                         &is_leading_cxn);

    if(pcxn == ACIT_NULL)
    {
        //
        // The target ws is present, but no pcxn structure is returned. This 
        // can happen if offload has failed for a persistent target and all
        // the backpointer that gets setup during offload completion has not
        // happened. Initialize the target with whatever data was returned with
        // async notification
        //

        if(iscsi_is_session_open(ptarget))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p session %d has notification on already open cxn but is un-init yet\n", 
                            ptarget, ptarget->piscsi_ws->session_handle));
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p has notification on un-init cxn!\n", ptarget));

            iscsi_update_session_info_from_async(pcontroller,
                                                 ptarget,
                                                 session_handle,
                                                 connection_handle);

            // Now call reinit target workspace
            ap_reinit_target_ws(pcontroller, 
                                ptarget, 
                                ptarget->piscsi_ws->session_handle);
        }
    }
    else
    {
        if(is_leading_cxn)
        {
            if(iscsi_is_session_open(ptarget))
            {
                //
                // The connection exists and is already opened. This is a 
                // duplicate notification, ignore it
                //
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p session %d has notification on already open cxn\n", 
                                ptarget, ptarget->piscsi_ws->session_handle));
            }
            else
            {

                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p session %d has notification on leading cxn\n", 
                            ptarget, ptarget->piscsi_ws->session_handle));
                // The leading connection is coming back online, reinit target ws
                ap_reinit_target_ws(pcontroller, 
                                    ptarget, 
                                    ptarget->piscsi_ws->session_handle);
            }
       }
        else
        {
            //
            // A non-leading connection is coming back online. Add this connection
            // to the session and initiate offload
            //
            // TODO: MCS case
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Target 0x%p has notification on non-leading cxn\n", ptarget));
            OSM_ASSERT(0);
            //ap_init_connection(pcontroller, ptarget, connection_handle);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_get_cid_from_connection_handle (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           connection_handle)
{
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           is_leading_cxn = 0;

    pcxn = iscsi_get_cxn_from_cxn_handle(pcontroller, 
                                         ptarget, 
                                         connection_handle,
                                         &is_leading_cxn);

    return ((pcxn == 0) ? 0xffffffff : pcxn->cid );

}

// No MCS support
#if 0
ACIT_UEXACT32
iscsi_close_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_VOID              mgmt_ioctl_request)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           is_leading_cxn = 0;
    ACIT_UEXACT32           status;
    PCLOSE_CONNECTION_PARAMS_IN pclose_cxn_in;

    /*
     * This API is called for logging out of a connection by MGMT layer
     * or when called from SNMP or iscsicli
     */
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    piscsi_target->mgmt_ioctl_request = mgmt_ioctl_request;
    pclose_cxn_in    = (PCLOSE_CONNECTION_PARAMS_IN)((PACIT_IOCTL_REQUEST)mgmt_ioctl_request)->pdata_buffer;

    ACIT_BIT_SET(ptarget->flags, TGT_FLAG_SKIP_TIMEOUT_TICKS_COUNTDOWN);

    pcxn = iscsi_get_cxn_from_cxn_handle(pcontroller, 
                                         ptarget, 
                                         pclose_cxn_in->connection_handle,
                                         &is_leading_cxn);

    if (piscsi_target->num_connections == 1){
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Can't remove last cxn 0x%p:: cid %x\n", pcxn, pcxn->cid));
        return MGMT_STATUS_CANT_REMOVE_LAST_CONNECTION; 
    }
    if (pcxn == ACIT_NULL) {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR, ("Invalid connection\n"));
        return MGMT_STATUS_INVALID_CONNECTION;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Sending Logout on cxn 0x%p:: cid %x\n", pcxn, pcxn->cid));

    // Mark target as busy and initiate a reopen a connection IOCTL
    ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_REMOVE_CONNECTION);

    if ((status = ap_send_logout (pcontroller, 
                            ptarget, 
                            pcxn->cri, 
                            ISCSI_LOGOUT_REASON_CLOSE_CONNECTION)) != ACIT_STATUS_SUCCESS)
    {
        OSM_ASSERT(0); // TOREVIEW Why are the flags cleared?
        ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_REMOVE_CONNECTION);
    }
   
    return status;

}
#endif

ACIT_UEXACT32
iscsi_check_and_close_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           recover)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           status, cleanup_flag, i;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    if(recover)
        cleanup_flag = TGT_CLEANUP_FLAG_FOR_RECOVERY;
    else
        cleanup_flag = TGT_CLEANUP_FLAG_FOR_LOGOUT;

    ACIT_BIT_SET(ptarget->flags, TGT_FLAG_SKIP_TIMEOUT_TICKS_COUNTDOWN);

    //
    // If the link is down, or IOs are outstanding on the target
    // issue a reset. Else if one or more connection
    // is in error, follow the abortive cleanup path. Do not invalidate
    // outstanding commands, just invalidate connections with or without
    // a reset depending on the state. 
    // Lastly, if no connection is in error only then issue a graceful
    // logout
    //
    if(ACIT_IS_CONTROLLER_LINK_DOWN(pcontroller, piscsi_target->port_num))
    {
        status = ap_setup_target_for_cleanup(pcontroller, 
                                             ptarget, 
                                             ACIT_U32_MINUS_ONE, 
                                             cleanup_flag,
                                             ACIT_TRUE);

        if(status == ACIT_STATUS_SUCCESS)
        {
            ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_INVALIDATION_UPLOAD_IN_PROGRESS);
            iscsi_invalidate_all_connections(pcontroller,
                                            ptarget,
                                            ACIT_TRUE);    // force rst
        }
    }
    else 
    {
        //if(ISCSI_IS_ABORTIVE_CLEANUP(pcontroller, ptarget, piscsi_target))
        if(iscsi_is_abortive_cleanup(pcontroller, ptarget, piscsi_target))
        {
            status = ap_setup_target_for_cleanup(pcontroller, 
                                                 ptarget, 
                                                 ACIT_U32_MINUS_ONE, 
                                                 cleanup_flag,
                                                 ACIT_TRUE);

            if(status == ACIT_STATUS_SUCCESS)
            {
                ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_INVALIDATION_UPLOAD_IN_PROGRESS);
                iscsi_invalidate_all_connections(pcontroller,
                                                ptarget,
                                                ACIT_FALSE);   // rst on need
            }
        }
        else
        {
            status = ap_setup_target_for_cleanup(pcontroller, 
                                                ptarget, 
                                                ACIT_U32_MINUS_ONE, 
                                                cleanup_flag,
                                                ACIT_FALSE);

            if(status == ACIT_STATUS_SUCCESS)
            {
                ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_LOGOUT_PENDING);
                for(i=0; i<MAX_CXNS_PER_SESSION; i++)
                {
                    if(piscsi_target->pcxn_list[i])
                    {
                        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Sending Logout on cxn 0x%p\n", piscsi_target->pcxn_list[i]));

                        ap_send_logout (pcontroller, 
                                        ptarget, 
                                        piscsi_target->pcxn_list[i]->cri, 
                                        ISCSI_LOGOUT_REASON_CLOSE_SESSION);
                        break;
                    }
                }
            }
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_logout_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_VOID              mgmt_ioctl_request)
{
    PISCSI_TARGET_WS        piscsi_target;

    /*
     * This API is called for logging out of a session by MGMT layer
     * or when ASYNC logout is received
     */
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    piscsi_target->mgmt_ioctl_request = mgmt_ioctl_request;

    iscsi_check_and_close_session(pcontroller, ptarget, ACIT_FALSE);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_recover_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_VOID              mgmt_ioctl_request)
{
    PISCSI_TARGET_WS        piscsi_target;

    //
    // Session recovery can be invoked because
    // a. TCPIP settings changed on the fly (from MGMT)
    // b. There was an unexpected TCP Reset on a connection
    // c. A TMF did not complete
    // d. EP firmware reported errors on the CQ which needs a connection reset
    // e. Target logged out asynchronously
    // f. An IO exceeded IO retry count
    //

    //
    // 05/21/09 - According to the new implementation of handling 0 window update from
    // targets, the ping timer might be running when we've been asked to recover the
    // session. Stop the ping timer first
    //
    ap_clear_target_event(pcontroller, ptarget, EVENT_TGT_PING_WAIT);

    if(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Session recovery invoked multiple times for target 0x%p\n", ptarget));
        // return pendng status since recovery is already in progress
        return ACIT_STATUS_PENDING;
    }

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    piscsi_target->mgmt_ioctl_request = mgmt_ioctl_request;
    pcontroller->eh_stats.target_recoveries++;
    ptarget->eh_stats.recoveries++;
    pcontroller->counters.num_targets_in_recovery++;

    iscsi_check_and_close_session(pcontroller, ptarget, ACIT_TRUE);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_recover_all_sessions(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_VOID              mgmt_ioctl_request)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PACIT_TARGET_WS         ptarget;
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_TARGET_ENTRY      ptarget_entry = ACIT_NULL;
    ACIT_UEXACT32           index, num_targets = 0, status;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    piscsi_controller->mgmt_ioctl_request = mgmt_ioctl_request;
    ACIT_BIT_SET(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_SESSION_RECOVERY);

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
                continue;

            piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
            status = iscsi_recover_session(pcontroller, 
                                  ptarget, 
                                  piscsi_target->mgmt_ioctl_request);

            if(status == ACIT_STATUS_SUCCESS)
            {
                num_targets++;
            }
        }
    }

    if(num_targets == 0)
    {
        // There are no sessions that have upload pending. Call the completion routine
        if(piscsi_controller->mgmt_ioctl_request)
        {
            mgmt_all_targets_close_notify(pcontroller, piscsi_controller->mgmt_ioctl_request);
            piscsi_controller->mgmt_ioctl_request = ACIT_NULL;
        }
        else
        {
            mgmt_reopen_sessions(pcontroller, 
                                 IOCTL_ISCSI_REOPEN_ALL_SESSIONS, 
                                 ACIT_U32_MINUS_ONE,
                                 0, // domain 0
                                 ACIT_FALSE);
        }

        ACIT_BIT_CLR(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_SESSION_RECOVERY);
    }
    else
    {
        piscsi_controller->num_uploads_pending = num_targets;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
iscsi_write_logout_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(plogout_handle->ptarget);
    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, plogout_handle->cri));
    if (plogout_handle->reason_code)
        pcontroller->eh_stats.ssn_logout_others++;
    else
        pcontroller->eh_stats.ssn_logout_normals++;

    pdu_logout_build ((PPDU_LOGOUT)&plogout_handle->noncmd_bhs.iscsi_hdr, 
                      1,                    // immediate
                      (ACIT_UEXACT8)plogout_handle->reason_code, 
                      (ACIT_UEXACT16)pcxn->iscsi_cid, 
                      plogout_handle->itt,
                      plogout_handle->cmd_sn
                      );   

}

ACIT_VOID
iscsi_write_nop_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle)
{  
    PISCSI_TARGET_WS        piscsi_target;

    piscsi_target = GET_ISCSI_TARGET_WS(pnop_handle->ptarget);

    pdu_build_nop_out ((PPDU_NOP_OUT)&pnop_handle->noncmd_bhs.iscsi_hdr, 
                       1,                           // immediate bit==1, implies cmd_sn 
                                                    // not incremented
                       &pnop_handle->optional_lun,  // lun
                       pnop_handle->issued_itt, 
                       pnop_handle->ttt,  
                       pnop_handle->cmd_sn);
}    


ACIT_VOID
iscsi_complete_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT8            iresponse,
    ACIT_UEXACT32           reason_code,
    ACIT_UEXACT8            timeout)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT16           issue_reset;
    PISCSI_CONNECTION       pcxn;
    
    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    if (iresponse != ISCSI_PDU_LOGOUT_RESPONSE_SUCCESS) 
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("Bad logout status 0x%x\n", iresponse));
    }
    
    issue_reset = timeout ? 1 : 0;

    if (reason_code == ISCSI_LOGOUT_REASON_CLOSE_SESSION)
        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_INVALIDATION_UPLOAD_IN_PROGRESS);

    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, cri));
    if(pcxn)
    {
        iscsi_invalidate_connection(pcontroller, 
                                    piscsi_target,
                                    pcxn,
                                    issue_reset);
    }
}

ACIT_UEXACT32
iscsi_process_target_event(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           async)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_U32                i;
    ACIT_UEXACT32           status = ACIT_STATUS_SUCCESS;

    /*
     * Currently 2 target events are handled in iSCSI, one for logout and
     * other for connection invalidation
     */

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    OSM_ASSERT(piscsi_target);

    if(ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_LOGOUT_PENDING))
    {
        iscsi_check_and_close_session(pcontroller, ptarget, ACIT_FALSE);
    }
    else if(ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_RECOVERY_LOGOUT_PENDING))
    {
        iscsi_check_and_close_session(pcontroller, ptarget, ACIT_TRUE);
    }
    else if(ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_CXN_EVENT_PENDING))
    {
        // walk thru the different connections, check any processing
        // is waiting and start them now

        for(i = 0; i < MAX_CXNS_PER_SESSION; i++)
        {
            if((piscsi_target->pcxn_list[i]) == 0)
                continue;

            if(piscsi_target->pcxn_list[i]->status & ISCSI_CONN_STATUS_NEED_INVALIDATION)
            {
                ACIT_BIT_CLR(piscsi_target->pcxn_list[i]->status, 
                             ISCSI_CONN_STATUS_NEED_INVALIDATION);

                status |= iscsi_invalidate_connection(pcontroller, 
                                                      piscsi_target, 
                                                      piscsi_target->pcxn_list[i],
                                                      ACIT_FALSE);
            }
            else if(piscsi_target->pcxn_list[i]->status & ISCSI_CONN_STATUS_NEED_UPLOAD)
            {
                ACIT_BIT_CLR(piscsi_target->pcxn_list[i]->status, 
                             ISCSI_CONN_STATUS_NEED_UPLOAD);

                status |= iscsi_upload_connection(pcontroller, 
                                                  piscsi_target, 
                                                  piscsi_target->pcxn_list[i]->cri);
            }
            else if(piscsi_target->pcxn_list[i]->status & ISCSI_CONN_STATUS_NEED_RESET)
            {
                ACIT_BIT_CLR(piscsi_target->pcxn_list[i]->status, 
                             ISCSI_CONN_STATUS_NEED_RESET);

                status |= iscsi_reset_connection(pcontroller, 
                                                 piscsi_target->pcxn_list[i]);
            }
        }

        // all events are cleared?
        if(async & (status == ACIT_STATUS_SUCCESS))
        {
            ap_clear_target_event(pcontroller, 
                                  ptarget, 
                                  EVENT_TGT_CXN_EVENT_PENDING);
        }
    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
iscsi_init_lun_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_LUN_WS            plun)
{

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_init_io_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_WS             pio)
{
    PACIT_IO_REQUEST        pio_request = pio->pio_request;
    ACIT_UEXACT8            read_bit = FALSE, write_bit = FALSE;

    if (pio_request->flags.inbound_data) 
    {
        read_bit = TRUE;
    } 
    else if(pio_request->flags.outbound_data)
    {
        write_bit = TRUE;
    }

    pdu_build_command ((PPDU_COMMAND)&pio->cmd_bhs.u1.iscsi_hdr,
                       (ACIT_UEXACT8)pio_request->task_attribute,
                       pio_request->data_transfer_length,      // expected transfer length for SCSI op
                       pio_request->cdb, 
                       &(pio_request->lun.u.lun_bytes),
                       read_bit,
                       write_bit);

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_init_tmf_ws(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_TMF_WS            ptmf)
{
    return ISCSI_STATUS_SUCCESS;
}

#if 0
// load based connection scheduling, not needed for 1 connection/session case
INLINE ACIT_UEXACT32
iscsi_is_cxn_least_loaded(
    PISCSI_CONNECTION       pcxn,
    PACIT_UEXACT32          pmin_load)
{
    if ((pcxn->total_pending_bytes * pcxn->num_of_active_cmds) < *pmin_load)
    {
        *pmin_load = pcxn->total_pending_bytes * pcxn->num_of_active_cmds;
        return ACIT_TRUE;
    }

    return ACIT_FALSE;
}
#endif

INLINE ACIT_UEXACT8
iscsi_assign_best_cxn_for_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target)
{
    ACIT_UEXACT8            i=0;

    for (i = 0; i < piscsi_target->num_connections; i++)
    {
        if(piscsi_target->pcxn_list[i] && 
            piscsi_target->pcxn_list[i]->num_tmfs < pcontroller->params.num_tmfs_per_cxn &&
            (!piscsi_target->pcxn_list[i]->busy))
        {
            return i;
        }
    }

    return 0xff;
}

#if 0
INLINE ACIT_UEXACT8
iscsi_assign_best_cxn_for_io(
    PISCSI_TARGET_WS        piscsi_target,
    ACIT_UEXACT32           outbound,
    PACIT_UEXACT32          pnum_io_credit_busy)
{
    ACIT_UEXACT8            idx, i = 0;

    /*
     * Select the connection based on round robin policy
     */
    idx = (piscsi_target->last_cxn + 1) % piscsi_target->num_connections;
    //idx = (piscsi_target->last_cxn + 1) % MAX_CXNS_PER_SESSION;
    for (i = 0; i < piscsi_target->num_connections; 
            idx = ((idx + 1) % piscsi_target->num_connections), i++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[idx]);
        {
            if(piscsi_target->pcxn_list[idx]->busy)
                continue;

            if(outbound)
            {
                if(ACIT_IS_BIT_SET(piscsi_target->pcxn_list[idx]->status, 
                                   ISCSI_CONN_STATUS_WRITE_REQ_LIMIT))
                {
                    continue;
                }
            }

            if(piscsi_target->pcxn_list[idx]->num_of_active_cmds >=
                                        piscsi_target->max_ios_per_cxn)
            {
                (*pnum_io_credit_busy)++;
                continue;
            }
#if 0
            // Do not worry about the min load right now just use 
            // round robin approach for selecting connection
            if(iscsi_is_cxn_least_loaded(piscsi_target->pcxn_list[idx], 
                                            &piscsi_target->min_load))
            {
                return i;
            }
#endif
            piscsi_target->last_cxn = idx;
            return idx;
        }
        //idx = ((idx + 1) % piscsi_target->num_connections);       
    }

    return 0xff;

}
/*
 * This API needs to handle the following cases
 * a) check if all connections are busy and mark the target as busy. Connections
 *    go busy if they reach the threshold of cmds_per_cxn. Connections also go
      busy when there is no enough credits on them
 * b) check for link status on a connection while deciding on the connection.
 *    Controller is NOT marked busy unless both port links are down.
 * c) Logic for checking the write queue depth limitation. Ideally this needs to
 *    go to HWI module, but since its needed at the time of deciding the connection
 *    to send out the request on, it probably needs to be done here
 */

// this is a MCS version, not needed
ACIT_UEXACT32
iscsi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_HANDLE         pio_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_IO_REQUEST        pio_request;
    PACIT_IO_WS             pio;
    ACIT_UEXACT8            connection_index = 0xff;
    ACIT_UEXACT32           num_io_credit_busy = 0;

    pio = pio_handle->pio;
    pio_request = pio->pio_request;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    OSM_ASSERT(piscsi_target);
    OSM_ASSERT(!ptarget->busy);

    connection_index = iscsi_assign_best_cxn_for_io(piscsi_target,
                                                    pio_request->flags.outbound_data,
                                                    &num_io_credit_busy);

    if(connection_index == 0xff)
    {
        if(num_io_credit_busy == piscsi_target->num_connections)
        {
            /*
            * Mark target as busy since no connections are free
            */
            ap_flag_busy_target(pcontroller,
                                ptarget,
                                TARGET_BUSY_ISCSI_IO_CREDITS);
        }

        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    if(pio_request->flags.outbound_data)
    {
        pio_handle->is_write = ACIT_TRUE;
        piscsi_target->pcxn_list[connection_index]->num_of_active_writes++;
        piscsi_target->pcxn_list[connection_index]->pending_xfer_bytes += 
                                        pio_request->data_transfer_length;

        // setup state for next time
        if(piscsi_target->pcxn_list[connection_index]->num_of_active_writes >= 
                                    BE_MAX_R2T_CREDITS)
        {
            ACIT_BIT_SET(piscsi_target->pcxn_list[connection_index]->status,
                         ISCSI_CONN_STATUS_WRITE_REQ_LIMIT);
        }
    }
    else
    {
        pio_handle->is_write = ACIT_FALSE;
        piscsi_target->pcxn_list[connection_index]->pending_read_bytes += 
                                        pio_request->data_transfer_length;
    }

    piscsi_target->pcxn_list[connection_index]->num_of_active_cmds++;
    piscsi_target->pcxn_list[connection_index]->total_pending_bytes += 
                                    pio_request->data_transfer_length;

    piscsi_target->min_load = \
        piscsi_target->pcxn_list[connection_index]->num_of_active_cmds *
        piscsi_target->pcxn_list[connection_index]->total_pending_bytes;

    // Return the connection
    pio_handle->cri = (ACIT_UEXACT16)piscsi_target->pcxn_list[connection_index]->cri;
    //pio_handle->connection_index = connection_index;
    pio_handle->pconnection_handle = piscsi_target->pcxn_list[connection_index];

    return ACIT_STATUS_SUCCESS;
}
#endif

ACIT_UEXACT32
iscsi_alloc_io_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_HANDLE         pio_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_IO_REQUEST        pio_request;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    OSM_ASSERT(piscsi_target);
    OSM_ASSERT(!ptarget->busy);
    OSM_ASSERT(piscsi_target->num_connections == 1);

    // no MCS support, just jump to connection 0
    pio_request = pio_handle->pio->pio_request;
    pio_handle->is_write = (ACIT_UEXACT8)pio_request->flags.outbound_data;

    //
    // if task attribute is untagged, then issue the command only
    // if no other command is outstanding
    //

    // ACIT_TASK_ATTRIBUTE_UNTAGGED has value 0
    pio_handle->is_untagged = (ACIT_UEXACT8)!(pio_request->task_attribute);

    if(pio_handle->is_untagged)
    {
        if(piscsi_target->pcxn_list[0]->num_of_active_cmds)
        {
            //
            // if there are commands pending, then mark target busy so that
            // when active command count goes to 0, the target is made ready
            // and IOs flushed
            //
            ap_flag_busy_target(pcontroller,
                                ptarget, 
                                TARGET_BUSY_FLUSH_UNTAGGED_CMD);

            return ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
        else
        {
            //
            // mark the target as busy so that no further commands will
            // be issued
            //
            ap_flag_busy_target(pcontroller,
                                ptarget,
                                TARGET_BUSY_UNTAGGED_CMD);

            pcontroller->stats.untagged_cmd++;
            ptarget->stats.untagged_cmd++;
        }
    }

    if(pio_handle->is_write)
    {
        if(ACIT_IS_BIT_CLR(piscsi_target->pcxn_list[0]->status, 
                           ISCSI_CONN_STATUS_WRITE_REQ_LIMIT))
        {
            piscsi_target->pcxn_list[0]->num_of_active_writes++;
            piscsi_target->pcxn_list[0]->pending_xfer_bytes += 
                                            pio_request->data_transfer_length;

            // setup state for next time
            if(piscsi_target->pcxn_list[0]->num_of_active_writes >= 
                                        BE_MAX_R2T_CREDITS)
            {
                ACIT_BIT_SET(piscsi_target->pcxn_list[0]->status,
                             ISCSI_CONN_STATUS_WRITE_REQ_LIMIT);
            }
        }
        else
        {
            return ACIT_STATUS_INSUFFICIENT_RESOURCE;
        }
    }
    else
    {
        piscsi_target->pcxn_list[0]->pending_read_bytes += 
                                        pio_request->data_transfer_length;
    }

    piscsi_target->pcxn_list[0]->num_of_active_cmds++;
    piscsi_target->pcxn_list[0]->total_pending_bytes += 
                                    pio_request->data_transfer_length;

    // Return the connection
    pio_handle->cri = (ACIT_UEXACT16)piscsi_target->pcxn_list[0]->cri;
    pio_handle->pconnection_handle = piscsi_target->pcxn_list[0];

    if((piscsi_target->pcxn_list[0]->num_of_active_cmds >=
                            piscsi_target->max_ios_per_cxn))
    {
        ap_flag_busy_target(pcontroller,
                            ptarget,
                            TARGET_BUSY_ISCSI_IO_CREDITS);

    }

    return ACIT_STATUS_SUCCESS;
}

INLINE
ACIT_VOID
iscsi_update_io_counters(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_IO_REQUEST        pio_request;
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(pio_handle->ptarget);
    pio_request = pio_handle->pio->pio_request;

    //pcxn = piscsi_target->pcxn_list[pio_handle->connection_index];
    pcxn = pio_handle->pconnection_handle;

    OSM_ASSERT (pcxn);

    if(pio_request->data_transfer_length)
    {
        OSM_ASSERT(pcxn->total_pending_bytes);
    }

    pcxn->total_pending_bytes -= pio_request->data_transfer_length;

    if (pio_handle->is_write == ACIT_FALSE)
    {
        pcxn->pending_read_bytes -= pio_request->data_transfer_length;
    }
    else
    {
        pcxn->pending_xfer_bytes -= pio_request->data_transfer_length;
    }
}

ACIT_UEXACT32
iscsi_free_io_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_IO_HANDLE         pio_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(pio_handle->ptarget);
    pcxn = pio_handle->pconnection_handle;

    OSM_ASSERT(pcxn->num_of_active_cmds);
    pcxn->num_of_active_cmds--;

    if(pio_handle->is_write)
    {
        //pcxn = piscsi_target->pcxn_list[pio_handle->connection_index];
        pcxn->num_of_active_writes--;

        if(ACIT_IS_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_WRITE_REQ_LIMIT))
        {
            //
            // this is a internal state in iSCSI set to prevent the write Q
            // from overflowing beyond 128. But we don't mark the target busy
            // which means that the flush_queued_ios flag must be explicitly
            // turned on if there are pending IOs
            //
            ACIT_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_WRITE_REQ_LIMIT);

            if(pcontroller->counters.queued_ios)
            {
                pcontroller->state.flush_queued_ios = ACIT_TRUE;
            }
        }
    }

    //
    // Note!!! Assumes MCS is not supported
    //
    if(piscsi_target->pcxn_list[0]->num_of_active_cmds == 0)
    {
        //
        // All active commands have been completed, check if target has any
        // tagged command to flush
        //
        if(ACIT_IS_BIT_SET(pio_handle->ptarget->busy, TARGET_BUSY_FLUSH_UNTAGGED_CMD))
        {
            ap_flag_ready_target(   pcontroller,
                                    pio_handle->ptarget, 
                                    TARGET_BUSY_FLUSH_UNTAGGED_CMD);
        }
    }

    if(ACIT_IS_BIT_SET(pio_handle->ptarget->busy, TARGET_BUSY_ISCSI_IO_CREDITS))
    {
        ap_flag_ready_target(   pcontroller,
                                pio_handle->ptarget, 
                                TARGET_BUSY_ISCSI_IO_CREDITS);
    }

    if(pio_handle->is_untagged)
    {
        ap_flag_ready_target(   pcontroller,
                                pio_handle->ptarget, 
                                TARGET_BUSY_UNTAGGED_CMD);
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_start_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    PACIT_IO_HANDLE         pio_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PPDU_COMMAND            pdu;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    pdu = (PPDU_COMMAND)&pio_handle->pio->cmd_bhs.u1.iscsi_hdr;

    //pcxn = piscsi_target->pcxn_list[pio_handle->connection_index];

    pio_handle->cmd_sn = piscsi_target->cmd_sn;

    // BYTES 24 to 27
    ACIT_PUT_BE32(pdu->cmd_sn, pio_handle->cmd_sn);    

    // BYTES 16 to 19
    ACIT_PUT_BE32(pdu->task_tag, pio_handle->itt);

    /* Vivek: added change below to assign new CmdSn to piscsi_target->cmd_sn */
    piscsi_target->cmd_sn = SNA_ADD(piscsi_target->cmd_sn, 1);

#if ACIT_CFG_DEBUG_BUILD
    if(SNA_FORMER_IS_LESSER(piscsi_target->max_cmd_sn,
                            pio_handle->cmd_sn) ||
       SNA_FORMER_IS_LESSER(pio_handle->cmd_sn,
                            piscsi_target->exp_cmd_sn))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("cmdsn mismatch! max_cmd_sn %d cmd_cn %d exp_cmd_sn %d\n"),
            piscsi_target->max_cmd_sn, pio_handle->cmd_sn, piscsi_target->exp_cmd_sn));
    }
#endif

    //
    // if the cmd_sn is greater than maximum after an
    // increment, mark target busy so that the next I/O
    // is queued
    //
    if(SNA_FORMER_IS_GREATER(piscsi_target->cmd_sn,
                             piscsi_target->max_cmd_sn))
    {
        ap_flag_busy_target(pcontroller,
                            ptarget,
                            TARGET_BUSY_ISCSI_WINDOW);

        // also start a timer to handle the case where some targets
        // do not update cmd_sn either by NOP-In or during TM responses
        iscsi_start_cmd_window_timer(pcontroller, ptarget);
    }

    pio_handle->pconnection_handle->iscsi_cmd_pdus++;
    //iscsi_write_cmd_pdu (pcontroller, pio_handle);

    return ISCSI_STATUS_SUCCESS;
}

/*
 * This function returns ACIT_STATUS_SUCCESS if the I/O request can be
 * returned back to OSM. 
 */

ACIT_UEXACT32
iscsi_complete_io(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle,
    PI_SOL_CQE              psol,
    ACIT_UEXACT32           invalidated)
{
    pio_handle->pconnection_handle->iscsi_rsp_pdus++;

    // Update cmd_sns. If this is in CEV, then we already have the exp_cmd_sn
    // in little endian format
    iscsi_update_sn_by_offset (pcontroller, 
                               GET_ISCSI_TARGET_WS(pio_handle->ptarget), 
                               psol->i_cmd_wnd, 
                               psol->i_exp_cmd_sn);

    if(!invalidated)
    {
        PACIT_VOID              psense_buffer;
        PACIT_IO_REQUEST        pio_request;

        pio_request = pio_handle->pio->pio_request;

        //
        // Copy sense data if necessary.
        //
        if (ACIT_SCSI_STATUS_CHECK_CONDITION == psol->i_sts) {

            ACIT_UEXACT32 sense_length;

            psense_buffer = &pio_handle->pio->cmd_bhs.u2.sense_info[0];

            // The start of the payload contains the sense data length as a UINT16.
            sense_length = ACIT_GET_BE16 ((PACIT_UEXACT16)psense_buffer);
            OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("SCSI check condition. Length:0x%x\n", sense_length));            

            // Hardware may truncate sense data to fit in the header fragment.
            if (sense_length > ACIT_SENSE_INFO_SIZE) {
                sense_length = ACIT_SENSE_INFO_SIZE;
                OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Sense data truncated by header frag size to 0x%x bytes\n", sense_length));
            }

            // The OS buffer may truncate the sense data as well.
            if (pio_request->sense_buffer_length < sense_length) {
                sense_length = pio_request->sense_buffer_length;
                OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Sense data truncated to 0x%x bytes\n", sense_length));
            }

            // Copy the sense data into the io request.
            OSM_MEMCPY (pio_request->psense_buffer_virtual, 
                        (PACIT_UEXACT8)psense_buffer + sizeof (ACIT_UEXACT16), // skip the sense length u16
                        sense_length);

            // update OSM request sense_len 
            pio_request->sense_buffer_length = sense_length;
        }
        
        // For read command EP will provide the total bytes read till now. This byte count can be 
        // for this command or all the commands executed since last update.
        if (psol->s)
            pio_handle->pconnection_handle->total_bytes_read += psol->i_res_cnt;

        iscsi_update_io_counters(pcontroller, pio_handle);

        iscsi_update_io_request(pio_request,
                                (ACIT_UEXACT8)psol->i_resp,
                                (ACIT_UEXACT8)psol->i_sts,
                                (ACIT_UEXACT8)psol->i_flags,
                                psol->i_res_cnt);
        
        if (pio_handle->is_write == ACIT_TRUE)
        {
            pio_handle->pconnection_handle->total_bytes_xferred += pio_request->data_transfer_length;
            pio_handle->pconnection_handle->total_bytes_xferred += pio_request->overrun_length;
            pio_handle->pconnection_handle->total_bytes_xferred -= pio_request->underrun_length;
        }

        if(pio_request->io_status != ACIT_STATUS_SUCCESS)
        {
            ACIT_SET_TARGET_SUBSTATE(pio_handle->ptarget, TARGET_SUB_STATE_IO_ERROR);

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Response for IO in target 0x%p error, will trigger session recovery..\n"),
                                                    pio_handle->ptarget));

            pcontroller->eh_stats.io_status_errors++;
            pio_handle->ptarget->eh_stats.io_status_errors++;

            // FixMeBE2: Do not trigger session recovery, only retry the command??
            iscsi_recover_session(pcontroller, pio_handle->ptarget, ACIT_NULL);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_start_logout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_LOGOUT_HANDLE     plogout_handle)
{
    PISCSI_TARGET_WS        piscsi_target;

    piscsi_target = GET_ISCSI_TARGET_WS(plogout_handle->ptarget);
    plogout_handle->cmd_sn = piscsi_target->cmd_sn;
    iscsi_write_logout_pdu (pcontroller, plogout_handle);
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_alloc_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT8            connection_index;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    if(ptmf_handle->pio_handle)
    {
        PISCSI_CONNECTION pcxn;
        //
        // The abort IO has a requirement that the TMF MUST be
        // sent on the same connection that the IO is pending
        //

        ptmf_handle->cri = ptmf_handle->pio_handle->cri;
        pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, ptmf_handle->cri));

        if(pcxn->num_tmfs < pcontroller->params.num_tmfs_per_cxn &&
           !(pcxn->busy))
        {
            ACIT_UEXACT32   i = 0, cxn_busy = 0;
            pcxn->num_tmfs++;

            for(i=0; i<MAX_CXNS_PER_SESSION; i++)
            {
                if(piscsi_target->pcxn_list[i] == ACIT_NULL)
                    continue;

                if(piscsi_target->pcxn_list[i]->busy ||
                   piscsi_target->pcxn_list[i]->num_tmfs >= pcontroller->params.num_tmfs_per_cxn)
                {
                    cxn_busy++;
                }

                if(cxn_busy == piscsi_target->num_connections)
                {
                    ACIT_BIT_SET(ptarget->flags, TGT_FLAG_TMF_QUEUE_FULL);
                }
            }
                
            return ACIT_STATUS_SUCCESS;
        }

        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    connection_index = iscsi_assign_best_cxn_for_tmf(pcontroller,
                                                     piscsi_target);

    if(connection_index == 0xff)
    {
        ACIT_BIT_SET(ptarget->flags, TGT_FLAG_TMF_QUEUE_FULL);
        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    piscsi_target->pcxn_list[connection_index]->num_tmfs++;
    ptmf_handle->cri = (ACIT_UEXACT16)piscsi_target->pcxn_list[connection_index]->cri;
    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
iscsi_free_tmf_resources(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PISCSI_CONNECTION       pcxn;

    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, ptmf_handle->cri));
    OSM_ASSERT(pcxn);
    OSM_ASSERT(pcxn->num_tmfs);

    pcxn->num_tmfs--;

    if(ACIT_IS_BIT_SET(ptmf_handle->ptarget->flags, TGT_FLAG_TMF_QUEUE_FULL))
    {
        ACIT_BIT_CLR(ptmf_handle->ptarget->flags, TGT_FLAG_TMF_QUEUE_FULL);
    }
}

ACIT_UEXACT32
iscsi_start_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;

    piscsi_target = GET_ISCSI_TARGET_WS(ptmf_handle->ptarget);
    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, ptmf_handle->cri));

    ptmf_handle->cmd_sn = piscsi_target->cmd_sn;
    ptmf_handle->port_num = (ACIT_UEXACT16)piscsi_target->port_num;

    iscsi_write_tmf_pdu(pcontroller, ptmf_handle);

    return ISCSI_STATUS_SUCCESS;

}

ACIT_UEXACT32
iscsi_start_nopout(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    PPDU_NOP_IN             pdu,
    PACIT_NOP_HANDLE        poptional_handle)
{
    PISCSI_TARGET_WS        piscsi_target;

    piscsi_target = GET_ISCSI_TARGET_WS(pnop_handle->ptarget);

    /* Its incorrect for both pdu and poptional_handle to be a NULL */
    OSM_ASSERT((pdu != ACIT_NULL) || (poptional_handle != ACIT_NULL));

     /* If we're responding to a NOPIn, then ITT is always a 0xffffffff */
    if(poptional_handle)
    {
        OSM_MEMCPY(&pnop_handle->optional_lun, &poptional_handle->optional_lun, sizeof(ACIT_LUN));
        pnop_handle->issued_itt = pnop_handle->itt;
    }
    else
    {
        pnop_handle->issued_itt = 0xffffffff;
    }

    /* 
     * The TTT is either copied from the incoming PDU in case its valid or
     * is 0xffffffff
     */
    if(pdu)
    {
        OSM_MEMCPY(&pnop_handle->optional_lun, &pdu->lun, sizeof(ACIT_LUN));
        pnop_handle->ttt = pdu->ttt;
    }
    else
    {
        pnop_handle->ttt = 0xffffffff;
    }

    pnop_handle->cmd_sn = piscsi_target->cmd_sn;
    iscsi_write_nop_pdu (pcontroller, pnop_handle);

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
iscsi_write_cmd_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_IO_HANDLE         pio_handle)
{

}

ACIT_VOID
iscsi_write_tmf_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle)
{
    ACIT_UEXACT32           ref_cmd_sn = 0;
    ACIT_UEXACT32           ref_task_tag = ACIT_U32_MINUS_ONE;
    ACIT_LUN                lun = {{{0,0,0,0, 0,0,0,0}}};
    PACIT_LUN_WS            plun;
    PACIT_TMF_REQUEST       ptmf_request = ACIT_NULL;

    switch(ptmf_handle->tmf_function)
    {
    case TMF_LOGICAL_UNIT_RESET:
    
        /* We never issue a reset internally, so tmf_request is always valid */
        plun = ptmf_handle->plun;
        ptmf_request = plun->plun_reset_tmf_request;
        OSM_ASSERT(ptmf_request);
        lun.u.lun_complete = ptmf_request->lun.u.lun_complete;
        break;

    case TMF_ABORT_TASK:

        OSM_ASSERT(ptmf_handle->pio_handle);
        ref_task_tag = ptmf_handle->pio_handle->itt;
        ref_cmd_sn = ptmf_handle->pio_handle->cmd_sn;
        OSM_ASSERT(ptmf_handle->pio_handle);
        lun.u.lun_complete = ptmf_handle->pio_handle->lun.u.lun_complete;
        break;
    }
      
    pdu_build_tmf ((PPDU_TASK_MGMT)&ptmf_handle->noncmd_bhs.iscsi_hdr, 
                   (ACIT_UEXACT8)ptmf_handle->tmf_function,  // function
                   1,                          // immediate bit==1, implies cmd_sn not incremented
                   &lun,                       // lun
                   ptmf_handle->itt,           // itt
                   ref_task_tag,               // ref_task_tag - itt of aborted cmd
                   ptmf_handle->cmd_sn,
                   ref_cmd_sn);
}

ACIT_UEXACT32
iscsi_complete_tmf(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TMF_HANDLE        ptmf_handle,
    ACIT_UEXACT8            iresponse,           
    ACIT_UEXACT32           max_cmd_sn_offset,
    ACIT_UEXACT32           exp_cmd_sn)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           status;

    piscsi_target = GET_ISCSI_TARGET_WS(ptmf_handle->ptarget);

    iscsi_update_sn_by_offset(pcontroller, piscsi_target, max_cmd_sn_offset, exp_cmd_sn);

    switch(iresponse)
    {
    case TMR_FUNCTION_COMPLETE:

        status = ACIT_STATUS_SUCCESS;
        break;

    case TMR_TASK_NOT_IN_SET:

        status = ACIT_TMF_STATUS_INVALID_TASK;
        break;

    case TMR_TASK_MGMT_FUNCTION_UNSUPPORTED:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Error in TMF response - function unsupported for TMF %d in target 0x%p\n", 
            ptmf_handle->tmf_function, ptmf_handle->ptarget));
        status = ACIT_TMF_STATUS_NOT_SUPPORTED;
        break;

    case TMR_FUNCTION_REJECTED:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Error in TMF response - function rejected for TMF %d in target 0x%p\n", 
            ptmf_handle->tmf_function, ptmf_handle->ptarget));
        status = ACIT_TMF_STATUS_FUNCTION_REJECTED;
        break;


    case TMR_TASK_STILL_ALLEGIANT:
    case TMR_FUNCTION_AUTH_FAILED:
    case TMR_TASK_FAILOVER_UNSUPPORTED:
    default:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Error in TMF response - %d for TMF %d in target 0x%p\n", 
            iresponse, ptmf_handle->tmf_function, ptmf_handle->ptarget));

        status = ACIT_TMF_STATUS_NOT_SUPPORTED;
        break;
    }

    return status;
}


PISCSI_CONNECTION
iscsi_alloc_cxn (
    PACIT_CONTROLLER_WS     pcontroller)
{
    PISCSI_CONTROLLER_WS    piscsi_controller = ACIT_NULL;
    PISCSI_CONNECTION       pcxn = ACIT_NULL;
    ACIT_UEXACT32           index = 0;
    
    piscsi_controller = GET_ISCSI_CONTROLLER_WS (pcontroller);

    /*
     * Walk the connection table and get a free connection
     */
#ifdef ISCSI_ALLOC_CXN_IN_REVERSE
    for(index = (pcontroller->params.num_cxn - 1); index >= 0; index++)
#else
    for(index = 0; index < pcontroller->params.num_cxn; index++)
#endif
    {
        if(piscsi_controller->connection_table[index] == ACIT_NULL)
        {
            pcxn = iscsi_get_cxn_from_pool(pcontroller, index);

            OSM_MEMSET (pcxn, 0, sizeof (*pcxn));

#ifndef BE_EMULATION
            pcxn->cri = index;
            pcxn->cid = CID_FROM_CRI(pcontroller, index);
            OSM_ASSERT (pcontroller->params.num_cmd_per_cxn <= BE_MAX_CMD_PER_CXN);

            piscsi_controller->connection_table[index] = pcxn;

#else
            pcxn->cid = emul_get_cid(pcontroller, pcontroller->poffload_params);
            pcxn->cri = pcxn->cid;  // this is needed since the HWI is expecting
                                    // the same CID at the CRI

            OSM_ASSERT (pcontroller->params.num_cmd_per_cxn <= BE_MAX_CMD_PER_CXN);

            piscsi_controller->connection_table[index] = pcxn;

#endif
            break;
        }
    }

    return pcxn;
}

ACIT_VOID
iscsi_free_cxn (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    
#if defined (BE_EMULATION)// || (FPGA_ARM)
    ACIT_UEXACT32           index;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    for(index = 0; index < pcontroller->params.num_cxn; index++)
    {
        if(piscsi_controller->connection_table[index] == pcxn)
        {
            piscsi_controller->connection_table[index] = ACIT_NULL;
            break;
        }
    }
#else
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    OSM_ASSERT(piscsi_controller->connection_table[pcxn->cri]);

    piscsi_controller->connection_table[pcxn->cri] = ACIT_NULL;
#endif
    // Clear the object
    OSM_MEMSET (pcxn, 0, sizeof (*pcxn));

}

ACIT_STATUS
iscsi_process_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT16           cid,
    PPDU_BASE               ppdu,
    ACIT_UEXACT32           pdu_len,
    PVOID                   pbuffer,
    ACIT_UEXACT32           buf_len)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PACIT_TARGET_WS         ptarget;
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;
    ACIT_STATUS             status = ACIT_STATUS_SUCCESS;

    //
    // async pdus include 
    // a. unsolicited NOP-In (target initiated NOP-In)
    // b. Async Messages
    // c. Reject PDU
    // These headers arrive unprocessed by the EP firmware and iSCSI layer
    // process them
    
    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);
    pcxn = iscsi_get_cxn(pcontroller, cid);
    piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;

    if(!piscsi_target)
    {
        // if we see this, then the session is being uploaded and therefore we
        // do not want to process the PDU now
        OSM_ASSERT(0);
        return ACIT_STATUS_ERROR;
    }

    ptarget = piscsi_target->pacit_ws;

    // Update last async received at
    ACIT_UPDATE_TARGET_LAST_ASYNC_TS(pcontroller, ptarget);

    if(!ACIT_IS_TARGET_INITIALIZED(ptarget))
    {
        //
        // If target is not initialized, then hold on to the buffer. This is a 
        // very rare occurance, can happen only between the time we issued a 
        // offload IOCTL and the offload IOCTL completing back to the driver
        //
        pcxn->async_pending = 1;
        piscsi_controller->async_pending++;

        return ACIT_STATUS_INSUFFICIENT_RESOURCE;
    }

    switch(ppdu->opcode)
    {
    case ITOC_NOP_IN:

        iscsi_process_nop_in(pcontroller, 
                             ACIT_NULL, 
                             pcxn, 
                             (PPDU_NOP_IN)ppdu, 
                             pdu_len);
        break;

    case ITOC_ASYNC_EVENT:

        iscsi_process_async_event(pcontroller, 
                                  pcxn, 
                                  (PPDU_ASYNC_MSG)ppdu, 
                                  pdu_len,
                                  (PPDU_BASE)pbuffer,
                                  buf_len);
        break;

    case ITOC_REJECT:

        OSM_ASSERT(pbuffer);
        OSM_ASSERT(buf_len == 48);

        iscsi_process_reject(pcontroller, 
                             pcxn, 
                             (PPDUH_REJECT)ppdu, 
                             pdu_len,
                             (PPDU_BASE)pbuffer,
                             buf_len);
        break;

    default:
        
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Unrecognized opcode 0x%x in async msg for target 0x%p, triggering session recovery\n",	
                                           ppdu->opcode, 
                                           ptarget));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_IO_ERROR);

        pcontroller->eh_stats.unknown_defpdu++;
        ptarget->eh_stats.unknown_defpdu++;

        iscsi_recover_session(pcontroller, 
                              ptarget, 
                              ACIT_NULL);

        //
        // Feb 03 - 2012
        // Note! We can free the DPDU entry since it does not interfere with
        // the session cleanup. Not doing otherwise will lead to a DPDU entry
        // leak. So comment out status = ACIT_STATUS_ERROR
        //

        //
        // FixmeBE2: Warning!! Assumes default buffer cleanup is done as part of
        // session recovery
        //status = ACIT_STATUS_ERROR;

        break;
    }

    return status;
}

ACIT_UEXACT32
iscsi_check_async_pdu (
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri,
    PPDU_BASE               ppdu,
    PACIT_UEXACT32          pbytes_needed)
{
    PISCSI_CONNECTION       pcxn;

    //
    // This function checks the async PDU header for correct digest value 
    // if header digest is enabled. It also returns the number of bytes
    // of data to expect for the async PDU
    //

    OSM_ASSERT(pbytes_needed);

    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, cri));
    OSM_ASSERT(pcxn);

    *pbytes_needed = (ppdu->data_len_hi << 16) |
                      ACIT_GET_BE16(&ppdu->data_len_lo);

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
iscsi_process_nop_in (
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_NOP_HANDLE        pnop_handle,
    PISCSI_CONNECTION       pcxn,
    PPDU_NOP_IN             pdu,
    ACIT_UEXACT32           pdu_len)
{
    PISCSI_TARGET_WS        piscsi_target;    
    ACIT_STATUS             status;

    if(pdu_len < ISCSI_PDU_HEADER_SIZE)
    {
        OSM_ASSERT(0);
        return ACIT_STATUS_PENDING;
    }

    piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;

    // Fix endianess
    pdu_nop_in_swap_bytes (pdu);

    if (pnop_handle) 
    {
        // If pnop_handle exists, this is a reply to our nop.

        //
        // Spec, "When a target receives the NOP-Out with a valid Initiator Task Tag
        // (not the reserved value 0xffffffff), it MUST respond with a NOP-In
        // with the same Initiator Task Tag that was provided in the NOP-Out
        // request."
        //
        if (pdu->itt == PDU_TAG_RESERVED) 
        {
            // Avoid sending a response by setting ITT to zero.  
            // Incorrectly responding could lead to an infinite loop of NOPs.
            OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Bad itt in nop-in response: 0x%x\n", pdu->itt));
            pdu->itt = 0; 
        }

        //
        // Spec, "If the target is responding to a NOP-Out, this [ttt] is set to the
        // reserved value 0xffffffff."
        //
        if (pdu->ttt != PDU_TAG_RESERVED) 
        {
            OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Bad ttt in nop-in response: 0x%x\n", pdu->ttt));
        }

        ap_free_nop_resources(pcontroller, pnop_handle);

        // signal not to send reply later.
        pnop_handle = NULL; 

        // Update maxCmdSn 
        iscsi_update_sn (pcontroller, piscsi_target, pdu->max_cmd_sn, pdu->exp_cmd_sn);

        return ACIT_STATUS_SUCCESS;
    } 
    else
    {
        // Update maxCmdSn 
        iscsi_update_sn (pcontroller, piscsi_target, pdu->max_cmd_sn, pdu->exp_cmd_sn);

        if (pdu->itt == PDU_TAG_RESERVED && pdu->ttt != PDU_TAG_RESERVED) 
        {

            // Target initiated ping, reply required.
            OSM_ASSERT (pnop_handle == NULL);

            // Spec, "Otherwise, when a target sends a NOP-In that is not a response to a 
            // Nop-Out received from the initiator, the Initiator Task Tag MUST be set to 
            // 0xffffffff and the Data Segment MUST NOT contain any data (DataSegmentLength 
            // MUST be 0).
            if ((pdu->data_len_hi | pdu->data_len_lo ) > 0)
            {
                OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, 
                                ("Protocol Error: nop in received with dataSegLength == 0x%x.\n", 
                                (pdu->data_len_hi | pdu->data_len_lo )));
            }

            //
            // Target will not respond again.  This command completes with TCP ack.  Therefore,
            // we need to use the "driverMessage" wrb.
            //
            status = ap_send_nopout(pcontroller, 
                                    piscsi_target->pacit_ws, 
                                    pcxn->cri, 
                                    pdu,
                                    ACIT_NULL);

            // Note that we do not look at if nop out was actually sent or not. If
            // there were not enough resources then nopout send could fail. But we
            // still return ACIT_STATUS_SUCCESS since we are not interested in
            // retrying the NOPOut. It should be ok since it should not happen
            // normally
        }
        else
        {
            OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Ignoring NOPIn from target 0x%p itt 0x%x ttt 0x%x\n", 
                piscsi_target->pacit_ws, pdu->itt, pdu->ttt));
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
iscsi_process_async_event (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn,
    PPDU_ASYNC_MSG          pasync_msg,
    ACIT_UEXACT32           async_msg_len,
    PVOID                   pbuffer,
    ACIT_UEXACT32           buf_len)
{
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_TARGET_WS         ptarget;
    
    piscsi_target = pcxn->piscsi_target_ws;
    OSM_ASSERT(piscsi_target);
    ptarget = piscsi_target->pacit_ws;
    
    // NOTE! The pasync_msg and pbuffer are in network byte order!!

    // Fix endianess
    pdu_async_msg_swap_bytes(pasync_msg);

    switch(pasync_msg->async_event)
    {
    case ASYNC_SCSI_EVENT:

        pcontroller->stats.async_scsi_recvd++;

        if(ACIT_IS_BIT_SET(piscsi_target->flags, ISCSI_DATA_DIGEST_ERR))
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Err flag set, ignoring the async scsi event pdu because of digest error!!\n"));
            // One DPDU buffer has to be discarded because of the above flag. 
            // We can clear the flag now..
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Clearing Digest Error in PDU bit for cid %x...\n", pcxn->cid));
            ACIT_BIT_CLR(piscsi_target->flags, ISCSI_DATA_DIGEST_ERR);
        }
        else
        {
            OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR, ("ASYNC_SCSI_EVENT, parsing sense data..\n"));
            ap_process_scsi_event(pcontroller, 
                                  ptarget, 
                                  (PACIT_SCSI_SENSE_DATA)((PACIT_UEXACT8)pbuffer + 0x2), 
                                  ACIT_GET_BE16 ((PACIT_UEXACT16)pbuffer));
        }

        break;

    case ASYNC_LOGOUT_REQUEST:

        //
        //PRB - TODO - spec says we must send logout on the same
        //connection that received the logout request
        //

        // Save connection index where logout should happen
        //piscsi_target->logout_cid = pcxn->cid;

        //
        // when target asks for a logout, we will logout and try
        // to login again. Since it is not a logout initiated by
        // the driver, we should not drop the session. The firmware
        // will then get a chance to login again the next time target
        // is back online
        //

        OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR,("target requested async logout\n"));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ASYNC_MSG_LOGOUT);

        pcontroller->eh_stats.async_logouts++;
        ptarget->eh_stats.async_logouts++;

        iscsi_recover_session(pcontroller, 
                              ptarget, 
                              ACIT_NULL);

        break;

    case ASYNC_PARAM_NEGOTIATION_REQUEST:

        OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR,("target requested parameter renego\n"));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ASYNC_MSG_LOGOUT);

        pcontroller->eh_stats.async_param_renego++;
        ptarget->eh_stats.async_param_renego++;

        iscsi_recover_session(pcontroller, 
                              ptarget, 
                              ACIT_NULL);

        break;

    case ASYNC_DROP_CONNECTION:
    case ASYNC_DROP_ALL_CONNECTIONS:

        //
        // The ARM fimware is resposible for the login negotiation. So
        // trigger a session recovery 
        //
        OSM_DEBUG_PRINT (ACIT_DBGMSG_ERROR,("target dropping connection(s)\n"));

        ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_ASYNC_MSG_ERROR);

        pcontroller->eh_stats.async_reopens++;
        ptarget->eh_stats.async_reopens++;

        iscsi_recover_session(pcontroller, 
                              ptarget, 
                              ACIT_NULL);
        break;

    case ASYNC_VENDOR_SPECIFIC_ISCSI_EVENT:
    default:
        OSM_DEBUG_PRINT (ACIT_DBGMSG_WARNING, ("Unknown/vendor-specific event 0x%x, ignoring..\n", 
                                pasync_msg->async_event));
        break;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_STATUS
iscsi_process_reject (
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_CONNECTION       pcxn,
    PPDUH_REJECT            preject_msg,
    ACIT_UEXACT32           reject_msg_len,
    PPDU_BASE               ppdu_hdr,
    ACIT_UEXACT32           pdu_hdr_len)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           action = REJECT_ACTION_INVALIDATE_CMD;
    PACIT_IO_HANDLE         pio_handle = ACIT_NULL;

    piscsi_target = pcxn->piscsi_target_ws;
    OSM_ASSERT(piscsi_target);
    pcontroller->eh_stats.rejects++;
    piscsi_target->pacit_ws->eh_stats.rejects++;

    // NOTE! The preject_msg and ppdu_hdr are in network byte order!!

    // default action is to invalidate the command if the rejected PDU
    // is a command pdu and ignore the reject if the rejected PDU is
    // a non-command PDU.
    OSM_ASSERT(ppdu_hdr);

    if(ACIT_IS_BIT_SET(piscsi_target->flags, ISCSI_DATA_DIGEST_ERR))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Err flag set, ignoring the rejected pdu because of digest error!!\n"));
        // Normal command timeout will take care of recovering the command.
        action = REJECT_ACTION_DO_NOTHING;

        // One DPDU buffer has to be discarded because of the above flag. 
        // We can clear the flag now..
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Clearing Digest Error in PDU bit for cid %x...\n", pcxn->cid));
        ACIT_BIT_CLR(piscsi_target->flags, ISCSI_DATA_DIGEST_ERR);
    }
    else if(ppdu_hdr->opcode == IIOC_SCSI_CMD)
    {
        // find the io handle from the command BHS 
        pio_handle = hwi_io_handle_from_itt(pcontroller, 
                                            pcxn->cri, 
                                            ACIT_GET_BE32(&ppdu_hdr->itt));

        if(pio_handle)
        {
            action = REJECT_ACTION_INVALIDATE_CMD;
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Invalid ITT in rejected PDU, ignoring!! 0x%x\n",
                                    ppdu_hdr->itt));
            action = REJECT_ACTION_DO_NOTHING;
        }
    }
    else
    {
        // non command PDUs are ignored. These may include TMFs and logouts.
        // Ignoring reject msg for these is ok since there are timers
        // running on them which will expire eventually
        action = REJECT_ACTION_DO_NOTHING;
    }

    // specific reject reason may over-ride the above default behaviour

    OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("reject - %s - reason %d on cid 0x%x\n",
                                        rrc_string[preject_msg->reason],
                                        preject_msg->reason, 
                                        pcxn->cid));

    switch(preject_msg->reason)
    {
    case RRC_RESERVED:
    case RRC_PROTOCOL_ERROR:
    case RRC_INVALID_DATA_ACK:
    case RRC_INVALID_PDU_FIELD:
    case RRC_LONG_OPERATION_REJECT:
    case RRC_DATA_DIGEST_ERROR:

        break;

    case RRC_WAITING_FOR_LOGOUT:
    case RRC_TASK_IN_PROGRESS:
    case RRC_COMMAND_UNSUPPORTED:
    case RRC_REJECT_IMMD_COMMAND:

        action = REJECT_ACTION_DO_NOTHING;

        break;

    case RRC_SNACK_REJECT:
    case RRC_NEGOTIATION_RESET:
    default:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Unexpected/unknown reject 0x%x\n", preject_msg->reason));
        action = REJECT_ACTION_INVALIDATE_CXN;
        break;
    }

    switch(action) {

    case REJECT_ACTION_DO_NOTHING:
        break;

    case REJECT_ACTION_INVALIDATE_CMD:
        OSM_ASSERT(pio_handle);
        ap_complete_io_error(pcontroller, pio_handle, FLAG_IOH_REJECTED_BY_TARGET);
        break;

    case REJECT_ACTION_INVALIDATE_CXN:

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Session recovery due to reject on target 0x%p\n", 
                                piscsi_target->pacit_ws));

        ACIT_SET_TARGET_SUBSTATE(piscsi_target->pacit_ws, 
                                 TARGET_SUB_STATE_REJECT_MSG);

        iscsi_recover_session(pcontroller,
                              piscsi_target->pacit_ws,
                              ACIT_NULL);
        break;
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_rst_send_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid)
{
    PISCSI_CONNECTION       pcxn = iscsi_get_cxn (pcontroller, cid);
    PISCSI_TARGET_WS        piscsi_target = ACIT_NULL;

    //
    // This API is called whenever the lower layers detect error in transport
    // layer. Right now we will do a session recovery, when we support lower
    // error handling than ERL0, we can do connection level recovery
    //
    if(pcxn)
    {
        piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;
        OSM_ASSERT(piscsi_target);
    }

    if(piscsi_target)
    {
        // Ignore the notification if we're in the middle of logout or recovery
        if(!(ACIT_IS_TARGET_BUSY(piscsi_target->pacit_ws, TARGET_BUSY_SESSION_LOGOUT)) &&
            (!(ACIT_IS_TARGET_BUSY(piscsi_target->pacit_ws, TARGET_BUSY_SESSION_RECOVERY))) &&
            (ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS)))
        {
            //
            // Ignore the notification if we've already in the process of resetting 
            // the connection
            //
            if(ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_TCP_RESET_PENDING) &&
            ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_RESET))
            {
                ACIT_BIT_SET(pcxn->status, ISCSI_CONN_CQ_ERROR);
                piscsi_target->num_connections_reset_needed++;

                iscsi_recover_session(pcontroller, piscsi_target->pacit_ws, ACIT_NULL);
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Ignoring RST send on target 0x%p already in cleanup...\n", piscsi_target->pacit_ws));
        }
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Ignoring RST send on non-existant target...\n"));
    }
    
    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_rst_recv_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;

    //
    // This API is called whenever the lower layers detect a TCP reset
    // being received on the connection.
    //

    pcxn = iscsi_get_cxn (pcontroller, cid);

    OSM_ASSERT(pcxn);
    piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;
    OSM_ASSERT(piscsi_target);

    if(piscsi_target)
    {
        PACIT_TARGET_WS     ptarget;

        ptarget = piscsi_target->pacit_ws;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("connection status 0x%x\n", pcxn->status ));

        // Ignore the notification if we're in the middle of logout or recovery
        if(!(ACIT_IS_TARGET_BUSY(piscsi_target->pacit_ws, TARGET_BUSY_SESSION_LOGOUT)) &&
            (!(ACIT_IS_TARGET_BUSY(piscsi_target->pacit_ws, TARGET_BUSY_SESSION_RECOVERY))) &&
            (ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS)))
        {
            //
            // close the connection with ARM firmware. Since we can receive
            // RST multiple times, ignore it if already handled
            //
            if(ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATE_PENDING) &&
                ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATED))
            {
                ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_RESET);
                piscsi_target->num_connections_reset++;

                iscsi_recover_session(pcontroller, ptarget, ACIT_NULL);
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,(("Received RST on cxn 0x%p which is in invalidation. Ignoring...\n"),
                                                        pcxn));
                OSM_ASSERT(0);
            }
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Ignoring RST recv on target 0x%p already in cleanup...\n", piscsi_target->pacit_ws));
        }
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Ignoring RST recv on non-existant target... cid 0x%x\n", cid));
    }

    return ISCSI_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_invalidate_all_connections(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           force_rst)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT16           i, inv_reset;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);
    OSM_ASSERT(piscsi_target);

    for(i=0; i<MAX_CXNS_PER_SESSION; i++)
    {
        if(piscsi_target->pcxn_list[i])
        {
            if(ACIT_IS_BIT_SET(piscsi_target->pcxn_list[i]->status, 
                            ISCSI_CONN_STATUS_RESET) && (!force_rst))
            {
                inv_reset = ACIT_FALSE;
            }
            else
            {
                inv_reset = ACIT_TRUE;
            }

            iscsi_invalidate_connection(pcontroller, 
                                        piscsi_target, 
                                        piscsi_target->pcxn_list[i], 
                                        inv_reset);
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_invalidate_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    PISCSI_TARGET_WS        piscsi_target,
    PISCSI_CONNECTION       pcxn,
    ACIT_UEXACT16           issue_reset)
{
    ACIT_UEXACT16           savecfg_flag = 0;
    MGMT_STATUS             mgmt_status;

    // already in invalidation?
    OSM_ASSERT(ACIT_IS_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATE_PENDING));
    OSM_ASSERT(pcxn->piscsi_target_ws);
    OSM_ASSERT(piscsi_target->num_connections);
    ACIT_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_VALID);

    savecfg_flag = iscsi_get_cfg_flag(pcontroller, piscsi_target->pacit_ws);

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, 
        ("target 0x%p sh %d cid 0x%x issue_reset %d\n", piscsi_target->pacit_ws,
                                                        piscsi_target->session_handle, 
                                                        pcxn->cid, 
                                                        issue_reset));

    mgmt_status = mgmt_invalidate_connection(pcontroller,
                                             piscsi_target->pacit_ws,
                                             pcxn->cid,
                                             issue_reset,
                                             savecfg_flag);

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        pcontroller->eh_stats.cxn_invalidations++;
        ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_INVALIDATE_PENDING);
        return ACIT_STATUS_SUCCESS;
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Deferring connection invalidate on target 0x%p cid 0x%x\n",
            piscsi_target->pacit_ws, pcxn->cid));

        if(issue_reset)
            ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_NEED_RESET);
        else
            ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_NEED_INVALIDATION);

        ap_set_target_event (pcontroller, 
                             piscsi_target->pacit_ws, 
                             EVENT_TGT_CXN_EVENT_PENDING);
    }

    return ACIT_STATUS_PENDING;
}

ACIT_UEXACT32
iscsi_connection_invalidate_notify(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT16           cid)
{
    PISCSI_CONNECTION       pcxn = ACIT_NULL;
    ACIT_UEXACT32           pid;

    //
    // For every invalidation response, check the CQ again to make
    // sure there are no responses that have come in 
    // This is to close the window of time when
    // CQs could coming in just after we looked at the CQ but before
    // we look into the MCC CQ
    //
    if(ACIT_IS_BIT_CLR(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS) &&
        ((pcontroller->state.init_type == ACIT_INIT_TYPE_POWER_MGMT) ||
        (pcontroller->state.init_type == ACIT_INIT_TYPE_NORMAL)))
    {
        for(pid = 0; pid < pcontroller->params.num_cq; pid++)
        {
            hwi_process_cq(pcontroller, pid); 
        }
    }

    pcxn = iscsi_get_cxn(pcontroller, cid);
    OSM_ASSERT(pcxn);

    if(ACIT_IS_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_TCP_RESET_PENDING))
    {
        ACIT_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_TCP_RESET_PENDING);
        ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_RESET);
        pcxn->piscsi_target_ws->num_connections_reset++;
    }

    ACIT_BIT_CLR(pcxn->status, ISCSI_CONN_STATUS_INVALIDATE_PENDING);
    ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_INVALIDATED);
    pcxn->piscsi_target_ws->num_connections_invalidated ++;

    iscsi_upload_connection(pcontroller, pcxn->piscsi_target_ws, pcxn->cri);

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_upload_connection(
    PACIT_CONTROLLER_WS     pcontroller, 
    PISCSI_TARGET_WS        piscsi_target, 
    ACIT_UEXACT32           cri)
{
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           upload_flag;
    MGMT_STATUS             mgmt_status;

    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, cri));

    //
    // Depending on the state of the connection, the upload is going
    // to be either graceful or non-abortive upload.
    // If we've earlier issued a reset, upload is non-abortive
    // Else upload is graceful.
    // The 3rd case, abortive upload would send the RST on the 
    // wire while uploading, but iSCSI drivers never use it
    //

    if(ACIT_IS_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_RESET))
    {
        upload_flag = CONNECTION_UPLOAD_ABORT;
    }
    else
    {
        upload_flag = CONNECTION_UPLOAD_GRACEFUL;
    }

    OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, 
                    ("Target 0x%p sh %d cid 0x%x upload flag %d\n", 
                    piscsi_target->pacit_ws, piscsi_target->session_handle, pcxn->cid, upload_flag));

    mgmt_status = mgmt_upload_connection(pcontroller,
                                         piscsi_target->pacit_ws,
                                         pcxn->cid,
                                         upload_flag);  // issue_reset is always false.

    if(mgmt_status == MGMT_STATUS_SUCCESS)
    {
        pcontroller->eh_stats.cxn_uploads++;
        return ACIT_STATUS_SUCCESS;
    }
    else
    {
        OSM_ASSERT(mgmt_status == MGMT_STATUS_IOCTLHANDLE_ALLOC_FAILED);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Deferring connection upload on target 0x%p cid 0x%x\n",
            piscsi_target->pacit_ws, pcxn->cid));

        ACIT_BIT_SET(pcxn->status, ISCSI_CONN_STATUS_NEED_UPLOAD);

        ap_set_target_event (pcontroller, 
                             piscsi_target->pacit_ws, 
                             EVENT_TGT_CXN_EVENT_PENDING);
    }

    return ACIT_STATUS_PENDING;
}

ACIT_UEXACT32
iscsi_connection_upload_notify(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid)
{
    PACIT_TARGET_WS         ptarget;
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn = ACIT_NULL;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);


    //ACIT_BIT_SET(ptarget->flags, FLAG_TGT_UPLOAD_COMPLETE);

    pcxn = iscsi_get_cxn(pcontroller, cid);
    OSM_ASSERT(pcxn);

    piscsi_target = pcxn->piscsi_target_ws;
    piscsi_target->num_connections_uploaded++;

    // Set the status to ISCSI_CONN_STATUS_UPLOADED, since at this time connection
    // state should be uploaded. If the connection needs to be reopened then later on
    // connection will be initialized an will become valid.
    pcxn->status = ISCSI_CONN_STATUS_UPLOADED;

    ptarget = piscsi_target->pacit_ws;

    if(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_LOGOUT) ||
      (ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS)))
    {
        if(piscsi_target->num_connections_uploaded == piscsi_target->num_connections)
        {
            ptarget->initialized = 0;

            ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_CONX_UPLOAD_COMPLETE);
            //
            // for now, lets assume there's no need for cmd free IOCTL, so directly free
            // the handles and call iscsi upload
            //
            ap_free_pending_handles(pcontroller, ptarget, ACIT_NULL, ACIT_U32_MINUS_ONE);
            iscsi_delete_session_info(pcontroller, ptarget);
        }
        else
        {
            OSM_ASSERT(0);
        }
    }
    else if(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_REMOVE_CONNECTION))
    {
        ACIT_UEXACT16 i,j = 0;

        pcxn->status = ISCSI_CONN_STATUS_NULL;
        pcxn->piscsi_target_ws = ACIT_NULL;

        /* Remove the connection from target's connection array */
        for(i=0; i<MAX_CXNS_PER_SESSION; i++)
        {
            if(piscsi_target->pcxn_list[i] &&
                piscsi_target->pcxn_list[i] == pcxn)
            {
                piscsi_target->pcxn_list[i] = ACIT_NULL;
                break;
            }
        }

        /* Rearrange the array to avoid the holes */
        for(j=i+1; j<MAX_CXNS_PER_SESSION; j++)
        {
            if(piscsi_target->pcxn_list[j])
            {
                piscsi_target->pcxn_list[i] = piscsi_target->pcxn_list[j];
                piscsi_target->pcxn_list[j] = ACIT_NULL;
                i = j;
            }
        }
        
        iscsi_free_cxn (pcontroller, pcxn);
        piscsi_target->num_connections--; 
        piscsi_target->num_connections_uploaded = 0;
        /* this flag got set in the iscsi_close_connection */
        ap_flag_ready_target(pcontroller, ptarget, TARGET_BUSY_REMOVE_CONNECTION);

        piscsi_target->last_cxn = (piscsi_target->num_connections) - 1;

        if (piscsi_target->mgmt_ioctl_request)
            mgmt_return_ioctl_request(pcontroller, 
                                    piscsi_target->mgmt_ioctl_request,
                                    ACIT_STATUS_SUCCESS);
            
    }
    else if(ACIT_IS_TARGET_BUSY(ptarget, TARGET_BUSY_SESSION_RECOVERY))
    {
        if(piscsi_target->num_connections_uploaded == piscsi_target->num_connections)
        {
            ptarget->initialized = 0;

            ACIT_SET_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_CONX_UPLOAD_COMPLETE);
            //
            // for now, lets assume there's no need for cmd free IOCTL, so directly free
            // the handles and call iscsi upload
            //
            ap_free_pending_handles(pcontroller, ptarget, ACIT_NULL, ACIT_U32_MINUS_ONE);

            // target is not deleted yet, so do not free IOs..
            ap_free_queued_ws_resources(pcontroller, ptarget, ACIT_FALSE);

            if(ACIT_IS_BIT_SET(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_SESSION_RECOVERY))
            {
                OSM_ASSERT(piscsi_controller->num_uploads_pending);
                piscsi_controller->num_uploads_pending--;

                if(piscsi_controller->num_uploads_pending == 0)
                {
                    if(piscsi_controller->mgmt_ioctl_request)
                    {
                        mgmt_all_targets_close_notify(pcontroller, 
                                                      piscsi_controller->mgmt_ioctl_request);

                        piscsi_controller->mgmt_ioctl_request = ACIT_NULL;
                    }
                    else
                    {
                        mgmt_reopen_sessions(pcontroller, 
                                             IOCTL_ISCSI_REOPEN_ALL_SESSIONS, 
                                             ACIT_U32_MINUS_ONE,
                                             0, // domain 0
                                             ACIT_FALSE);
                    }

                    ACIT_BIT_CLR(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_SESSION_RECOVERY);
                }
            }
            else
            {
                mgmt_target_upload_complete(pcontroller, 
                                            ptarget, 
                                            piscsi_target->session_handle, 
                                            piscsi_target->mgmt_ioctl_request);

                piscsi_target->mgmt_ioctl_request = ACIT_NULL;
            }
        }
        else
        {
            OSM_ASSERT(0);
        }
    }
    else
    {
        OSM_ASSERT(0);
        // why are we here??
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_VOID
iscsi_update_statistics(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid,
    ACIT_UEXACT32           code)
{
    PISCSI_TARGET_WS        piscsi_target = ACIT_NULL;
    PISCSI_CONNECTION       pcxn;
    PACIT_TARGET_WS         ptarget;

    pcxn = iscsi_get_cxn (pcontroller, cid);

    if(pcxn) 
    {
        piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;
        OSM_ASSERT(piscsi_target);
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("Cxn notification 0x%x on CID 0x%x thats freed!\n", code, cid));
    }

    if(piscsi_target)
    {
        ptarget = piscsi_target->pacit_ws;
        pcontroller->eh_stats.ssn_failure_count++;
        
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE, ("Update the statistics for the connection, target and controller.\n"));

        switch(code)
        {
        case SOL_CMD_COMPLETE:
        case DRIVERMSG_NOTIFY:
        case UNSOL_HDR_NOTIFY:
        case UNSOL_DATA_NOTIFY:
        case CMD_INVALIDATED_NOTIFY:
        case CXN_INVALIDATE_NOTIFY:

            // success or in response to an earlier invalidation. We get responses
            // from ARM fw for invalidations, so ignore these

            break;

        case CXN_INVALIDATE_INDEX_NOTIFY:

            // FixmeBE2: Not implemented!!
            OSM_ASSERT(0);
            break;

        case SOL_CMD_KILLED_DATA_DIGEST_ERR:
        case CXN_KILLED_HDR_DIGEST_ERR:
        case UNSOL_DATA_DIGEST_ERROR_NOTIFY:

            pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_DIGEST_ERRORS;
            pcontroller->eh_stats.ssn_digest_errors++;
            ptarget->eh_stats.ssn_digest_errors++;
            pcxn->digest_errors++;
            
            break;

        case CMD_KILLED_INVALID_STATSN_RCVD:
        case CMD_KILLED_INVALID_R2T_RCVD:
        case CMD_CXN_KILLED_LUN_INVALID:
        case CMD_CXN_KILLED_ICD_INVALID:
        case CMD_CXN_KILLED_ITT_INVALID:
        case CMD_CXN_KILLED_SEQ_OUTOFORDER:
        case CMD_CXN_KILLED_INVALID_DATASN_RCVD:

        case CXN_KILLED_PDU_SIZE_EXCEEDS_DSL:
        case CXN_KILLED_BURST_LEN_MISMATCH:
        case CXN_KILLED_AHS_RCVD:
        case CXN_KILLED_UNKNOWN_HDR:
        case CXN_KILLED_STALE_ITT_TTT_RCVD:
        case CXN_KILLED_INVALID_ITT_TTT_RCVD:
        case CXN_KILLED_BAD_UNSOL_PDU_RCVD:
        case CXN_KILLED_BAD_WRB_INDEX_ERROR:
        case CXN_KILLED_OVER_RUN_RESIDUAL:
        case CXN_KILLED_UNDER_RUN_RESIDUAL:
        case CXN_KILLED_CMND_DATA_NOT_ON_SAME_CONN:

            pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_FORMAT_ERRORS;
            pcontroller->eh_stats.ssn_format_errors++;
            ptarget->eh_stats.ssn_format_errors++;
            pcxn->format_errors++;

            break;

        case CXN_KILLED_RST_SENT:        
        case CXN_KILLED_RST_RCVD:
        case CXN_KILLED_TCP_PROT_ERR:
        case CXN_KILLED_FIN_RCVD:

			// Update statistics counter only if it is
            // a. Not a case of TMF time out
            // b. Not a case of logout in progress (Bug27675)
			if(ACIT_IS_TARGET_SUBSTATE(ptarget, TARGET_SUB_STATE_TMF_TIMEOUT) ||
               ACIT_IS_BIT_SET(ptarget->busy, TARGET_BUSY_SESSION_LOGOUT))
			{
				pcontroller->eh_stats.ssn_failure_count--;
            }
            else
            {
				pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_CXNTIMEOUT_ERRORS;
				pcontroller->eh_stats.ssn_cxn_timeout_errors++;
				ptarget->eh_stats.ssn_cxn_timeout_errors++;
				pcxn->timeout_errors++;
			}

            if((code == CXN_KILLED_RST_RCVD) ||
               (code == CXN_KILLED_FIN_RCVD))
            {
                // Update last async received at
                ACIT_UPDATE_TARGET_LAST_CQERR_TS(pcontroller, ptarget);
            }

            break;

        default:

            pcontroller->eh_stats.last_ssn_failure_type = MIB_SESSION_FAILURE_OTHER_ERRORS;

            break;
        }
    }
}

ACIT_VOID
iscsi_flag_connection(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cri,
    ACIT_UEXACT32           cxn_reason,
    ACIT_UEXACT32           is_ready)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;
    ACIT_UEXACT32           i, affected_cxns = 0, target_reason;

    pcxn = iscsi_get_cxn(pcontroller, CID_FROM_CRI(pcontroller, cri));

    // Mark status on the connection
    if(is_ready)
    {
        ACIT_BIT_CLR(pcxn->busy, cxn_reason);
    }
    else
    {
        ACIT_BIT_SET(pcxn->busy, cxn_reason);
    }

    piscsi_target = pcxn->piscsi_target_ws;
    for(i = 0; i < MAX_CXNS_PER_SESSION; i++)
    {
        if(piscsi_target->pcxn_list[i])
        {
            if(is_ready)
            {
                if(ACIT_IS_BIT_CLR(piscsi_target->pcxn_list[i]->busy, cxn_reason))
                    affected_cxns++;
            }
            else
            {
                if(ACIT_IS_BIT_SET(piscsi_target->pcxn_list[i]->busy, cxn_reason))
                    affected_cxns++;
            }
        }
    }

    if(affected_cxns == piscsi_target->num_connections)
    {
        // map the connection busy state to target busy state
        switch(cxn_reason)
        {
        case CONNECTION_BUSY_WRB_HANDLE:
            target_reason = TARGET_BUSY_WRB_HANDLE;
            break;

        default:
            target_reason = 0;
            OSM_ASSERT(0);
        }

        // signal the target event appropriately
        if(is_ready)
        {
            ap_flag_ready_target(pcontroller, 
                                piscsi_target->pacit_ws, 
                                target_reason);
        }
        else
        {
            ap_flag_busy_target(pcontroller, 
                                piscsi_target->pacit_ws, 
                                target_reason);
        }
    }
}

#ifdef NO_ARM_PORT_FAILOVER
ACIT_UEXACT32
iscsi_process_linkdown_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           port_num)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i, cxn_reason;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    cxn_reason = (port_num == 0)? CONNECTION_BUSY_PORT_A_DOWN : CONNECTION_BUSY_PORT_B_DOWN;

    for(i = 0; i < piscsi_target->num_connections; i++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[i]);
        if (piscsi_target->pcxn_list[i]->port_num == port_num)
        {
            iscsi_flag_connection(pcontroller, 
                                  piscsi_target->pcxn_list[i]->cri,
                                  cxn_reason,
                                  ACIT_FALSE);  // ready = FALSE
        }
    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_process_linkup_event(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget,
    ACIT_UEXACT32           port_num)
{
    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           i, cxn_reason;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    cxn_reason = (port_num == 0)? CONNECTION_BUSY_PORT_A_DOWN : CONNECTION_BUSY_PORT_B_DOWN;

    for(i = 0; i < piscsi_target->num_connections; i++)
    {
        OSM_ASSERT(piscsi_target->pcxn_list[i]);
        if (piscsi_target->pcxn_list[i]->port_num == port_num)
        {
            iscsi_flag_connection(pcontroller, 
                                  piscsi_target->pcxn_list[i]->cri,
                                  cxn_reason,
                                  ACIT_TRUE);   // ready = TRUE
        }
    }

    return ACIT_STATUS_SUCCESS;
}

#endif

ACIT_VOID
iscsi_unitt_init_controller(
    PACIT_CONTROLLER_WS     pcontroller)
{
#ifdef ACIT_UNIT_TEST
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PISCSI_CONNECTION       pcxn[BE_MAX_CXNS + 1] = {ACIT_NULL};
    ACIT_U32                i=0;

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    // test all iscsi connection allocations
    for(i = 0; i < pcontroller->params.num_cxn; i++)
    {
        pcxn[i] = iscsi_alloc_cxn(pcontroller);
        OSM_ASSERT(pcxn[i]);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,
                        ("UnitTest: iscsi_alloc_cxn for cxn %d - 0x%p\n", i, pcxn[i]));
    }

    // the next allocation MUST fail
    pcxn[BE_MAX_CXNS] = iscsi_alloc_cxn(pcontroller);
    OSM_ASSERT(pcxn[BE_MAX_CXNS] == ACIT_NULL);

    for(i = 0; i < pcontroller->params.num_cxn; i++)
    {
        iscsi_free_cxn(pcontroller, pcxn[i]);
    }

    // at the end of it, the connection table must all show nulls
    for(i = 0; i < BE_MAX_CXNS; i++)
    {
        OSM_ASSERT(!piscsi_controller->connection_table[i]);
    }

#endif
}

PACIT_TARGET_WS
iscsi_get_target_from_cid(
    PACIT_CONTROLLER_WS     pcontroller,
    ACIT_UEXACT32           cid)
{
    PISCSI_TARGET_WS        piscsi_target;
    PISCSI_CONNECTION       pcxn;

    pcxn = iscsi_get_cxn (pcontroller, cid);

    OSM_ASSERT(pcxn);
    piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;
    OSM_ASSERT(piscsi_target);

    if(piscsi_target)
    {
        return piscsi_target->pacit_ws;
    }

    return ACIT_NULL;
}

ACIT_UEXACT32
iscsi_start_cmd_window_timer(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    //
    // This routine is called when target command window is closed. Ideally we
    // can start a timer which periodically issues a NOP Out to check on the
    // target's command window. But for now we start a timer which when runs to
    // its completion will trigger a session recovery. The timer is set to huge
    // value (atleast 25 secs) and will be stopped when target opens the command
    // window. We've seen some targets forgetting to open command window and
    // thats why this implementation.
    //
    if(ACIT_IS_BIT_CLR(ptarget->event_flag, EVENT_TGT_PING_WAIT))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Target 0x%p Session %d updated 0 cmd window\n", 
            ptarget, GET_SESSION_HANDLE(pcontroller, ptarget)));

        ptarget->ping_wait_ticks = 
                        ACIT_MAX(ptarget->eto.original_ticks / 4,
                        DEFAULT_PING_WAIT_TICKS);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Ping wait timer started for %d seconds...\n", 
            ptarget->ping_wait_ticks / TICKS_PER_SEC(pcontroller)));

        ap_set_target_event(pcontroller, 
                            ptarget, 
                            EVENT_TGT_PING_WAIT);

    }

    return ACIT_STATUS_SUCCESS;
}

ACIT_UEXACT32
iscsi_stop_cmd_window_timer(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    if(ACIT_IS_BIT_SET(ptarget->event_flag, EVENT_TGT_PING_WAIT))
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Target 0x%p Session %d \
                                           cleared 0 cmd window condition\n", 
                                           ptarget, 
                                           GET_SESSION_HANDLE(pcontroller, ptarget)));

        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Stopping ping wait timer %d \
                seconds before expiry\n", 
                ptarget->ping_wait_ticks / TICKS_PER_SEC(pcontroller)));

        ptarget->ping_wait_ticks = 0;

        ap_clear_target_event(pcontroller, 
                              ptarget, 
                              EVENT_TGT_PING_WAIT);

    }

    return ACIT_STATUS_SUCCESS;
}


ACIT_UEXACT32
iscsi_cleanup_fw_owned_session(
    PACIT_CONTROLLER_WS     pcontroller,
    PACIT_TARGET_WS         ptarget)
{
    PISCSI_TARGET_WS        piscsi_target;

    // this routine is called by MGMT when offload was failed because
    // some parameters in login response was incorrect. 
    mgmt_logout_fw_session(pcontroller,
                           ptarget);

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    OSM_ASSERT(piscsi_target->num_connections);

    iscsi_delete_session_info(pcontroller, ptarget);

    return ACIT_STATUS_SUCCESS;
}

MGMT_STATUS
iscsi_process_ws_init_failure(
    PACIT_CONTROLLER_WS     pcontroller, 
    PACIT_TARGET_WS         ptarget, 
    MGMT_STATUS             mgmt_status)
{

    PISCSI_TARGET_WS        piscsi_target;
    ACIT_UEXACT32           login_ioctl;

    // If offoad has been failing continuosly, do not attempt to
    // ofload again
    ptarget->counters.offload_err++;

    login_ioctl = 
        pcontroller->ptarget_table->target_entry[ptarget->target_table_index].is_login_pending;

    piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

    OSM_ASSERT(piscsi_target);

    piscsi_target->pcxn_list[0]->status = ISCSI_CONN_OFFLOAD_ERROR;

    // if offload fails for targets in recovery, initiate a reopen session in the next
    // timer tick. But do not issue reopen if we're shutting down...
    
    //
    // For Palau2.3 - try to recover from offload failures for even the 
    // persistent targets that are initializing the first time
    //
    
    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS))
    {
        // ignore offload failure here. simply return
        return mgmt_status;
    }

    // If offload failed for a target that was logged in via an
    // IOCTL request, then cleanup and return. There's no recovery
    // or retry attempt
    if(login_ioctl)
    {
        // If its a driver returned status, then do not close the session.
        // This is because, technically the offload has succeeded. But because
        // the ARM fw didn't validate the login response thoroughly, some parameter
        // is invalid. Fail the offload, but don't issue a logout yet since
        // iscsi_cleanup_fw_owned_session assumes the session is with ARM fw.
        //
        if(mgmt_status != MGMT_STATUS_ERROR_ACITISCSI)
        {
            // The offload was failed for a login IOCTL. Close the connection with ARM fw 
            // so that the session is not valid
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Cleaning up target 0x%p that failed offload\n",
                                                    ptarget));

            iscsi_cleanup_fw_owned_session(pcontroller, ptarget);
        }

        return mgmt_status;
    }


    if(ptarget->counters.offload_err <= pcontroller->params.offload_failure_threshold)
    {
        // This is either a persistent target or a target in recovery.                

        // If its the driver that failed the parameter validation for
        // offload, then from ARM fw/hw point of view, the offload
        // is successful and driver owns the session. So the driver
        // has to issue a logout to clear up the target before
        // asking fw to reopen session again
        //
        if(mgmt_status == MGMT_STATUS_ERROR_ACITISCSI)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Parameter error during offload\n"));
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Setting recovery pending event for target 0x%p\n",
                        ptarget));

            //
            // If this target is not already in recovery, then this is a
            // persistent target that is initializing the first time. 
            // Return status success from this call so that OSM initialization 
            // for the target will be successful.
            //
            if(ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY))
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Return offload with success for target 0x%p\n",
                            ptarget));
                mgmt_status = MGMT_STATUS_SUCCESS;  // see !!Alert!! below

                // Since its a persistent target init the first time, set recovery
                // flag and invalidate the connection. By doing this, we chain the
                // target to its normal course of session recovery..
                //
                ptarget->busy |= TARGET_BUSY_SESSION_RECOVERY;
            }
            
            // No MCS support
            ACIT_BIT_SET(piscsi_target->pcxn_list[0]->status, ISCSI_CONN_STATUS_NEED_RESET);

            ap_set_target_event (pcontroller, 
                                 ptarget, 
                                 EVENT_TGT_CXN_EVENT_PENDING);
        }
        else
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Setting reopen pending event for target 0x%p\n",
                        ptarget));

            ap_set_target_event(pcontroller, ptarget, EVENT_TGT_REOPEN_PENDING);

            if(ACIT_IS_BIT_CLR(ptarget->busy, TARGET_BUSY_SESSION_RECOVERY))
            {
                // !!Alert!!
                // mgmt_status is hardcoded to success for those cases when there is no
                // login IOCTL. This is to let OSMs not make change to handle offload
                // failure the first time a persistent target is initializing. The 
                /// change I'm implying is to do with target state intialization. 
                // Normally when acit_init_target_ws fails, OSMs will not initialize 
                // the target state. It may not report the target availability to OS 
                // either. If such targets come online later and offload successfully, 
                // there is no good way to notify OSMs to report the target or re-init
                // the target again. ACIT could notify of target ready, but OSMs may 
                // not (do not!) maintain state and see if the target has never been
                // reported. 
                // To simplify the OSM handling, we will return acit_init_target_ws
                // with success, but make the target busy so that we don't process
                // anything on this target. We will also report the target as
                // busy (once we are out of acit_init_target_ws context) so that
                // from OSM point of view its like any other session recovery.
                // 
                mgmt_status = MGMT_STATUS_SUCCESS;

                //
                // mark the target as busy directly here. The API to mark target
                // busy will notify the OSM and OSMs may not be ready for target
                // event when it is still being initialized. The flag is marked
                // busy so that we won't process the target. If we wait for the
                // next timer tick to mark target busy it may not be early enough
                //
                ptarget->busy |= TARGET_BUSY_SESSION_RECOVERY;
            }
        }
    }
    else
    {
        // set target flag to indicate that offload failures exceed
        // configured thresholds
        ACIT_BIT_SET(ptarget->flags, TGT_FLAG_OFFLOAD_RETRY_EXCEEDED);

        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Offload fail threshold reached, stopping recovery on target 0x%p\n",
                                                ptarget));

        // The threshold for tolerating offload failures has reached. Do not
        // retry reopen again. If it is a target that was previously in recovery,
        // expire ETO and return all IOs belonging to the target

        if(mgmt_status == MGMT_STATUS_ERROR_ACITISCSI)
        {
            // The offload was failed by the driver. Close the connection with EP fw 
            // so that the session is not valid. The target flags will be used during
            // upload so that a recovery will not be attempted again

            // No MCS support
            ACIT_BIT_SET(piscsi_target->pcxn_list[0]->status, ISCSI_CONN_STATUS_NEED_RESET);

            ap_set_target_event (pcontroller, 
                                 ptarget, 
                                 EVENT_TGT_CXN_EVENT_PENDING);
        }
        else
        {
            // expire ETO. Since we're directly invoking the expiry routine, the
            // target flags will give a hint that ETO expiry occurred because
            // of offload failures
            ap_force_eto_expiry(pcontroller, ptarget);
        }
    }

    return mgmt_status;
}

ACIT_VOID
iscsi_process_unsol_data_digest_err_notify(
    PACIT_CONTROLLER_WS     pcontroller, 
    ACIT_UEXACT32           cid)
{
    PISCSI_CONNECTION       pcxn = iscsi_get_cxn (pcontroller, cid);
    PISCSI_TARGET_WS        piscsi_target = ACIT_NULL;

    if(pcxn)
    {
        piscsi_target = (PISCSI_TARGET_WS)pcxn->piscsi_target_ws;
        OSM_ASSERT(piscsi_target);
    }

    if(piscsi_target)
    {
        piscsi_target->flags |= ISCSI_DATA_DIGEST_ERR;
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Setting Digest Error in PDU bit for cid %x...\n", pcxn->cid));
    }
    else
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("iscsi target on cid %x invalid\n", cid));
    }

    return;
}

ACIT_UEXACT32
iscsi_cleanup_all_sessions(
    PACIT_CONTROLLER_WS     pcontroller)
{
    PISCSI_CONTROLLER_WS    piscsi_controller;
    PACIT_TARGET_WS         ptarget;
    PISCSI_TARGET_WS        piscsi_target;
    PACIT_TARGET_ENTRY      ptarget_entry = ACIT_NULL;
    ACIT_UEXACT32           index;

    //
    // This routine is called in response to a UE detection. It iterates thru
    // all current targets and cleans up the resources associated
    //

    piscsi_controller = GET_ISCSI_CONTROLLER_WS(pcontroller);

    ACIT_BIT_SET(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_UE_RECOVERY);

    ptarget_entry = &pcontroller->ptarget_table->target_entry[0];

    for(index = 0; 
        index < pcontroller->params.targets_per_controller;
        index ++)
    {
        if(ACIT_IS_BIT_SET(ptarget_entry[index].acit_tflags,
                            ACIT_TFLAG_IN_USE))
        {
            ptarget = GET_TARGET_WS_SAFE(ptarget_entry[index].ptarget_key);

            if(!ptarget)
                continue;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_ERROR,("Cleanup target %p..\n", ptarget));

            piscsi_target = GET_ISCSI_TARGET_WS(ptarget);

            if (piscsi_target->mgmt_ioctl_request)
            {
                mgmt_return_ioctl_request(pcontroller,
                                        piscsi_target->mgmt_ioctl_request,
                                        ACIT_STATUS_ERROR);
            }

            ap_cleanup_target(pcontroller, ptarget, CL_REASON_UE_TOLERANCE);

            OSM_ASYNC_EVENT(pcontroller->pcontroller_key,
                            ACIT_EVENT_CLASS_TARGET,
                            ACIT_EVENT_TARGET_MISSING,
                            ptarget->target_table_index);
        }
    }

    if(piscsi_controller->mgmt_ioctl_request)
    {
        mgmt_return_ioctl_request(pcontroller,
                                piscsi_controller->mgmt_ioctl_request,
                                ACIT_STATUS_ERROR);

        piscsi_controller->mgmt_ioctl_request = ACIT_NULL;
    }

    ACIT_BIT_CLR(piscsi_controller->flags, FLAG_ISCSI_ALL_TGTS_UE_RECOVERY);

    return ACIT_STATUS_SUCCESS;
}
