/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __checksum_h__
#define __checksum_h__

SA_CPP_HEADER

#include "sa_network_packet.h"

// Computes TCP checksum for an IPV4 packet assuming
// all tcp data is contiguous after the tcp header.
u16 sa_ipv4_tcp_checksum (
    SA_IP_HEADER *ip,     
    SA_TCP_HEADER *tcp);

// Computes TCP checksum for an IPV6 packet assuming
// all tcp data is contiguous after the tcp header.
u16 sa_ipv6_tcp_checksum (
    SA_IPV6_HEADER *ip,   // ptr to IPV6 header
    SA_TCP_HEADER *tcp);  // ptr to TCP header

// Performs the IPV4 UDP checksum assuming the UDP data is contiguous
// with the udp header.
u16 sa_ipv4_udp_checksum(
    SA_IP_HEADER *ip,   
    SA_UDP_HEADER *udp);

// Performs the IPV6 UDP checksum assuming the UDP data is contiguous
// with the udp header.
u16 sa_ipv6_udp_checksum (
    SA_IPV6_HEADER *ip,   // ptr to IPV6 header
    SA_UDP_HEADER *udp);   // ptr to udp header and data

// Performs the IPV4 header checksum.
u16 sa_ipv4_checksum(
    SA_IP_HEADER *ip);

// ------------------------------------------------------------------
// Legacy versions -- from before we had structs for the header types.
// ------------------------------------------------------------------

u16 sa_udp_checksum(
    u8 *pheader,     // ptr to pseudo header (only src & dst IP addresses!)
    u8 *udpheader,   // ptr to UDP header and data
    u32 udplen);     // UDP header + data length

u16 sa_ip_checksum(
    u8 *pdata,   // pointer to data IP Header to checksum
    u32 len);    // length in bytes

u16 sa_tcp_checksum(
    u8 *pheader,     // ptr to pseudo header (only src & dst IP addresses!)
    u8 *tcpheader,   // ptr to tcp header and data
    u32 tcplen);     // tcp frame length

typedef struct _SA_CHECKSUM_INFO {
    u32 len;
    u32 sum;
} SA_CHECKSUM_INFO, *PSA_CHECKSUM_INFO;

// Start TCP checksum calculation, saving intermediate results in the
// opaque info struct.  The caller should provide an info struct, but
// it should never need to to look at individual fields within the struct.
// More data may be added to the checksum with the sa_tcp_checksum_continue
// function.  The final checksum is returned with the sa_tcp_checksum_end 
// function.
void sa_tcp_checksum_start (
    u8 *pheader,     // ptr to pseudo header
    u8 *tcpheader,   // ptr to tcp header and data
    u32 tcplen,      // tcp frame length
    SA_CHECKSUM_INFO *info);  // opaque info used during computation

void sa_tcp_checksum_continue (
         u8*               tcpdata, // ptr to tcp header and data
         u32               tcplen,  // tcp frame length
         SA_CHECKSUM_INFO* info);   // opaque info used during computation
    
u16 sa_tcp_checksum_end (SA_CHECKSUM_INFO* info);

SA_CPP_TRAILER
#endif

