/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_convert.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_convert.h,v 1.2 2006/02/28 21:24:30 menz Exp $
*/
#ifndef __sa_convert_h__
#define __sa_convert_h__

#include "sa_types.h"
#include "sa_assert.h"
SA_CPP_HEADER

//  -----------------------------------------------------------------------
// Simple conversions are located in sa_defines.h.
// These are functional conversions that require asserts (not just casting).
//  -----------------------------------------------------------------------

// Safely converts u32 to a i16.  Checking that the upper bits are either
// 0 or Fs for a postive or negative number respectively.  This is particularly
// useful to avoid warnings about an assignment or cast truncating information.
// This routine asserts that the truncated information is the expected value.
STATIC FORCEINLINE i16
SA_U32_TO_I16 (u32 value) {
    ASSERT ((value & 0xFFFF0000UL) == 0xFFFF0000UL ||
            (value & 0xFFFF0000UL) == 0UL); // positive or negative
    return (i16)(value & 0xFFFF);
}

// Same as above, but for i8
STATIC FORCEINLINE i8
SA_U32_TO_I8 (u32 value) {
    ASSERT ((value & 0xFFFFFF00UL) == 0xFFFFFF00UL ||
            (value & 0xFFFFFF00UL) == 0UL); // positive or negative
    return (i8)(value & 0xFF);
}


STATIC FORCEINLINE i16
SA_U64_TO_I16 (u64 value) {
    ASSERT ((value & 0xFFFFFFFFFFFF0000ULL) == 0xFFFFFFFFFFFF0000ULL ||
            (value & 0xFFFFFFFFFFFF0000ULL) == 0ULL); // positive or negative
    return (i16)(value & 0xFFFFULL);
}

// Same as above, but for i8
STATIC FORCEINLINE i8
SA_U64_TO_I8 (u64 value) {
    ASSERT ((value & 0xFFFFFFFFFFFF0000ULL) == 0xFFFFFFFFFFFF0000ULL ||
            (value & 0xFFFFFFFFFFFF0000ULL) == 0ULL); // positive or negative
    return (i8)(value & 0xFFULL);
}

STATIC FORCEINLINE i16
SA_I32_TO_I16 (i32 value) {
    ASSERT ((value & 0xFFFF0000L) == 0xFFFF0000L ||
            (value & 0xFFFF0000L) == 0UL); // positive or negative
    return (i16)(value & 0xFFFF);
}

// Same as above, but for i8
STATIC FORCEINLINE i8
SA_I32_TO_I8 (i32 value) {
    ASSERT ((value & 0xFFFFFF00L) == 0xFFFFFF00L ||
            (value & 0xFFFFFF00L) == 0L); // positive or negative
    return (i8)(value & 0xFF);
}

SA_CPP_TRAILER    
#endif // __sa_convert_h__



