/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_platform.h

@brief

    Defines preprocessor macros based on the current operating system
    and environment.  All other files should use these macros.

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_platform.h,v 1.15 2006/03/10 20:22:40 menz Exp $
*/
#ifndef __sa_platform_h__
#define __sa_platform_h__


// ------------------------------------------------------------
// Detect operating system and environment.
// ------------------------------------------------------------
#ifdef _MSC_VER

    #define SA_OS "windows"
    #define SA_MICROSOFT    
        
    // DETECT windows version.
    #if defined(_WINXP)        
        #define SA_VER_OS_STR "Microsoft\251 Windows XP"
        #define SA_WIN_VERSION 2600
    #elif defined (_WNET)       
        #define SA_VER_OS_STR "Microsoft\251 Windows 2003"
        #define SA_WIN_VERSION 3663
    #elif defined (_NT_5)
        #define SA_VER_OS_STR "Microsoft\251 Windows 2000(TM)"
        #define SA_WIN_VERSION 2195
    #elif defined(_NT_40)        
        #define SA_VER_OS_STR "Microsoft\251 Windows (TM) NT 4.0"
        #define SA_WIN_VERSION 4000
    #else        
        #define SA_VER_OS_STR  "Microsoft\Unknown"
        #define SA_WIN_VERSION VER_PRODUCTBUILD
    #endif


    // Detect DDK target OS version
    #if (WINVER==0x0500)
        #ifndef SA_TARGET_WIN2000
            #define SA_TARGET_WIN2000
        #endif
    #elif (WINVER==0x0502)
        #ifndef SA_TARGET_WIN2003
            #define SA_TARGET_WIN2003
        #endif
    #elif (WINVER==0x0600)
        #ifndef SA_TARGET_WIN2008
            #define SA_TARGET_WIN2008
        #endif
    #elif (WINVER==0x0602)
        #ifndef SA_TARGET_WIN2008
            #define SA_TARGET_WIN8
        #endif
    #endif
        
    // Detect windows environment.    
    #if defined(_SA_STORPORT)
        #define SA_ENV "storport"
        #define SA_KERNEL 
        #define SA_STORPORT
        #define SA_STORAGE_MINIPORT

        // Operating system define to allow us to use scsiport functions that are
        // alised to the storport versions.
        #define STOR_USE_SCSI_ALIASES

        #if !defined (_WIN8)
            #define SA_FAKE_SAFE_PRINTS
        #elif defined (_WIN8)
            #define _SA_MODULE_NAME "storport"
        #endif

    #elif defined(_SA_SCSIPORT)
        #define SA_ENV "scsiport"
        #define SA_KERNEL
        #define SA_SCSIPORT
        #define SA_STORAGE_MINIPORT

        // TODO: Figure out a way to know build environment. And fake safe prints
        //       for Windows 2008 or lesser OS versions
        #define SA_FAKE_SAFE_PRINTS
        
    #elif defined(_SA_NDIS)
    
        #if NDIS50_MINIPORT
            #define SA_NDIS50
            #define SA_NDIS_VERSION "5.0"
            // Fake safe prints for Windows 2008 or lesser OS versions
            #define SA_FAKE_SAFE_PRINTS
        #elif NDIS51_MINIPORT
            #define SA_NDIS51
            #define SA_NDIS_VERSION "5.1"
            // Fake safe prints for Windows 2008 or lesser OS versions
            #define SA_FAKE_SAFE_PRINTS
        #elif NDIS52_MINIPORT
            #define SA_NDIS52
            #define SA_NDIS_VERSION "5.2"
            // Fake safe prints for Windows 2008 or lesser OS versions
            #define SA_FAKE_SAFE_PRINTS
        #elif NDIS60_MINIPORT
            #define SA_NDIS60
            #define SA_NDIS_VERSION "6.0"
            // Fake safe prints for Windows 2008 or lesser OS versions
            #define SA_FAKE_SAFE_PRINTS
        #elif NDIS61_MINIPORT
            #define SA_NDIS61
            #define SA_NDIS_VERSION "6.1"
            #define __SA_PREFAST_SUPPORTED
            // Fake safe prints for Windows 2008 or lesser OS versions
            #define SA_FAKE_SAFE_PRINTS
        #elif NDIS620_MINIPORT
            #define SA_NDIS620
            #define SA_NDIS_VERSION "6.20"
            #define __SA_PREFAST_SUPPORTED
        #elif NDIS630_MINIPORT
            #define SA_NDIS630
            #define SA_NDIS_VERSION "6.30"
            #define _SA_MODULE_NAME "be2nd63"
            // Code Analysis for Win8 doesn't like function annotations
            //#define __SA_PREFAST_SUPPORTED
        #else
            #error Undefined NDIS miniport version.
            #define SA_NDIS_VERSION "?"
        #endif
        
        #define SA_ENV      "NDIS " SA_NDIS_VERSION
        #define SA_KERNEL
        #define SA_NDIS             
        
    #elif defined(_SA_WIN32K)
        #define SA_ENV "win kernel"
        #define SA_KERNEL         
        #define SA_WIN_KERNEL
        #define SA_WIN32K
        #define SA_FAKE_SAFE_PRINTS
                
    #elif defined(_SA_WIN32) || defined(_SA_WIN32U)
        #define SA_ENV "win user"
        #define SA_USER
        #define SA_WIN_USER
        #define SA_WIN32U
        #define SA_FAKE_SAFE_PRINTS
    #endif

#elif defined(__linux__) || defined (LINUX_OS)

    #define SA_OS "linux"
    #define SA_LINUX

    #if defined(_SA_LINUXK)
        #define SA_ENV "linux kernel"
        #define SA_KERNEL         
        #define SA_LINUX_KERNEL
        #define SA_LINUXK
        #ifndef __KERNEL__
            #define __KERNEL__
        #endif
        #ifndef LINUX_OS
            #define LINUX_OS
        #endif

    #elif defined(_SA_LINUXU)
        #define SA_ENV "linux user"
        #define SA_USER
        #define SA_LINUX_USER
        #define SA_LINUXU
    #endif

#elif defined(_SA_DOS)

    #define SA_OS "dos"
    
    #define SA_DOS
    #define SA_USER
    #define SA_ENV "dos"
    
#else
    #error Undefined SA environment. e.g. _SA_STORPORT, _SA_NDIS, _SA_WIN32k, _SA_WIN32, _SA_LINUXK, _SA_LINUXU, _SA_DOS.
#endif



// ------------------------------------------------------------
// Detect a simulation build.  This modifies the OS routines slightly
// to support simulating some of the memory funcitonality (like
// virtual to physical memory mapping).
// ------------------------------------------------------------
#if defined(_SA_SIM)
    #define SA_SIM
#endif

// ------------------------------------------------------------
// Detect build type
// ------------------------------------------------------------
#if (DBG==1) || defined(_DEBUG) || defined(_SA_DEBUG)
    #define DBG 1
    #define DEBUG
    #define SA_DEBUG
#else
    #define SA_RELEASE
#endif


// ------------------------------------------------------------
//  Detect integer size
// ------------------------------------------------------------
#if defined(WIN32) || defined(_SA_32BIT)
    #define SA_32BIT
#else
    #define SA_64BIT
#endif


// ------------------------------------------------------------
// Detect application type
// ------------------------------------------------------------
#if defined(_CONSOLE) || defined(_SA_CONSOLE)
    #define SA_CONSOLE
#endif   


// ------------------------------------------------------------
// Detect Language
// ------------------------------------------------------------
#if defined(__cplusplus)
   #define  SA_CPP
   #define  SA_CPP_HEADER         extern "C" {
   #define  SA_CPP_TRAILER        }   
#else
   #define  SA_C
   #define  SA_CPP_HEADER
   #define  SA_CPP_TRAILER   
#endif

// ------------------------------------------------------------
//  Detect processor.
// ------------------------------------------------------------
#if defined(_X86_) || defined(_SA_X86) || defined(_M_IX86)
    #define SA_X86
#elif defined(_AMD64_) || defined(AMD64) || defined(_SA_AMD64)
    #define SA_AMD64
#elif defined(_XSCALE_)
    #define SA_XSCALE
#elif defined(_PPC_)
    #define SA_PPC
#elif defined(_IA64_)
    #define SA_IA64
#else
    #error Undefined processor. e.g. _SA_X86
#endif

// ------------------------------------------------------------
//  Disable some warnings.
// ------------------------------------------------------------
#ifdef SA_MICROSOFT

    #pragma warning (disable:4201)      // nameless struct/union
    #pragma warning (disable:4214)      // bit field types other than int
    #pragma warning (disable:4115)      // named type definition in parentheses
    #pragma warning (disable:4514)      // unreferenced inline function has been removed
    #pragma warning (disable:4115)      // 'type' : named type definition in parentheses
    #pragma warning (disable:4296)      // '>=' : expression is always true

    #pragma warning (disable:4101)  // unreferenced local variable
    #pragma warning (disable:4100)  // unreferenced parameter

    //#if defined(SA_DEBUG)
        #pragma warning (disable:4702) //  unreachable code
        #pragma warning (disable:4505)  // unreferenced local function has been removed
        #pragma warning (disable:4189) // local variable is initialized but not referenced
    //#endif

    #if defined (SA_SCSIPORT)
        // This warning exists in windows headers for win2000 ddk.        
        #pragma warning (disable:4200) // nonstandard extension used : zero-sized array in struct/union
    #endif

#endif

// ------------------------------------------------------------
// Include platform header files.
// ------------------------------------------------------------
#if defined(SA_WIN_KERNEL)
    #include <ntddk.h>
#elif defined (SA_STORPORT)
    #include <ntddk.h>
    #include <storport.h>
#elif defined (SA_SCSIPORT)
    #include <ntddk.h>
    #include <scsi.h>
    #include <srb.h>    
#elif defined (SA_NDIS)
    #include <ntddk.h>
    #include <ndis.h>
#elif defined(SA_WIN_USER)
    #include <windows.h>
#elif defined(SA_LINUXU)
    #include <sys/types.h>
#elif defined(SA_LINUXK)       
    #include <linux/kernel.h>
#elif defined(SA_DOS)
    #include <sys/types.h>
#else
    #error "No platform header files defined"
#endif

//
// There are no PREFast annotaions for Linux or Windows < Windows Vista
// Create wrapper around PREFast annotations.
//
#undef __SA_PREFAST

#if defined(__SA_PREFAST_SUPPORTED)
# define __SA_PREFAST(_notation_) _notation_
#else
# define __SA_PREFAST(_notation_)
#endif

#endif // __sa_platform_h__


