/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_std.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_std.h,v 1.9 2006/03/15 22:09:21 menz Exp $
*/
#ifndef __sa_std_h__
#define __sa_std_h__

// All environments need stdarg.
#include <stdarg.h>

SA_CPP_HEADER

#define SA_PRINTF  printf
#define SA_SPRINTF sprintf
#define SA_FPRINTF fprintf

#if defined(SA_MICROSOFT)

    #include <stdio.h>

    SA_CPP_HEADER
    //LONG  __cdecl _InterlockedIncrement(LONG volatile *Addend);
    //LONG  __cdecl _InterlockedDecrement(LONG volatile *Addend);
    LONG  __cdecl _InterlockedCompareExchange(LONG volatile* Dest, LONG Exchange, LONG Comp);
    //LONG  __cdecl _InterlockedExchange(LPLONG volatile Target, LONG Value);
    //LONG  __cdecl _InterlockedExchangeAdd(LPLONG volatile Addend, LONG Value);
    SA_CPP_TRAILER

    #pragma intrinsic(_InterlockedCompareExchange)
    #pragma intrinsic(memcmp, memcpy, memset)

    //#define InterlockedCompareExchange _InterlockedCompareExchange
    #define SA_SNPRINTF   _snprintf 
    #define SA_VSNPRINTF  _vsnprintf
    #define SA_VSNWPRINTF  _vsnwprintf
    #define SA_FORMAT_U64 "%I64u"
    #define WSA_FORMAT_U64 L"%I64u"
    #define SA_FORMAT_I64 "%I64d"
    #define WSA_FORMAT_I64 L"%I64d"
    #define SA_FORMAT_X64 "%I64x"
    #define WSA_FORMAT_X64 L"%I64x"
    #define SA_FORMAT_XX64 "%I64X"
    #define WSA_FORMAT_XX64 L"%I64X"
    #define SA_FORMAT64(_digits_, _type_) "%" #_digits_ "I64" #_type_

    #if defined(SA_FAKE_SAFE_PRINTS)        
        STATIC INLINE int SA_SNPRINTF_S (char* buf, size_t sizeOfBuffer, size_t count, const char* format, ...) {
            int value;
            va_list args;
            va_start (args, format);            
            value = SA_VSNPRINTF (buf, count, format, args);
            va_end (args);
            return value;
        }

        #define SA_VSNPRINTF_S(_buf, _size, _count, _format, _va_list)  SA_VSNPRINTF(_buf, _count, _format, _va_list)

        STATIC INLINE int SA_SPRINTF_S (char* buf, size_t sizeOfBuffer, const char* format, ...) {
            int value;
            va_list args;
            va_start (args, format);            
            value = SA_SPRINTF (buf, format, args);
            va_end (args);
            return value;
        }

    #else
        #define SA_SNPRINTF_S   _snprintf_s
        #define SA_VSNPRINTF_S  _vsnprintf_s
        #define SA_SPRINTF_S    sprintf_s
    #endif

    #include <stdio.h>
    #include <stdlib.h>

    #if defined(SA_USER) 
        #include <conio.h>  // for _getche in user mode asserts.
        #define SA_GETC()   _getche() // Gets a char from stdin without pressing return
    #else
        // kernel mode
    #endif

#else // Linux, etc.
    
            
    #if defined (SA_USER) 
	    #include <string.h> // for memset, memcpy, etc.
        #include <stdio.h>
        #include <stdlib.h>
        #include <unistd.h> 
        STATIC INLINE char sa_getc (void) { 
            char c = getc (stdin);  // Gets a char from stdin, requires pressing return.

            if (c == EOF) {                
                return c;
            }
            
            if (c == '\n') {
                return c;
            }

            getc (stdin);           // Get the newline char
            return c;
        }
        #define SA_GETC            sa_getc
    #else
        // kernel mode
        #include <linux/slab.h>
    #endif

    #if defined(SA_DOS)
        // DJGPP doesn't have snprintf in std library. These are defined in sa_dos_snprintf.c
        SA_SIZE_T sa_vsnprintf (c8 *str, SA_SIZE_T len, const c8* format, va_list args);
        SA_SIZE_T sa_snprintf (c8 *str, SA_SIZE_T len, const c8* format, ...);

        #define SA_VSNPRINTF sa_vsnprintf
        #define SA_SNPRINTF  sa_snprintf
        #define SA_VSNPRINTF_S(_buffer, _size, _count, _format, _arguments) SA_VSNPRINTF(_buffer, _count, _format, _arguments)
        //#define SA_SNPRINTF_S
        //#define SA_SPRINTF_S
    #else
        #define SA_SNPRINTF   snprintf 
        #define SA_VSNPRINTF  vsnprintf
        #define SA_VSNPRINTF_S(_buffer, _size, _count, _format, _arguments) vsnprintf(_buffer, _count, _format, _arguments)
        #define SA_SNPRINTF_S(_buffer, _size, _count, _format, args...) snprintf(_buffer, _count, _format, ##args)
        #define SA_SPRINTF_S(_buffer, _size, _format, args...) sprintf(_buffer, _format, ##args)
    #endif

    #define SA_FORMAT_U64 "%llu"
    #define SA_FORMAT_I64 "%lld"
    #define SA_FORMAT_X64 "%llx"
    #define SA_FORMAT_XX64 "%llX"
    #define SA_FORMAT64(_digits_, _type_) "%" #_digits_ "ll" #_type_

#endif



SA_CPP_TRAILER

#endif // __sa_std_h__

